#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "group.h"
#include "pmt.h"
#include "group_check.h"
#include "group_all_cjg.h"

// ----------------------------------------------------------------
// xxx to lib func

int set_find(
	void * set_buf,
	int nelts,
	group_t * pgroup,
	void * sought)
{
	int k;
	void * pa;

	for (k = 0; k < nelts; k++) {
		pa = VARRAY_ELEMENT(set_buf, k, pgroup->pspec->element_size);
		if (pgroup->pspec->pvequal(pa, sought, pgroup->pvaux))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
// Form the union of gHg^-1 for all g in G.  Unlike the *intersection* of
// gHg^-1 for all g in G (i.e. the "core" of H in G), this is not a group.

void group_all_cjg(
	group_t * pG,
	group_t * pH,
	void    * cjg_buf,
	int       max_cjg_bytes,
	int     * pncjg,
	int       verbosity)
{
	int i, j;
	void * pg, * ph, * pdst;
	unsigned char gh[VOP_MAX_ELEMENT_SIZE];
	unsigned char gi[VOP_MAX_ELEMENT_SIZE];
	unsigned char ghgi[VOP_MAX_ELEMENT_SIZE];
	int max_cjg_elts = max_cjg_bytes / pG->pspec->element_size;
	int ncjg = 0;
	int already_there;

	// xxx to lib func.
	if (pG->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pG->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pG->num_elements; i++) {
		pg = VGROUP_ELEMENT(pG, i);
		for (j = 0; j < pH->num_elements; j++) {
			ph = VGROUP_ELEMENT(pH, j);

			pG->pspec->pvop(gh, pg, ph, pG->pvaux);
			group_element_inverse(pG, pg, gi);
			pG->pspec->pvop(ghgi, gh, gi, pG->pvaux);

			already_there = set_find(cjg_buf, ncjg, pG, ghgi);

			if (verbosity) {
				printf("%sdding: ",
					already_there ? "Not a" : "A");
				pG->pspec->pvprint(pg, pG->pvaux);
				printf(" * ");
				pG->pspec->pvprint(ph, pG->pvaux);
				printf(" * ");
				pG->pspec->pvprint(gi, pG->pvaux);
				printf(" = ");
				pG->pspec->pvprint(ghgi, pG->pvaux);
				printf("\n");
			}

			if (already_there)
				continue;

			if (ncjg >= max_cjg_elts) {
				fprintf(stderr, "group_all_cjg: "
					"too many elements; max %d.\n",
					max_cjg_elts);
				exit(1);
			}
			pdst = VARRAY_ELEMENT(cjg_buf, ncjg,
				pG->pspec->element_size);
			memcpy(pdst, ghgi, pG->pspec->element_size);
			ncjg++;
		}
	}
	*pncjg = ncjg;
}
