#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "group.h"
#include "pmt.h"
#include "group_check.h"
#include "group_cayley_sn.h"

void make_cayley_Sn_subgroup(
	group_t * pgroup,
	group_t * psn_subgroup,
	void    * sgr_elements_buffer,
	int       max_sn_sgr_bytes)
{
	pmt_t * pSns;
	int i, j, k;
	void * px;
	void * pg;
	unsigned char xg[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}


	psn_subgroup->pvelements = sgr_elements_buffer;
	psn_subgroup->num_elements = pgroup->num_elements;

	psn_subgroup->pvaux = &psn_subgroup->num_elements;
	psn_subgroup->pspec = &sn_group_spec;

	pSns = (pmt_t *)sgr_elements_buffer;

	// for each x [i] in G
	//	for each g [j] in G
	//		find index [k] of xg in G
	//		cayley_Sns[i].images[j] = k

	for (i = 0; i < pgroup->num_elements; i++) {
		px = VGROUP_ELEMENT(pgroup, i);
		for (j = 0; j < pgroup->num_elements; j++) {
			pg = VGROUP_ELEMENT(pgroup, j);
			pgroup->pspec->pvop(xg, px, pg, pgroup->pvaux);

			if (!group_find_index(pgroup, xg, &k)) {
				fprintf(stderr,
				"make_cayley_Sn_subgroup: group not closed.\n");
				exit(1);
			}
			pSns[i].images[j] = k;
		}
	}
}
