#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "group.h"
#include "pmt.h"
#include "group_check.h"
#include "group_centralizer.h"

void make_centralizer(
	group_t * pG,
	group_t * pH,
	group_t * pcentralizer,
	void    * centralizer_elements_buffer,
	int       max_centralizer_bytes)
{
	int    i, j, all_commute;
	void * pg, * ph, * pc;
	unsigned char hg[VOP_MAX_ELEMENT_SIZE];
	unsigned char gh[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pG->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pG->pspec->element_size);
		exit(1);
	}

	pcentralizer->pvelements = centralizer_elements_buffer;
	pcentralizer->num_elements = 0; // Increment below
	pcentralizer->pvaux = pG->pvaux;
	pcentralizer->pspec = pG->pspec;

	//for each g [i] in G
	//  for each h [j] in H
	//    if hg != gh
	//      next g
	//  add g to the centralizer

	for (i = 0; i < pG->num_elements; i++) {
		pg = VGROUP_ELEMENT(pG, i);

		all_commute = 1;
		for (j = 0; j < pH->num_elements; j++) {
			ph = VGROUP_ELEMENT(pH, j);

			pG->pspec->pvop(gh, pg, ph, pG->pvaux);
			pG->pspec->pvop(hg, ph, pg, pG->pvaux);

			if (!pG->pspec->pvequal(gh, hg, pG->pvaux)) {
				all_commute = 0;
				break;
			}
		}

		if (all_commute) {
			pc = VGROUP_ELEMENT(pcentralizer,
				pcentralizer->num_elements);
			memcpy(pc, pg, pG->pspec->element_size);
			pcentralizer->num_elements++;
		}
	}
}
