#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "group.h"
#include "group_check.h"

// ----------------------------------------------------------------
int is_group(
	group_t * pgroup,
	int verbosity)
{
	int closed;
	int assoc;
	int id;
	int invs;

	if (verbosity)
		printf("\n");


	if (verbosity) {
		printf("  Closed:          ");
		fflush(stdout);
	}
	closed = is_closed(pgroup, verbosity);
	if (verbosity) {
		printf("%s\n", closed ? "yes" : "no");
	}


	if (verbosity) {
		printf("  Associative:     ");
		fflush(stdout);
	}
	assoc  = is_associative(pgroup, verbosity);
	if (verbosity) {
		printf("%s\n", assoc  ? "yes" : "no");
	}


	if (verbosity) {
		printf("  Unique identity: ");
		fflush(stdout);
	}
	id     = has_unique_identity(pgroup, verbosity);
	if (verbosity) {
		printf("%s\n", id     ? "yes" : "no");
	}


	if (verbosity) {
		printf("  Unique inverses: ");
		fflush(stdout);
	}
	invs   = has_unique_inverses(pgroup, verbosity);
	if (verbosity) {
		printf("%s\n", invs   ? "yes" : "no");
	}

	return closed && assoc && id && invs;
}

// ----------------------------------------------------------------
int is_closed(
	group_t * pgroup,
	int verbosity)
{
	void * pa;
	void * pb;
	unsigned char ab[VOP_MAX_ELEMENT_SIZE];
	int i, j;
	int rv = 1;
	int index;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);

		for (j = 0; j < pgroup->num_elements; j++) {
			pb = VGROUP_ELEMENT(pgroup, j);

			pgroup->pspec->pvop(ab, pa, pb, pgroup->pvaux);
			if (!group_find_index(pgroup, ab, &index)) {
				if (verbosity >= 2) {
					printf("Closure: ");
					pgroup->pspec->pvprint(pa,
						pgroup->pvaux);
					printf(" * ");
					pgroup->pspec->pvprint(pb,
						pgroup->pvaux);
					printf(" = ");
					pgroup->pspec->pvprint(ab,
						pgroup->pvaux);
					printf(" not in group.\n");
				}
				rv = 0;
			}
		}
	}
	return rv;
}

// ----------------------------------------------------------------
int is_associative(
	group_t * pgroup,
	int verbosity)
{
	void * pa;
	void * pb;
	void * pc;
	unsigned char ab[VOP_MAX_ELEMENT_SIZE];
	unsigned char bc[VOP_MAX_ELEMENT_SIZE];
	unsigned char ab_c[VOP_MAX_ELEMENT_SIZE];
	unsigned char a_bc[VOP_MAX_ELEMENT_SIZE];
	int i, j, k;
	int match;
	int rv = 1;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);

		for (j = 0; j < pgroup->num_elements; j++) {
			pb = VGROUP_ELEMENT(pgroup, j);

			for (k = 0; k < pgroup->num_elements; k++) {
				pc = VGROUP_ELEMENT(pgroup, k);

				pgroup->pspec->pvop(ab, pa, pb, pgroup->pvaux);
				pgroup->pspec->pvop(bc, pb, pc, pgroup->pvaux);
				pgroup->pspec->pvop(ab_c, ab, pc, pgroup->pvaux);
				pgroup->pspec->pvop(a_bc, pa, bc, pgroup->pvaux);

				match = pgroup->pspec->pvequal(a_bc, ab_c,
					pgroup->pvaux);

				if (verbosity >= 4) {

					printf("    (");
					pgroup->pspec->pvprint(pa, pgroup->pvaux);
					printf(" ");
					pgroup->pspec->pvprint(pb, pgroup->pvaux);
					printf(") ");
					pgroup->pspec->pvprint(pc, pgroup->pvaux);
					printf(" = ");

					pgroup->pspec->pvprint(ab_c, pgroup->pvaux);
					printf(" %s ", match ? "==" : "!=");
					pgroup->pspec->pvprint(a_bc, pgroup->pvaux);

					printf(" = ");
					pgroup->pspec->pvprint(pa, pgroup->pvaux);
					printf(" (");
					pgroup->pspec->pvprint(pb, pgroup->pvaux);
					printf(" ");
					pgroup->pspec->pvprint(pc, pgroup->pvaux);
					printf(") ");
					printf(".\n");
				}

				if (!match) {
					rv = 0;

					if (verbosity >= 2) {
						printf("Assoc: ");
						pgroup->pspec->pvprint(pa,
							pgroup->pvaux);
						printf(", ");
						pgroup->pspec->pvprint(pb,
							pgroup->pvaux);
						printf(", ");
						pgroup->pspec->pvprint(pc,
							pgroup->pvaux);
						printf(".\n");
					}

				}
			}
		}
	}
	return rv;
}

// ----------------------------------------------------------------
int has_unique_identity(
	group_t * pgroup,
	int verbosity)
{
	void * pg;
	void * pe;
	unsigned char ge[VOP_MAX_ELEMENT_SIZE];
	unsigned char eg[VOP_MAX_ELEMENT_SIZE];
	int i, j;
	int num_identities = 0;
	int this_is_an_identity;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pgroup->num_elements; i++) {
		pe = VGROUP_ELEMENT(pgroup, i);

		this_is_an_identity = 1;

		if (verbosity >= 2) {
			printf("Identity:  testing ");
			pgroup->pspec->pvprint(pe, pgroup->pvaux);
			printf("\n");
		}

		for (j = 0; j < pgroup->num_elements; j++) {
			pg = VGROUP_ELEMENT(pgroup, j);

			pgroup->pspec->pvop(eg, pe, pg, pgroup->pvaux);
			pgroup->pspec->pvop(ge, pg, pe, pgroup->pvaux);

			if (verbosity >= 2) {
				printf("  g: ");
				pgroup->pspec->pvprint(pg, pgroup->pvaux);
				printf("   eg: ");
				pgroup->pspec->pvprint(eg, pgroup->pvaux);
				printf("   ge: ");
				pgroup->pspec->pvprint(ge, pgroup->pvaux);
				printf("\n");
			}

			if ((!pgroup->pspec->pvequal(pg, eg, pgroup->pvaux))
			|| (!pgroup->pspec->pvequal(pg, ge, pgroup->pvaux)))
				this_is_an_identity = 0;
		}

		if (this_is_an_identity) {
			num_identities++;
		}
		if (verbosity >= 2) {
			if (this_is_an_identity)
				printf("Is an identity.\n");
			else
				printf("Is not an identity.\n");
			printf("\n");
		}
	}
	if (verbosity >= 2) {
		printf("Number of identities: %d.\n", num_identities);
	}

	return (num_identities == 1);
}

// ----------------------------------------------------------------
int has_unique_inverses(
	group_t * pgroup,
	int verbosity)
{
	void * pg;
	void * pi;
	unsigned char e [VOP_MAX_ELEMENT_SIZE];
	unsigned char gi[VOP_MAX_ELEMENT_SIZE];
	unsigned char ig[VOP_MAX_ELEMENT_SIZE];
	int i, j;
	int rv = 1;
	int this_num_invs;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	if (!find_identity(pgroup, e)) {
		fprintf(stderr, "has_unique_inverses: no identity found!\n");
		return 0;
	}

	for (i = 0; i < pgroup->num_elements; i++) {
		pg = VGROUP_ELEMENT(pgroup, i);

		if (verbosity >= 2) {
			printf("Inverses:  Testing ");
			pgroup->pspec->pvprint(pg, pgroup->pvaux);
			printf("\n");
		}

		this_num_invs = 0;
		for (j = 0; j < pgroup->num_elements; j++) {
			pi = VGROUP_ELEMENT(pgroup, j);

			pgroup->pspec->pvop(gi, pg, pi, pgroup->pvaux);
			pgroup->pspec->pvop(ig, pi, pg, pgroup->pvaux);

			if (verbosity >= 2) {
				printf("  g: ");
				pgroup->pspec->pvprint(pg, pgroup->pvaux);
				printf("   i: ");
				pgroup->pspec->pvprint(pi, pgroup->pvaux);
				printf("   gi: ");
				pgroup->pspec->pvprint(gi, pgroup->pvaux);
				printf("   ig: ");
				pgroup->pspec->pvprint(ig, pgroup->pvaux);
				printf("\n");
			}

			if ((pgroup->pspec->pvequal(gi, e, pgroup->pvaux))
			&& (pgroup->pspec->pvequal(ig, e, pgroup->pvaux)))
				this_num_invs++;
		}

		if (verbosity >= 2) {
			printf("Number of inverses: %d\n", this_num_invs);
		}
		if (this_num_invs != 1)
			rv = 0;
	}

	return rv;
}

// ----------------------------------------------------------------
int is_abelian(
	group_t * pgroup,
	int verbosity)
{
	void * pa;
	void * pb;
	unsigned char ab[VOP_MAX_ELEMENT_SIZE];
	unsigned char ba[VOP_MAX_ELEMENT_SIZE];
	int i, j;
	int match;
	int rv = 1;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);

		for (j = 0; j < pgroup->num_elements; j++) {
			pb = VGROUP_ELEMENT(pgroup, j);

			pgroup->pspec->pvop(ab, pa, pb, pgroup->pvaux);
			pgroup->pspec->pvop(ba, pb, pa, pgroup->pvaux);

			match = pgroup->pspec->pvequal(ab, ba, pgroup->pvaux);
			if (!match)
				rv = 0;
			if (verbosity) {

				pgroup->pspec->pvprint(pa, pgroup->pvaux);
				printf(" * ");
				pgroup->pspec->pvprint(pb, pgroup->pvaux);
				printf(" = ");
				pgroup->pspec->pvprint(ab, pgroup->pvaux);

				if (match)
					printf(" == ");
				else
					printf(" != ");

				pgroup->pspec->pvprint(pb, pgroup->pvaux);
				printf(" * ");
				pgroup->pspec->pvprint(pa, pgroup->pvaux);
				printf(" = ");
				pgroup->pspec->pvprint(ba, pgroup->pvaux);
				printf("\n");
			}
		}
		if (verbosity)
			printf("\n");
	}
	return rv;
}

// ----------------------------------------------------------------
// Find an element equal to its own square:
//   a^2 = a
//   aa  = a
//   aa  = ae
//   a^{-1}aa  = a^{-1}ae
//   a   = e
// Such an element will be the group identity.
//
// This function returns only the first such found, so if in doubt if the
// putative group is really a group, please call has_unique_identity() first.

int find_identity(
	group_t * pgroup,
	void    * pe)
{
	void * pa;
	unsigned char aa[VOP_MAX_ELEMENT_SIZE];
	int i;

	// xxx to lib func
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);
		pgroup->pspec->pvop(aa, pa, pa, pgroup->pvaux);
		if (pgroup->pspec->pvequal(pa, aa, pgroup->pvaux)) {
			memcpy(pe, pa, pgroup->pspec->element_size);
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
// Returns 1 if found, 0 if not.
int group_find_index(
	group_t * pgroup,
	void    * sought,
	int     * pindex)
{
	int i;
	void * other;

	for (i = 0; i < pgroup->num_elements; i++) {
		other = VGROUP_ELEMENT(pgroup, i);
		if (pgroup->pspec->pvequal(sought, other, pgroup->pvaux)) {
			*pindex = i;
			return 1;
		}
	}
	return 0;
}


// ----------------------------------------------------------------
int is_subset(
	group_t * pH,
	group_t * pG,
	int       verbosity)
{
	int i;
	void * ph;
	int index;
	int rv = 1;

	for (i = 0; i < pH->num_elements; i++) {
		ph = VGROUP_ELEMENT(pH, i);

		if (!group_find_index(pG, ph, &index)) {
			rv = 0;
			break;
		}
	}
	return rv;
}

// ----------------------------------------------------------------
int is_subgroup(
	group_t * pH,
	group_t * pG,
	int       verbosity)
{
	if (!is_subset(pH, pG, verbosity)) {
		if (verbosity > 1) {
			printf("is_subgroup:  H is not a subset of G.\n");
		}
		return 0;
	}

	// Don't need to check associativity for subgroups.
	if (!has_unique_identity(pH, 0))
		return 0;

	if (!has_unique_inverses(pH, 0))
		return 0;

	return 1;
}

// ----------------------------------------------------------------
int is_normal_subgroup(
	group_t * pH,
	group_t * pG,
	int       verbosity)
{
	int space_needed;
	void * pgH, * pHg;
	int i, j;
	void * pg, * ph, * pgh, * phg;
	int rv = 1;
	int equal;

	if (!is_subgroup(pH, pG, verbosity)) {
		if (verbosity > 0) {
			printf("is_normal_subgroup:  H is not a "
				"subgroup of G.\n");
		}
		return 0;
	}

	// Allocate memory for a left coset and a right coset.
	space_needed = pH->num_elements * pH->pspec->element_size;
	pgH = malloc(space_needed);
	pHg = malloc(space_needed);
	if (!pgH || !pHg) {
		fprintf(stderr, "malloc(%d) failed.\n", space_needed);
		exit(1);
	}

	if (verbosity)
		printf("Checking subgroup normalcy.\n");

	for (i = 0; i < pG->num_elements; i++) {
		pg = VGROUP_ELEMENT(pG, i);

		// For each element g in G, form the left and right cosets
		// gH and Hg.
		for (j = 0; j < pH->num_elements; j++) {
			ph = VGROUP_ELEMENT(pH, j);

			pgh = VARRAY_ELEMENT(pgH, j, pG->pspec->element_size);
			phg = VARRAY_ELEMENT(pHg, j, pG->pspec->element_size);

			pH->pspec->pvop(pgh, pg, ph, pH->pvaux);
			pH->pspec->pvop(phg, ph, pg, pH->pvaux);
		}

		// Dump.
		if (verbosity) {
			printf("  Testing g = ");
			pG->pspec->pvprint(pg, pG->pvaux);
			printf("\n");

			printf("    gH =");
			for (j = 0; j < pH->num_elements; j++) {
				pgh = VARRAY_ELEMENT(pgH, j,
					pG->pspec->element_size);
				printf(" ");
				pG->pspec->pvprint(pgh, pG->pvaux);
			}
			printf("\n");

			printf("    Hg =");
			for (j = 0; j < pH->num_elements; j++) {
				phg = VARRAY_ELEMENT(pHg, j,
					pG->pspec->element_size);
				printf(" ");
				pG->pspec->pvprint(phg, pG->pvaux);
			}
			printf("\n");

			printf("\n");
		}

		// Sort gH and Hg to facilitate equality testing.
		sort_varray(pgH, pH->num_elements, pH->pspec->element_size,
			pH->pspec->pvless, pH->pvaux);
		sort_varray(pHg, pH->num_elements, pH->pspec->element_size,
			pH->pspec->pvless, pH->pvaux);

		if (verbosity >= 2) {
			printf("    sorted gH =");
			for (j = 0; j < pH->num_elements; j++) {
				pgh = VARRAY_ELEMENT(pgH, j,
					pG->pspec->element_size);
				printf(" ");
				pG->pspec->pvprint(pgh, pG->pvaux);
			}
			printf("\n");

			printf("    sorted Hg =");
			for (j = 0; j < pH->num_elements; j++) {
				phg = VARRAY_ELEMENT(pHg, j,
					pG->pspec->element_size);
				printf(" ");
				pG->pspec->pvprint(phg, pG->pvaux);
			}
			printf("\n");

			printf("\n");
		}

		// See if gH == Hg.
		equal = 1;
		for (j = 0; j < pH->num_elements; j++) {
			pgh = VARRAY_ELEMENT(pgH, j, pG->pspec->element_size);
			phg = VARRAY_ELEMENT(pHg, j, pG->pspec->element_size);
			if (!pH->pspec->pvequal(pgh, phg, pH->pvaux)) {
				equal = 0;
				break;
			}
		}
		if (!equal)
			rv = 0;
		if (verbosity) {
			if (equal)
				printf("  gH == Hg\n");
			else
				printf("  gH != Hg\n");
		}
	}
	if (verbosity)
		printf("\n");

	free(pgH);
	free(pHg);

	return rv;
}

// ----------------------------------------------------------------
int close_group(
	group_t * pgroup,
	int max_group_bytes,
	int verbosity)
{
	void * pa;
	void * pb;
	unsigned char ab[VOP_MAX_ELEMENT_SIZE];
	unsigned char ba[VOP_MAX_ELEMENT_SIZE];
	int i, j;
	int index;
	int m = max_group_bytes;
	int round = 0;
	int prev_size;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	do {
		round++;
		if (verbosity >= 1) {
			printf("\n");
			printf("close_group round %d\n", round);
		}
		prev_size = pgroup->num_elements;
		for (i = 0; i < prev_size; i++) {
			pa = VGROUP_ELEMENT(pgroup, i);

			for (j = 0; j < prev_size; j++) {
				pb = VGROUP_ELEMENT(pgroup, j);

				pgroup->pspec->pvop(ab, pa, pb, pgroup->pvaux);
				pgroup->pspec->pvop(ba, pb, pa, pgroup->pvaux);

				if (!group_find_index(pgroup, ab, &index)) {
					if (!group_append_element(pgroup, ab, m))
						return 0;
					if (verbosity >= 1) {
						printf("  ");
						pgroup->pspec->pvprint(pa,
							pgroup->pvaux);
						printf(" * ");
						pgroup->pspec->pvprint(pb,
							pgroup->pvaux);
						printf(" = ");
						pgroup->pspec->pvprint(ab,
							pgroup->pvaux);
						printf("\n");
					}
				}
				if (!group_find_index(pgroup, ba, &index)) {
					if (!group_append_element(pgroup, ba, m))
						return 0;
					if (verbosity >= 1) {
						printf("  ");
						pgroup->pspec->pvprint(pa,
							pgroup->pvaux);
						printf(" * ");
						pgroup->pspec->pvprint(pb,
							pgroup->pvaux);
						printf(" = ");
						pgroup->pspec->pvprint(ba,
							pgroup->pvaux);
						printf("\n");
					}
				}
			}
		}
	} while (pgroup->num_elements > prev_size);

	if (verbosity >= 1) {
		printf("\n");
		printf("close_group round end\n");
	}
	//if (!is_group(pgroup, verbosity)) {
	//	if (verbosity >= 1) {
	//		printf("close_group:  not a group\n");
	//	}
	//	return 0;
	//}
	return 1;
}
