#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "group.h"
#include "group_check.h"
#include "group_commutator.h"

// ----------------------------------------------------------------
int make_commutator_subgroup(
	group_t * pgroup,
	group_t * pcommutator_subgroup,
	void    * sgr_elements_buffer,
	int       max_group_bytes,
	int       verbosity)
{
	int i, j;
	void * px, * py;
	unsigned char xi       [VOP_MAX_ELEMENT_SIZE];
	unsigned char yi       [VOP_MAX_ELEMENT_SIZE];
	unsigned char x_y      [VOP_MAX_ELEMENT_SIZE];
	unsigned char x_y_xi   [VOP_MAX_ELEMENT_SIZE];
	unsigned char x_y_xi_yi[VOP_MAX_ELEMENT_SIZE];

	pcommutator_subgroup->pvelements = sgr_elements_buffer;
	pcommutator_subgroup->num_elements = 0; // increment below

	pcommutator_subgroup->pvaux = pgroup->pvaux;
	pcommutator_subgroup->pspec = pgroup->pspec;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pgroup->num_elements; i++) {
		px = VGROUP_ELEMENT(pgroup, i);
		for (j = 0; j < pgroup->num_elements; j++) {
			py = VGROUP_ELEMENT(pgroup, j);

			pgroup->pspec->pvop(x_y, px, py, pgroup->pvaux);
			group_element_inverse(pgroup, px, xi);
			group_element_inverse(pgroup, py, yi);
			pgroup->pspec->pvop(x_y_xi, x_y, xi, pgroup->pvaux);
			pgroup->pspec->pvop(x_y_xi_yi, x_y_xi, yi,
				pgroup->pvaux);

			if (!group_maybe_add_element(pcommutator_subgroup,
					x_y_xi_yi, max_group_bytes))
				return 0;
		}
	}

	if (!close_group(pcommutator_subgroup, max_group_bytes, verbosity))
		return 0;
	return 1;
}
