#include <string.h>
#include <ctype.h>
#include "group_file.h"

#define LINE_MAX 256

// ----------------------------------------------------------------
int fill_group_from_file(
	group_t * pgroup,           // input
	group_spec_t * pspec,       // input
	void    * pvaux,            // input
	char    * elements_file,    // input
	void    * elements_buffer,  // output
	int       max_group_bytes)  // output
{
	FILE * fp;
	int rv;

	if (strcmp(elements_file, "-") == 0) {
		fp = stdin;
	}
	else {
		fp = fopen(elements_file, "r");
		if (fp == NULL) {
			fprintf(stderr,
				"Couldn't open input file \"%s\".\n",
				elements_file);
			perror("Reason");
			return 0;
		}
	}

	rv = fill_group_from_stream(pgroup, pspec, pvaux, fp,
		elements_buffer, max_group_bytes);

	if (fp != stdin)
		fclose(fp);
	return rv;
}

// ----------------------------------------------------------------
int fill_group_from_stream(
	group_t * pgroup,           // input
	group_spec_t * pspec,       // input
	void    * pvaux,            // input
	FILE    * fp,               // input
	void    * elements_buffer,  // output
	int       max_group_bytes)  // output
{
	int max_num_elements;
	char line[LINE_MAX];
	char * pline;
	char * pcmt;
	char * pend;
	int line_number = 0;
	void * pg;
	int len;

	pgroup->pvelements = elements_buffer;
	pgroup->pvaux = pvaux;
	pgroup->pspec = pspec;
	pgroup->num_elements = 0; // increment below

	max_num_elements = max_group_bytes / pspec->element_size;

	while (fgets(line, sizeof(line), fp)) {
		line_number++;

		// Strip trailing carriage return, if any.
		len = strlen(line);
		if (line[len-1] == '\n')
			line[len-1] = 0;

		pline = line;

		// Strip comment characters.
		for (pcmt = pline; *pcmt; pcmt++) {
			if (*pcmt == '#') {
				*pcmt = 0;
				break;
			}
		}

		// Strip leading whitespace.
		while ((*pline == ' ') || (*pline == '\t'))
			pline++;

		// Strip trailing whitespace.
		len = strlen(line);
		if (len > 0) {
			pend = &line[len-1];
			while ((*pend == ' ') || (*pend == '\t')) {
				if (pend <= pline)
					break;
				*pend = 0;
				pend--;
			}
		}

		// Skip whitespace (or comment-only) lines
		if (strcmp(pline, "") == 0)
			continue;

		//tokenize line count strings
		//for each token {
			pg = VGROUP_ELEMENT(pgroup, pgroup->num_elements);
			if (pspec->pvscan(pg, pline, pgroup->pvaux,
				pspec->element_size))
			{
				if (pgroup->num_elements >= max_num_elements) {
					fprintf(stderr, "fill_group_from_file: "
					"too many elements; max %d.\n",
					max_num_elements);
					return 0;
				}
				pgroup->num_elements++;
			}
			else {
				fprintf(stderr, "fill_group_from_file: "
				"scan failure at line %d.\n", line_number);
				return 0;
			}
		//}
	}

	return 1;
}
