#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "group.h"
#include "group_check.h"
#include "group_orders.h"

// ----------------------------------------------------------------
void show_orders(
	group_t * pgroup)
{
	int * orders;
	int i;
	void * pa;

	orders = (int *) malloc(pgroup->num_elements * sizeof(int));
	if (orders == 0) {
		fprintf(stderr, "malloc(%d) failed.\n",
			pgroup->num_elements * sizeof(int));
		exit(1);
	}

	group_find_orders(pgroup, orders, 0);

	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);
		pgroup->pspec->pvprint(pa, pgroup->pvaux);
		printf("  %3d\n", orders[i]);
	}

	free(orders);
}

// ----------------------------------------------------------------
void group_find_orders(
	group_t * pgroup,
	int     * orders,
	int       verbosity)
{
	unsigned char e[VOP_MAX_ELEMENT_SIZE];
	void * pa;
	unsigned char apower[VOP_MAX_ELEMENT_SIZE];
	int i;

	// xxx to lib func.
	if (pgroup->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: group element size %d too big.\n",
				__FILE__, pgroup->pspec->element_size);
		exit(1);
	}

	// First, locate the identity.
	if (!find_identity(pgroup, e)) {
		fprintf(stderr,
			"group_find_orders:  Couldn't find identity "
			"element in group.\n");
		exit(1);
	}

// xxx use group_element_order() instead.
	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);
		memcpy(apower, pa, pgroup->pspec->element_size);
		for (orders[i] = 1; ; orders[i]++) {
			if (orders[i] > pgroup->num_elements) {
				printf("group_find_orders: "
					"non-invertible element ");
				pgroup->pspec->pvprint(pa, pgroup->pvaux);
				printf(".\n");
				exit(1);
			}
			if (pgroup->pspec->pvequal(apower, e, pgroup->pvaux))
				break;
			pgroup->pspec->pvop(apower, apower, pa, pgroup->pvaux);
		}
	}
}

// ----------------------------------------------------------------
int find_maximal_element_order(
	group_t * pgroup)
{
	int i;
	void * pa;
	int max = 1;
	int current;

	for (i = 0; i < pgroup->num_elements; i++) {
		pa = VGROUP_ELEMENT(pgroup, i);
		current = group_element_order(pgroup, pa);
		if (current > max)
			max = current;

	}
	return max;
}
