#ifndef VRING_H
#define VRING_H

#include "vops.h"

#define VRING_ELEMENT(pr,i) \
	VARRAY_ELEMENT((pr)->pvelements, (i), (pr)->pspec->element_size)

// ith coset, jth element in coset
#define VRCOSETS_ELEMENT(pq,i,j) \
	VARRAY_ELEMENT((pq)->pvelements, \
		(i) * (pq)->elements_per_coset + (j), \
		(pq)->pspec->element_size)

// ----------------------------------------------------------------
// Returns 0 if the ring would be too big.  Caller must handle this.
typedef int  fill_ring_t(
	void * pvelements,
	void * pvaux,
	int    max_ring_bytes,
	int  * pnum_elements);

// ----------------------------------------------------------------
typedef struct _ring_spec_t {
	int                element_size;
	int                naux;

	vop_t            * pvadd;
	vop_t            * pvmul;

	vequal_t         * pvequal;
	vless_t          * pvless;

	vprint_t         * pvprint;
	vauxprint_t      * pvauxprint;

	vscan_t          * pvscan;
	vauxscan_t       * pvauxscan;

	fill_ring_t      * pfill_ring;
	char             * desc;
} ring_spec_t;

typedef struct _ring_t {
	void             * pvelements;
	int                num_elements;
	void             * pvaux;
	ring_spec_t      * pspec;
} ring_t;

typedef struct _ring_cosets_t {
	void             * pvelements;
	int                num_cosets;
	int                elements_per_coset;
	void             * pvaux;
	ring_spec_t      * pspec;
} ring_cosets_t;

// ----------------------------------------------------------------
int find_ring_spec(
	char * desc,
	ring_spec_t ** ppspec);

void list_ring_specs(
	void);

int make_ring(
	ring_t * pring,             // input
	ring_spec_t * pring_spec,   // input
	void    * pvaux,            // input
	char    * elements_file,    // input -- null for default elts.
	void    * elements_buffer,  // output
	int       max_ring_bytes);  // output

void describe_ring(
	ring_t * pring);

void print_ring_elements(
	ring_t * pring);

void ring_print_cosets(
	ring_cosets_t * pcosets);

// ----------------------------------------------------------------
// Adds the element unless already present.  This entails a linear search
// through the ring elements on every function call.
int ring_maybe_add_element(
	ring_t * pring,
	void    * psrc,
	int       max_ring_bytes);

// Appends the new element unconditionally.
int ring_append_element(
	ring_t * pring,
	void    * psrc,
	int       max_ring_bytes);

// ----------------------------------------------------------------
void ring_element_power(
	ring_t       * pring,
	void         * pvin,
	int            power,
	void         * pvout);

void ring_element_inverse(
	ring_t * pring,
	void    * pvin,
	void    * pvout);

// ----------------------------------------------------------------
void sort_varray(
	void * pvelements,
	int    num_elements,
	int    element_size,
	vless_t * pvless,
	void * pvaux);

#endif // VRING_H
