#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "ring_check.h"
#include "ring_cap.h"

void make_cap(
	ring_t * pH,
	ring_t * pK,
	ring_t * pH_cap_K,
	void    * H_cap_K_elts_buffer,
	int       max_ring_bytes,
	int       verbosity)
{
	int hi, ki;
	void * ph;
	void * pk;
	void * pcap;
	int    index;
	int    size;

	pH_cap_K->pvelements = H_cap_K_elts_buffer;
	pH_cap_K->num_elements = 0; // increment below

	// xxx needs miscibility check.
	pH_cap_K->pvaux = pH->pvaux;
	pH_cap_K->pspec = pH->pspec;

	// xxx to lib func.
	if (pH->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pH->pspec->element_size);
		exit(1);
	}

	for (hi = 0; hi < pH->num_elements; hi++) {
		ph = VRING_ELEMENT(pH, hi);

		if (!ring_find_index(pH, ph, &index))
			// Element not in H.
			continue;
		if (!ring_find_index(pK, ph, &index))
			// Element not in K.
			continue;
		if (ring_find_index(pH_cap_K, ph, &index))
			// Element already in H cap K.
			continue;

		size = (pH_cap_K->num_elements + 1)
			* pH_cap_K->pspec->element_size;
		if (size > max_ring_bytes) {
			fprintf(stderr,
				"make_cap: out of space.\n");
			exit(1);
		}
		pcap = VRING_ELEMENT(pH_cap_K, pH_cap_K->num_elements);
		memcpy(pcap, ph, pH_cap_K->pspec->element_size);
		pH_cap_K->num_elements++;
	}

	for (ki = 0; ki < pK->num_elements; ki++) {
		pk = VRING_ELEMENT(pK, ki);

		if (!ring_find_index(pH, pk, &index))
			// Element not in H.
			continue;
		if (!ring_find_index(pK, pk, &index))
			// Element not in K.
			continue;
		if (ring_find_index(pH_cap_K, pk, &index))
			// Element already in H cap K.
			continue;

		size = (pH_cap_K->num_elements + 1)
			* pH_cap_K->pspec->element_size;
		if (size > max_ring_bytes) {
			fprintf(stderr,
				"make_cap: out of space.\n");
			exit(1);
		}
		pcap = VRING_ELEMENT(pH_cap_K, pH_cap_K->num_elements);
		memcpy(pcap, pk, pH_cap_K->pspec->element_size);
		pH_cap_K->num_elements++;
	}

}
