#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "pmt.h"
#include "ring_check.h"
#include "ring_cayley_sn.h"

void make_cayley_Sn_subring(
	ring_t * pring,
	ring_t * psn_subring,
	void    * sgr_elements_buffer,
	int       max_sn_sgr_bytes)
{
	pmt_t * pSns;
	int i, j, k;
	void * px;
	void * pg;
	unsigned char xg[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}


	psn_subring->pvelements = sgr_elements_buffer;
	psn_subring->num_elements = pring->num_elements;

	psn_subring->pvaux = &psn_subring->num_elements;
	psn_subring->pspec = &sn_ring_spec;

	pSns = (pmt_t *)sgr_elements_buffer;

	// for each x [i] in G
	//	for each g [j] in G
	//		find index [k] of xg in G
	//		cayley_Sns[i].images[j] = k

	for (i = 0; i < pring->num_elements; i++) {
		px = VRING_ELEMENT(pring, i);
		for (j = 0; j < pring->num_elements; j++) {
			pg = VRING_ELEMENT(pring, j);
			pring->pspec->pvop(xg, px, pg, pring->pvaux);

			if (!ring_find_index(pring, xg, &k)) {
				fprintf(stderr,
				"make_cayley_Sn_subring: ring not closed.\n");
				exit(1);
			}
			pSns[i].images[j] = k;
		}
	}
}
