#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "pmt.h"
#include "ring_check.h"
#include "ring_center.h"

void make_center(
	ring_t * pring,
	ring_t * pcenter,
	void    * center_elements_buffer,
	int       max_center_bytes)
{
	int i, j;
	void * px;
	void * pg;
	void * pc;
	unsigned char xg[VOP_MAX_ELEMENT_SIZE];
	unsigned char gx[VOP_MAX_ELEMENT_SIZE];
	int ok;

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}


	pcenter->pvelements = center_elements_buffer;
	pcenter->num_elements = 0; // Increment below
	pcenter->pvaux = pring->pvaux;
	pcenter->pspec = pring->pspec;

	// for each x [i] in G
	//	for each g [j] in G
	//		if xg == gx
	//			add x to the center

	for (i = 0; i < pring->num_elements; i++) {
		px = VRING_ELEMENT(pring, i);
		ok = 1;
		for (j = 0; j < pring->num_elements; j++) {
			pg = VRING_ELEMENT(pring, j);
			pring->pspec->pvop(xg, px, pg, pring->pvaux);
			pring->pspec->pvop(gx, pg, px, pring->pvaux);

			if (!pring->pspec->pvequal(xg, gx, pring->pvaux)) {
				ok = 0;
				break;
			}
		}
		if (ok) {
			pc = VRING_ELEMENT(pcenter, pcenter->num_elements);
			memcpy(pc, px, pring->pspec->element_size);
			pcenter->num_elements++;
		}
	}
}
