#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "ring_check.h"
#include "ring_commutator.h"

// ----------------------------------------------------------------
int make_commutator_subring(
	ring_t * pring,
	ring_t * pcommutator_subring,
	void    * sgr_elements_buffer,
	int       max_ring_bytes,
	int       verbosity)
{
	int i, j;
	void * px, * py;
	unsigned char xi       [VOP_MAX_ELEMENT_SIZE];
	unsigned char yi       [VOP_MAX_ELEMENT_SIZE];
	unsigned char x_y      [VOP_MAX_ELEMENT_SIZE];
	unsigned char x_y_xi   [VOP_MAX_ELEMENT_SIZE];
	unsigned char x_y_xi_yi[VOP_MAX_ELEMENT_SIZE];

	pcommutator_subring->pvelements = sgr_elements_buffer;
	pcommutator_subring->num_elements = 0; // increment below

	pcommutator_subring->pvaux = pring->pvaux;
	pcommutator_subring->pspec = pring->pspec;

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}

	for (i = 0; i < pring->num_elements; i++) {
		px = VRING_ELEMENT(pring, i);
		for (j = 0; j < pring->num_elements; j++) {
			py = VRING_ELEMENT(pring, j);

			pring->pspec->pvop(x_y, px, py, pring->pvaux);
			ring_element_inverse(pring, px, xi);
			ring_element_inverse(pring, py, yi);
			pring->pspec->pvop(x_y_xi, x_y, xi, pring->pvaux);
			pring->pspec->pvop(x_y_xi_yi, x_y_xi, yi,
				pring->pvaux);

			if (!ring_maybe_add_element(pcommutator_subring,
					x_y_xi_yi, max_ring_bytes))
				return 0;
		}
	}

	if (!close_ring(pcommutator_subring, max_ring_bytes, verbosity))
		return 0;
	return 1;
}
