#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "pmt.h"
#include "ring_check.h"
#include "ring_core.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// intersection {g in G} g^-1 H g

void make_core(
	ring_t * pG,
	ring_t * pH,
	ring_t * pcore,
	void    * core_elements_buffer,
	int       max_core_bytes,
	int       verbosity)
{
	int i, j;
	void * pg;
	void * ph;
	int  * counts;
	int    conj_index;
	unsigned char ginv    [VOP_MAX_ELEMENT_SIZE];
	unsigned char h_g     [VOP_MAX_ELEMENT_SIZE];
	unsigned char ginv_h_g[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pG->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pG->pspec->element_size);
		exit(1);
	}

	pcore->pvelements = core_elements_buffer;
	pcore->num_elements = 0; // Increment below
	pcore->pvaux = pG->pvaux;
	pcore->pspec = pG->pspec;

	counts = malloc_check(sizeof(int) * pG->num_elements);
	for (i = 0; i < pG->num_elements; i++)
		counts[i] = 0;

	// for each g [i] in G
	//	for each h [j] in H
	//		make g^-1 h g
	//			mark g^-1 h g in G
	// for each g in G
	//   if # marks == |G|
	//     add to core

	for (i = 0; i < pG->num_elements; i++) {
		pg = VRING_ELEMENT(pG, i);
		for (j = 0; j < pH->num_elements; j++) {
			ph = VRING_ELEMENT(pH, j);
			ring_element_inverse(pG, pg, ginv);
			pG->pspec->pvop(h_g, ph, pg, pG->pvaux);
			pG->pspec->pvop(ginv_h_g, ginv, h_g, pG->pvaux);
			if (!ring_find_index(pG, ginv_h_g, &conj_index)) {
				fprintf(stderr,
					"make_core:  ring not closed.\n");
				exit(1);
			}
			counts[conj_index]++;
		}
	}

	for (i = 0; i < pG->num_elements; i++) {
		if (counts[i] != pG->num_elements)
			continue;
		pg = VRING_ELEMENT(pG, i);

		if (!ring_append_element(pcore, pg, max_core_bytes)) {
			fprintf(stderr, "make_core:  core full.\n");
			exit(1);
		}
	}

	free(counts);
}
