#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "ring_check.h"
#include "ring_cosets.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
void ring_print_cosets(
	ring_cosets_t * pcosets)
{
	int i, j, k;
	void * pa;

	for (k = 0, i = 0; i < pcosets->num_cosets; i++) {
		printf("  ");
		for (j = 0; j < pcosets->elements_per_coset; j++, k++) {
			pa = VRING_ELEMENT(pcosets, k);
			if (j > 0)
				printf(" ");
			pcosets->pspec->pvprint(pa, pcosets->pvaux);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
// Note:  As of this writing I am not yet supporting fully recursive quotient
// rings, nor does most of the code in this directory handle quotient rings
// at this time.  This function is ad hoc, intended solely for computing
// ascending central series.

int make_cosets(
	ring_t  * pG,
	ring_t  * pH,
	int        direction,
	ring_cosets_t * pcosets,
	void     * quot_elements_buffer,
	int        max_cosets_bytes,
	int        verbosity)
{
	int i, j;
	int oG, oH, iGH;
	int max_coset_bytes;
	char * marks;
	int num_marks = 0;
	ring_t coset;
	char * pc_coset_buf;
	int cosetno = 0;
	int gi_h0_loc, gi_hj_loc;
	void * pgi, * ph0, * phj;
	unsigned char gi_hj[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pG->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pG->pspec->element_size);
		exit(1);
	}

	oG = pG->num_elements;
	oH = pH->num_elements;
	if ((oG % oH) != 0) {
		fprintf(stderr,
			"make_cosets:  Incompatible sizes %d, %d.\n",
				oG, oH);
		return 0;
	}
	iGH = oG / oH;
	if (iGH < 1) {
		fprintf(stderr,
			"make_cosets:  Empty quotient.\n");
		return 0;
	}

	marks = malloc_check(sizeof(char) * oG);
	for (i = 0; i < oG; i++)
		marks[i] = 0;

	max_coset_bytes = max_cosets_bytes / iGH;
	pcosets->pvelements = quot_elements_buffer;
	pcosets->pvaux = pG->pvaux;
	pcosets->pspec = pG->pspec;

	pcosets->elements_per_coset = oH;
	pcosets->num_cosets = iGH;

	pc_coset_buf = (unsigned char *)pcosets->pvelements;
	coset.pvelements = pc_coset_buf;
	coset.num_elements = 0;
	coset.pvaux = pG->pvaux;
	coset.pspec = pG->pspec;

	ph0 = VRING_ELEMENT(pH, 0);

	for (i = 0; i < oG; i++) {
		pgi = VRING_ELEMENT(pG, i);
		if (direction == LEFT_COSETS)
			pG->pspec->pvop(gi_hj, pgi, ph0, pG->pvaux);
		else
			pG->pspec->pvop(gi_hj, ph0, pgi, pG->pvaux);
		if (!ring_find_index(pG, gi_hj, &gi_h0_loc)) {
			fprintf(stderr,
				"make_cosets: ring not closed.\n");
			return 0;
		}
		if (marks[gi_h0_loc])
			continue;

		if (cosetno >= iGH) {
			fprintf(stderr,
				"make_cosets: too many cosets (%d)\n",
					cosetno + 1);
			return 0;
		}

		coset.num_elements = 0;
		for (j = 0; j < oH; j++) {
			phj = VRING_ELEMENT(pH, j);
			if (direction == LEFT_COSETS)
				pG->pspec->pvop(gi_hj, pgi, phj, pG->pvaux);
			else
				pG->pspec->pvop(gi_hj, phj, pgi, pG->pvaux);

			if (!ring_append_element(&coset, gi_hj,
				max_coset_bytes))
			{
				fprintf(stderr,
				"make_cosets: coset %d overflow.\n",
					cosetno);
				return 0;
			}

			if (!ring_find_index(pG, gi_hj, &gi_hj_loc)) {
				fprintf(stderr,
				"make_cosets: ring not closed.\n");
				return 0;
			}
			marks[gi_hj_loc] = 1;
		}

		num_marks += oH;
		if (num_marks > oG) {
			fprintf(stderr,
				"make_cosets: too many elements (%d)\n",
					num_marks);
			return 0;
		}
		else if (num_marks == oG)
			goto break2;
		cosetno++;

		pc_coset_buf += oH * pH->pspec->element_size;
		coset.pvelements = pc_coset_buf;
	}
break2:

	free(marks);

	return 1;
}
