#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "ring_check.h"
#include "ring_dirprod.h"

void make_direct_product(
	ring_t * pH,
	ring_t * pK,
	ring_t * pHK,
	void    * HK_elts_buffer,
	int       max_ring_bytes,
	int       verbosity)
{
	int hi, ki;
	void * ph;
	void * pk;
	unsigned char hk[VOP_MAX_ELEMENT_SIZE];
	void * phk;
	int    index;
	int    size;

	pHK->pvelements = HK_elts_buffer;
	pHK->num_elements = 0; // increment below

	// xxx needs miscibility check.
	pHK->pvaux = pH->pvaux;
	pHK->pspec = pH->pspec;

	// xxx to lib func.
	if (pH->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pH->pspec->element_size);
		exit(1);
	}

	for (hi = 0; hi < pH->num_elements; hi++) {
		ph = VRING_ELEMENT(pH, hi);

		for (ki = 0; ki < pK->num_elements; ki++) {
			pk = VRING_ELEMENT(pK, ki);

			pH->pspec->pvop(hk, ph, pk, pH->pvaux);

			if (ring_find_index(pHK, hk, &index))
				// Element is already in there.
				continue;

			size = (pHK->num_elements + 1)
				* pH->pspec->element_size;
			if (size > max_ring_bytes) {
				fprintf(stderr,
					"make_direct_product: out of space.\n");
				exit(1);
			}
			phk = VRING_ELEMENT(pHK, pHK->num_elements);
			memcpy(phk, hk, pH->pspec->element_size);
			pHK->num_elements++;
		}
	}
}
