#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "pmt.h"
#include "ring_check.h"
#include "ring_inn.h"

// ----------------------------------------------------------------
void make_inn(
	ring_t * pring,
	ring_t * pinn,
	void    * inn_elements_buffer,
	int       max_inn_bytes,
	int       verbosity)
{
	pmt_t   autom;
	pmt_t * pautoms;
	int i, j, k;
	void * px;
	void * pg;
	unsigned char g_x     [VOP_MAX_ELEMENT_SIZE];
	unsigned char ginv    [VOP_MAX_ELEMENT_SIZE];
	unsigned char g_x_ginv[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}

	// xxx need to check ring size < 2^(num bits in pmt_elt_t).

	pinn->pvelements = inn_elements_buffer;
	pinn->num_elements = 0; // increment below
	pinn->pvaux = &pring->num_elements;
	pinn->pspec = &sn_ring_spec;

	pautoms = (pmt_t *)inn_elements_buffer;

	// for each g [i] in G
	//	for each x [j] in G
	//		find index [k] of g x g^-1 in G
	//		autom.images[j] = k
	//	add autom to inn if not already present
	if (verbosity)
		printf("make inn enter\n");

	for (i = 0; i < pring->num_elements; i++) {
		pg = VRING_ELEMENT(pring, i);
		ring_element_inverse(pring, pg, ginv);

		for (j = 0; j < pring->num_elements; j++) {
			autom.images[j] = 255;
		}
		for (j = 0; j < pring->num_elements; j++) {
			px = VRING_ELEMENT(pring, j);

			pring->pspec->pvop(g_x, pg, px, pring->pvaux);
			pring->pspec->pvop(g_x_ginv, g_x, ginv, pring->pvaux);

			if (!ring_find_index(pring, g_x_ginv, &k)) {
				fprintf(stderr,
				"make_cayley_Sn_subring: ring not closed.\n");
				exit(1);
			}
			autom.images[j] = k;
		}
		for (j = 0; j < pring->num_elements; j++) {
			if (autom.images[j] == 255) {
				fprintf(stderr,
					"make_inn: coding error: "
					"non-isomorphic conjugation.\n");
				exit(1);
			}
		}
		if (verbosity) {
			pmt_print(&autom, pring->num_elements);
			printf(" ");
			pring->pspec->pvprint(pg, pring->pvaux);
			printf("\n");
		}

		if (!ring_maybe_add_element(pinn, &autom, max_inn_bytes)) {
			fprintf(stderr, "make_inn:  inn buf overflow.\n");
			exit(1);
		}
	}
	if (verbosity)
		printf("make inn exit\n");
}
