#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "ring_check.h"
#include "ring_orders.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// xxx would prefer a partition-ring-into-conj-classes function,
// with a separate printer function.

void list_conj_cls(
	ring_t * pring,
	int       verbosity)
{
	int    i, j, k;
	void * pconjs;
	char * marks;
	int    nconj, cjgidx;
	int    ring_num_bytes;
	void * px, * pg, * pc;
	unsigned char xinv  [VOP_MAX_ELEMENT_SIZE];
	unsigned char xg    [VOP_MAX_ELEMENT_SIZE];
	unsigned char xgxinv[VOP_MAX_ELEMENT_SIZE];
	int    found;

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}

	ring_num_bytes = pring->num_elements * pring->pspec->element_size;
	pconjs = malloc_check(ring_num_bytes);
	marks  = malloc_check(pring->num_elements);
	for (i = 0; i < pring->num_elements; i++) {
		marks[i] = 0;
	}

	for (i = 0; i < pring->num_elements; i++) {
		if (marks[i])
			continue;
		pg = VRING_ELEMENT(pring, i);

		if (verbosity >= 2) {
			printf("[");
			pring->pspec->pvprint(pg, pring->pvaux);
			printf("] = ");
		}

		nconj = 0;
		for (j = 0; j < pring->num_elements; j++) {
			px = VRING_ELEMENT(pring, j);
			ring_element_inverse(pring, px, xinv);
			pring->pspec->pvop(xg, px, pg, pring->pvaux);
			pring->pspec->pvop(xgxinv, xg, xinv, pring->pvaux);

			// xxx really need an add-to-set lib function.
			found = 0;
			for (k = 0; k < nconj; k++) {
				pc = VARRAY_ELEMENT(pconjs, k,
					pring->pspec->element_size);
				if (pring->pspec->pvequal(pc, xgxinv,
					pring->pvaux))
				{
					found = 1;
					break;
				}
			}
			if (!found) {
				pc = VARRAY_ELEMENT(pconjs, nconj,
					pring->pspec->element_size);
				memcpy(pc, xgxinv,
					pring->pspec->element_size);

				if (!ring_find_index(pring, xgxinv,&cjgidx)) {
					fprintf(stderr,
					"list_conj_cls:  ring not closed.\n");
					exit(1);
				}
				marks[cjgidx] = 1;
				nconj++;
			}
		}

		for (j = 0; j < nconj; j++) {
			pc = VARRAY_ELEMENT(pconjs, j,
				pring->pspec->element_size);
			if (j > 0)
				printf(" ");
			pring->pspec->pvprint(pc, pring->pvaux);
		}

		if (verbosity >= 1) {
			printf(" (%d)", ring_element_order(pring, pg));
		}

		printf("\n");
	}

	free(marks);
	free(pconjs);
}
