#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "pmt.h"
#include "ring_check.h"
#include "ring_normalizer.h"

void make_normalizer(
	ring_t * pG,
	ring_t * pH,
	ring_t * pnormalizer,
	void    * normalizer_elements_buffer,
	int       max_normalizer_bytes)
{
	int    i, j, dummy, all_cjg;
	void * pg, * ph, * pc;
	unsigned char gh[VOP_MAX_ELEMENT_SIZE];
	unsigned char gi[VOP_MAX_ELEMENT_SIZE];
	unsigned char ghgi[VOP_MAX_ELEMENT_SIZE];

	// xxx to lib func.
	if (pG->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pG->pspec->element_size);
		exit(1);
	}

	pnormalizer->pvelements = normalizer_elements_buffer;
	pnormalizer->num_elements = 0; // Increment below
	pnormalizer->pvaux = pG->pvaux;
	pnormalizer->pspec = pG->pspec;

	//for each g [i] in G
	//  for each h [j] in H
	//    if ghg^-1 not in H
	//      next g
	//  add g to the normalizer

	for (i = 0; i < pG->num_elements; i++) {
		pg = VRING_ELEMENT(pG, i);
		ring_element_inverse(pG, pg, gi);

		all_cjg = 1;
		for (j = 0; j < pH->num_elements; j++) {
			ph = VRING_ELEMENT(pH, j);

			pG->pspec->pvop(gh, pg, ph, pG->pvaux);
			pG->pspec->pvop(ghgi, gh, gi, pG->pvaux);

			if (!ring_find_index(pH, ghgi, &dummy)) {
				all_cjg = 0;
				break;
			}
		}

		if (all_cjg) {
			pc = VRING_ELEMENT(pnormalizer,
				pnormalizer->num_elements);
			memcpy(pc, pg, pG->pspec->element_size);
			pnormalizer->num_elements++;
		}
	}
}
