#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ring.h"
#include "ring_check.h"
#include "ring_orders.h"

// ----------------------------------------------------------------
void show_orders(
	ring_t * pring)
{
	int * orders;
	int i;
	void * pa;

	orders = (int *) malloc(pring->num_elements * sizeof(int));
	if (orders == 0) {
		fprintf(stderr, "malloc(%d) failed.\n",
			pring->num_elements * sizeof(int));
		exit(1);
	}

	ring_find_orders(pring, orders, 0);

	for (i = 0; i < pring->num_elements; i++) {
		pa = VRING_ELEMENT(pring, i);
		pring->pspec->pvprint(pa, pring->pvaux);
		printf("  %3d\n", orders[i]);
	}

	free(orders);
}

// ----------------------------------------------------------------
void ring_find_orders(
	ring_t * pring,
	int     * orders,
	int       verbosity)
{
	unsigned char e[VOP_MAX_ELEMENT_SIZE];
	void * pa;
	unsigned char apower[VOP_MAX_ELEMENT_SIZE];
	int i;

	// xxx to lib func.
	if (pring->pspec->element_size > VOP_MAX_ELEMENT_SIZE) {
		fprintf(stderr,
			"%s: ring element size %d too big.\n",
				__FILE__, pring->pspec->element_size);
		exit(1);
	}

	// First, locate the identity.
	if (!find_identity(pring, e)) {
		fprintf(stderr,
			"ring_find_orders:  Couldn't find identity "
			"element in ring.\n");
		exit(1);
	}

// xxx use ring_element_order() instead.
	for (i = 0; i < pring->num_elements; i++) {
		pa = VRING_ELEMENT(pring, i);
		memcpy(apower, pa, pring->pspec->element_size);
		for (orders[i] = 1; ; orders[i]++) {
			if (orders[i] > pring->num_elements) {
				printf("ring_find_orders: "
					"non-invertible element ");
				pring->pspec->pvprint(pa, pring->pvaux);
				printf(".\n");
				exit(1);
			}
			if (pring->pspec->pvequal(apower, e, pring->pvaux))
				break;
			pring->pspec->pvop(apower, apower, pa, pring->pvaux);
		}
	}
}

// ----------------------------------------------------------------
int find_maximal_element_order(
	ring_t * pring)
{
	int i;
	void * pa;
	int max = 1;
	int current;

	for (i = 0; i < pring->num_elements; i++) {
		pa = VRING_ELEMENT(pring, i);
		current = ring_element_order(pring, pa);
		if (current > max)
			max = current;

	}
	return max;
}
