#ifndef VOPS_H
#define VOPS_H

#define VOP_MAX_ELEMENT_SIZE 2048

#define VARRAY_ELEMENT(v,i,elt_size) \
	((void *) ((unsigned char *)(v) + (i)*(elt_size)))

// Example:
//
// A matrix is declared as int M[5][5]. In a particular situation, that space
// is used for a 3x3 matrix.  "lda" is the leading dimension of the array,
// that is, the distance between adjacent rows in terms of elements (not
// bytes) -- in this example, lda is 5.  We want to return a pointer to
// M[2][1], i.e. i = 2 and j = 1.
//
// The address of the cell in question is i times lda times element size,
// plus j times element size.  Or, ((i*lda)+j) * element_size.  In this
// example, (2 * 5 + 1) * 4, or 44 bytes from the start of the matrix.
//
//      j = 1
// **** **** **** ---- ----
// **** **** **** ---- ----
// **** @*** **** ---- ---- i = 2
// ---- ---- ---- ---- ----
// ---- ---- ---- ---- ----

#define VMATRIX_ELEMENT(v,i,j,lda,elt_size) \
	((void *) ((unsigned char *)((v) + ((i)*(lda)+(j)) *(elt_size))))

#define VARRAY_ASSIGN(dst,di,src,si,elt_size) \
	memcpy( \
		VARRAY_ELEMENT((dst),(di),(elt_size)), \
		VARRAY_ELEMENT((src),(si),(elt_size)), \
		elt_size)

// E.g. out = in1 op in2 mod aux
typedef void vop_t(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux);

typedef int vequal_t(
	void * pvin1,
	void * pvin2,
	void * pvaux);

typedef int vless_t(
	void * pvin1,
	void * pvin2,
	void * pvaux);

typedef void vmake_id_t(
	void * pvin,
	void * pvaux);

typedef void vprint_t(
	void * pvin,
	void * pvaux);

typedef void vauxprint_t(
	void * pvaux);

typedef int  vscan_t(
	void * pvout,
	char * in,
	void * pvaux,
	int    max_elt_bytes);

typedef int vauxscan_t(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes);

void sort_varray(
	void * pvelements,
	int    num_elements,
	int    element_size,
	vless_t * pvless,
	void * pvaux);

// Field operations (e.g. for vmatrix), all in one bag.
typedef struct _vfops_t {
	int          element_size;
	vop_t      * pvadd;
	vop_t      * pvsub;
	vop_t      * pvmul;
	vop_t      * pvdiv;
	vequal_t   * pvequal;
	vless_t    * pvless;
	vprint_t   * pvprint;
	vmake_id_t * pvmake_add_id;
	vmake_id_t * pvmake_mul_id;
} vfops_t;

#endif // VOPS_H
