#include <stdio.h>
#include "lgf.h"
#include "power.h"

#define MAX_Q  101

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int n = 2, q = 2;
	int nmat;
	lgf_t mm;
	int i, j, k;
	int C11, C12, det;
	int A, B, C, D;
	int max_print = 2;
	int fill_width = 4;

	int C11_bins[MAX_Q+1];
	int C12_bins[MAX_Q+1][MAX_Q+1];
	int det_bins[MAX_Q+1];
	int gcd;

	if (argc >= 2) {
		(void)sscanf(argv[1], "%d", &q);
	}
	if (argc >= 3)
		max_print = q;

	if (q > MAX_Q) {
		fprintf(stderr, "q (%d) too big.\n", q);
		exit(1);
	}
	nmat = calc_power(q, n*n);

	printf("\n");
	printf("%dx%d's over GF(%d):\n", n, n, q);

	for (i = 0; i < q; i++) {
		det_bins[i] = 0;
	}
	for (i = 0; i < q; i++) {
		C11_bins[i] = 0;
		for (j = 0; j < q; j++) {
			C12_bins[i][j] = 0;
		}
	}

	for (k = 0; k < nmat; k++) {
		if ((k % 100000) == 0) {
			if (isatty(fileno(stdout))) {
				printf("\r%3.0f%%",
					(double)k * 100.0 / (double)nmat);
				fflush(stdout);
			}
		}
		kth_nqmat(&mm, k, n, q);
		A = mm.mat[0][0]; B = mm.mat[0][1];
		C = mm.mat[1][0]; D = mm.mat[1][1];

		C11 = posmod(A * D, q);
		C12 = posmod(B * C, q);

		det = posmod(C11 - C12, q);

		C11_bins[C11]++;
		C12_bins[C11][C12]++;

		det_bins[det]++;
	}

	if (isatty(fileno(stdout))) {
		printf("\r       \r");
		fflush(stdout);
	}

	for (i = 0; i < max_print; i++) {
		printf("C11 %d = %7d",
			i, C11_bins[i]);
		if (i == 0) {
			gcd = calc_gcd(
				C11_bins[0],
				C11_bins[1]);
			printf("          %d/%d",
				C11_bins[0]/gcd,
				C11_bins[1]/gcd);
		}
		printf("\n");
		for (j = 0; j < max_print; j++) {
			printf("%*sC12 %d = %7d",
				fill_width, " ", j, C12_bins[i][j]);
			if (j == 0) {
				gcd = calc_gcd(
					C12_bins[i][0],
					C12_bins[i][1]);
				printf("          %d/%d",
					C12_bins[i][0]/gcd,
					C12_bins[i][1]/gcd);
			}
			printf("\n");
		}
	}
	printf("\n");
	printf("Total: %d\n", nmat);
	for (i = 0; i < max_print; i++) {
		printf("  det %d = %7d\n", i, det_bins[i]);
	}

	return 0;
}
