#include <stdio.h>
#include "lgf.h"
#include "power.h"

#define MAX_Q  11

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int n = 4, q = 2;
	int nmat;
	lgf_t mm;
	int i, j, k, l;
	int C11, C12, C13, C14;
	int A, B, C, D;
	int E, F, G, H;
	int I, J, K, L;
	int M, N, O, P;
	int IN_MJ, IO_MK, IP_ML, JO_NK, JP_NL, KP_OL;
	int det;
	int max_print = 2;
	int fill_width = 4;
	int gcd;

	int C11_bins[MAX_Q+1];
	int C12_bins[MAX_Q+1][MAX_Q+1];
	int C13_bins[MAX_Q+1][MAX_Q+1][MAX_Q+1];
	int C14_bins[MAX_Q+1][MAX_Q+1][MAX_Q+1][MAX_Q+1];
	int det_bins[MAX_Q+1];

	if (argc >= 2) {
		(void)sscanf(argv[1], "%d", &q);
	}
	if (argc >= 3)
		max_print = q;

	if (q > MAX_Q) {
		fprintf(stderr, "q (%d) too big.\n", q);
		exit(1);
	}
	nmat = calc_power(q, n*n);

	printf("\n");
	printf("%dx%d's over GF(%d):\n", n, n, q);

	for (i = 0; i < q; i++) {
		det_bins[i] = 0;
	}
	for (i = 0; i < q; i++) {
		C11_bins[i] = 0;
		for (j = 0; j < q; j++) {
			C12_bins[i][j] = 0;
			for (k = 0; k < q; k++) {
				C13_bins[i][j][k] = 0;
				for (l = 0; l < q; l++) {
					C14_bins[i][j][k][l] = 0;
				}
			}
		}
	}

	for (k = 0; k < nmat; k++) {
		if ((k % 100000) == 0) {
			if (isatty(fileno(stdout))) {
				printf("\r%3.0f%%",
					(double)k * 100.0 / (double)nmat);
				fflush(stdout);
			}
		}
		kth_nqmat(&mm, k, n, q);
		A=mm.mat[0][0];B=mm.mat[0][1];C=mm.mat[0][2];D=mm.mat[0][3];
		E=mm.mat[1][0];F=mm.mat[1][1];G=mm.mat[1][2];H=mm.mat[1][3];
		I=mm.mat[2][0];J=mm.mat[2][1];K=mm.mat[2][2];L=mm.mat[2][3];
		M=mm.mat[3][0];N=mm.mat[3][1];O=mm.mat[3][2];P=mm.mat[3][3];

		IN_MJ = posmod(I*N-M*J, q);
		IO_MK = posmod(I*O-M*K, q);
		IP_ML = posmod(I*P-M*L, q);
		JO_NK = posmod(J*O-N*K, q);
		JP_NL = posmod(J*P-N*L, q);
		KP_OL = posmod(K*P-O*L, q);

		C11 = posmod( A * (F*KP_OL-G*JP_NL+H*JO_NK), q);
		C12 = posmod(-B * (E*KP_OL-G*IP_ML+H*IO_MK), q);
		C13 = posmod( C * (E*JP_NL-F*IP_ML+H*IN_MJ), q);
		C14 = posmod(-D * (E*JO_NK-F*IO_MK+G*IN_MJ), q);

		det = (C11 + C12 + C13 + C14) % q;

		C11_bins[C11]++;
		C12_bins[C11][C12]++;
		C13_bins[C11][C12][C13]++;
		C14_bins[C11][C12][C13][C14]++;

		det_bins[det]++;
	}

	if (isatty(fileno(stdout))) {
		printf("\r       \r");
		fflush(stdout);
	}

	for (i = 0; i < max_print; i++) {
		printf("C11 %d = %7d",
			i, C11_bins[i]);
		if (i == 0) {
			gcd = calc_gcd(
				C11_bins[0],
				C11_bins[1]);
			printf("          %d/%d",
				C11_bins[0]/gcd,
				C11_bins[1]/gcd);
		}
		printf("\n");
		for (j = 0; j < max_print; j++) {
			printf("%*sC12 %d = %7d",
				fill_width, " ", j, C12_bins[i][j]);
			if (j == 0) {
				gcd = calc_gcd(
					C12_bins[i][0],
					C12_bins[i][1]);
				printf("          %d/%d",
					C12_bins[i][0]/gcd,
					C12_bins[i][1]/gcd);
			}
			printf("\n");
			for (k = 0; k < max_print; k++) {
				printf("%*s%*sC13 %d = %7d",
					fill_width, " ", fill_width, " ",
					k, C13_bins[i][j][k]);
				if (k == 0) {
					gcd = calc_gcd(
						C13_bins[i][j][0],
						C13_bins[i][j][1]);
					printf("          %d/%d",
						C13_bins[i][j][0]/gcd,
						C13_bins[i][j][1]/gcd);
				}
				printf("\n");
				for (l = 0; l < max_print; l++) {
					printf("%*s%*s%*sC14 %d = %7d",
						fill_width, " ",
						fill_width, " ",
						fill_width, " ",
						l, C14_bins[i][j][k][l]);
					if (l == 0) {
						gcd = calc_gcd(
							C14_bins[i][j][k][0],
							C14_bins[i][j][k][1]);
						printf("          %d/%d",
							C14_bins[i][j][k][0]/gcd,
							C14_bins[i][j][k][1]/gcd);
					}
					printf("\n");
				}
			}
		}
	}
	printf("\n");
	printf("Total: %d\n", nmat);
	for (i = 0; i < max_print; i++) {
		printf("  det %d = %7d\n", i, det_bins[i]);
	}

	return 0;
}
