#include <stdio.h>
#include <stdlib.h>
#include "gint.h"

// ----------------------------------------------------------------
gint_t gint_add(gint_t a, gint_t b)
{
	gint_t c;
	c.re = a.re + b.re;
	c.im = a.im + b.im;
	return c;
}

// ----------------------------------------------------------------
gint_t gint_sub(gint_t a, gint_t b)
{
	gint_t c;
	c.re = a.re - b.re;
	c.im = a.im - b.im;
	return c;
}

// ----------------------------------------------------------------
gint_t gint_mul(gint_t a, gint_t b)
{
	gint_t c;
	c.re = a.re * b.re - a.im * b.im;
	c.im = a.re * b.im + a.im * b.re;
	return c;
}

// ----------------------------------------------------------------
// a = qb + r:
// q = a/b (using integer division)
// r = a - qb

// (u+vi)(x-yi)   (ux+vy)+(vx-uy)i
// ------------ = ------------
// (x+yi)(x-yi)   x^2 + y^2

// ((a.re*b.re + a.im*b.im), (a.im*b.re - a.re*b.im))
// --------------------------------------------------
// b.re * b.re + b.im * b.im

gint_t gint_quot(gint_t a, gint_t b)
{
	gint_t q;
	int denom = b.re * b.re + b.im * b.im;
	if (denom == 0) {
		fprintf(stderr, "gint:  divide by zero.\n");
		exit(1);
	}
	q.re = a.re*b.re + a.im*b.im;
	q.im = a.im*b.re - a.re*b.im;
	q.re /= denom;
	q.im /= denom;
	return q;
}

gint_t gint_rem(gint_t a, gint_t b)
{
	gint_t q  = gint_quot(a, b);
	gint_t qb = gint_mul(q, b);
	return gint_sub(a, qb);
}

// ----------------------------------------------------------------
void gint_quot_and_rem(gint_t a, gint_t b, gint_t * pquot, gint_t * prem)
{
	gint_t q  = gint_quot(a, b);
	gint_t qb = gint_mul(q, b);
	*pquot = q;
	*prem = gint_sub(a, qb);
}

// ----------------------------------------------------------------
int gint_equal(gint_t a, gint_t b)
{
	if (a.re != b.re)
		return 0;
	else if (a.im != b.im)
		return 0;
	else
		return 1;
}

// ----------------------------------------------------------------
int gint_norm(gint_t a)
{
	return a.re * a.re + a.im * a.im;
}

// ----------------------------------------------------------------
int gint_is_zero(gint_t a)
{
	if (a.re != 0)
		return 0;
	if (a.im != 0)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int gint_is_unit(gint_t a)
{
	if (a.re == 0) {
		if ((a.im == 1) || (a.im == -1))
			return 1;
	}
	if (a.im == 0) {
		if ((a.re == 1) || (a.re == -1))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
int gint_scan(
	gint_t * pout,
	char   * pin)
{
	gint_t g;

	if (sscanf(pin, "%d,%d", &g.re, &g.im) != 2) {
		g.im = 0;
		if (sscanf(pin, "%d", &g.re) != 1) {
			return 0;
		}
	}
	return 1;
}

// ----------------------------------------------------------------
void gint_print(
	gint_t g)

{
	printf("%d,%d", g.re, g.im);
}
