// SL_n(GF(q)) (det == 1) and GL_n(GF(q)) (det != 0).
#include <stdio.h>
#include <stdlib.h>
#include "lgf.h"
#include "posmod.h"

#include "phi.h"
#include "gcd.h"
#include "power.h"
#include "isprime.h"

// ----------------------------------------------------------------
void lgf_mul(
	lgf_t * pa,
	lgf_t * pb,
	lgf_t * pc,
	int     n,
	int     q)
{
	lgf_t    c;
	int i, j, k;

	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			c.mat[i][j] = 0;
			for (k = 0; k < n; k++) {
				c.mat[i][j] += posmod(
					pa->mat[i][k] * pb->mat[k][j],
					q);
				if (c.mat[i][j] >= q)
					c.mat[i][j] -= q;
			}
		}
	}
	*pc = c;
}

// ----------------------------------------------------------------
void lgf_print(
	lgf_t * pa,
	int     n,
	int     q)
{
	int i, j;
#if 1
	printf("[");
	for (i = 0; i < n; i++) {
		printf("[");
		for (j = 0; j < n; j++) {
			if (j > 0)
				printf(",");
			printf("%d", pa->mat[i][j]);
		}
		printf("]");
	}
	printf("]");
#else
	for (i = 0; i < n; i++) {
		if (i > 0)
			printf("  ");
		printf("[");
		for (j = 0; j < n; j++) {
			if (j > 0)
				printf(",");
			printf("%d", pa->mat[i][j]);
		}
		printf("]");
		printf("\n");
	}
#endif
}

// ----------------------------------------------------------------
static void lgf_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	lgf_t * pa = (lgf_t *)pvin1;
	lgf_t * pb = (lgf_t *)pvin2;
	lgf_t * pc = (lgf_t *)pvout;
	lgf_aux_t * paux= (lgf_aux_t *)pvaux;

	lgf_mul(pa, pb, pc, paux->n, paux->q);
}

// ----------------------------------------------------------------
static int lgf_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	lgf_t * pa = (lgf_t *)pvin1;
	lgf_t * pb = (lgf_t *)pvin2;
	lgf_aux_t * paux= (lgf_aux_t *)pvaux;
	int i, j;
	for (i = 0; i < paux->n; i++)
		for (j = 0; j < paux->n; j++)
			if (pa->mat[i][j] != pb->mat[i][j])
				return 0;
	return 1;
}

// ----------------------------------------------------------------
static int lgf_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	lgf_t * pa = (lgf_t *)pvin1;
	lgf_t * pb = (lgf_t *)pvin2;
	lgf_aux_t * paux= (lgf_aux_t *)pvaux;
	int i, j;
	for (i = 0; i < paux->n; i++) {
		for (j = 0; j < paux->n; j++) {
			if (pa->mat[i][j] < pb->mat[i][j])
				return 1;
			else if (pa->mat[i][j] > pb->mat[i][j])
				return 0;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
static void lgf_vprint(
	void * pvin,
	void * pvaux)
{
	lgf_t * pa = (lgf_t *)pvin;
	lgf_aux_t * paux= (lgf_aux_t *)pvaux;

	lgf_print(pa, paux->n, paux->q);
}

// ----------------------------------------------------------------
static void lgf_vauxprint(
	void * pvaux)
{
	lgf_aux_t * paux = (lgf_aux_t *)pvaux;
	printf("%d,%d", paux->n, paux->q);
}

// ----------------------------------------------------------------
int lgf_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	if (sizeof(lgf_t) > max_elt_bytes) {
		fprintf(stderr, "lgf:  Need %d bytes for element.\n",
			sizeof(lgf_t));
		exit(1);
	}

	printf("lgf_vscan stub.\n");
	return 0;
}

// ----------------------------------------------------------------
int lgf_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	lgf_aux_t * paux = (lgf_aux_t *)pvaux;

	if (max_aux_bytes < sizeof(lgf_aux_t)) {
		fprintf(stderr, "lgf:  Need %d bytes for aux.\n",
			sizeof(lgf_aux_t));
		return 0;
	}

	if (argc != 2) {
		fprintf(stderr, "lgf:  Need n and q aux arguments.\n");
		return 0;
	}

	if (sscanf(argv[0], "%d", &paux->n) != 1) {
		fprintf(stderr, "lgf: Couldn't scan \"%s\" as int.\n",
			argv[0]);
		return 0;
	}
	if (sscanf(argv[1], "%d", &paux->q) != 1) {
		fprintf(stderr, "lgf: Couldn't scan \"%s\" as int.\n",
			argv[0]);
		return 0;
	}
	if (paux->n < 1) {
		fprintf(stderr, "lgf: Dimension must be >= 1; got %d.\n",
			paux->n);
		return 0;
	}
	if (paux->n > LGF_MAX_DIM) {
		fprintf(stderr, "lgf: Dimension must be <= %d; got %d.\n",
			LGF_MAX_DIM, paux->n);
		return 0;
	}
	if (paux->q < 1) {
		fprintf(stderr, "lgf: Modulus must be >= 1; got %d.\n",
			paux->q);
		return 0;
	}
	if (!isprime(paux->q)) {
		fprintf(stderr, "lgf: Modulus must be prime; got %d.\n",
			paux->q);
		return 0;
	}
	return 1;
}

// ----------------------------------------------------------------
void kth_nqmat(
	lgf_t * pa,
	int     k,
	int     n,
	int     q)
{
	int i, j;
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			pa->mat[i][j] = k % q;
			k /= q;
		}
	}
}

// ----------------------------------------------------------------
int lgf_calc_det(
	lgf_t * pa,
	int     N,
	int     q)
{
	int a, b, c, d;
	int e, f, g, h;
	int i, j, k, l;
	int m, n, o, p;

	if (N == 1) {
		return pa->mat[0][0];
	}

	else if (N == 2) {
		a = pa->mat[0][0]; b = pa->mat[0][1];
		c = pa->mat[1][0]; d = pa->mat[1][1];
		return posmod(a * d - b * c, q);
	}

	else if (N == 3) {
		int aeihf, bdifg, cdheg;
		int det;
		a = pa->mat[0][0]; b = pa->mat[0][1]; c = pa->mat[0][2];
		d = pa->mat[1][0]; e = pa->mat[1][1]; f = pa->mat[1][2];
		g = pa->mat[2][0]; h = pa->mat[2][1]; i = pa->mat[2][2];

		aeihf = (a * posmod(e * i - h * f, q)) % q;
		bdifg = (b * posmod(d * i - f * g, q)) % q;
		cdheg = (c * posmod(d * h - e * g, q)) % q;

		det = posmod(aeihf - bdifg + cdheg, q);

		return det;
	}
	else if (N == 4) {
		int in_mj, io_mk, ip_ml, jo_nk, jp_nl, kp_ol;
		int row1, row2, row3, row4;

		a=pa->mat[0][0];b=pa->mat[0][1];c=pa->mat[0][2];d=pa->mat[0][3];
		e=pa->mat[1][0];f=pa->mat[1][1];g=pa->mat[1][2];h=pa->mat[1][3];
		i=pa->mat[2][0];j=pa->mat[2][1];k=pa->mat[2][2];l=pa->mat[2][3];
		m=pa->mat[3][0];n=pa->mat[3][1];o=pa->mat[3][2];p=pa->mat[3][3];

		// |a b c d|
		// |e f g h|
		// |i j k l|
		// |m n o p|

		//   |f g h|     |e g h|     |e f h|     |e f g|
		// a |j k l| - b |i k l| + c |i j l| + d |i j k|
		//   |n o p|     |m o p|     |m n p|     |m n o|

		//   a [f(kp-ol)-g(jp-nl)+h(jo-nk)]
		// - b [e(kp-ol)-g(ip-ml)+h(io-mk)]
		// + c [e(jp-nl)-f(ip-ml)+h(in-mj)]
		// - d [e(jo-nk)-f(io-mk)+g(in-mj)]

		in_mj = posmod(i*n-m*j, q);
		io_mk = posmod(i*o-m*k, q);
		ip_ml = posmod(i*p-m*l, q);
		jo_nk = posmod(j*o-n*k, q);
		jp_nl = posmod(j*p-n*l, q);
		kp_ol = posmod(k*p-o*l, q);

		row1 = posmod(a * (f*kp_ol-g*jp_nl+h*jo_nk), q);
		row2 = posmod(b * (e*kp_ol-g*ip_ml+h*io_mk), q);
		row3 = posmod(c * (e*jp_nl-f*ip_ml+h*in_mj), q);
		row4 = posmod(d * (e*jo_nk-f*io_mk+g*in_mj), q);

		return posmod(row1 - row2 + row3 - row4, q);

	}
	else {
		// Really, need to implement row reduction & take the product
		// along the diagonal.
		fprintf(stderr, "Pteh!\n");
		exit(1);
	}
}

// ----------------------------------------------------------------
// Diagonal subgroup of GL_n(GF(q)).
static int dgf_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	lgf_t * pelements = (lgf_t *) pvelements;
	lgf_aux_t * paux = (lgf_aux_t *)pvaux;
	int k, i, j, which;
	int num_elements = calc_power(paux->q - 1, paux->n);
	int max_group_elements = max_group_bytes / sizeof(lgf_t);

	if (num_elements >= max_group_elements) {
		fprintf(stderr,
			"dgf_fill_group:  Out of space.\n");
		exit(1);
	}

	for (k = 0; k < num_elements; k++) {
		for (i = 0; i < paux->n; i++)
			for (j = 0; j < paux->n; j++)
				pelements[k].mat[i][j] = 0;
		which = k;
		for (i = 0; i < paux->n; i++) {
			pelements[k].mat[i][i] = 1 + which % (paux->q - 1);
			which /= paux->q - 1;
		}
	}
	*pnum_elements = num_elements;
	return 1;
}

// ----------------------------------------------------------------
static int lgf_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements,
	int    do_sl)
{
	lgf_t * pelements = (lgf_t *) pvelements;
	lgf_aux_t * paux = (lgf_aux_t *)pvaux;
	int k;
	int num_elements = 0; // Increment below.
	int max_group_elements = max_group_bytes / sizeof(lgf_t);
	// There must be a formula for |SL_n(GF(q))| and |GL_n(GF(q))|,
	// better than this brute-force search.
	int qn2 = calc_power(paux->q, paux->n * paux->n);
	int det;

	for (k = 0; k < qn2; k++) {
		kth_nqmat(&pelements[num_elements], k,
			paux->n, paux->q);
		det = lgf_calc_det(&pelements[num_elements],
			paux->n, paux->q);
		if (do_sl) {
			if (det != 1)
				continue;
		}
		else {
			if (det == 0)
				continue;
		}

		if (num_elements >= max_group_elements) {
			fprintf(stderr,
				"%clgf_fill_group:  Out of space.\n",
				do_sl ? 's' : 'g');
			exit(1);
		}
		num_elements++;
	}
	*pnum_elements = num_elements;
	return 1;
}

// ----------------------------------------------------------------
int  slgf_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	return lgf_fill_group(pvelements, pvaux, max_group_bytes,
		pnum_elements, 1);
}

// ----------------------------------------------------------------
int  glgf_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	return lgf_fill_group(pvelements, pvaux, max_group_bytes,
		pnum_elements, 0);
}

// ================================================================
group_spec_t dgf_group_spec = {
	element_size  : sizeof(lgf_t),
	naux          : 2,
	pvop          : lgf_vmul,
	pvequal       : lgf_vequal,
	pvless        : lgf_vless,
	pvprint       : lgf_vprint,
	pvauxprint     : lgf_vauxprint,
	pvscan        : lgf_vscan,
	pvauxscan      : lgf_vauxscan,
	pfill_group   : dgf_fill_group,
	desc          : "dgf"
};

group_spec_t slgf_group_spec = {
	element_size  : sizeof(lgf_t),
	naux          : 2,
	pvop          : lgf_vmul,
	pvequal       : lgf_vequal,
	pvless        : lgf_vless,
	pvprint       : lgf_vprint,
	pvauxprint     : lgf_vauxprint,
	pvscan        : lgf_vscan,
	pvauxscan      : lgf_vauxscan,
	pfill_group   : slgf_fill_group,
	desc          : "slgf"
};

group_spec_t glgf_group_spec = {
	element_size  : sizeof(lgf_t),
	naux          : 2,
	pvop          : lgf_vmul,
	pvequal       : lgf_vequal,
	pvless        : lgf_vless,
	pvprint       : lgf_vprint,
	pvauxprint     : lgf_vauxprint,
	pvscan        : lgf_vscan,
	pvauxscan      : lgf_vauxscan,
	pfill_group   : glgf_fill_group,
	desc          : "glgf"
};
