#include <stdio.h>
#include <stdlib.h>
#include "miquat.h"
#include "mod.h"
#include "phi.h"
#include "gcd.h"

typedef struct _smiquat_aux_t {
	int modulus;
	int stride;
} smiquat_aux_t;

// ----------------------------------------------------------------
miquat_t miquat_add(miquat_t a, miquat_t b, int m)
{
	miquat_t c;
	c.H = mod_add(a.H, b.H, m);
	c.I = mod_add(a.I, b.I, m);
	c.J = mod_add(a.J, b.J, m);
	c.K = mod_add(a.K, b.K, m);
	return c;
}

// ----------------------------------------------------------------
miquat_t miquat_sub(miquat_t a, miquat_t b, int m)
{
	miquat_t c;
	c.H = mod_sub(a.H, b.H, m);
	c.I = mod_sub(a.I, b.I, m);
	c.J = mod_sub(a.J, b.J, m);
	c.K = mod_sub(a.K, b.K, m);
	return c;
}

// ----------------------------------------------------------------
miquat_t miquat_mul(miquat_t a, miquat_t b, int m)
{
	miquat_t c;

	c.H = a.H * b.H  - a.I  * b.I - a.J  * b.J - a.K  * b.K;
	c.I = a.I * b.H  + a.H  * b.I - a.K  * b.J + a.J  * b.K;
	c.J = a.J * b.H  + a.K  * b.I + a.H  * b.J - a.I  * b.K;
	c.K = a.K * b.H  - a.J  * b.I + a.I  * b.J + a.H  * b.K;

	c.H %= m;
	c.I %= m;
	c.J %= m;
	c.K %= m;

	if (c.H < 0)
		c.H += m;
	if (c.I < 0)
		c.I += m;
	if (c.J < 0)
		c.J += m;
	if (c.K < 0)
		c.K += m;

	return c;
}

// ----------------------------------------------------------------
int miquat_equal(miquat_t a, miquat_t b, int m)
{
	if (a.H != b.H)
		return 0;
	if (a.I != b.I)
		return 0;
	if (a.J != b.J)
		return 0;
	if (a.K != b.K)
		return 0;
	return 1;
}

// ================================================================
static void miquat_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	miquat_t a = *(miquat_t *)pvin1;
	miquat_t b = *(miquat_t *)pvin2;
	int m = *(int *)pvaux;
	miquat_t c = miquat_add(a, b, m);
	*(miquat_t *)pvout = c;
}

// ================================================================
/*static xxx*/ void miquat_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	miquat_t a = *(miquat_t *)pvin1;
	miquat_t b = *(miquat_t *)pvin2;
	int m = *(int *)pvaux;
	miquat_t c = miquat_sub(a, b, m);
	*(miquat_t *)pvout = c;
}

// ----------------------------------------------------------------
static void miquat_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	miquat_t a = *(miquat_t *)pvin1;
	miquat_t b = *(miquat_t *)pvin2;
	int m = *(int *)pvaux;
	miquat_t c = miquat_mul(a, b, m);
	*(miquat_t *)pvout = c;
}

// ----------------------------------------------------------------
static int miquat_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	miquat_t a = *(miquat_t *)pvin1;
	miquat_t b = *(miquat_t *)pvin2;
	int m = *(int *)pvaux;
	return miquat_equal(a, b, m);
}

// ----------------------------------------------------------------
static int miquat_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	miquat_t a = *(miquat_t *) pvin1;
	miquat_t b = *(miquat_t *) pvin2;

	if (a.H < b.H)
		return 1;
	else if (a.H > b.H)
		return 0;
	if (a.I < b.I)
		return 1;
	else if (a.I > b.I)
		return 0;
	if (a.J < b.J)
		return 1;
	else if (a.J > b.J)
		return 0;
	if (a.K < b.K)
		return 1;
	else if (a.K > b.K)
		return 0;

	return 0;
}

// ----------------------------------------------------------------
static void miquat_vprint(
	void * pvin,
	void * pvaux)
{
	miquat_t x = *(miquat_t *)pvin;
	int m = *(int *)pvaux;
	char * fmt;

	if (m < 10)
		fmt = "%d,%d,%d,%d";
	else if (m < 100)
		fmt = "%2d,%2d,%2d,%2d";
	else
		fmt = "%3d,%3d,%3d,%3d";

	printf(fmt, x.H, x.I, x.J, x.K);
}

// ----------------------------------------------------------------
int miquat_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	miquat_t * pout = (miquat_t *)pvout;

	if (sizeof(miquat_t) > max_elt_bytes) {
		fprintf(stderr, "miquat:  Need %d bytes for element.\n",
			sizeof(miquat_t));
		exit(1);
	}

	return sscanf(pin, "%d,%d,%d,%d",
		&pout->H, &pout->I, &pout->J, &pout->K);
}

// ----------------------------------------------------------------
static void miquat_vauxprint(
	void * pvaux)
{
	printf("%d", *(int *)pvaux);
}

// ----------------------------------------------------------------
static int miquat_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	int modulus;

	if (max_aux_bytes < sizeof(int)) {
		fprintf(stderr, "miquat:  Need %d bytes for aux.\n",
			sizeof(int));
		exit(1);
	}

	if (argc != 1)
		return 0;
	if (sscanf(argv[0], "%d", &modulus) != 1) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			argv[0]);
		exit(1);
	}
	if (modulus < 1) {
		fprintf(stderr, "Modulus must be >= 1; got %d.\n",
			modulus);
		exit(1);
	}
	*(int *)pvaux = modulus;
	return 1;
}

// ----------------------------------------------------------------
int  miquatadd_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	miquat_t * pelements = (miquat_t *) pvelements;
	int m = *(int *)pvaux;
	int num_elements = m * m * m * m;
	int didx, h, i, j, k;

	if (num_elements > max_group_bytes/sizeof(int))
		return 0;

	didx = 0;
	for (h = 0; h < m; h++) {
		for (i = 0; i < m; i++) {
			for (j = 0; j < m; j++) {
				for (k = 0; k < m; k++) {
					pelements[didx].H = h;
					pelements[didx].I = i;
					pelements[didx].J = j;
					pelements[didx].K = k;
					didx++;
				}
			}
		}
	}

	*pnum_elements = num_elements;
	return 1;
}

// ----------------------------------------------------------------
int miquatmul_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	int m = *(int *)pvaux;
	int zi, zmi;

// xxx port me
	*pnum_elements = calc_phi_signed(m);
	if (*pnum_elements > max_group_bytes)
		return 0;

	for (zi = 0, zmi = 0; zi < m; zi++) {
		if (calc_gcd_signed(zi, m) == 1) {
			pelements[zmi] = zi;
			zmi++;
		}
	}

	return 1;
}

// ----------------------------------------------------------------
static void smiquat_vauxprint(
	void * pvaux)
{
	smiquat_aux_t * paux = pvaux;
	printf("%d:%d", paux->modulus, paux->stride);
}

// ----------------------------------------------------------------
static int smiquat_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	smiquat_aux_t * paux = pvaux;

	if (max_aux_bytes < sizeof(smiquat_aux_t)) {
		fprintf(stderr, "miquat:  Need %d bytes for aux.\n",
			sizeof(smiquat_aux_t));
		exit(1);
	}

	if (argc != 1)
		return 0;
	if (sscanf(argv[0], "%d:%d", &paux->modulus, &paux->stride) == 2) {
	}
	else if (sscanf(argv[0], "%d", &paux->modulus) == 1) {
		paux->stride = 1;
	}
	else {
		fprintf(stderr, "smiquat: Couldn't scan \"%s\".\n",
			argv[0]);
		exit(1);
	}
	if (paux->modulus < 1) {
		fprintf(stderr, "Modulus must be >= 1; got %d.\n",
			paux->modulus);
		exit(1);
	}
	if (paux->stride < 1) {
		fprintf(stderr, "Stride must be >= 1; got %d.\n",
			paux->stride);
		exit(1);
	}
	return 1;
}

// ----------------------------------------------------------------
int smiquatadd_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	miquat_t * pelements = (miquat_t *) pvelements;
	smiquat_aux_t * paux = pvaux;
	int m = paux->modulus;
	int s = paux->stride;
	int num_elements = m * m * m * m / s / s / s / s;
	int didx, h, i, j, k, vh, vi, vj, vk;

	if (num_elements > max_group_bytes/sizeof(int))
		return 0;

	didx = 0;
	for (vh = 0, h = 0; h < m; h++, vh += s) {
		for (vi = 0, i = 0; i < m; i++, vi += s) {
			for (vj = 0, j = 0; j < m; j++, vj += s) {
				for (vk = 0, k = 0; k < m; k++, vk += s) {
					pelements[didx].H = vh;
					pelements[didx].I = vi;
					pelements[didx].J = vj;
					pelements[didx].K = vk;
					didx++;
				}
			}
		}
	}

	*pnum_elements = num_elements;
	return 1;
}
// ----------------------------------------------------------------
int smiquatmul_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	miquat_t * pelements = (miquat_t *) pvelements;
	smiquat_aux_t * paux = pvaux;
	int m = paux->modulus;
	int s = paux->stride;
	int num_elements = (m * m * m * m / s / s / s / s) - 1;
	int didx, h, i, j, k, vh, vi, vj, vk;

	if (num_elements > max_group_bytes/sizeof(int))
		return 0;

	didx = 0;
	for (vh = 0, h = 0; h < m; h++, vh += s) {
		for (vi = 0, i = 0; i < m; i++, vi += s) {
			for (vj = 0, j = 0; j < m; j++, vj += s) {
				for (vk = 0, k = 0; k < m; k++, vk += s) {
					if ((h==0)&&(i==0)&&(j==0)&&(k==0))
						continue;
					pelements[didx].H = vh;
					pelements[didx].I = vi;
					pelements[didx].J = vj;
					pelements[didx].K = vk;
					didx++;
				}
			}
		}
	}

	*pnum_elements = num_elements;
	return 1;
}
// ================================================================
group_spec_t miquat_add_group_spec = {
	element_size  : sizeof(miquat_t),
	naux          : 1,
	pvop          : miquat_vadd,
	pvequal       : miquat_vequal,
	pvless        : miquat_vless,
	pvprint       : miquat_vprint,
	pvscan        : miquat_vscan,
	pvauxprint    : miquat_vauxprint,
	pvauxscan     : miquat_vauxscan,
	pfill_group   : miquatadd_fill_group,
	desc          : "miquatadd"
};

group_spec_t miquat_mul_group_spec = {
	element_size  : sizeof(miquat_t),
	naux          : 1,
	pvop          : miquat_vmul,
	pvequal       : miquat_vequal,
	pvless        : miquat_vless,
	pvprint       : miquat_vprint,
	pvscan        : miquat_vscan,
	pvauxprint    : miquat_vauxprint,
	pvauxscan     : miquat_vauxscan,
	pfill_group   : miquatmul_fill_group,
	desc          : "miquatmul"
};

group_spec_t smiquat_add_group_spec = {
	element_size  : sizeof(miquat_t),
	naux          : 1,
	pvop          : miquat_vadd,
	pvequal       : miquat_vequal,
	pvless        : miquat_vless,
	pvprint       : miquat_vprint,
	pvscan        : miquat_vscan,
	pvauxprint    : smiquat_vauxprint,
	pvauxscan     : smiquat_vauxscan,
	pfill_group   : smiquatadd_fill_group,
	desc          : "smiquatadd"
};

group_spec_t smiquat_mul_group_spec = {
	element_size  : sizeof(miquat_t),
	naux          : 1,
	pvop          : miquat_vmul,
	pvequal       : miquat_vequal,
	pvless        : miquat_vless,
	pvprint       : miquat_vprint,
	pvscan        : miquat_vscan,
	pvauxprint    : smiquat_vauxprint,
	pvauxscan     : smiquat_vauxscan,
	pfill_group   : smiquatmul_fill_group,
	desc          : "smiquatmul"
};
