#include <stdio.h>
#include "lgf.h"
#include "power.h"

#define MAX_ELEMENTS 1000000
#define MAX_Q  11

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int n = 2, q = 3;
	static lgf_t elements[MAX_ELEMENTS];
	static int dets[MAX_ELEMENTS];
	static int det_bins[MAX_Q];
	static int add_bins[MAX_Q];
	static int mul_bins[MAX_Q];
	int qn2, k, j;
	int verbose = 0;

	if (argc >= 3) {
		(void)sscanf(argv[1], "%d", &n);
		(void)sscanf(argv[2], "%d", &q);
	}
	if (argc > 3)
		verbose = 1;

	if (q > MAX_Q) {
		printf("q too big.\n");
		exit(1);
	}
	if (q < 2) {
		printf("q too small.\n");
		exit(1);
	}

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	printf("n = %2d q = %2d\n", n, q);

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	for (j = 0; j < q; j++) {
		add_bins[j] = 0;
	}
	for (j = 0; j < q; j++) {
		for (k = 0; k < q; k++) {
			int l = (j + k) % q;
			add_bins[l]++;
		}
	}
	printf("\n");
	for (j = 0; j < q; j++) {
		printf("  # add = %2d: %4d\n",
			j, add_bins[j]);
	}
	printf("\n");

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	for (j = 0; j < q; j++) {
		mul_bins[j] = 0;
	}
	for (j = 0; j < q; j++) {
		for (k = 0; k < q; k++) {
			int l = (j * k) % q;
			mul_bins[l]++;
		}
	}
	printf("\n");
	for (j = 0; j < q; j++) {
		printf("  # mul = %2d: %4d\n",
			j, mul_bins[j]);
	}
	printf("\n");

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	qn2 = calc_power(q, n * n);
	if (qn2 > MAX_ELEMENTS) {
		printf("q^(n^2) too big.\n");
		exit(1);
	}

	for (k = 0; k < q; k++) {
		det_bins[k] = 0;
	}

	for (k = 0; k < qn2; k++) {
		kth_nqmat(&elements[k], k, n, q);
		dets[k] = lgf_calc_det(&elements[k], n, q);
		det_bins[dets[k]]++;

		if (verbose) {
			lgf_print(&elements[k], n, q);
			printf(" %d\n", dets[k]);
		}
	}

	if (verbose)
		printf("\n");
	for (k = 0; k < q; k++) {
		printf("  # det = %2d: %4d (%4d * %4d rem %4d)\n",
			k, det_bins[k], det_bins[k]/q, q, det_bins[k] % q);
	}
	printf("\n");


	return 0;
}
