#include <stdio.h>
#include <stdlib.h>
#include "padic.h"

static void padic_fold_digit(
	padic_t * pa,
	int       digit,
	int       pos,
	int       p);

// ----------------------------------------------------------------
padic_t padic_zero(void)
{
	int i;
	padic_t rv;

	for (i = 0; i < PADIC_NUM_TOTAL; i++)
		rv.digits[i] = 0;
	return rv;
}

// ----------------------------------------------------------------
void padic_print(padic_t * pc, int p)
{
	int i;

	for (i = 0; i < PADIC_NUM_TOTAL; i++) {
		if (i == PADIC_NUM_FRACT)
			printf(".");
		else if ((p > 10) && (i > 0))
			printf(",");
		printf("%d", pc->digits[i]);
	}
}

// ----------------------------------------------------------------
padic_t padic_add(padic_t * pa, padic_t * pb, int p)
{
	padic_t rv = padic_zero();
	int i, psum;
	int carry = 0;

	for (i = 0; i < PADIC_NUM_TOTAL; i++) {
		psum = pa->digits[i] + pb->digits[i] + carry;
		rv.digits[i] = psum % p;
		carry = psum / p;
	}
	return rv;
}

// ----------------------------------------------------------------
padic_t padic_uneg(padic_t * pa, int p)
{
	padic_t rv = *pa;
	int i;
	int got_non_zero = 0;
	int flip_it = 0;

	// xxx find MS pos
	// invert from there to the right
	// fold digit 1 at pos

	for (i = 0; i < PADIC_NUM_TOTAL; i++) {
		if (rv.digits[i] != 0)
			got_non_zero = 1;
		if (flip_it)
			rv.digits[i] = p - 1 - rv.digits[i];
		if (got_non_zero == 1)
			flip_it = 1;
	}
	return rv;
}

// ----------------------------------------------------------------
padic_t padic_sub(padic_t * pa, padic_t * pb, int p)
{
	padic_t negb = padic_uneg(pb, p);
	padic_t diff = padic_add(pa, &negb, p);
	return diff;
}

// ----------------------------------------------------------------
//                  a1   a0 . b0   b1
//                * c1   c0 . d0   d1
//                 --------------------
//                  d1a1   d1a0 d1b0 d1b1
//             d0a1 d0a0   d0b0 d0b1
//        c0a1 c0a0 c0b0   c0b1
//   c1a1 c1a0 c1b0 c1b1
//   ------------------------------------
// = **** **** **** **** . **** **** ****
//
// Check for overflow on left; underflow on right is disregarded.

padic_t padic_mul(padic_t * pa, padic_t * pb, int p)
{
	padic_t rv = padic_zero();
	int i, j, k;
	int izpos, jzpos, kzpos;
	int digit;

	for (i = 0; i < PADIC_NUM_TOTAL; i++) {
		izpos = i - PADIC_NUM_FRACT;
		for (j = 0; j < PADIC_NUM_TOTAL; j++) {
			jzpos = j - PADIC_NUM_FRACT;
			kzpos = izpos + jzpos;
			k = kzpos + PADIC_NUM_FRACT;

			digit = pa->digits[i] * pb->digits[j];

			if (digit == 0)
				continue;

			// Overflow is fatal.
			if (k < 0) {
				fprintf(stderr, "padic_mul overflow.\n");
				exit(1);
			}

			// Underflow is ignored.
			if (k >= PADIC_NUM_TOTAL)
				continue;

			padic_fold_digit(&rv, digit, k, p);
		}
	}

	return rv;
}

// ----------------------------------------------------------------
padic_t padic_div(padic_t * pa, padic_t * pb, int p)
{
	padic_t rv = padic_zero();
	return rv;
}

// ----------------------------------------------------------------
padic_t padic_from_int(int n, int p)
{
	padic_t rv = padic_zero();
	int i, q, r;

	for (i = PADIC_NUM_FRACT; i < PADIC_NUM_TOTAL; i++) {
		q = n / p;
		r = n % p;

		if (r < 0) {
			// Silly C mod!
			q--;
			r += p;
		}

		rv.digits[i] = r;
		if (q == 0)
			break;
		n = q;
	}

	return rv;
}

// ----------------------------------------------------------------
padic_t padic_from_rat(rat_t n, int p)
{
	padic_t numer, denom, quot;
	numer = padic_from_int(n.numer, p);
	denom = padic_from_int(n.denom, p);
	quot  = padic_div(&numer, &denom, p);
	return quot;
}

// ----------------------------------------------------------------
static void padic_fold_digit(
	padic_t * pa,
	int       digit,
	int       pos,
	int       p)
{
	int i, psum;
	int carry = digit;

	for (i = pos; i < PADIC_NUM_TOTAL; i++) {
		psum = pa->digits[i] + carry;
		pa->digits[i] = psum % p;
		carry = psum / p;
	}
}
