#ifndef PMT_H
#define PMT_H

#include "group.h"

#define PMT_MAX_N 256

// Styles for pmt_print_cd_style()
#define PMT_CD_PARENS      4
#define PMT_CD_COMMA_COLON 5

typedef unsigned char pmt_elt_t;

typedef struct _pmt_t {
	pmt_elt_t images[PMT_MAX_N];
} pmt_t;

pmt_t pmt_mul  (pmt_t * pa, pmt_t * pb, int n);
int   pmt_equal(pmt_t * pa, pmt_t * pb, int n);
int   pmt_less (pmt_t * pa, pmt_t * pb, int n);
void  pmt_make_id(pmt_t * pa, int n);
void  pmt_invert(pmt_t * pin, pmt_t * pout, int n);
int   pmt_parity(pmt_t * pa, int n);

void  pmt_vmul  (void * pvout, void * pvin1, void * pvin2, void * pvaux);
int   pmt_vequal(void * pvin1, void * pvin2, void * pvaux);
int   pmt_vless (void * pvin1, void * pvin2, void * pvaux);
int   sn_fill_group(void * pvelements, void * pvaux, int max_bytes, int * pn);

int   pmt_validate_element(pmt_t * ppmt, int n);
int   an_validate_element(pmt_t * ppmt, int n);
int   an_fill_group(void * pvelements, void * pvaux, int max_bytes, int * pn);

void  pmt_print(pmt_t * pa, int n);
void  pmt_print_cd(pmt_t * pa, int n);
void  pmt_print_cd_style(pmt_t * pa, int n, int style);

void  pmt_print_cycle_type(pmt_t * pa, int n);

// xxx make a non-void one
int   pmt_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes);

int  pmt_vscan_cd(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes);

extern group_spec_t sn_group_spec;  // Sn with image-list I/O
extern group_spec_t an_group_spec;  // An with image-list I/O
extern group_spec_t snc_group_spec; // Sn with cycle-decomposition I/O
extern group_spec_t anc_group_spec; // An with cycle-decomposition I/O

#endif // PMT_H
