#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "q8.h"

static int cayley_table[8][8] = {
	/*              1     -1      i     -i      j     -j      k     -k */
	/*  1 */   { Q8_EP, Q8_EN, Q8_IP, Q8_IN, Q8_JP, Q8_JN, Q8_KP, Q8_KN },
	/* -1 */   { Q8_EN, Q8_EP, Q8_IN, Q8_IP, Q8_JN, Q8_JP, Q8_KN, Q8_KP },
	/*  i */   { Q8_IP, Q8_IN, Q8_EN, Q8_EP, Q8_KP, Q8_KN, Q8_JN, Q8_JP },
	/* -i */   { Q8_IN, Q8_IP, Q8_EP, Q8_EN, Q8_KN, Q8_KP, Q8_JP, Q8_JN },
	/*  j */   { Q8_JP, Q8_JN, Q8_KN, Q8_KP, Q8_EN, Q8_EP, Q8_IP, Q8_IN },
	/* -j */   { Q8_JN, Q8_JP, Q8_KP, Q8_KN, Q8_EP, Q8_EN, Q8_IN, Q8_IP },
	/*  k */   { Q8_KP, Q8_KN, Q8_JP, Q8_JN, Q8_IN, Q8_IP, Q8_EN, Q8_EP },
	/* -k */   { Q8_KN, Q8_KP, Q8_JN, Q8_JP, Q8_IP, Q8_IN, Q8_EP, Q8_EN },
};

static char * names[] = {
	" 1", // Q8_EP
	"-1", // Q8_EN
	" i", // Q8_IP
	"-i", // Q8_IN
	" j", // Q8_JP
	"-j", // Q8_JN
	" k", // Q8_KP
	"-k" // Q8_KN
};

// ----------------------------------------------------------------
int q8_mul(int a, int b)
{
	return cayley_table[a & 7][b & 7];
}

// ----------------------------------------------------------------
int q8_invert(int a)
{
	int c = Q8_EP;

	switch (a & 7) {
	case Q8_EP: c = Q8_EP; break;
	case Q8_EN: c = Q8_EN; break;
	case Q8_IP: c = Q8_IN; break;
	case Q8_IN: c = Q8_IP; break;
	case Q8_JP: c = Q8_JN; break;
	case Q8_JN: c = Q8_JP; break;
	case Q8_KP: c = Q8_KN; break;
	case Q8_KN: c = Q8_KP; break;
	}
	return c;
}

// ----------------------------------------------------------------
int q8_equal(int a, int b)
{
	return (a == b);
}

// ----------------------------------------------------------------
static void q8_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int c = q8_mul(a, b);
	*(int *)pvout = c;
}

// ----------------------------------------------------------------
static int q8_vequal(
	void * pin1,
	void * pin2,
	void * paux)
{
	return (*(int *)pin1 == *(int *)pin2);
}

// ----------------------------------------------------------------
static int q8_vless(
	void * pin1,
	void * pin2,
	void * paux)
{
	return (*(int *)pin1 < *(int *)pin2);
}

// ----------------------------------------------------------------
void q8_print(
	int a)
{
	printf("%s", names[a & 7]);
}

// ----------------------------------------------------------------
static void q8_vprint(
	void * pin,
	void * paux)
{
	int a = *(int *)pin;
	printf("%s", names[a & 7]);
}

// ----------------------------------------------------------------
int q8_scan(
	int  * pout,
	char * pin)
{
	int i;

	if (strcmp(pin, "1") == 0) {
		*pout = Q8_EP;
		return 1;
	}
	if (strcmp(pin, "i") == 0) {
		*pout = Q8_IP;
		return 1;
	}
	if (strcmp(pin, "j") == 0) {
		*pout = Q8_JP;
		return 1;
	}
	if (strcmp(pin, "k") == 0) {
		*pout = Q8_KP;
		return 1;
	}

	for (i = 0; i < 8; i++) {
		if (strcmp(pin, names[i]) == 0) {
			*pout = i;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
static int q8_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	if (sizeof(int) > max_elt_bytes) {
		fprintf(stderr, "q8:  Need %d bytes for element.\n",
			sizeof(int));
		exit(1);
	}

	return q8_scan(pvout, pin);
}

// ----------------------------------------------------------------
int q8_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (argc != 0)
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int  q8_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	int i;
	int m = 8;

	if (m > max_group_bytes/sizeof(int))
		return 0;
	for (i = 0; i < m; i++)
		pelements[i] = i;
	*pnum_elements = m;
	return 1;
}

// ================================================================
group_spec_t q8_group_spec = {
	element_size  : sizeof(int),
	naux          : 0,
	pvop          : q8_vmul,
	pvequal       : q8_vequal,
	pvless        : q8_vless,
	pvprint       : q8_vprint,
	pvauxprint    : 0,
	pvscan        : q8_vscan,
	pvauxscan     : q8_vauxscan,
	pfill_group   : q8_fill_group,
	desc          : "q8"
};
