#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "qnfe.h"

// ----------------------------------------------------------------
qnfe_t qnfe_add(qnfe_t a, qnfe_t b, qnfe_aux_t m)
{
	qnfe_t c;
	c.r = rat_add(a.r, b.r);
	c.i = rat_add(a.i, b.i);
	return c;
}

// ----------------------------------------------------------------
qnfe_t qnfe_sub(qnfe_t a, qnfe_t b, qnfe_aux_t m)
{
	qnfe_t c;
	c.r = rat_sub(a.r, b.r);
	c.i = rat_sub(a.i, b.i);
	return c;
}

// ----------------------------------------------------------------
//   (ar + ai sqrt(D)) * (br + bi sqrt(D))
// = (ar * br + ai * bi * D) + (ar * bi + ai * br) sqrt(D)

qnfe_t qnfe_mul(qnfe_t a, qnfe_t b, qnfe_aux_t m)
{
	qnfe_t c;
	c.r = rat_add(rat_mul(a.r, b.r), rat_mul(rat_mul(a.i, b.i), m.D));
	c.i = rat_add(rat_mul(a.r, b.i), rat_mul(a.i, b.r));
	return c;
}

// ----------------------------------------------------------------
// a = qb + r:
// q = a/b (using integer division)
// r = a - qb

// (u+vi)(x-yi)   (ux+vy)+(vx-uy)i
// ------------ = ------------
// (x+yi)(x-yi)   x^2 + y^2

// ((a.r*b.r + a.i*b.i), (a.i*b.r - a.r*b.i))
// --------------------------------------------------
// b.r * b.r + b.i * b.i

qnfe_t qnfe_quot(qnfe_t a, qnfe_t b, qnfe_aux_t m)
{
	qnfe_t q;
	rat_t denom = rat_add(rat_mul(b.r, b.r), rat_mul(b.i, b.i));
	if (rat_is_zero(denom)) {
		fprintf(stderr, "qnfe:  divide by zero.\n");
		exit(1);
	}
	q.r = rat_add(rat_mul(a.r, b.r), rat_mul(a.i, b.i));
	q.i = rat_sub(rat_mul(a.i, b.r), rat_mul(a.r, b.i));
	q.r = rat_div(q.r, denom);
	q.i = rat_div(q.i, denom);
	return q;
}

qnfe_t qnfe_rem(qnfe_t a, qnfe_t b, qnfe_aux_t m)
{
	qnfe_t q  = qnfe_quot(a, b, m);
	qnfe_t qb = qnfe_mul(q, b, m);
	return qnfe_sub(a, qb, m);
}

// ----------------------------------------------------------------
void qnfe_quot_and_rem(qnfe_t a, qnfe_t b, qnfe_aux_t m,
	qnfe_t * pquot, qnfe_t * prem)
{
	qnfe_t q  = qnfe_quot(a, b, m);
	qnfe_t qb = qnfe_mul(q, b, m);
	*pquot = q;
	*prem = qnfe_sub(a, qb, m);
}

// ----------------------------------------------------------------
int qnfe_equal(qnfe_t a, qnfe_t b, qnfe_aux_t m)
{
	if (!rat_equal(a.r, b.r))
		return 0;
	if (!rat_equal(a.i, b.i))
		return 0;
	return 1;
}

// ----------------------------------------------------------------
int qnfe_scan(
	qnfe_t * pout,
	char   * pin)
{
	qnfe_t g;
	char   temp[256];
	char * pcomma;

	strcpy(temp, pin);
	pcomma = strchr(temp, ',');
	if (pcomma) {
		*pcomma = 0;
		if (!rat_sscan(temp, &g.r))
			return 0;
		if (!rat_sscan(&pcomma[1], &g.i))
			return 0;
	}
	else {
		if (!rat_sscan(temp, &g.r))
			return 0;
		g.i = rat_from_int(0);
	}
	return 1;
}

// ----------------------------------------------------------------
void qnfe_print(
	qnfe_t g)
{
	rat_print(g.r);
	printf(",");
	rat_print(g.i);
}
