#include <stdio.h>
#include <stdlib.h>
#include "qnre.h"

// ----------------------------------------------------------------
qnre_t qnre_add(qnre_t a, qnre_t b, qnre_aux_t m)
{
	qnre_t c;
	c.r = a.r + b.r;
	c.i = a.i + b.i;
	return c;
}

// ----------------------------------------------------------------
qnre_t qnre_sub(qnre_t a, qnre_t b, qnre_aux_t m)
{
	qnre_t c;
	c.r = a.r - b.r;
	c.i = a.i - b.i;
	return c;
}

// ----------------------------------------------------------------
//   (ar + ai sqrt(D)) * (br + bi sqrt(D))
// = (ar * br + ai * bi * D) + (ar * bi + ai * br) sqrt(D)

qnre_t qnre_mul(qnre_t a, qnre_t b, qnre_aux_t m)
{
	qnre_t c;
	c.r = a.r * b.r + a.i * b.i * m.D;
	c.i = a.r * b.i + a.i * b.r;
	return c;
}

// ----------------------------------------------------------------
// a = qb + r:
// q = a/b (using integer division)
// r = a - qb

// (u+vi)(x-yi)   (ux+vy)+(vx-uy)i
// ------------ = ------------
// (x+yi)(x-yi)   x^2 + y^2

// ((a.r*b.r + a.i*b.i), (a.i*b.r - a.r*b.i))
// --------------------------------------------------
// b.r * b.r + b.i * b.i

qnre_t qnre_quot(qnre_t a, qnre_t b, qnre_aux_t m)
{
	qnre_t q;
	int denom = b.r * b.r + b.i * b.i;
	if (denom == 0) {
		fprintf(stderr, "qnre:  divide by zero.\n");
		exit(1);
	}
	q.r = a.r*b.r + a.i*b.i;
	q.i = a.i*b.r - a.r*b.i;
	q.r /= denom;
	q.i /= denom;
	return q;
}

qnre_t qnre_rem(qnre_t a, qnre_t b, qnre_aux_t m)
{
	qnre_t q  = qnre_quot(a, b, m);
	qnre_t qb = qnre_mul(q, b, m);
	return qnre_sub(a, qb, m);
}

// ----------------------------------------------------------------
void qnre_quot_and_rem(qnre_t a, qnre_t b, qnre_aux_t m,
	qnre_t * pquot, qnre_t * prem)
{
	qnre_t q  = qnre_quot(a, b, m);
	qnre_t qb = qnre_mul(q, b, m);
	*pquot = q;
	*prem = qnre_sub(a, qb, m);
}

// ----------------------------------------------------------------
int qnre_equal(qnre_t a, qnre_t b, qnre_aux_t m)
{
	if (a.r != b.r)
		return 0;
	else if (a.i != b.i)
		return 0;
	else
		return 1;
}

// ----------------------------------------------------------------
int qnre_scan(
	qnre_t * pout,
	char   * pin)
{
	qnre_t g;

	if (sscanf(pin, "%d,%d", &g.r, &g.i) != 2) {
		g.i = 0;
		if (sscanf(pin, "%d", &g.r) != 1) {
			return 0;
		}
	}
	return 1;
}

// ----------------------------------------------------------------
void qnre_print(
	qnre_t g)

{
	printf("%d,%d", g.r, g.i);
}
