#include <iostream.h>
#include "qpoly.h"
#include "gcd.h"
#include "power.h"

// ----------------------------------------------------------------
// Default ctor
qpoly::qpoly(void)
{
	this->degree = 0;
	this->coeffs[0] = 0;
}

// ----------------------------------------------------------------
qpoly::qpoly(rat c0)
{
	this->degree = 0;
	this->coeffs[0] = c0;
	this->normalize();
}
qpoly::qpoly(rat c0, rat c1)
{
	this->degree = 1;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->normalize();
}
qpoly::qpoly(rat c0, rat c1, rat c2)
{
	this->degree = 2;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->coeffs[2] = c2;
	this->normalize();
}
qpoly::qpoly(rat c0, rat c1, rat c2, rat c3)
{
	this->degree = 3;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->coeffs[2] = c2;
	this->coeffs[3] = c3;
	this->normalize();
}
qpoly::qpoly(rat c0, rat c1, rat c2, rat c3, rat c4)
{
	this->degree = 4;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->coeffs[2] = c2;
	this->coeffs[3] = c3;
	this->coeffs[4] = c4;
	this->normalize();
}
qpoly::qpoly(rat c0, rat c1, rat c2, rat c3, rat c4, rat c5)
{
	this->degree = 5;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->coeffs[2] = c2;
	this->coeffs[3] = c3;
	this->coeffs[4] = c4;
	this->coeffs[5] = c5;
	this->normalize();
}
qpoly::qpoly(rat c0, rat c1, rat c2, rat c3, rat c4, rat c5, rat c6)
{
	this->degree = 6;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->coeffs[2] = c2;
	this->coeffs[3] = c3;
	this->coeffs[4] = c4;
	this->coeffs[5] = c5;
	this->coeffs[6] = c6;
	this->normalize();
}
qpoly::qpoly(int arg_degree, rat * arg_coeffs)
{
	this->degree = arg_degree;
	this->check_degree();
	for (int i = 0; i <= this->degree; i++)
		this->coeffs[i] = arg_coeffs[i];
	this->normalize();
}

qpoly::qpoly(int c0)
{
	this->degree = 0;
	this->coeffs[0] = c0;
	this->normalize();
}
qpoly::qpoly(int c0, int c1)
{
	this->degree = 1;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->normalize();
}
qpoly::qpoly(int c0, int c1, int c2)
{
	this->degree = 2;
	this->coeffs[0] = c0;
	this->coeffs[1] = c1;
	this->coeffs[2] = c2;
	this->normalize();
}

// ----------------------------------------------------------------
// Copy ctor
qpoly::qpoly(const qpoly & that)
{
	this->degree = that.degree;
	for (int i = 0; i <= that.degree; i++)
		this->coeffs[i] = that.coeffs[i];
};

// ----------------------------------------------------------------
void qpoly::operator=(const qpoly& that)
{
	this->degree = that.degree;
	for (int i = 0; i <= that.degree; i++)
		this->coeffs[i] = that.coeffs[i];
}

// ----------------------------------------------------------------
int qpoly::operator==(const qpoly & that)
{
	if (this->degree != that.degree)
		return 0;
	for (int i = 0; i <= this->degree; i++)
		if (this->coeffs[i] != that.coeffs[i])
			return 0;
	return 1;
}

// ----------------------------------------------------------------
int qpoly::operator!=(const qpoly & that)
{
	return !(*this == that);
}

// ----------------------------------------------------------------
qpoly qpoly::operator+(qpoly & that)
{
	qpoly rv;
	int i;

	if (this->degree >= that.degree) {
		rv.degree = this->degree;
		for (i = 0; i <= that.degree; i++) {
			rv.coeffs[i] = this->coeffs[i] + that.coeffs[i];
		}
		for (i = that.degree + 1; i <= this->degree; i++)
			rv.coeffs[i] = this->coeffs[i];
	}
	else {
		rv.degree = that.degree;
		for (i = 0; i <= this->degree; i++) {
			rv.coeffs[i] = this->coeffs[i] + that.coeffs[i];
		}
		for (i = this->degree + 1; i <= that.degree; i++)
			rv.coeffs[i] = that.coeffs[i];
	}

	rv.normalize();
	return rv;
}

// ----------------------------------------------------------------
qpoly qpoly::operator-(void)
{
	qpoly rv(*this);
	for (int i = 0; i <= rv.degree; i++) {
		rv.coeffs[i] = -rv.coeffs[i];
	}
	return rv;
}

// ----------------------------------------------------------------
qpoly qpoly::operator-(qpoly & that)
{
	qpoly neg = -that;
	return (*this) + neg;
}

// ----------------------------------------------------------------
qpoly qpoly::operator*(qpoly & that)
{
	qpoly rv;
	int i, j;
	rat prod;

	rv.degree = this->degree + that.degree;
	if (rv.degree > QPOLY_MAX_DEGREE) {
		cerr << "qpoly::*():  Overflow in multiplication." << endl;
		exit(1);
	}

	for (i = 0; i <= rv.degree; i++) {
		rv.coeffs[i] = 0;
	}

	for (i = 0; i <= this->degree; i++) {
		for (j = 0; j <= that.degree; j++) {
			prod = this->coeffs[i] * that.coeffs[j];
			rv.coeffs[i+j] = rv.coeffs[i+j] + prod;
		}
	}
	rv.normalize();
	return rv;
}

// ----------------------------------------------------------------
qpoly qpoly::operator/(qpoly & that)
{
	qpoly quot, rem;
	this->quot_and_rem(that, quot, rem);
	return quot;
};

// ----------------------------------------------------------------
qpoly qpoly::operator%(qpoly & that)
{
	qpoly quot, rem;
	this->quot_and_rem(that, quot, rem);
	return rem;
};

// ----------------------------------------------------------------
// E.g.
// this = 1,2,3,4 (1 + 2x + 3x^2 + 4x^3)
// that = 1,1,2   (1 +  x + 2x^2)
// modulus = 7
//
//             q=4,2   r = 4,3
//        +----------
// 1,1,2  |  1,2,3,4
//        |    2,2,4 shift = 1.  4/2 mod 7 = 2.  1,1,2 * 2 = 2,2,4.
//        +----------
//        |  1 0 1
//        |  4 4 1   shift = 0.  1/2 mod 7 = 4   1,1,2 * 4 = 4,4,1
//        +----------
//        |  4 3

void qpoly::quot_and_rem(qpoly & divisor, qpoly & quot, qpoly & rem)
{
	int i;
	int max_shift = this->degree - divisor.degree;
	int shift;
	rat rem_leader;
	rat divisor_leader;
	rat multiplier;
	qpoly divisor_with_mul;
	int stop;

	if (this->degree < divisor.degree) {
		quot.degree = 0;
		quot.coeffs[0] = 0;
		rem = *this;
		return;
	}

	quot.degree = this->degree - divisor.degree;
	for (i = 0; i <= quot.degree; i++)
		quot.coeffs[i] = 0;

	rem = *this;
	divisor_leader = divisor.coeffs[divisor.degree];
	for (shift = max_shift; shift >= 0; shift--) {
		rem_leader = rem.coeffs[shift + divisor.degree];
		multiplier = rem_leader / divisor_leader;
		divisor_with_mul = divisor * multiplier;
		quot.coeffs[shift] = multiplier;

		stop = shift + divisor_with_mul.degree;
		for (i = shift; i <= stop; i++) {
			rem.coeffs[i] = rem.coeffs[i] -
				divisor_with_mul.coeffs[i-shift];
		}
	}
	rem.normalize();
}

// ----------------------------------------------------------------
// Input: a, b
//   a is *this.
//   b is the first argument.
// Output: GCD g, and m, n such that am + bn = d

qpoly qpoly::ext_gcd(
	qpoly & b, // 2nd input; *this is the 1st.
	qpoly & m, // Output coefficient on a
	qpoly & n) // Output coefficient on b
{
	qpoly mprime, nprime, c, q, r, t, qm, qn;
	qpoly d;   // Return value.

	// Initialize
	mprime = qpoly(1);
	n      = qpoly(1);
	m      = qpoly(0);
	nprime = qpoly(0);
	c = *this;
	d = b;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		c.quot_and_rem(d, q, r);
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		if (r == 0) {
			break;
		}

		// Recycle
		c = d;
		d = r;

		t = mprime;
		mprime = m;
		qm = q * m;
		m = t - qm;

		t = nprime;
		nprime = n;
		qn = q * n;
		n = t - qn;
	}
	return d;
}

// ----------------------------------------------------------------
// Input: a, b
//   a is *this.
//   b is the first argument.
// Output: GCD g.

qpoly qpoly::gcd(
	qpoly & b) // 2nd input; *this is the 1st.
{
	qpoly c, q, r;
	qpoly d;   // Return value.

	// Initialize
	c = *this;
	d = b;

	// xxx temp
	if (d == 0)
		return c;

	while (1) {

		// Divide
		// q = c / d, r = c % d
		c.quot_and_rem(d, q, r);
		// Note:  now c = qd + r and 0 <= r < d
		// Remainder zero?
		if (r == 0) {
			break;
		}

		// Recycle
		c = d;
		d = r;
	}
	return d;
}

// ----------------------------------------------------------------
qpoly qpoly::operator*(rat m)
{
	int i;
	qpoly rv(*this);
	for (i = 0; i <= this->degree; i++)
		rv.coeffs[i] = this->coeffs[i] * m;
	return rv;
}

// ----------------------------------------------------------------
// Polynomial divided by scalar.

qpoly qpoly::operator/(rat m)
{
	int i;
	qpoly rv(*this);

	for (i = 0; i <= this->degree; i++)
		rv.coeffs[i] = this->coeffs[i] / m;
	return rv;
}

// ----------------------------------------------------------------
ostream & operator<<(
	ostream & s,
	const qpoly & r)
{
	int i;

	for (i = 0; i <= r.degree; i++) {
		s << r.coeffs[i];
		if (i < r.degree)
			s << ",";
	}

//	s << "(";
//	for (i = 0; i <= r.degree; i++) {
//		if ((i != 0) && (r.coeffs[i].is_positive()))
//			s << "+" << r.coeffs[i];
//		else
//			s << r.coeffs[i];
//		if (i == 0)
//			;
//		else if (i == 1)
//			s << "x";
//		else
//			s << "x^" << i;
//	}
//	s << ")";

	return s;
}

// ----------------------------------------------------------------
void qpoly::check_degree(void)
{
	if (this->degree < 0) {
		cerr << "qpoly::qpoly():  degree must be non-negative; "
			"got " << this->degree << endl;
		exit(1);
	}
	if (this->degree > QPOLY_MAX_DEGREE) {
		cerr << "qpoly::qpoly():  degree must be <= "
			<< QPOLY_MAX_DEGREE
			<< "; got " << this->degree << endl;
		exit(1);
	}
}

// ----------------------------------------------------------------
int qpoly::find_actual_degree(void)
{
	int i;
	for (i = this->degree; i >= 0; i--) {
		if (this->coeffs[i] != 0)
			return i;
	}
	return 0;
}

// ----------------------------------------------------------------
void qpoly::normalize(void)
{
	this->degree = this->find_actual_degree();
}

// ================================================================
// Field methods.

// ----------------------------------------------------------------
// E.g. a.mul(b, r) for a * b mod r.
qpoly qpoly::mul(qpoly & b, qpoly & r)
{
	qpoly c = *this * b;
	c.reduce(r);
	return c;
}

// ----------------------------------------------------------------
// E.g. a.reduce(r) to reduce a mod r.
void qpoly::reduce(qpoly & r)
{
	qpoly quot, rem;
	this->quot_and_rem(r, quot, rem);
	*this = rem;
}

// ----------------------------------------------------------------
// E.g. a.div(b, r) for a / b mod r.
qpoly qpoly::div(qpoly & b, qpoly & r)
{
	qpoly binv = b.recip(r); // 1/b mod r.
	return this->mul(binv, r);   // a * (1/b) mod r.
};

// ----------------------------------------------------------------
// E.g. b.recip(r) for 1/b mod r.
//
// Finite-field division uses extended Euclidean GCD, with ring division
// (quotient and remainder) as a substep.
//
// a / b is a * 1/b.  To compute 1/b, find the GCD d.  (Since the reductor
// is irreducible -- although we don't test this on every division -- and
// if b is monic, the GCD had better be 1).  Then, compute 1 = m*b + n*r.
// Since r = 0 mod r, 1 = m*b and therefore m = 1/b.  The extended Euclidean
// algorithm (see Knuth vol. 1) computes d, m and n all at once.

qpoly qpoly::recip(qpoly & r)
{
	qpoly g, m, n;
	g = this->ext_gcd(r, m, n);
	return m / g.coeffs[0];
};

// ----------------------------------------------------------------
qpoly qpoly::exp(int power, qpoly & r)
{
	unsigned upower = power;
	qpoly p2 =  *this;
	qpoly rv(1);

	while (upower != 0) {
		if (upower & 1)
			rv = rv.mul(p2, r);
		upower >>= 1;
		p2 = p2.mul(p2, r);
	}
	return rv;
}
