// Elements of the ring Q[x].

#ifndef QPOLY_H
#define QPOLY_H

#include "rat.h"

#define QPOLY_MAX_DEGREE 64

typedef struct _qpoly_t {
	rat_t coeffs[QPOLY_MAX_DEGREE+1];
	int degree;
} qpoly_t;

// ----------------------------------------------------------------
qpoly_t qpoly_add(
	qpoly_t * pin1,
	qpoly_t * pin2);

qpoly_t qpoly_sub(
	qpoly_t * pin1,
	qpoly_t * pin2);

qpoly_t qpoly_usub(
	qpoly_t * pin);

qpoly_t qpoly_mul(
	qpoly_t * pin1,
	qpoly_t * pin2);

qpoly_t qpoly_div(
	qpoly_t * pin1,
	qpoly_t * pin2);

qpoly_t qpoly_mod(
	qpoly_t * pin1,
	qpoly_t * pin2);

void qpoly_quot_and_rem(
	qpoly_t * pdividend,
	qpoly_t * pdivisor,
	qpoly_t * pquot,
	qpoly_t * prem);

qpoly_t qpoly_ext_gcd(
	qpoly_t * pa,  // 1st input.
	qpoly_t * pb,  // 2nd input.
	qpoly_t * pm,  // Output coefficient on a
	qpoly_t * pn); // Output coefficient on b

qpoly_t qpoly_gcd(
	qpoly_t * pa,  // 1st input.
	qpoly_t * pb); // 2nd input.

qpoly_t qpoly_power(
	qpoly_t * px,
	unsigned  power);

int qpoly_equal(
	qpoly_t * pin1,
	qpoly_t * pin2);

// ----------------------------------------------------------------
// Construction

qpoly_t qpoly_from_coeffs(
	int     degree,
	rat_t * coeffs);

qpoly_t qpoly_from_coeffs0(
	rat_t   coeff0);

qpoly_t qpoly_from_coeffs1(
	rat_t   coeff1,
	rat_t   coeff0);

qpoly_t qpoly_from_coeffs2(
	rat_t   coeff2,
	rat_t   coeff1,
	rat_t   coeff0);

qpoly_t qpoly_binomial_from_coeffs(
	rat_t   coeffa,
	int     dega,
	rat_t   coeffb,
	int     degb);

// ----------------------------------------------------------------
// I/O

void qpoly_print(
	qpoly_t * pin);

void qpoly_comma_print(
	qpoly_t * pin);

void qpoly_x_print(
	qpoly_t * pin);

int  qpoly_scan(
	qpoly_t * pout,
	char    * pin);

// ----------------------------------------------------------------
// Factorization, evaluation and classification.

rat_t qpoly_eval(
	qpoly_t * pz,
	rat_t     x);

int qpoly_is_zero(
	qpoly_t * pz);

int qpoly_is_one(
	qpoly_t * pz);

#endif // QPOLY_H
