// xxx irred check

// xxx static qpoly::from_string() method.
// test out irred <==> prime-as-int conjecture.

// make-monic method?

#ifndef QPOLY_H
#define QPOLY_H

#include <iostream.h>
#include <stdlib.h>

#include "rat.h"

#define QPOLY_MAX_DEGREE 36

// ================================================================
class qpoly
{
public:
	// Default constructor; value is set to zero.
	qpoly(void);

	qpoly(rat c0);
	qpoly(rat c0, rat c1);
	qpoly(rat c0, rat c1, rat c2);
	qpoly(rat c0, rat c1, rat c2, rat c3);
	qpoly(rat c0, rat c1, rat c2, rat c3, rat c4);
	qpoly(rat c0, rat c1, rat c2, rat c3, rat c4, rat c5);
	qpoly(rat c0, rat c1, rat c2, rat c3, rat c4, rat c5, rat c6);
	qpoly(int arg_degree, rat * coeffs);

	qpoly(int c0);
	qpoly(int c0, int c1);
	qpoly(int c0, int c1, int c2);

	// Copy ctor
	qpoly::qpoly(const qpoly & that);

	// Dtor not needed.
	// ~qpoly(void);

	// Operators
	void   operator=(const qpoly& that);
	int    operator==(const qpoly & that);
	int    operator!=(const qpoly & that);

	qpoly operator+(qpoly & that);
	qpoly operator-(qpoly & that);
	qpoly operator-(void);
	qpoly operator*(qpoly & that);
	qpoly operator/(qpoly & that);
	qpoly operator%(qpoly & that);
	void  quot_and_rem(qpoly & that, qpoly & quot, qpoly & rem);

	// E.g. g = a.ext_gcd(b, m, n).  a and b are inputs; g, m and n are
	// outputs.  g is the GCD; m and n are polynomials such that
	// am + bn = g.  This is used by the field-element reciprocator.
	qpoly ext_gcd(qpoly & b, qpoly & m, qpoly & n);

	// E.g. g = a.gcd(b).
	qpoly gcd(qpoly & b);

	qpoly operator*(rat m);      // Polynomial times scalar
	qpoly operator/(rat m);      // Polynomial divided by scalar

	friend ostream & operator<<(ostream & s, const qpoly & r);

	int    get_degree(void) { return this->degree; };

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Field operations.
	// E.g. a.mul(b, r) for a * b rat r.
	qpoly mul(qpoly & b, qpoly & r);

	// E.g. a.div(b, r) for a / b rat r.
	qpoly div(qpoly & b, qpoly & r);

	// E.g. b.recip(r) for 1 / b rat r.
	qpoly recip(qpoly & r);

	// E.g. b.exp(10, r) for b^10 rat r.  Exponentiation is done using
	// a logarithmic-time algorithm.
	qpoly exp(int power, qpoly & r);

protected:
	int    degree;
	rat    coeffs[QPOLY_MAX_DEGREE + 1];

	void   check_degree(void);
	void   normalize(void);
	int    find_actual_degree(void);

	//  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// For field use.
	// E.g. a.reduce(r) to reduce a rat r.
	// Could be written as an operator%, but I want it to be an in-place
	// operation for the sake of (a negligible gain in) efficiency.
	void   reduce(qpoly & r);
};

#endif // QPOLY_H
