#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "testing.h"

#define E 0
#define A 1
#define B 2
#define C 3

static char * names[] = {
	"e",
	"a",
	"b",
	"c",
};
#define NUM_NAMES (sizeof(names)/sizeof(names[0]))

#if 0
static int cayley_table[NUM_NAMES][NUM_NAMES] = {
	{ E, A, B, C },
	{ A, B, C, E },
	{ B, C, E, A },
	{ C, E, A, B },
};
#endif

#if 0
static int cayley_table[NUM_NAMES][NUM_NAMES] = {
	{ E, A, B, C },
	{ A, E, C, B },
	{ B, C, E, A },
	{ C, B, A, E },
};
#endif

// ----------------------------------------------------------------

#if 0
static int cayley_table[NUM_NAMES][NUM_NAMES] = {
	{ E, A, B, C },
	{ A, C, E, B },
	{ B, E, C, A },
	{ C, B, A, E },
};
#endif

#if 0
static int cayley_table[NUM_NAMES][NUM_NAMES] = {
	{ E, A, B, C },
	{ A, E, C, B },
	{ B, C, A, E },
	{ C, B, E, A },
};
#endif

#if 1
static int cayley_table[NUM_NAMES][NUM_NAMES] = {
	{ E, A, B, C },
	{ A, E, C, B },
	{ B, C, E, A },
	{ C, B, A, E },
};
#endif

// e a b c    % e a b c
// a e c b    % a e c b
// b c a e    % b c e a
// c b e a    % c b a e


// ----------------------------------------------------------------
static void check(int a)
{
	if ((a < 0) || (a >= NUM_NAMES)) {
		fprintf(stderr, "Test group: %d out of bounds 0-%d.\n",
			a, NUM_NAMES - 1);
		exit(1);
	}
}

// ----------------------------------------------------------------
int testing_mul(int a, int b)
{
	check(a);
	check(b);
	return cayley_table[a][b];
}

// ----------------------------------------------------------------
int testing_equal(int a, int b)
{
	return (a == b);
}

// ----------------------------------------------------------------
static void testing_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int c = testing_mul(a, b);
	*(int *)pvout = c;
}

// ----------------------------------------------------------------
static int testing_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	return (*(int *)pvin1 == *(int *)pvin2);
}

// ----------------------------------------------------------------
static int testing_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	return (*(int *)pvin1 < *(int *)pvin2);
}

// ----------------------------------------------------------------
static void testing_vprint(
	void * pvin,
	void * pvaux)
{
	int a = *(int *)pvin;
	check(a);
	printf("%s", names[a]);
}

// ----------------------------------------------------------------
int testing_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	int i;

	if (sizeof(int) > max_elt_bytes) {
		fprintf(stderr, "testing:  Need %d bytes for element.\n",
			sizeof(int));
		exit(1);
	}

	for (i = 0; i < NUM_NAMES; i++) {
		if (strcmp(pin, names[i]) == 0) {
			*(int *)pvout = i;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
int testing_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (argc != 0)
		return 0;
	return 1;
}


// ----------------------------------------------------------------
int  testing_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	int i;
	int m = NUM_NAMES;

	if (m > max_group_bytes/sizeof(int))
		return 0;
	for (i = 0; i < m; i++)
		pelements[i] = i;
	*pnum_elements = m;
	return 1;
}

// ================================================================
group_spec_t testing_group_spec = {
	element_size  : sizeof(int),
	naux          : 0,
	pvop          : testing_vmul,
	pvequal       : testing_vequal,
	pvless        : testing_vless,
	pvprint       : testing_vprint,
	pvauxprint    : 0,
	pvscan        : testing_vscan,
	pvauxscan     : testing_vauxscan,
	pfill_group   : testing_fill_group,
	desc          : "testing"
};
