#ifndef UBIG_H
#define UBIG_H

#include <stdio.h>
#include "cmps.h"

// ----------------------------------------------------------------
// 16 16-bit integers => 256 bits (~= 1e77).  A single integer of this
// size pretty much fills a line in your terminal window (assuming 80
// columns).
#define UBIG_NUM_PARTS 18//64
#define UBIG_PART_MASK  15
#define UBIG_PART_LOG    4

typedef unsigned short ubig_part_t;
typedef unsigned       ubig_pprod_t;
#define UBIG_PART_MSB  0x8000
#define UBIG_CARRY_BIT 0x10000

#define UBIG_NYBBLES_PER_PART (sizeof(ubig_part_t) * 2)
#define UBIG_NUM_NYBBLES      (UBIG_NUM_PARTS * UBIG_NYBBLES_PER_PART)

#define UBIG_BITS_PER_PART    (sizeof(ubig_part_t) * 8)
#define UBIG_NUM_BITS         (UBIG_NUM_PARTS * UBIG_BITS_PER_PART)

// ----------------------------------------------------------------
typedef struct _ubig {
	ubig_part_t parts[UBIG_NUM_PARTS];
} ubig;

typedef struct _ubig_opts_t {
	int  hex_in;
	int  hex_out;
	int  zero_fill;
	int  underscore_out;
} ubig_opts_t;

// ----------------------------------------------------------------
void ubig_print(FILE * fp, ubig * pc, ubig_opts_t * pubig_opts);
void ubig_xprint(ubig * pc);
void ubig_dprint(ubig * pc);
int  ubig_fscan (FILE * fp,  ubig * pc, ubig_opts_t * pubig_opts);
int  ubig_sscan (char * pin, ubig * pc, ubig_opts_t * pubig_opts);
int  ubig_sdscan(char * pin, ubig * pc);

ubig ubig_or  (ubig * pa, ubig * pb);
ubig ubig_xor (ubig * pa, ubig * pb);
ubig ubig_and (ubig * pa, ubig * pb);

void ubig_set_bit(ubig * pc, unsigned pos);
unsigned ubig_get_bit(ubig * pc, unsigned pos);
unsigned ubig_find_msb(ubig * pc);

ubig ubig_lsh (ubig * pa, ubig * pb);
ubig ubig_ilsh(ubig * pa, unsigned shamt);
ubig ubig_rsh (ubig * pa, ubig * pb);
ubig ubig_irsh(ubig * pa, unsigned shamt);

ubig ubig_add (ubig * pa, ubig * pb);
ubig ubig_sub (ubig * pa, ubig * pb);
ubig ubig_mul (ubig * pa, ubig * pb);
ubig ubig_div (ubig * pa, ubig * pb);
ubig ubig_mod (ubig * pa, ubig * pb);

ubig ubig_power(ubig * pa, ubig * pb);
ubig ubig_ipower(ubig * pa, unsigned e);
ubig ubig_uadd(ubig * pa);
ubig ubig_usub(ubig * pa);
ubig ubig_unot(ubig * pa);
void ubig_incr(ubig * pc);
void ubig_incr2(ubig * pc);
void ubig_decr(ubig * pc);

int  ubig_cmp(ubig * pa, int cmp, ubig * pb);
int  ubig_lt(ubig * pa, ubig * pb);
int  ubig_le(ubig * pa, ubig * pb);
int  ubig_gt(ubig * pa, ubig * pb);
int  ubig_ge(ubig * pa, ubig * pb);
int  ubig_eq(ubig * pa, ubig * pb);
int  ubig_ne(ubig * pa, ubig * pb);

// Suitable for qsort(3):
int ubig_qcmp(const void * pv1, const void * pv2);

ubig ubig_from_unsigned(unsigned i);
int  ubig_equals_unsigned(ubig * pa, unsigned i);

#endif // UBIG_H

