#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "v4.h"

static char * names[] = {
	"e", // V4E
	"a", // V4A
	"b", // V4B
	"c", // V4C
};

// ----------------------------------------------------------------
int v4_mul(int a, int b)
{
	// This is true since V4 ~= (Z2 x Z2, +).
	return (a ^ b) & 3;
}

// ----------------------------------------------------------------
int v4_equal(int a, int b)
{
	return (a == b);
}

// ----------------------------------------------------------------
static void v4_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int c = v4_mul(a, b);
	*(int *)pvout = c;
}

// ----------------------------------------------------------------
static int v4_vequal(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	return (*(int *)pvin1 == *(int *)pvin2);
}

// ----------------------------------------------------------------
static int v4_vless(
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	return (*(int *)pvin1 < *(int *)pvin2);
}

// ----------------------------------------------------------------
static void v4_vprint(
	void * pvin,
	void * pvaux)
{
	int a = *(int *)pvin;
	printf("%s", names[a & 3]);
}

// ----------------------------------------------------------------
int v4_vscan(
	void * pvout,
	char * pin,
	void * pvaux,
	int    max_elt_bytes)
{
	int i;

	if (sizeof(int) > max_elt_bytes) {
		fprintf(stderr, "v4:  Need %d bytes for element.\n",
			sizeof(int));
		exit(1);
	}

	for (i = 0; i < 4; i++) {
		if (strcmp(pin, names[i]) == 0) {
			*(int *)pvout = i;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
int v4_vauxscan(
	void *  pvaux,
	int     argc,
	char ** argv,
	int     max_aux_bytes)
{
	if (argc != 0)
		return 0;
	return 1;
}


// ----------------------------------------------------------------
int  v4_fill_group(
	void * pvelements,
	void * pvaux,
	int    max_group_bytes,
	int  * pnum_elements)
{
	int * pelements = (int *) pvelements;
	int i;
	int m = 4;

	if (m > max_group_bytes/sizeof(int))
		return 0;
	for (i = 0; i < m; i++)
		pelements[i] = i;
	*pnum_elements = m;
	return 1;
}

// ================================================================
group_spec_t v4_group_spec = {
	element_size  : sizeof(int),
	naux          : 0,
	pvop          : v4_vmul,
	pvequal       : v4_vequal,
	pvless        : v4_vless,
	pvprint       : v4_vprint,
	pvauxprint     : 0,
	pvscan        : v4_vscan,
	pvauxscan      : v4_vauxscan,
	pfill_group   : v4_fill_group,
	desc          : "v4"
};
