// Elements of the ring Z_2[x], implemented using a single 64-bit integer.

#ifndef Z2IPOLY_H
#define Z2IPOLY_H

#include "group.h"
#include "factorize.h"

// ----------------------------------------------------------------
typedef unsigned long long z2ipoly_t;
#define Z2IPOLY_NUM_BITS (8 * sizeof(z2ipoly_t))
#define Z2IPOLY_MAX_DEGREE (Z2IPOLY_NUM_BITS - 1)

#define Z2IPOLY_BIT_AT(p,j)    ( ((p) >> (j)) & 1 )
#define Z2IPOLY_SET_BIT(p,j)   ( (p) |=  ((z2ipoly_t)1 << (j)) )
#define Z2IPOLY_CLEAR_BIT(p,j) ( (p) &= ~((z2ipoly_t)1 << (j)) )

// ----------------------------------------------------------------
z2ipoly_t z2ipoly_add(
	z2ipoly_t a,
	z2ipoly_t b);

z2ipoly_t z2ipoly_sub(
	z2ipoly_t a,
	z2ipoly_t b);

z2ipoly_t z2ipoly_mul(
	z2ipoly_t a,
	z2ipoly_t b);

z2ipoly_t z2ipoly_div(
	z2ipoly_t a,
	z2ipoly_t b);

z2ipoly_t z2ipoly_mod(
	z2ipoly_t a,
	z2ipoly_t b);

void    z2ipoly_quot_and_rem(
	z2ipoly_t   a,
	z2ipoly_t   b,
	z2ipoly_t * pquot,
	z2ipoly_t * prem);

z2ipoly_t z2ipoly_gcd(
	z2ipoly_t a,
	z2ipoly_t b);

z2ipoly_t z2ipoly_ext_gcd(
	z2ipoly_t   a,
	z2ipoly_t   b,
	z2ipoly_t * pm,
	z2ipoly_t * pn);

z2ipoly_t z2ipoly_power(
	z2ipoly_t a,
	unsigned  power);

// Assumes the input is already known to be a square.
z2ipoly_t z2ipoly_square_root(
	z2ipoly_t in);

// ----------------------------------------------------------------
// I/O

void z2ipoly_print(
	z2ipoly_t a);

void z2ipoly_dprint(
	z2ipoly_t a,
	int       min_deg);

void z2ipoly_binary_print(
	z2ipoly_t a);

void z2ipoly_binary_width_print(
	z2ipoly_t a,
	int       width);

void z2ipoly_pretty_print(
	z2ipoly_t a);

int z2ipoly_scan(
	z2ipoly_t * pz,
	char    * pin);

// ----------------------------------------------------------------
// Factorization, evaluation and classification.

int z2ipoly_eval(
	z2ipoly_t r,
	int x);

z2ipoly_t z2ipoly_compose(
	z2ipoly_t r,
	z2ipoly_t z);

z2ipoly_t z2ipoly_deriv(
	z2ipoly_t z);

int z2ipoly_is_rootless(
	z2ipoly_t r);

int z2ipoly_count_bits(
	z2ipoly_t z);

int z2ipoly_get_degree(
	z2ipoly_t a);

// Used to check for overflow.
void z2ipoly_check_degree(
	int degree);

// ----------------------------------------------------------------
// For use with qsort(3).
int z2ipoly_qcmp(
	const void * pv1,
	const void * pv2);

#endif // Z2IPOLY_H
