// Elements of the ring Z_2[x] / (r(x)) (a field when r(x) is irreducible).

#ifndef Z2IPOLYMOD_H
#define Z2IPOLYMOD_H

#include "z2ipoly.h"

// ----------------------------------------------------------------
z2ipoly_t z2ipolymod_mul(
	z2ipoly_t a,
	z2ipoly_t b,
	z2ipoly_t r);

z2ipoly_t z2ipolymod_div(
	z2ipoly_t a,
	z2ipoly_t b,
	z2ipoly_t rr);

z2ipoly_t z2ipolymod_recip(
	z2ipoly_t a,
	z2ipoly_t r);

z2ipoly_t z2ipolymod_power(
	z2ipoly_t z,
	unsigned power,
	z2ipoly_t  r);

z2ipoly_t z2ipolymod_power64(
	z2ipoly_t z,
	unsigned long long power,
	z2ipoly_t  r);

z2ipoly_t z2ipolymod_compose(
	z2ipoly_t r,
	z2ipoly_t z,
	z2ipoly_t m);

int z2ipolymod_element_order(
	z2ipoly_t z,
	z2ipoly_t r);

z2ipoly_t z2ipolymod_find_generator(
	z2ipoly_t r);

int z2ipolymod_log(
	z2ipoly_t z,  // Input
	z2ipoly_t g,  // Base, assumed (!) primitive mod r.
	z2ipoly_t r); // Modulus for FF arithmetic

// These two are not strictly necessary:  there are already z2ipoly_add and
// z2ipoly_sub, which take p as their final argument.  The monic irreducible
// contained in the aux struct isn't used for addition or subtraction.
// However, these functions give the pleasant result of a uniform interface for
// z2ipolymod functions, which is useful when an application makes tables of
// function pointers.

z2ipoly_t z2ipolymod_add(
	z2ipoly_t a,
	z2ipoly_t b,
	z2ipoly_t r);

z2ipoly_t z2ipolymod_sub(
	z2ipoly_t a,
	z2ipoly_t b,
	z2ipoly_t r);

z2ipoly_t z2ipolymod_trace(
	z2ipoly_t a,
	z2ipoly_t r);

z2ipoly_t z2ipolymod_norm(
	z2ipoly_t a,
	z2ipoly_t r);

// ----------------------------------------------------------------
void z2ipolymod_print(
	z2ipoly_t a,
	z2ipoly_t r);

int z2ipolymod_auxscan(
	z2ipoly_t * pr,
	char    * pin);

// ----------------------------------------------------------------
// Finds the numerically lowest primitive monic irreducible for a given
// Galois field.
z2ipoly_t z2ipoly_find_irreducible(
	int m);

int z2ipoly_is_irreducible(
	z2ipoly_t r);

// Assumes r is known to be irreducible.
int z2ipoly_is_primitive(
	z2ipoly_t r);

// z2ipoly_is_primitive() factors 2^m-1 each time it is called.  If you are
// calling it multiple times with the same m, you can avoid redundant effort
// by factoring 2^m-1 yourself and passing in the factorization.
int z2ipoly_is_primitive_f(
	z2ipoly_t r,
	unsigned long long pm1,
	factor_info_ull_t * pfinfo);

// ----------------------------------------------------------------
// Finds all irreducibles of degree d dividing n.  That is, this is the
// factorization of X^(p^n) - X, without actually factoring (or requiring
// storage for) that huge-degree polynomial.

void z2ipoly_find_dirreds(
	unsigned    n,
	z2ipoly_t * pirred_factors,
	unsigned  * pnum_irred_factors,
	unsigned    max_irred_factors);

// ----------------------------------------------------------------
// The caller must free the return value.
z2ipoly_t * z2ipolymod_get_elements(
	int n,
	int * pnel);

// ----------------------------------------------------------------
// GF(2^m) with addition; GF(2^m)^* with multiplication.
extern group_spec_t z2ipolymodadd_group_spec;
extern group_spec_t z2ipolymodmul_group_spec;

#endif // Z2IPOLYMOD_H
