#include <stdio.h>
#include <stdlib.h>
#include "z2n_convert.h"
#include "z2npoly_factor.h"

// ----------------------------------------------------------------
int z2n_convert_find_generators(
	z2poly_t * prs, // Smaller field
	z2poly_t * prl, // Larger field
	z2poly_t * pgs,
	z2poly_t * pgl)
{
	z2poly_t * roots;
	z2poly_t ms;
	int ds;

	*pgs = z2polymod_find_generator(prs);

	// Get min poly for gs in smaller field
	ms = z2poly_get_min_poly(pgs, prs);

	// Factor it over larger field
	roots = z2poly_get_roots(&ms, prl);
	ds = z2poly_get_degree(&ms);

	// Take any root; aesthetically, the lexically smallest is preferred.
	qsort(roots, ds, sizeof(z2poly_t), z2poly_qcmp);
	*pgl = roots[0];
	free(roots);
	return 1;
}

// ----------------------------------------------------------------
int z2n_convert_scalar(
	z2poly_t * pu1,
	z2poly_t * pr1,
	z2poly_t * pg1,
	z2poly_t * pu2,
	z2poly_t * pr2,
	z2poly_t * pg2)
{
	int l1 = 0, l2 = 0;

	if (z2poly_is_zero(pu1)) {
		*pu2 = z2poly_from_base_rep(0);
		return 1;
	}

	l1 = z2polymod_log(pu1, pg1, pr1);
	l2 = l1;
	*pu2 = z2polymod_power(pg2, l2, pr2);

	return 1;
}

// ----------------------------------------------------------------
int z2n_convert_poly(
	z2npoly_t * pf1,
	z2poly_t  * pr1,
	z2poly_t  * pg1,
	z2npoly_t * pf2,
	z2poly_t  * pr2,
	z2poly_t  * pg2)
{
	int i;
	z2poly_t u1, u2;

	*pf2 = *pf1;

	for (i = 0; i <= pf2->degree; i++) {
		u1 = pf1->coeffs[i];
		u1 = z2poly_mod(&u1, pr1);
		if (!z2n_convert_scalar(&u1, pr1, pg1, &u2, pr2, pg2))
			return 0;
		pf2->coeffs[i] = u2;
	}
	return 1;
}
