#include <stdio.h>
#include <stdlib.h>
#include "z2ni_convert.h"
#include "z2nipoly_factor.h"

// ----------------------------------------------------------------
int z2ni_convert_find_generators(
	z2ipoly_t   rs, // Smaller field
	z2ipoly_t   rl, // Larger field
	z2ipoly_t * pgs,
	z2ipoly_t * pgl)
{
	z2ipoly_t * roots;
	z2ipoly_t ms;
	int ds;

	*pgs = z2ipolymod_find_generator(rs);

	// Get min poly for gs in smaller field
	ms = z2ipoly_get_min_poly(*pgs, rs);

	// Factor it over larger field
	roots = z2ipoly_get_roots(ms, rl);
	ds = z2ipoly_get_degree(ms);

	// Take any root; aesthetically, the lexically smallest is preferred.
	qsort(roots, ds, sizeof(z2ipoly_t), z2ipoly_qcmp);
	*pgl = roots[0];
	free(roots);
	return 1;
}

// ----------------------------------------------------------------
int z2ni_convert_scalar(
	z2ipoly_t   u1,
	z2ipoly_t   r1,
	z2ipoly_t   g1,
	z2ipoly_t * pu2,
	z2ipoly_t   r2,
	z2ipoly_t   g2)
{
	int l1 = 0, l2 = 0;

	if (u1 == 0LL) {
		*pu2 = 0LL;
		return 1;
	}

	l1 = z2ipolymod_log(u1, g1, r1);
	l2 = l1;
	*pu2 = z2ipolymod_power(g2, l2, r2);

	return 1;
}

// ----------------------------------------------------------------
int z2ni_convert_poly(
	z2nipoly_t * pf1,
	z2ipoly_t    r1,
	z2ipoly_t    g1,
	z2nipoly_t * pf2,
	z2ipoly_t    r2,
	z2ipoly_t    g2)
{
	int i;
	z2ipoly_t u1, u2;

	*pf2 = *pf1;

	for (i = 0; i <= pf2->degree; i++) {
		u1 = pf1->coeffs[i];
		u1 = z2ipoly_mod(u1, r1);
		if (!z2ni_convert_scalar(u1, r1, g1, &u2, r2, g2))
			return 0;
		pf2->coeffs[i] = u2;
	}
	return 1;
}
