// Elements of the ring F_3^n[x] (polynomials with coefficients in a
// Galois field).

#ifndef Z2NPOLY_H
#define Z2NPOLY_H

#include "z2polymod.h"

// ----------------------------------------------------------------
#define Z2NPOLY_MAX_DEGREE Z2POLY_MAX_DEGREE
typedef struct _z2npoly_t {
	z2poly_t coeffs[Z2NPOLY_MAX_DEGREE+1];
	int degree;
} z2npoly_t;

// ----------------------------------------------------------------
z2npoly_t z2npoly_add(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t  * pr);

z2npoly_t z2npoly_sub(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t  * pr);

z2npoly_t z2npoly_mul(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t  * pr);

void z2npoly_quot_and_rem(
	z2npoly_t * pdividend,
	z2npoly_t * pdivisor,
	z2npoly_t * pquot,
	z2npoly_t * prem,
	z2poly_t  * pr);

z2npoly_t z2npoly_div(
	z2npoly_t * pa,
	z2npoly_t * pb,
	z2poly_t  * pr);

z2npoly_t z2npoly_mod(
	z2npoly_t * pa,
	z2npoly_t * pb,
	z2poly_t  * pr);

z2npoly_t z2npoly_power(
	z2npoly_t * pin,
	int       power,
	z2poly_t  * pr);

z2npoly_t z2npoly_square_root(
	z2npoly_t * pin,
	z2poly_t  * pr);

z2npoly_t z2npoly_gcd(
	z2npoly_t * pa,
	z2npoly_t * pb,
	z2poly_t  * pr);

z2npoly_t z2npoly_extend_z2poly(
	z2poly_t  * pz2);

z2poly_t z2poly_unextend_z2npoly(
	z2npoly_t * pznp,
	z2poly_t  * pr);

// Example:
// * a_aux   = a^2 + a + 1 = 0x07 which defines an F_2^2.
// * b_red   = b^6 + b + 1 = 0x43 which defines an F_2^6.
// * ab_poly = x^3 + x^2 + (a+1) x + a = 1:1:3:2
//   which is a cubic polynomial in F_2^2[x].
// * ab_exp  = 21 since (2^6-1) / (2^2-1) = 21.
//
// This function substitutes x with b and a with b^exp, then reduces mod
// b_red.  The above example ab_poly reduces to 0 mod b^6 + b + 1.

z2poly_t z2npoly_flatten(
	z2npoly_t * pab_poly,
	z2poly_t  * pa_red,
	z2poly_t  * pb_red,
	int         ab_exp);

// For general use, just examine pin->degree.  This function is used
// to recompute the degree after the poly has changed.
int z2npoly_find_degree(
	z2npoly_t * pin,
	int         max_deg);

// Used to check for overflow in polynomial multiplication.
void z2npoly_check_degree(
	int degree);

z2npoly_t z2npoly_from_coeffs0(
	z2poly_t * pcoeff0);

z2npoly_t z2npoly_from_coeffs1(
	z2poly_t * pcoeff1,
	z2poly_t * pcoeff0);

z2npoly_t z2npoly_from_int(
	int a);

z2npoly_t z2npoly_from_base_rep(
	int rep,
	int intermediate_degree);

z2npoly_t z2npoly_from_two_ints(
	int a1,
	int a0);

// ----------------------------------------------------------------
z2npoly_t z2npoly_deriv(
	z2npoly_t * pz,
	z2poly_t  * pr);

z2poly_t z2poly_get_min_poly(
	z2poly_t * pz,
	z2poly_t * pr);

// ----------------------------------------------------------------
int z2npoly_is_zero(z2npoly_t * pz);
int z2npoly_is_one (z2npoly_t * pz);

void z2npoly_make_monic(
	z2npoly_t * pz,
	z2poly_t  * pr);

// ----------------------------------------------------------------
int z2npoly_equal(
	z2npoly_t * pin1,
	z2npoly_t * pin2,
	z2poly_t  * pr);

// Fpn[x] is *not* totally ordered.  Nonetheless, these functions are
// handy for looping and sorting.
int z2npoly_lt(z2npoly_t * pa, z2npoly_t * pb);
int z2npoly_gt(z2npoly_t * pa, z2npoly_t * pb);
int z2npoly_le(z2npoly_t * pa, z2npoly_t * pb);
int z2npoly_ge(z2npoly_t * pa, z2npoly_t * pb);
int z2npoly_eq(z2npoly_t * pa, z2npoly_t * pb);
int z2npoly_ne(z2npoly_t * pa, z2npoly_t * pb);

int z2npoly_cmp(
	z2npoly_t * pa,
	int         cmp, // see cmps.h
	z2npoly_t * pb);

// ----------------------------------------------------------------
void z2npoly_print(
	z2npoly_t * pin,
	z2poly_t  * pr);

int  z2npoly_scan(
	z2npoly_t * pvout,
	char      * pin,
	z2poly_t  * pr);

#endif // Z2NPOLY_H
