// Elements of the ring Z_2[x] / (r(x)) (a field when r(x) is irreducible).

#ifndef Z2POLYMOD_H
#define Z2POLYMOD_H

#include "z2poly.h"

// ----------------------------------------------------------------
z2poly_t z2polymod_mul(
	z2poly_t * pin1,
	z2poly_t * pin2,
	z2poly_t * paux);

z2poly_t z2polymod_div(
	z2poly_t * pin1,
	z2poly_t * pin2,
	z2poly_t * paux);

z2poly_t z2polymod_recip(
	z2poly_t * pin,
	z2poly_t * paux);

z2poly_t z2polymod_power(
	z2poly_t * pin,
	unsigned   power,
	z2poly_t * paux);

z2poly_t z2polymod_power64(
	z2poly_t * pin,
	unsigned long long power,
	z2poly_t * paux);

int z2polymod_element_order(
	z2poly_t * pinz,
	z2poly_t * paux);

z2poly_t z2polymod_find_generator(
	z2poly_t * pr);

int z2polymod_log(
	z2poly_t * pz,  // Input
	z2poly_t * pg,  // Generator, assumed (!) primitive mod r.
	z2poly_t * pr); // Modulus for FF arithmetic

// ----------------------------------------------------------------
void z2polymod_print(
	z2poly_t * pin,
	z2poly_t * paux);

int z2polymod_auxscan(
	z2poly_t * pout,
	char     * pin);

// ----------------------------------------------------------------
z2poly_t z2polymod_compose(
	z2poly_t * pr,
	z2poly_t * pz,
	z2poly_t * pm);

// ----------------------------------------------------------------
// Finds the numerically lowest primitive monic irreducible for a given
// Galois field.
z2poly_t z2poly_find_irreducible(
	int m);

int z2poly_is_irreducible(
	z2poly_t * pin);

// Assumes r is known to be irreducible.
int z2poly_is_primitive(
	z2poly_t * pin);

#endif // Z2POLYMOD_H
