// Elements of the ring F_3^n[x] (polynomials with coefficients in a
// Galois field).

#ifndef Z3NPOLY_H
#define Z3NPOLY_H

#include "z3polymod.h"

// ----------------------------------------------------------------
#define Z3NPOLY_MAX_DEGREE Z3POLY_MAX_DEGREE
typedef struct _z3npoly_t {
	z3poly_t coeffs[Z3NPOLY_MAX_DEGREE+1];
	int degree;
} z3npoly_t;

// ----------------------------------------------------------------
z3npoly_t z3npoly_add(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t  * pr);

z3npoly_t z3npoly_sub(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t  * pr);

z3npoly_t z3npoly_mul(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t  * pr);

void z3npoly_quot_and_rem(
	z3npoly_t * pdividend,
	z3npoly_t * pdivisor,
	z3npoly_t * pquot,
	z3npoly_t * prem,
	z3poly_t  * pr);

z3npoly_t z3npoly_div(
	z3npoly_t * pa,
	z3npoly_t * pb,
	z3poly_t  * pr);

z3npoly_t z3npoly_mod(
	z3npoly_t * pa,
	z3npoly_t * pb,
	z3poly_t  * pr);

z3npoly_t z3npoly_power(
	z3npoly_t * pin,
	int       power,
	z3poly_t  * pr);

z3npoly_t z3npoly_cube_root(
	z3npoly_t * pin,
	z3poly_t  * pr);

z3npoly_t z3npoly_gcd(
	z3npoly_t * pa,
	z3npoly_t * pb,
	z3poly_t  * pr);

z3npoly_t z3npoly_extend_z3poly(
	z3poly_t  * pz3);

z3poly_t z3poly_unextend_z3npoly(
	z3npoly_t * pznp,
	z3poly_t  * pr);

// For general use, just examine pin->degree.  This function is used
// to recompute the degree after the poly has changed.
int z3npoly_find_degree(
	z3npoly_t * pin,
	int         max_deg);

// Used to check for overflow in polynomial multiplication.
void z3npoly_check_degree(
	int degree);

z3npoly_t z3npoly_from_coeffs0(
	z3poly_t * pcoeff0);

z3npoly_t z3npoly_from_coeffs1(
	z3poly_t * pcoeff1,
	z3poly_t * pcoeff0);

z3npoly_t z3npoly_from_int(
	int a);

z3npoly_t z3npoly_from_two_ints(
	int a1,
	int a0);

// ----------------------------------------------------------------
z3npoly_t z3npoly_deriv(
	z3npoly_t * pz,
	z3poly_t  * pr);

z3poly_t z3poly_get_min_poly(
	z3poly_t * pz,
	z3poly_t * pr);

// ----------------------------------------------------------------
int z3npoly_is_zero(z3npoly_t * pz);
int z3npoly_is_one (z3npoly_t * pz);

void z3npoly_make_monic(
	z3npoly_t * pz,
	z3poly_t  * pr);

// ----------------------------------------------------------------
int z3npoly_equal(
	z3npoly_t * pin1,
	z3npoly_t * pin2,
	z3poly_t  * pr);

// Fpn[x] is *not* totally ordered.  Nonetheless, these functions are
// handy for looping and sorting.
int z3npoly_lt(z3npoly_t * pa, z3npoly_t * pb);
int z3npoly_gt(z3npoly_t * pa, z3npoly_t * pb);
int z3npoly_le(z3npoly_t * pa, z3npoly_t * pb);
int z3npoly_ge(z3npoly_t * pa, z3npoly_t * pb);
int z3npoly_eq(z3npoly_t * pa, z3npoly_t * pb);
int z3npoly_ne(z3npoly_t * pa, z3npoly_t * pb);

int z3npoly_cmp(
	z3npoly_t * pa,
	int         cmp, // see cmps.h
	z3npoly_t * pb);

// ----------------------------------------------------------------
void z3npoly_print(
	z3npoly_t * pin,
	z3poly_t  * pr);

int  z3npoly_scan(
	z3npoly_t * pvout,
	char      * pin,
	z3poly_t  * pr);

#endif // Z3NPOLY_H
