// Elements of the ring Z_p[x] / (r(x)) (a field when r(x) is irreducible).

#ifndef Z3POLYMOD_H
#define Z3POLYMOD_H

#include "group.h"
#include "ring.h"
#include "z3poly.h"

// ----------------------------------------------------------------
// Field arithmetic.

z3poly_t z3polymod_mul(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr);

z3poly_t z3polymod_div(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr);

z3poly_t z3polymod_recip(
	z3poly_t * pin,
	z3poly_t * pr);

z3poly_t z3polymod_power(
	z3poly_t * px,
	unsigned   power,
	z3poly_t * pr);

int z3polymod_element_order(
	z3poly_t * pz,
	z3poly_t * pr);

z3poly_t z3polymod_find_generator(
	z3poly_t * pr);

int z3polymod_log(
	z3poly_t * pz,  // Input
	z3poly_t * pg,  // Generator, assumed (!) primitive mod r.
	z3poly_t * pr); // Modulus for FF arithmetic

// These two are not strictly necessary:  there are already z3poly_add and
// z3poly_sub, which take p as their final argument.  The monic irreducible
// contained in the aux struct isn't used for addition or subtraction.
// However, these functions give the pleasant result of a uniform interface for
// z3polymod functions, which is useful when an application makes tables of
// function pointers.

z3poly_t z3polymod_add(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr);

z3poly_t z3polymod_sub(
	z3poly_t * pin1,
	z3poly_t * pin2,
	z3poly_t * pr);

// ----------------------------------------------------------------
// I/O

void z3polymod_print(
	z3poly_t * pin,
	z3poly_t * pr);

// ----------------------------------------------------------------
z3poly_t z3polymod_compose(
	z3poly_t * pr,
	z3poly_t * pz,
	z3poly_t * pm);

int z3poly_is_irreducible(
	z3poly_t * pz);

// Finds the numerically lowest primitive monic irreducible for a given Galois
// field.
z3poly_t z3poly_find_irreducible(
	int   m);

// Assumes z is known to be irreducible.
int z3poly_is_primitive(
	z3poly_t * pz,
	int   m);

// z3poly_is_primitive() factors p^m-1 each time it is called.
// If you are calling it multiple times with the same p and m,
// you can avoid redundant effort by factoring p^m-1 yourself
// and passing in the factorization.
int z3poly_is_primitive_f(
	z3poly_t * pz,
	int m,
	unsigned pm1,
	factor_info_t * pfinfo);

// ----------------------------------------------------------------
// The caller must free the return value.
z3poly_t * z3polymod_get_elements(
	int n,
	int * pnel);

// ----------------------------------------------------------------
// Finds all irreducibles of degree d dividing n.  That is, this is the
// factorization of X^(p^n) - X, without actually factoring (or requiring
// storage for) that huge-degree polynomial.
void z3poly_find_dirreds(
	unsigned   n,
	z3poly_t * pirred_factors,
	unsigned * pnum_irred_factors,
	unsigned   max_irred_factors);

#endif // Z3POLYMOD_H
