#ifndef ZPLINALG_H
#define ZPLINALG_H

#include <stdio.h>
#include "zppoly.h"

int ** zpla_alloc_matrix(
	int        nrows,
	int        ncols);

void zpla_free_matrix(
	int     ** M,
	int        nrows,
	int        ncols);

void zpla_print_matrix(
	int     ** M,
	int        nrows,
	int        ncols);

void zpla_print_vector(
	int      * v,
	int        length);

void zpla_matrix_from_zpolys(
	int     ** M,
	int        nrows,
	int        ncols,
	zpoly_t  * polys);

// Input of the form 1010:0010:0011:1110,
// or 1,0,1,0:0,0,1,0:0,0,1,1:1,1,1,0.
// Caller must free the return value.
int zpla_matrix_from_string(
	char     * string,
	int        p,
	int    *** pM,
	int      * pnrows,
	int      * pncols);

// Input of the form
//   1 0 1 0
//   0 0 1 0
//   0 0 1 1
//   1 1 1 0
// Caller must free the return value.
int zpla_matrix_from_stream(
	FILE     * fp,
	int        p,
	int    *** pM,
	int      * pnrows,
	int      * pncols);

// Input of the form
//   1 0 1 0
//   0 0 1 0
//   0 0 1 1
//   1 1 1 0
// Caller must free the return value.
int zpla_matrix_from_file(
	char     * file_name,
	int        p,
	int    *** pM,
	int      * pnrows,
	int      * pncols);

void zpla_vector_from_base_rep(
	int      * v,
	int        len,
	int        p,
	int        base_rep);

void zpla_zpolys_from_matrix(
	zpoly_t  * polys,
	int     ** M,
	int        nrows,
	int        ncols);

void zpla_row_from_zpoly(
	int      * prow,
	zpoly_t  * pzpoly,
	int        ncols);

void zpla_transpose(
	int     ** Mout,
	int     ** Min,
	int        n);

void zpla_row_echelon_form(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols);

void zpla_row_reduce_below(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols);

// Assumes z2ila_row_reduce_below has already been done.
void zpla_row_reduce_above(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols);

void zpla_make_matrix_monic(
	int     ** M,
	int        p,
	int        nrows,
	int        ncols);

// Assumes the matrix is already in row-echelon form.
int zpla_find_rank(
	int     ** M,
	int        nrows,
	int        ncols);

int zpla_row_is_zero(
	int      * prow,
	int        ncols);

// Assumes the matrix is already in row-echelon form.
void zpla_find_nullspace_basis(
	int     ** BIT,
	int        nrows,
	int        ncols,
	int        p,
	zpoly_t  * nullspace_basis,
	int        dimker);

// For square matrices.
void zpla_mats_mul(
	int     ** out,
	int     ** in1,
	int     ** in2,
	int        p,
	int        n);

// For square matrices.
void zpla_mats_copy(
	int     ** out,
	int     ** in,
	int        n);

// For square matrices.
int zpla_is_I(
	int     ** M,
	int        n);

int zpla_is_zero_matrix(
	int     ** M,
	int        nrows,
	int        ncols);

int zpla_is_zero_vector(
	int      * v,
	int        length);

// For square matrices.
void zpla_mats_times_vr(
	int ** M,
	int  * vin,
	int  * vout,
	int    n,
	int    p);

void zpla_mat_times_vr(
	int ** M,
	int  * vin,
	int  * vout,
	int    nrows, // = output vector length
	int    ncols, // = input  vector length
	int    p);

#endif // ZPLINALG_H
