#include <stdio.h>
#include <stdlib.h>
#include "zpn_convert.h"
#include "zpnpoly_factor.h"

// ----------------------------------------------------------------
int zpn_convert_find_generators(
	zppoly_aux_t * prs, // Smaller field
	zppoly_aux_t * prl, // Larger field
	zpoly_t      * pgs,
	zpoly_t      * pgl)
{
	zpoly_t * roots;
	zpoly_t ms;
	int ds;

	*pgs = zppolymod_find_generator(prs);

	// Get min poly for gs in smaller field
	ms = zppoly_get_min_poly(pgs, prs);

	// Factor it over larger field
	roots = zppoly_get_roots(&ms, prl);
	ds = ms.degree;

	// Take any root; aesthetically, the lexically smallest is preferred.
	qsort(roots, ds, sizeof(zpoly_t), zppoly_qcmp);
	*pgl = roots[0];
	free(roots);
	return 1;
}

// ----------------------------------------------------------------
int zpn_convert_scalar(
	zpoly_t      * pu1,
	zppoly_aux_t * pr1,
	zpoly_t      * pg1,
	zpoly_t      * pu2,
	zppoly_aux_t * pr2,
	zpoly_t      * pg2)
{
	int l1 = 0, l2 = 0;

	if (zppoly_is_zero(pu1)) {
		*pu2 = zppoly_from_base_rep(pr1->p, 0);
		return 1;
	}

	l1 = zppolymod_log(pu1, pg1, pr1);
	l2 = l1;
	*pu2 = zppolymod_power(pg2, l2, pr2);

	return 1;
}

// ----------------------------------------------------------------
int zpn_convert_poly(
	zpnpoly_t    * pf1,
	zppoly_aux_t * pr1,
	zpoly_t      * pg1,
	zpnpoly_t    * pf2,
	zppoly_aux_t * pr2,
	zpoly_t      * pg2)
{
	int i;
	zpoly_t u1, u2;

	*pf2 = *pf1;

	for (i = 0; i <= pf2->degree; i++) {
		u1 = pf1->coeffs[i];
		u1 = zppoly_mod(&u1, &pr1->r, pr1->p);
		if (!zpn_convert_scalar(&u1, pr1, pg1, &u2, pr2, pg2))
			return 0;
		pf2->coeffs[i] = u2;
	}
	return 1;
}
