#ifndef ZPNLINALG_H
#define ZPNLINALG_H

#include <stdio.h>
#include "zppolymod.h"
#include "zpnpoly.h"

zpoly_t ** zpnla_alloc_matrix(
	int         nrows,
	int         ncols);

void zpnla_free_matrix(
	zpoly_t  ** M,
	int         nrows,
	int         ncols);

void zpnla_print_matrix(
	zpoly_t  ** M,
	int         nrows,
	int         ncols,
	zppoly_aux_t * paux);

void zpnla_print_vector(
	zpoly_t   * v,
	int         n,
	zppoly_aux_t * paux);

void zpnla_matrix_from_zpnpolys(
	zpoly_t  ** M,
	int         nrows,
	int         ncols,
	zpnpoly_t * polys);

void zpnla_zpnpolys_from_matrix(
	zpnpoly_t * polys,
	zpoly_t  ** M,
	int         nrows,
	int         ncols);

void zpnla_row_from_zpnpoly(
	zpoly_t   * prow,
	zpnpoly_t * pzpoly,
	int         ncols);

void zpnla_transpose(
	zpoly_t  ** Mout,
	zpoly_t  ** Min,
	int         n);

void zpnla_row_echelon_form(
	zpoly_t  ** M,
	zppoly_aux_t * paux,
	int         nrows,
	int         ncols);

void zpnla_row_reduce_below(
	zpoly_t  ** M,
	zppoly_aux_t * paux,
	int         nrows,
	int         ncols);

// Assumes zpnla_row_reduce_below has already been done.
void zpnla_row_reduce_above(
	zpoly_t  ** M,
	zppoly_aux_t * paux,
	int         nrows,
	int         ncols);

void zpnla_make_matrix_monic(
	zpoly_t  ** M,
	zppoly_aux_t * paux,
	int         nrows,
	int         ncols);

// Assumes the matrix is already in row-echelon form.
int zpnla_find_rank(
	zpoly_t  ** M,
	int         nrows,
	int         ncols);

int zpnla_row_is_zero(
	zpoly_t   * prow,
	int         ncols);

// Assumes the matrix is already in row-echelon form.
void zpnla_find_nullspace_basis(
	zpoly_t  ** M,
	int        nrows,
	int        ncols,
	zppoly_aux_t * paux,
	zpnpoly_t * nullspace_basis,
	int        dimker);

// For square matrices.
void zpnla_mats_copy(
	zpoly_t ** out,
	zpoly_t ** in,
	int        n);

// Square matrix times vector
void zpnla_mx_times_vr(
	zpoly_t     ** M,
	zpoly_t      * vin,
	zpoly_t      * vout,
	int            n,
	zppoly_aux_t * paux);

void zpnla_sq_mx_times_mx(
	zpoly_t     ** out,
	zpoly_t     ** in1,
	zpoly_t     ** in2,
	int            n,
	zppoly_aux_t * paux);

// Power must be non-negative.
void zpnla_mx_power(
	zpoly_t     ** out,
	zpoly_t     ** in,
	int            n,
	int            power,
	zppoly_aux_t * paux);

#endif // ZPNLINALG_H
