// Elements of the ring F_p^n[x] (polynomials with coefficients in a
// Galois field).

#ifndef ZPNPOLY_H
#define ZPNPOLY_H

#include "zppolymod.h"

// ----------------------------------------------------------------
#define ZPNPOLY_MAX_DEGREE ZPOLY_MAX_DEGREE
typedef struct _zpnpoly_t {
	zpoly_t coeffs[ZPNPOLY_MAX_DEGREE+1];
	int degree;
} zpnpoly_t;

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_add(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_sub(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_mul(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t  * paux);

void zpnpoly_quot_and_rem(
	zpnpoly_t * pdividend,
	zpnpoly_t * pdivisor,
	zpnpoly_t * pquot,
	zpnpoly_t * prem,
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_div(
	zpnpoly_t * pa,
	zpnpoly_t * pb,
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_mod(
	zpnpoly_t * pa,
	zpnpoly_t * pb,
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_power(
	zpnpoly_t * pin,
	int       power,
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_pth_root(
	zpnpoly_t * pin,
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_gcd(
	zpnpoly_t * pa, // 1st input.
	zpnpoly_t * pb, // 2nd input.
	zppoly_aux_t * paux);

zpnpoly_t zpnpoly_extend_zppoly(
	zpoly_t  * pzp);

zpoly_t zppoly_unextend_zpnpoly(
	zpnpoly_t * pznp,
	zppoly_aux_t * paux);

// For general use, just examine pin->degree.  This function is used
// to recompute the degree after the poly has changed.
int zpnpoly_find_degree(
	zpnpoly_t * pin,
	int         max_deg);

// Used to check for overflow in polynomial multiplication.
void zpnpoly_check_degree(
	int degree);

zpnpoly_t zpnpoly_from_coeffs0(
	zpoly_t * pcoeff0);

zpnpoly_t zpnpoly_from_coeffs1(
	zpoly_t * pcoeff1,
	zpoly_t * pcoeff0);

zpnpoly_t zpnpoly_from_coeffs2(
	zpoly_t * pcoeff2,
	zpoly_t * pcoeff1,
	zpoly_t * pcoeff0);

zpnpoly_t zpnpoly_from_int(
	int p,
	int a);

// Example:
// * a_aux   = a^2 + a + 1 = 1,1,1 which defines an F_2^2.
// * b_red   = b^6 + b + 1 = 1,0,0,0,0,1,1 which defines an F_2^6.
// * ab_poly = x^3 + x^2 + (a+1) x + a = 1:1:1,1:1,0
//   which is a cubic polynomial in F_2^2[x].
// * ab_exp  = 21 since (2^6-1) / (2^2-1) = 21.
//
// This function substitutes x with b and a with b^exp, then reduces mod
// b_red.  The above example ab_poly reduces to 0 mod b^6 + b + 1.

zpoly_t zpnpoly_flatten(
	zpnpoly_t    * pab_poly,
	zppoly_aux_t * pa_aux,
	zpoly_t      * pb_red,
	int            ab_exp);

zpnpoly_t zpnpoly_from_base_rep(
	int rep,
	int p,
	int intermed_degree);

// ----------------------------------------------------------------
zpnpoly_t zpnpoly_deriv(
	zpnpoly_t * pz,
	zppoly_aux_t * paux);

zpoly_t zpnpoly_eval(
	zpoly_t      * pa,
	zpnpoly_t    * poutr,
	zppoly_aux_t * pinnaux);

int zpnpoly_is_rootless(
	zpnpoly_t    * poutr,
	zppoly_aux_t * pinnaux);

int zpnpoly_is_primitive(
	zpnpoly_t    * poutr,
	zppoly_aux_t * pinnaux);

int zpnpoly_is_primitive_f(
	zpnpoly_t     * poutr,
	zppoly_aux_t  * pinnaux,
	unsigned        pm1,
	factor_info_t * pfinfo);

zpoly_t zppoly_get_min_poly(
	zpoly_t      * pz,
	zppoly_aux_t * paux);

// ----------------------------------------------------------------
int zpnpoly_is_zero(zpnpoly_t * pz);
int zpnpoly_is_one (zpnpoly_t * pz);

void zpnpoly_make_monic(
	zpnpoly_t * pz,
	zppoly_aux_t * paux);

// ----------------------------------------------------------------
int zpnpoly_equal(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux);

// Fpn[x] is *not* totally ordered.  Nonetheless, these functions are
// handy for looping and sorting.
int zpnpoly_lt(zpnpoly_t * pa, zpnpoly_t * pb);
int zpnpoly_gt(zpnpoly_t * pa, zpnpoly_t * pb);
int zpnpoly_le(zpnpoly_t * pa, zpnpoly_t * pb);
int zpnpoly_ge(zpnpoly_t * pa, zpnpoly_t * pb);
int zpnpoly_eq(zpnpoly_t * pa, zpnpoly_t * pb);
int zpnpoly_ne(zpnpoly_t * pa, zpnpoly_t * pb);

int zpnpoly_cmp(
	zpnpoly_t * pa,
	int         cmp, // see cmps.h
	zpnpoly_t * pb);

// ----------------------------------------------------------------
void zpnpoly_print(zpnpoly_t * pin, zppoly_aux_t * paux);
int  zpnpoly_scan(void * pvout, char * pin, void * pvaux);

void zpnpoly_vprint(void * pvin, void * pvaux);
void zpnpoly_vprintsmall(void * pvin, void * pvaux);
int  zpnpoly_vscan(void * pvout, char * pin, void * pvaux, int max_elt_bytes);

#endif // ZPNPOLY_H
