// Elements of the ring Z[x].

#ifndef ZPOLY_H
#define ZPOLY_H

//#define ZPOLY_MAX_DEGREE 256
//#define ZPOLY_MAX_DEGREE 64
#define ZPOLY_MAX_DEGREE 16

typedef struct _zpoly_t {
	int coeffs[ZPOLY_MAX_DEGREE+1];
	int degree;
} zpoly_t;

// ----------------------------------------------------------------
zpoly_t zpoly_add(
	zpoly_t * pin1,
	zpoly_t * pin2);

zpoly_t zpoly_sub(
	zpoly_t * pin1,
	zpoly_t * pin2);

zpoly_t zpoly_usub(
	zpoly_t * pin);

zpoly_t zpoly_mul(
	zpoly_t * pin1,
	zpoly_t * pin2);

zpoly_t zpoly_div(
	zpoly_t * pin1,
	zpoly_t * pin2);

zpoly_t zpoly_mod(
	zpoly_t * pin1,
	zpoly_t * pin2);

void zpoly_quot_and_rem(
	zpoly_t * pdividend,
	zpoly_t * pdivisor,
	zpoly_t * pquot,
	zpoly_t * prem);

zpoly_t zpoly_power(
	zpoly_t * px,
	unsigned  power);

int zpoly_equal(
	zpoly_t * pin1,
	zpoly_t * pin2);

// ----------------------------------------------------------------
// Construction

zpoly_t zpoly_from_coeffs(
	int   degree,
	int * coeffs);

zpoly_t zpoly_from_coeffs0(
	int   coeff0);

zpoly_t zpoly_from_coeffs1(
	int   coeff1,
	int   coeff0);

zpoly_t zpoly_from_coeffs2(
	int   coeff2,
	int   coeff1,
	int   coeff0);

zpoly_t zpoly_monomial_from_coeff(
	int   coeffa,
	int   dega);

zpoly_t zpoly_binomial_from_coeffs(
	int   coeffa,
	int   dega,
	int   coeffb,
	int   degb);

// ----------------------------------------------------------------
// I/O

void zpoly_x_print(
	zpoly_t * pout);

void zpoly_print(
	zpoly_t * pout);

int  zpoly_scan(
	zpoly_t * pout,
	char    * pin);

// ----------------------------------------------------------------
// Factorization, evaluation and classification.

int zpoly_eval(
	zpoly_t * pz,
	int   x);

int zpoly_is_zero(
	zpoly_t * pz);

int zpoly_is_one(
	zpoly_t * pz);

#endif // ZPOLY_H
