// Elements of the ring Z_p[x].

#ifndef ZPPOLY_H
#define ZPPOLY_H

#include "zpoly.h"
#include "cmps.h"
#include "factorize.h"

// ----------------------------------------------------------------
zpoly_t zppoly_add(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p);

zpoly_t zppoly_sub(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p);

zpoly_t zppoly_usub(
	zpoly_t * pin,
	int       p);

zpoly_t zppoly_mul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p);

zpoly_t zppoly_scalar_mul(
	zpoly_t * pin1,
	int       in2,
	int       p);

zpoly_t zppoly_div(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p);

zpoly_t zppoly_mod(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p);

void zppoly_quot_and_rem(
	zpoly_t * pdividend,
	zpoly_t * pdivisor,
	zpoly_t * pquot,
	zpoly_t * prem,
	int       p);

zpoly_t zppoly_ext_gcd(
	zpoly_t * pa,  // 1st input.
	zpoly_t * pb,  // 2nd input.
	zpoly_t * pm,  // Output coefficient on a
	zpoly_t * pn,  // Output coefficient on b
	int       p);

zpoly_t zppoly_power(
	zpoly_t * pin,
	int       power,
	int       p);

// Assumes the input is already known to be a pth power.
zpoly_t zppoly_pth_root(
	zpoly_t * pin,
	int p);

zpoly_t zppoly_gcd(
	zpoly_t * pa,  // 1st input.
	zpoly_t * pb,  // 2nd input.
	int       p);

int zppoly_equal(
	zpoly_t * pin1,
	zpoly_t * pin2,
	int       p);

// Fp[x] is *not* totally ordered.  Nonetheless, these functions are
// handy for looping and sorting.
int zppoly_lt(zpoly_t * pa, zpoly_t * pb);
int zppoly_gt(zpoly_t * pa, zpoly_t * pb);
int zppoly_le(zpoly_t * pa, zpoly_t * pb);
int zppoly_ge(zpoly_t * pa, zpoly_t * pb);
int zppoly_eq(zpoly_t * pa, zpoly_t * pb);
int zppoly_ne(zpoly_t * pa, zpoly_t * pb);

int zppoly_cmp(
	zpoly_t * pa,
	int       cmp, // see cmps.h
	zpoly_t * pb);


// ----------------------------------------------------------------
// Construction

zpoly_t zppoly_from_base_rep(
	int   p,
	int   base_rep);

int zppoly_to_base_rep(
	zpoly_t * pz,
	int       p);

zpoly_t zppoly_from_coeffs(
	int   p,
	int   degree,
	int * coeffs);

zpoly_t zppoly_from_coeffs0(
	int   coeff0,
	int   p);

zpoly_t zppoly_from_coeffs1(
	int   coeff1,
	int   coeff0,
	int   p);

zpoly_t zppoly_from_coeffs2(
	int   coeff2,
	int   coeff1,
	int   coeff0,
	int   p);

zpoly_t zppoly_monomial_from_coeff(
	int   coeff0,
	int   power,
	int   p);

zpoly_t zppoly_binomial_from_coeffs(
	int   coeffa,
	int   dega,
	int   coeffb,
	int   degb,
	int   p);

zpoly_t zppoly_from_zpoly(
	zpoly_t * pz,
	int       p);

// ----------------------------------------------------------------
// I/O

int  zppoly_scan(
	zpoly_t * pout,
	char    * pin,
	int       p);

void zppoly_x_print(
	zpoly_t * pin,
	int       p);

void zppoly_print(
	zpoly_t * pin,
	int       p);

void zppoly_dprint(
	zpoly_t * pin,
	int       p,
	int       min_deg);

// ----------------------------------------------------------------
// Factorization, evaluation and classification.

int zppoly_eval(
	zpoly_t * pz,
	int   p,
	int   x);

zpoly_t zppoly_compose(
	zpoly_t * pr,
	zpoly_t * pz,
	int       p);

zpoly_t zppoly_deriv(
	zpoly_t * pz,
	int p);

int zppoly_is_rootless(
	zpoly_t * pz,
	int   p);

int zppoly_count_non_zero_terms(
	zpoly_t * pz);

int zppoly_is_zero(
	zpoly_t * pz);

int zppoly_is_one(
	zpoly_t * pz);

void zppoly_make_monic(
	zpoly_t * pz,
	int       p);

int zppoly_find_degree(
	zpoly_t * pin,
	int       max_deg);

void zppoly_check_degree(
	int degree);

// ----------------------------------------------------------------
// For use with qsort(3).
int zppoly_qcmp(
	const void * pv1,
	const void * pv2);

#endif // ZPPOLY_H
