// Elements of the ring Z_p[x] / (r(x)) (a field when r(x) is irreducible).

#ifndef ZPPOLYMOD_H
#define ZPPOLYMOD_H

#include "group.h"
//#include "ring.h"
#include "zppoly.h"

// ----------------------------------------------------------------
typedef struct _zppoly_aux_t {
	int     p;  // Prime p
	zpoly_t r;  // Irreducible monic polynomial
} zppoly_aux_t;

// ----------------------------------------------------------------
// Field arithmetic.

zpoly_t zppolymod_mul(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux);

zpoly_t zppolymod_div(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux);

zpoly_t zppolymod_recip(
	zpoly_t * pin,
	zppoly_aux_t * paux);

zpoly_t zppolymod_power(
	zpoly_t * px,
	unsigned  power,
	zppoly_aux_t * paux);

int zppolymod_element_order(
	zpoly_t * pz,
	zppoly_aux_t * paux);

zpoly_t zppolymod_find_generator(
	zppoly_aux_t * paux);

int zppolymod_log(
	zpoly_t * pz,  // Input
	zpoly_t * pg,  // Base, assumed (!) primitive mod r.
	zppoly_aux_t * paux); // Modulus for FF arithmetic

// These two are not strictly necessary:  there are already zppoly_add and
// zppoly_sub, which take p as their final argument.  The monic irreducible
// contained in the aux struct isn't used for addition or subtraction.
// However, these functions give the pleasant result of a uniform interface for
// zppolymod functions, which is useful when an application makes tables of
// function pointers.

zpoly_t zppolymod_add(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux);

zpoly_t zppolymod_sub(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux);

// ----------------------------------------------------------------
// I/O

void zppolymod_print(
	zpoly_t * pin,
	zppoly_aux_t * paux);

int zppolymod_auxscan(
	zppoly_aux_t *  paux,
	char * p_string,
	char * m_or_red_string);

// ----------------------------------------------------------------
// Construction
zppoly_aux_t zppoly_aux_from_base_rep(
	int   p,
	int   m,
	int   base_rep);

zppoly_aux_t zppoly_aux_from_coeffs(
	int   p,
	int   m,
	int * coeffs);

// ----------------------------------------------------------------
zpoly_t zppolymod_compose(
	zpoly_t * pr,
	zpoly_t * pz,
	zpoly_t * pm,
	int       p);

// ----------------------------------------------------------------
int zppoly_is_irreducible(
	zpoly_t * pz,
	int   p);

void zppoly_list_irreducibles(
	int   p,
	int   m);

// Finds the numerically lowest primitive monic irreducible for a given Galois
// field.
zppoly_aux_t zppoly_find_irreducible(
	int   p,
	int   m);

// Assumes r is known to be irreducible.
int zppoly_is_primitive(
	zpoly_t * pr,
	int   p,
	int   m);

// zppoly_is_primitive() factors p^m-1 each time it is called.
// If you are calling it multiple times with the same p and m,
// you can avoid redundant effort by factoring p^m-1 yourself
// and passing in the factorization.
int zppoly_is_primitive_f(
	zpoly_t * pr,
	int p,
	int m,
	unsigned pm1,
	factor_info_t * pfinfo);

// ----------------------------------------------------------------
// The caller must free the return value.
zpoly_t * zppolymod_get_elements(
	int p,
	int n,
	int * pnel);

// ----------------------------------------------------------------
// Finds all irreducibles of degree d dividing n.  That is, this is the
// factorization of X^(p^n) - X, without actually factoring (or requiring
// storage for) that huge-degree polynomial.
void zppoly_find_dirreds(
	int        p,
	unsigned   n,
	zpoly_t  * pirred_factors,
	unsigned * pnum_irred_factors,
	unsigned   max_irred_factors);

// ----------------------------------------------------------------
extern group_spec_t zppolymod_add_group_spec;
extern group_spec_t zppolymod_mul_group_spec;
//extern ring_spec_t  zppolymod_ring_spec;
extern vfops_t      zppolymod_vfops;

#endif // ZPPOLYMOD_H
