#include <stdio.h>
#include <stdlib.h>
#include "z2ipolymod.h"

#define PLOT

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s r a b c d e\n", argv0);
	fprintf(stderr, "Searches for solutions of y^2 + ay = x^3 + bx^2 + "
		"cxy + dx + e.\n");
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t r, a, b, c, d, e, pn;
	z2ipoly_t x, y, lhs, rhs;
	int n;
	int nin = 0, nout = 0;

	if (argc != 7)
		usage(argv[0]);
	if (!z2ipoly_scan(&r, argv[1])) usage(argv[0]);
	if (!z2ipoly_scan(&a, argv[2])) usage(argv[0]);
	if (!z2ipoly_scan(&b, argv[3])) usage(argv[0]);
	if (!z2ipoly_scan(&c, argv[4])) usage(argv[0]);
	if (!z2ipoly_scan(&d, argv[5])) usage(argv[0]);
	if (!z2ipoly_scan(&e, argv[6])) usage(argv[0]);

	a = z2ipoly_mod(a, r);
	b = z2ipoly_mod(b, r);
	c = z2ipoly_mod(c, r);
	d = z2ipoly_mod(d, r);
	e = z2ipoly_mod(e, r);

	n = z2ipoly_get_degree(r);
	pn = 1LL << n;

	for (x = 0; x < pn; x++) {
		for (y = 0; y < pn; y++) {
			// y^2 + ay = x^3 + bx^2 + cxy + dx + e
			// y * (y + a) = x*(x^2 + b*x + c*y + d) + e
			// y * (y + a) = x*[x*(x + b) + c*y + d] + e
			lhs = z2ipolymod_mul(y, y^a, r);
			rhs = z2ipolymod_mul(x, x^b, r)
				^ z2ipolymod_mul(c, y, r) ^ d;
			rhs = z2ipolymod_mul(x, rhs, r) ^ e;
#ifdef PLOT
			if (lhs == rhs) {
				printf("*");
				nin++;
			}
			else {
				printf(".");
				nout++;
			}
#else
			if (lhs == rhs) {
				z2ipolymod_print(x, r);
				printf(" ");
				z2ipolymod_print(y, r);
				printf("\n");
			}
#endif
		}
#ifdef PLOT
		printf("\n");
#endif
	}
#ifdef PLOT
	printf("in %d  out %d\n", nin, nout);
#endif

	return 0;
}

