#include <stdio.h>
#include <stdlib.h>
#include <sys/select.h>

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {m1} {m2} [ms delay]\n", argv0);
	fprintf(stderr, "  m1 is plotted along rows, m2 along columns.\n");
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	unsigned i1 = 1, i2 = 1;
	unsigned x1, x2;
	unsigned m1, m2;
	unsigned got_m1_m2 = 0;
	unsigned msdelay = 25;
	struct timeval tv, tvs;
	int dot = 1;
	int pause = 0;
	int argi;

	for (argi = 1; argi < argc; argi++) {
		if (sscanf(argv[argi], "m=%u,%u", &m1, &m2) == 2)
			got_m1_m2 = 1;
		else if (sscanf(argv[argi], "i=%u,%u", &i1, &i2) == 2)
			;
		else if (sscanf(argv[argi], "delay=%u", &msdelay) == 1)
			;
		else if (sscanf(argv[argi], "dot=%u", &dot) == 1)
			;
		else if (sscanf(argv[argi], "pause=%u", &pause) == 1)
			;
		else
			usage(argv[0]);
	}
	if (!got_m1_m2)
		usage(argv[0]);

	tv.tv_sec  = 0;
	tv.tv_usec = msdelay * 1000;
	tv.tv_sec  = tv.tv_usec / 1000000;
	tv.tv_usec = tv.tv_usec % 1000000;
	tvs = tv;

	printf("\033[H\033[J");
	for (x1 = 0; x1 < m1; x1++) {
		for (x2 = 0; x2 < m2; x2++) {
			if (dot)
				printf(". ");
			else
				printf("  .   ");
		}
		printf("\n");
	}
	printf("\033[H");
	x1 = i1;
	x2 = i2;
	while ((x1 != 0) || (x2 != 0)) {
		if (dot) {
			printf("\033[%u;%uH", x1+1, x2*2+1);
			printf("*");
		}
		else {
			printf("\033[%u;%uH", x1+1, x2*6+1);
			printf("%2u,%2u", x1, x2);
		}
		fflush(stdout);
		x1 = (x1 + i1) % m1;
		x2 = (x2 + i2) % m2;
		if (pause) {
			fflush(stdin);
			(void)getchar();
		}
		else if (msdelay) {
			tv = tvs;
			select(0, 0, 0, 0, &tv);
		}
	}
	printf("\033[%u;%uH", x1+1, x2*2+1);
	printf("*");
	printf("\033[%u;%uH\n", m1, 1);

	return 0;
}
