#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include "diffy.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n} {# samples}\n", argv0);
	fprintf(stderr, "n must be >= 2; # samples must be >= 1.\n");
	exit(1);
}

// ----------------------------------------------------------------
static double funifrand(void)
{
	double r = (1 << 30) * 2.0;
	double f = random() / r;
	return f;
}

// ----------------------------------------------------------------
static int igaussrand(void)
{
	double f = 0.0;
	int n = 6;
	int i;

	for (i = 0; i < n; i++)
		f += funifrand();
	return (f / n) * 0x00100000;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int n, i, nsamp, sampi, niter;
	int * px, * pt, nbytes;
	int rfd, verbose = 0;

	if ((argc != 3) && (argc != 4))
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &n) != 1)
		usage(argv[0]);
	if (sscanf(argv[2], "%d", &nsamp) != 1)
		usage(argv[0]);
	if (argc == 4)
		verbose = 1;

	if (n < 2)
		usage(argv[0]);
	if (nsamp < 1)
		usage(argv[0]);

	nbytes = n * sizeof(int);
	px = (int *)malloc(nbytes);
	pt = (int *)malloc(nbytes);
	if (!px || !pt) {
		fprintf(stderr, "%s:  malloc(%d) failed.\n", argv[0], nbytes);
		exit(1);
	}

	rfd = open("/dev/urandom", O_RDONLY);
	srandom(time(0) ^ getpid());

	for (sampi = 0; sampi < nsamp; sampi++) {
		if (read(rfd, px, nbytes) < 0) {
			perror("read");
			exit(1);
		}

#if 0
		{
			static int m = 0xcafefeed;
			for (i = 0; i < n; i++) {
				px[i] = m;
				m *= 0xdeadbeef;
			}
		}
#endif

#if 1
		for (i = 0; i < n; i++)
			px[i] = igaussrand();
#endif

#if 1
		for (i = 0; i < n; i++)
			px[i] &= 0x03ffffff;
#endif
		diffy_iterate(px, pt, n, verbose, &niter);
		//printf("niter=%d\n", niter);
		printf("%d\n", niter);
	}

	close(rfd);
	free(px);
	free(pt);
	return 0;
}
