// Solution to problem 3.1.6 of Knuth.
// Input:  Function f from 0..m-1, X0.
// Output: mu, lambda.

#include "fxn_period.h"

// ----------------------------------------------------------------
void get_fxn_period(
	fxn_func_t * f,
	int   X0,
	int * pmu,
	int * plambda)
{
	int a, b, Xn, i, n, x, mu, lambda;

	// (1) Find n such that Xn = X2n.
	// We will then have mu <= n < mu + lambda.
	a = f(X0); b = f(f(X0)); n = 1;
	while (a != b) {
		a  = f(a); b = f(f(b)); n++;
	}
	Xn = a;

	// (2) Find lambda by iterating Xn forward until a duplicate
	// is found.
	lambda = 1; x = f(Xn);
	while (x != Xn) {
		x = f(x); lambda++;
	}

	// (3) Find mu.
	a = X0;
	for (i = 0, b = X0; i < lambda; i++)
		b = f(b); // b = f^lambda(X0)
	mu = 0;
	while (a != b) {
		a = f(a); b = f(b); mu++;
	}

	*pmu = mu;
	*plambda = lambda;
}
