#include <stdio.h>
#include <stdlib.h>

#define ADD 2
#define MUL 3

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n} {m} {k}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int is_hom(
	unsigned n,
	unsigned m,
	unsigned k,
	int op)
{
	unsigned x, y, xy, phix, phiy, phixy, phixphiy;

	for (x = 0; x < m; x++) {
		phix = (k * x) % m;
		for (y = 0; y < m; y++) {
			phiy  = (k * y) % m;
			if (op == ADD)
				xy = (x + y) % n;
			else
				xy = (x * y) % n;
			phixy = (k * xy) % m;
			if (op == ADD)
				phixphiy = (phix + phiy) % m;
			else
				phixphiy = (phix * phiy) % m;
			if (phixphiy != phixy)
				return 0;
		}
	}
	return 1;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	unsigned n = 0;
	unsigned m = 0;
	unsigned k = 0;
	unsigned klo, khi;

	if (argc < 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%u", &n) != 1)
		usage(argv[0]);
	if (sscanf(argv[2], "%u", &m) != 1)
		usage(argv[0]);
	if (argc == 4) {
		if (sscanf(argv[3], "%u", &klo) != 1)
			usage(argv[0]);
		khi = klo;
	}
	else {
		klo = 0;
		khi = m - 1;
	}

	printf("Addition:\n");
	for (k = klo; k <= khi; k++) {
		if (is_hom(n, m, k, ADD))
			printf("%8u\n", k);
	}
	printf("\n");

	printf("Multiplication:\n");
	for (k = klo; k <= khi; k++) {
		if (is_hom(n, m, k, MUL))
			printf("%8u\n", k);
	}
	printf("\n");

	return 0;
}
