#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2nipoly_factor.h"
#include "power.h"
#include "gcd.h"
#include "mod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {f} {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
// q == a / b
// a == q * b

static int try_ring_quot(
	int a,
	int b,
	int m,
	int * pq)
{
	int q, check;

	for (q = 1; q < m; q++) {
		check = (q * b) % m;
		if (check == a) {
			*pq = q;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t f, r, g, fa;
	z2ipoly_t * fes;
	int n, nel, pn1, j, l1, l2, m;
	int q12, q21;

	if ((argc != 3) && (argc != 4))
		usage(argv[0]);
	if (!z2ipoly_scan(&f, argv[1]))
		usage(argv[0]);
	if (!z2ipoly_scan(&r, argv[2]))
		usage(argv[0]);
	g = z2ipolymod_find_generator(r);
	n = z2ipoly_get_degree(r);
	pn1 = (1 << n) - 1;
	if (argc == 4) {
		if (sscanf(argv[3], "%d", &m) != 1)
			usage(argv[0]);
	}
	else {
		m = pn1;
	}
	fes = z2ipolymod_get_elements_frobord(r, &nel);
	//fes = z2ipolymod_get_elements(r, &nel);
	for (j = 0; j < nel; j++) {
		fa = z2ipolymod_compose(f, fes[j], r);
		l1 = z2ipolymod_log(fes[j], g, r);
		l2 = z2ipolymod_log(fa, g, r);

		printf("%d %d", l1, l2);

		if (try_ring_quot(l1, l2, m, &q12))
			printf(" %d", q12);
		else
			printf(" ?");

		if (try_ring_quot(l2, l1, m, &q21))
			printf(" %d", q21);
		else
			printf(" ?");

		printf("\n");
	}

	return 0;
}
