#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z3polymod.h"
#include "z3npoly_factor.h"
#include "power.h"
#include "gcd.h"
#include "mod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {f} {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
// q == a / b
// a == q * b

static int try_ring_quot(
	int a,
	int b,
	int m,
	int * pq)
{
	int q, check;

	for (q = 1; q < m; q++) {
		check = (q * b) % m;
		if (check == a) {
			*pq = q;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z3poly_t f, r, g, fa;
	z3poly_t * fes;
	int n, nel, pn1, j, l1, l2, m;
	int q12, q21;

	if ((argc != 3) && (argc != 4))
		usage(argv[0]);
	if (!z3poly_scan(&f, argv[1]))
		usage(argv[0]);
	if (!z3poly_scan(&r, argv[2]))
		usage(argv[0]);
	g = z3polymod_find_generator(&r);
	n = z3poly_get_degree(&r);
	pn1 = calc_power(3, n) - 1;
	if (argc == 4) {
		if (sscanf(argv[3], "%d", &m) != 1)
			usage(argv[0]);
	}
	else {
		m = pn1;
	}
	fes = z3polymod_get_elements_frobord(&r, &nel);
	//fes = z3polymod_get_elements(r, &nel);
	for (j = 0; j < nel; j++) {
		fa = z3polymod_compose(&f, &fes[j], &r);
		l1 = z3polymod_log(&fes[j], &g, &r);
		l2 = z3polymod_log(&fa, &g, &r);

		printf("%d %d", l1, l2);

		if (try_ring_quot(l1, l2, m, &q12))
			printf(" %d", q12);
		else
			printf(" ?");

		if (try_ring_quot(l2, l1, m, &q21))
			printf(" %d", q21);
		else
			printf(" ?");

		printf("\n");
	}

	return 0;
}
