#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "zpnpoly_factor.h"
#include "power.h"
#include "gcd.h"
#include "mod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {f} {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
// q == a / b
// a == q * b

static int try_ring_quot(
	int a,
	int b,
	int m,
	int * pq)
{
	int q, check;

	for (q = 1; q < m; q++) {
		check = (q * b) % m;
		if (check == a) {
			*pq = q;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zpoly_t f, r, g, fa;
	zpoly_t * fes;
	zppoly_aux_t aux;
	int p, n, nel, pn1, j, l1, l2, m;
	int q12, q21;

	if ((argc != 4) && (argc != 5))
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	if (!zppoly_scan(&f, argv[2], p))
		usage(argv[0]);
	if (!zppoly_scan(&r, argv[3], p))
		usage(argv[0]);
	aux.p = p;
	aux.r = r;

	g = zppolymod_find_generator(&aux);
	n = r.degree;
	pn1 = calc_power(p, n) - 1;
	if (argc == 5) {
		if (sscanf(argv[4], "%d", &m) != 1)
			usage(argv[0]);
	}
	else {
		m = pn1;
	}
	fes = zppolymod_get_elements_frobord(&r, p, &nel);
	//fes = zppolymod_get_elements(n, p, &nel);
	for (j = 0; j < nel; j++) {
		fa = zppolymod_compose(&f, &fes[j], &r,  p);
		l1 = zppolymod_log(&fes[j], &g, &aux);
		l2 = zppolymod_log(&fa, &g, &aux);

		printf("%d %d", l1, l2);

		if (try_ring_quot(l1, l2, m, &q12))
			printf(" %d", q12);
		else
			printf(" ?");

		if (try_ring_quot(l2, l1, m, &q21))
			printf(" %d", q21);
		else
			printf(" ?");

		printf("\n");
	}

	return 0;
}
