#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "zpnpoly_factor.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-c] {p} {r ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void show_iterated_map(
	zpoly_t * pr,
	int       p,
	int       show_log)
{
	zpoly_t * fes;
	int n, nel, i, j;
	zppoly_aux_t aux;
	zpoly_t g;

	aux.r = *pr;
	aux.p = p;
	g = zppolymod_find_generator(&aux);

	fes = zppolymod_get_elements_frobord(pr, p, &nel);
	n = pr->degree;

	zppoly_print(pr, p);
	if (show_log) {
		printf(" ");
		zppoly_print(&g, p);
	}
	printf("\n");

	for (i = 0; i <= n+n; i++) {
		for (j = 0; j < nel; j++) {
			if (show_log)
				printf("%d", zppolymod_log(&fes[j], &g, &aux));
			else
				zppoly_print(&fes[j], p);
			if (j == (nel - 1))
				printf("\n");
			else
				printf(" ");
		}
		for (j = 0; j < nel; j++) {
			fes[j] = zppolymod_compose(pr, &fes[j], pr, p);
		}
	}
	free(fes);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zpoly_t ri;
	int p, argb, argi, show_log;

	if ((argc >= 3) && (strcmp(argv[1], "-l") == 0)) {
		show_log = 1;
		argb = 3;
	}
	else {
		show_log = 0;
		argb = 2;
	}
	if (argc <= argb)
		usage(argv[0]);

	if (sscanf(argv[argb-1], "%d", &p) != 1)
		usage(argv[0]);

	for (argi = argb; argi < argc; argi++) {
		if (!zppoly_scan(&ri, argv[argi], p))
			usage(argv[0]);
		show_iterated_map(&ri, p, show_log);
		if (argi < (argc - 1))
			printf("\n");
	}

	return 0;
}
