#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z3polymod.h"
#include "z3npoly_factor.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-c] {r ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void show_iterated_map(
	z3poly_t * pr,
	int show_log)
{
	z3poly_t * fes;
	int n, nel, i, j;
	z3poly_t g = z3polymod_find_generator(pr);

	fes = z3polymod_get_elements_frobord(pr, &nel);
	n = z3poly_get_degree(pr);

	z3poly_print(pr);
	if (show_log) {
		printf(" ");
		z3poly_print(&g);
	}
	printf("\n");

	for (i = 0; i <= n+n; i++) {
		for (j = 0; j < nel; j++) {
			if (show_log)
				printf("%d", z3polymod_log(&fes[j], &g, pr));
			else
				z3poly_print(&fes[j]);
			if (j == (nel - 1))
				printf("\n");
			else
				printf(" ");
		}
		for (j = 0; j < nel; j++) {
			fes[j] = z3polymod_compose(pr, &fes[j], pr);
		}
	}
	free(fes);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z3poly_t r;
	int argb, argi, show_log;

	if ((argc >= 2) && (strcmp(argv[1], "-l") == 0)) {
		show_log = 1;
		argb = 2;
	}
	else {
		show_log = 0;
		argb = 1;
	}
	if (argc <= argb)
		usage(argv[0]);

	for (argi = argb; argi < argc; argi++) {
		if (!z3poly_scan(&r, argv[argi]))
			usage(argv[0]);
		show_iterated_map(&r, show_log);
		if (argi < (argc - 1))
			printf("\n");
	}

	return 0;
}
