#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {b} {r} {+|-}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t  b, r, bs;
	z2ipoly_t * in, * out;
	int d, n, pn, i, j;
	int forward;

	if (argc != 4)
		usage(argv[0]);
	if (!z2ipoly_scan(&b, argv[1]))
		usage(argv[0]);
	if (!z2ipolymod_auxscan(&r, argv[2]))
		usage(argv[0]);
	if (argv[3][0] == '+')
		forward = 1;
	else if (argv[3][0] == '-')
		forward = 0;
	else
		usage(argv[0]);

	n = z2ipoly_get_degree(r);
	d = z2ipolymod_element_order(b, r);
	in = z2ipolymod_get_elements(n, &pn);
	out = malloc_check(pn * sizeof(z2ipoly_t));

	if (forward)
		bs = b;
	else
		bs = z2ipolymod_recip(b, r);

	// out_j = sum_{i=0}^{d-1} in_i b^{ij},  j = 0 .. d-1
	for (j = 0; j < d; j++) {
		out[j] = 0LL;
		for (i = 0; i < d; i++) {
			z2ipoly_t bij = z2ipolymod_power(bs, i*j, r);
			out[j] ^= z2ipolymod_mul(in[i], bij, r);
		}
	}

	for (i = 0; i < d; i++) {
		z2ipolymod_print(in[i], r);
		printf("  ");
		z2ipolymod_print(out[i], r);
		printf("\n");
	}

	free(in);
	free(out);
	return 0;
}
