// Lidl and Niederreiter p.81 #2.40.

// Given a basis alpha_i, the unique dual basis beta_j satisfies
//
//   Tr(alpha_i * beta_j) = delta_{ij} (Kronecker's)

// F2n is defined by fr.
// ma is min poly of a.
// ma / (x-a) = sum b_i x^i  in Fpn[x]
// c = ma'(a)
// dual basis of { 1, a, ..., a^{n-1} } is
// { b_0/c, b_1/c, ..., b_{n-1}/c }.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2nipoly.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r} {a}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t fr, a, apower;
	z2ipoly_t c, cinv, g;
	z2nipoly_t ma, maprime, xa, quot, rem;
	int n, argi, nargs, d, i, j;
	z2ipoly_t * as, * bs;

	if (argc < 3)
		usage(argv[0]);
	if (!z2ipolymod_auxscan(&fr, argv[1]))
		usage(argv[0]);

	n = z2ipoly_get_degree(fr);
	as = malloc_check(n * sizeof(z2ipoly_t));
	bs = malloc_check(n * sizeof(z2ipoly_t));
	nargs = argc - 2;
	for (argi = 2; argi < argc; argi++) {
		if (!z2ipoly_scan(&a, argv[argi]))
			usage(argv[0]);

		ma = z2nipoly_extend_z2ipoly(z2ipoly_get_min_poly(a, fr));

		xa = z2nipoly_from_coeffs1(1LL, a);
		quot = z2nipoly_div(&ma, &xa, fr);
		rem  = z2nipoly_mod(&ma, &xa, fr);
		if (!z2nipoly_is_zero(&rem)) {
			fprintf(stderr, "%s: coding error.\n", argv[0]);
			exit(1);
		}

		maprime = z2nipoly_deriv(&ma, fr);
		c = z2nipoly_eval(a, &maprime, fr);
		cinv = z2ipolymod_recip(c, fr);
		d = quot.degree;
		printf("[d=%d] ", d);

		// Compute polynomial basis of a.
		apower = 1LL;
		for (i = 0; i <= d; i++) {
			as[i] = apower;
			apower = z2ipolymod_mul(a, apower, fr);
		}

		// Compute dual basis.
		for (i = 0; i <= d; i++) {
			bs[i] = z2ipolymod_mul(quot.coeffs[i], cinv, fr);
		}

		// Print polynomial basis of a.
		for (i = 0; i <= d; i++) {
			if (i != 0)
				printf(" ");
			z2ipolymod_print(as[i], fr);
		}
		if (nargs > 1)
			printf(" / ");
		else
			printf("\n");

		// Print dual basis of a.
		for (i = 0; i <= d; i++) {
			if (i != 0)
				printf(" ");
			z2ipolymod_print(bs[i], fr);
		}
		if (nargs > 1)
			printf(" / ");
		else
			printf("\n");

		// Print logs of dual basis of a.
		g = z2ipolymod_find_generator(fr);
		for (i = 0; i <= d; i++) {
			if (i != 0)
				printf(" ");
			printf("%d", z2ipolymod_log(bs[i], g, fr));
		}
		printf("\n");

		// Print cross traces
		// What's wrong??
		printf("\n");
		for (i = 0; i <= d; i++) {
			for (j = 0; j <= d; j++) {
				z2ipoly_t ab, tr;
				if (j != 0)
					printf(" ");
				ab = z2ipolymod_mul(as[i], bs[j], fr);
				tr = z2ipolymod_trace(ab, fr);
				printf("[");
				z2ipolymod_print(as[i], fr);
				printf("*");
				z2ipolymod_print(bs[j], fr);
				printf("=");
				z2ipolymod_print(ab, fr);
				printf("]");
				z2ipolymod_print(tr, fr);
			}
			printf("\n");
		}
		printf("\n");

	}

	free(as);
	free(bs);
	return 0;
}
