// Computes Zech's (Jacobi's) log:  Given a generator g of the multiplicative
// group of a finite field:
//
//  1 + g^k = g^L(k)
// L(k) = log_g (g^k + 1)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r} {a}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t r, g, a;
	int n, pn, i;

	if ((argc != 2) && (argc < 3))
		usage(argv[0]);
	if (!z2ipolymod_auxscan(&r, argv[1]))
		usage(argv[0]);
	if (argc == 3) {
		if (!z2ipoly_scan(&g, argv[2]))
			usage(argv[0]);
	}
	else {
		g = z2ipolymod_find_generator(r);
	}
	n = z2ipoly_get_degree(r);
	pn = 1 << n;

	a = g;
	for (i = 0; i < pn-1; i++) {
		int L = z2ipolymod_log(a ^ 1LL, g, r);
		printf("%llx %d %d\n", a, i+1, L);
		a = z2ipolymod_mul(a, g, r);
	}

	return 0;
}
