#include <stdio.h>
#include <stdlib.h>
#include "z2ipolymod.h"
#include <signal.h>

#define MAX_IRRED_FACTORS 4116 // See f2nsplit.c.

static int got_sig = 0;
void sig_hdlr(int signo)
{
	got_sig = 1;
}

// ----------------------------------------------------------------
static void z2ipoly_brute_force_factorize(
	z2ipoly_t   r,
	z2ipoly_t * pirred_factors,
	unsigned  * pnum_irred_factors,
	unsigned    max_irred_factors)
{
	unsigned num_irred_factors = 0;
	z2ipoly_t f, quot, rem;

	if (r <= 1LL) {
		pirred_factors[0] = r;
		*pnum_irred_factors = 1;
		return;
	}

	for (f = 2LL; ; f++) {
		if (r <= 1)
			break;

		if (z2ipoly_is_irreducible(r)) {
			pirred_factors[num_irred_factors] = r;
			num_irred_factors++;
			break;
		}
		z2ipoly_quot_and_rem(r, f, &quot, &rem);
		if (got_sig) {
			int i;
			printf("r=%llx f=%llx\n", r, f);
			for (i = 0; i < num_irred_factors; i++)
				printf("  [%2d] %llx\n", i, pirred_factors[i]);
			printf("\n");
			got_sig = 0;
		}
		while (rem == 0) {

			if (num_irred_factors >= MAX_IRRED_FACTORS) {
				fprintf(stderr, "Too many factors.\n");
				exit(1);
			}
			pirred_factors[num_irred_factors] = f;
			num_irred_factors++;
			r = quot;
			z2ipoly_quot_and_rem(r, f, &quot, &rem);
		}
	}
	*pnum_irred_factors = num_irred_factors;
}

// ----------------------------------------------------------------
void try_irred(int argc, char ** argv)
{
	int        argi, i;
	z2ipoly_t  r;
	z2ipoly_t  irred_factors[MAX_IRRED_FACTORS];
	unsigned   num_irred_factors = 0;

	if (argc < 2) {
		fprintf(stderr, "Usage: %s {Z2 polys in hex ...}\n", argv[0]);
		exit(1);
	}

	for (argi = 1; argi < argc; argi++) {

		if (sscanf(argv[argi], "%llx", &r) != 1) {
			fprintf(stderr, "Scan failure on \"%s\".\n",
				argv[argi]);
			continue;
		}

		z2ipoly_print(r);
		printf(": ");

		z2ipoly_brute_force_factorize(r, irred_factors,
			&num_irred_factors, MAX_IRRED_FACTORS);

		for (i = 0; i < num_irred_factors; i++) {
			printf(" ");
			z2ipoly_print(irred_factors[i]);
		}
		printf("\n");
	}
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	signal(SIGQUIT, sig_hdlr);
	try_irred(argc, argv);

	return 0;
}
