#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "zppolymod.h"
#include "power.h"

#define MAX_IRRED_FACTORS 4000

static int got_sig = 0;
void sig_hdlr(int signo)
{
	got_sig = 1;
}

// ----------------------------------------------------------------
static void zppoly_brute_force_factorize(
	zpoly_t  * pr,
	int        p,
	zpoly_t  * pirred_factors,
	unsigned * pnum_irred_factors,
	unsigned   max_irred_factors)
{
	unsigned num_irred_factors = 0;
	zpoly_t f, quot, rem;
	int n, d, b, lo, hi;
	zppoly_aux_t aux;

	aux.p = p;
	aux.r = *pr;

	if (pr->degree <= 1) {
		pirred_factors[0] = *pr;
		*pnum_irred_factors = 1;
		return;
	}

	n = pr->degree;
	for (d = 1; d < n; d++) {
		lo = calc_power(p, d);
		hi = 2 * lo;
		// 00010 - 00016
		// 00100 - 00166
		// 01000 - 01666
		// 10000 - 16666
		for (b = lo; b < hi; b++) {
			f = zppoly_from_base_rep(p, b);
			if (pr->degree <= 0)
				goto break2;

			if (zppoly_is_irreducible(pr, p)) {
				pirred_factors[num_irred_factors] = *pr;
				num_irred_factors++;
				goto break2;
			}
			zppoly_quot_and_rem(pr, &f, &quot, &rem, p);
			if (got_sig) {
				int i;
				printf("r=");
				zppolymod_print(pr, &aux);
				printf(" r=");
				zppolymod_print(&f, &aux);
				for (i = 0; i < num_irred_factors; i++) {
					printf("  [%2d] ", i);
					zppolymod_print(&pirred_factors[i],
						&aux);
					printf("\n");
				}
				printf("\n");
				got_sig = 0;
			}
			while (zppoly_is_zero(&rem)) {
				if (num_irred_factors >= MAX_IRRED_FACTORS) {
					fprintf(stderr, "Too many factors.\n");
					exit(1);
				}
				pirred_factors[num_irred_factors] = f;
				num_irred_factors++;
				*pr = quot;
				zppoly_quot_and_rem(pr, &f, &quot, &rem, p);
			}
		}
	}
break2:
	*pnum_irred_factors = num_irred_factors;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int      p, argi, i;
	zpoly_t  r;
	zpoly_t  irred_factors[MAX_IRRED_FACTORS];
	unsigned num_irred_factors = 0;

	signal(SIGQUIT, sig_hdlr);

	if ((argc < 3) || (sscanf(argv[1], "%d", &p) != 1)) {
		fprintf(stderr, "Usage: %s {p} {Zp polys ...}\n", argv[0]);
		exit(1);
	}

	for (argi = 2; argi < argc; argi++) {

		if (!zppoly_scan(&r, argv[argi], p)) {
			fprintf(stderr, "Scan failure on \"%s\".\n",
				argv[argi]);
			continue;
		}

		zppoly_print(&r, p);
		printf(": ");

		zppoly_brute_force_factorize(&r, p, irred_factors,
			&num_irred_factors, MAX_IRRED_FACTORS);

		for (i = 0; i < num_irred_factors; i++) {
			printf(" ");
			zppoly_print(&irred_factors[i], p);
		}
		printf("\n");
	}
	return 0;
}
