#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "zplinalg.h"
#include "zpnpoly_factor.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r1} {r2}, of same degree\n", argv0);
	exit(1);
}

//// ----------------------------------------------------------------
//static void print_phi(
//	zpoly_t * phi,
//	int       p,
//	int       pn,
//	zppoly_aux_t * paux)
//{
//	int i;
//	zpoly_t z;
//
//	for (i = 0; i < pn; i++) {
//		z = zppoly_from_base_rep(p, i);
//		printf("  ");
//		zppolymod_print(&z, paux);
//		printf(" |-> ");
//		zppolymod_print(&phi[i], paux);
//		printf("\n");
//	}
//}
//
//// ----------------------------------------------------------------
//int show_isom_stuff(
//	zpoly_t * pr1,
//	zpoly_t * pg1,
//	zpoly_t * pr2,
//	zpoly_t * pg2,
//	int       p,
//	int       n,
//	int       verbosity)
//{
//	int pn = calc_power(p, n);
//	int i;
//	zpoly_t * phi;
//	zpoly_t g1p, g2p;
//	zppoly_aux_t aux1, aux2;
//	int rv = 0;
//
//	aux1.p = p;
//	aux1.r = *pr1;
//	aux2.p = p;
//	aux2.r = *pr2;
//
//	phi = malloc_check(pn * sizeof(zpoly_t));
//
//	phi[0] = zppoly_from_coeffs0(0, p);
//	g1p = zppoly_from_coeffs0(1, p);
//	g2p = zppoly_from_coeffs0(1, p);
//
//	for (i = 1; i < pn; i++) {
//		phi[zppoly_to_base_rep(&g1p, p)] = g2p;
//		g1p = zppolymod_mul(&g1p, pg1, &aux1);
//		g2p = zppolymod_mul(&g2p, pg2, &aux2);
//	}
//
//	printf("Isom: ");
//	zppoly_print(pg1, p);
//	printf(" mod ");
//	zppoly_print(pr1, p);
//	printf(" and ");
//	zppoly_print(pg2, p);
//	printf(" mod ");
//	zppoly_print(pr2, p);
//	printf("\n");
//
//	print_phi(phi, p, pn, &aux1);
//
//	printf("\n");
//
//	free(phi);
//	return rv;
//}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p, n, j, k, l;
	zpoly_t r1, r2, g1, m1, xp, phixp;
	zpoly_t * g2s;
	zppoly_aux_t aux1, aux2;
	int ** M;
	int  * logs;
	int verb = 0;

	if (argc < 4)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	if (!zppoly_scan(&r1, argv[2], p))
		usage(argv[0]);
	if (!zppoly_scan(&r2, argv[3], p))
		usage(argv[0]);
	if (!zppoly_is_irreducible(&r1, p)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[2]);
		exit(1);
	}
	if (!zppoly_is_irreducible(&r2, p)) {
		fprintf(stderr, "%s is not irreducible.\n", argv[3]);
		exit(1);
	}
	verb = argc - 4;

	if (r1.degree != r2.degree)
		usage(argv[0]);
	n = r1.degree;

	aux1.p = p;
	aux1.r = r1;
	aux2.p = p;
	aux2.r = r2;

	// Find a generator mod ri
	//printf("Finding generator for ri ... ");
	//fflush(stdout);
	g1 = zppolymod_find_generator(&aux1);
	//printf("done.\n");

	printf("g1="); zppoly_print(&g1, p); printf("\n");

	// Get min poly over ri, and split mod rj.  g2's are the roots.
	//printf("Splitting generator over rj ... ");
	//fflush(stdout);
	m1 = zppoly_get_min_poly(&g1, &aux1);
	printf("m1="); zppoly_print(&m1, p); printf("\n");
	g2s = zppoly_get_roots(&m1, &aux2);
	//printf("done.\n");

	// Find log-base-gi of x^k mod ri
	//printf("Computing logs of stanard basis over ri ... ");
	//fflush(stdout);
	logs = malloc_check(n * sizeof(logs[0]));
	for (k = 0; k < n; k++) {
		xp = zppoly_monomial_from_coeff(1, k, p);
		logs[k] = zppolymod_log(&xp, &g1, &aux1);
		//printf("log of ");
		//zppoly_print(&xp, p);
		//printf(" = %d\n", logs[k]);
	}
	//printf("done.\n");

	// Compute matrix for each generator g2
	//printf("Computing matrices ... ");
	//fflush(stdout);
	M = zpla_alloc_matrix(n, n);
	for (j = 0; j < n; j++) {
		printf("g2="); zppoly_print(&g2s[j], p); printf("\n");

		//show_isom_stuff(&r1, &g1, &r2, &g2s[j], p, n, verb);

		// Find the matrix that maps the standard basis.
		// If x^k = gi ^ lk, then phi(x^k) = gj ^ lk (mod rj).
		for (k = 0; k < n; k++) {
			phixp = zppolymod_power(&g2s[j], logs[k], &aux2);
			for (l = 0; l < n; l++)
				M[l][k] = 0;
			for (l = 0; l <= phixp.degree; l++)
				M[l][k] = phixp.coeffs[l];
		}

		zpla_print_matrix(M, n, n);

		//if (xxx) {
		//    test if gi, gj are really generators
		//    test if is really an isom
		//    make phi map from generators; verify against matrix.
		//    show full phi listing
		//    whatever
		//}
		printf("\n");
	}
	zpla_free_matrix(M, n, n);
	//printf("done.\n");

	free(g2s);

	return 0;
}
