#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2polymod.h"
#include "isprime.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2poly_t r, numer, denom, g, rem;
	z2poly_t x1 = z2poly_from_base_rep(3);
	z2poly_t one = z2poly_from_base_rep(1);
	int argi;
	int k, m, num_bits, prim_bits;

	if (argc < 2)
		usage(argv[0]);
	for (argi = 1; argi < argc; argi++) {
		if (!z2poly_scan(&r, argv[argi])) {
			fprintf(stderr, "Couldn't scan \"%s\".\n", argv[argi]);
			continue;
		}
		// Fitzgerald criterion for primitivity:
		// r(x) is a polynomial of degree k with coefficients in Fpn,
		// with q = p^n.  Let m = q^k - 1.  Define g(x) =
		// (x^m - 1) / (x - 1)r(x).  r(x) is primitive iff g(x)
		// has exactly (q - 1) q^(k-1) - 1 non-zero terms.

		// Here, p = 2, n = 1, q = 2, m = 2^k - 1, prim # bits =
		// 2^(k-1) - 1.
		k = z2poly_get_degree(&r);
		m = (1 << k) - 1;
		z2poly_promote_n(&numer, &one, m);
		z2poly_add(&numer, &numer, &one);
		z2poly_mul(&denom, &x1, &r);
		z2poly_div(&g, &numer, &denom);
		z2poly_mod(&rem, &numer, &denom);
		if (!z2poly_equals_unsigned(&rem, 0LL)) {
			z2poly_print(&r);
			printf(" not primitive.\n");
		}
		prim_bits = (1 << (k - 1)) - 1;
		num_bits = z2poly_count_bits(&g);

		printf("r=");
		z2poly_print(&r);
		printf(" k=%d m=%d", k, m);
		//printf(" g=");
		//z2poly_print(&g);
		printf(" #bits=%3d primbits=%3d diff=%3d%s\n",
			num_bits, prim_bits,
			num_bits - prim_bits,
			num_bits == prim_bits ? " *" : "");

	}

	return 0;
}
