#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "isprime.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t r, numer, denom, g, rem;
	int argi;
	int k, m, num_bits, prim_bits;

	if (argc < 2)
		usage(argv[0]);
	for (argi = 1; argi < argc; argi++) {
		if (!z2ipoly_scan(&r, argv[argi])) {
			fprintf(stderr, "Couldn't scan \"%s\".\n", argv[argi]);
			continue;
		}
		// Fitzgerald criterion for primitivity:
		// r(x) is a polynomial of degree k with coefficients in Fpn,
		// with q = p^n.  Let m = q^k - 1.  Define g(x) =
		// (x^m - 1) / (x - 1)r(x).  r(x) is primitive iff g(x)
		// has exactly (q - 1) q^(k-1) - 1 non-zero terms.

		// Here, p = 2, n = 1, q = 2, m = 2^k - 1, prim # bits =
		// 2^(k-1) - 1.
		k = z2ipoly_get_degree(r);
		m = (1 << k) - 1;
		numer = (1LL << m) | 1LL;
		denom = z2ipoly_mul(3LL, r);
		g   = z2ipoly_div(numer, denom);
		rem = z2ipoly_mod(numer, denom);
		if (rem != 0LL) {
			printf("%llx not primitive.\n", r);
		}
		prim_bits = (1 << (k - 1)) - 1;
		num_bits = z2ipoly_count_bits(g);
		printf("r=%llx k=%d m=%d g=%llx #bits=%d primbits=%d%s\n",
			r, k, m, g, num_bits, prim_bits,
			num_bits == prim_bits ? " *" : "");

	}

	return 0;
}
