#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "zpnpoly_factor.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s {p} {a red} {b red} {exponent} {polys ...}\n",
		argv0);
	fprintf(stderr, "Example:\n");
	fprintf(stderr, "* Coefficient field is F_2^2 with a^2 + a + 1 = 0\n");
	fprintf(stderr, "* Extension   field is F_2^6 with b^6 + a + 1 = 0\n");
	fprintf(stderr, "* Polynomial  is x^3 + x^2 + (a+1) x + a\n");
	fprintf(stderr, "* a = b^21\n");
	fprintf(stderr, "Invoke as: %s 2 111 1000011 21 1:1:11:10\n",
		argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zppoly_aux_t a_aux;
	zpoly_t b_red;
	int ab_exp;
	zpnpoly_t ab_poly;
	zpoly_t flat_poly;
	int argi;
	int j, i;

	if (argc < 6)
		usage(argv[0]);
	if (!zppolymod_auxscan(&a_aux, argv[1], argv[2]))
		usage(argv[0]);
	if (!zppoly_scan(&b_red, argv[3], a_aux.p))
		usage(argv[0]);
	if (sscanf(argv[4], "%d", &ab_exp) != 1)
		usage(argv[0]);

	for (argi = 5; argi < argc; argi++) {

		if (!zpnpoly_scan(&ab_poly, argv[argi], &a_aux))
			usage(argv[0]);

		// * Coefficient field is F_2^2 with a^2 + a + 1 = 0
		// * Extension   field is F_2^6 with b^6 + a + 1 = 0
		// * Polynomial  is x^3 + x^2 + (a+1) x + a
		// * a = b^21
		//
		// a => b^21
		// x => b
		// Reduce mod b reductor.

		flat_poly = zppoly_from_base_rep(a_aux.p, 0);

		for (j = 0; j <= ab_poly.degree; j++) {
			for (i = 0; i <= ab_poly.coeffs[j].degree; i++) {
				zpoly_t term;

				//printf("%d b^%d\n",
					//ab_poly.coeffs[j].coeffs[i],
					//fm);
				term = zppoly_monomial_from_coeff(
					ab_poly.coeffs[j].coeffs[i],
					j + ab_exp * i,
					a_aux.p);

				flat_poly = zppoly_add(&flat_poly, &term,
					a_aux.p);
			}
		}
		flat_poly = zppoly_mod(&flat_poly, &b_red, a_aux.p);
		printf("<<");
		zppoly_print(&flat_poly, a_aux.p);
		printf(">>");
		printf("\n");
	}

	return 0;
}
