#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2nipoly_factor.h"
#include "power.h"
#include "mod.h"
#include "malloc_check.h"

#define SHOW_ELTS 3
#define SHOW_LOGS 4
#define SHOW_LOG_QUOTS 5

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-c] {r ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void show_iterated_map(
	z2ipoly_t f,
	z2ipoly_t r,
	int style)
{
	z2ipoly_t * fes;
	int n, nel, i, j, max;
	z2ipoly_t g = z2ipolymod_find_generator(r);
	int l1 = 0, l2 = 0, quot;
	int pn1;

	fes = z2ipolymod_get_elements_frobord(r, &nel);
	n = z2ipoly_get_degree(r);
	pn1 = (1 << z2ipoly_get_degree(r)) - 1;
	max = n + 3;
	printf("%llx %llx", f, r);
	if (style != SHOW_ELTS)
		printf(" %llx", g);
	printf("\n");
	for (j = 0; j < nel; j++) {
		for (i = 0; i <= max; i++) {
			if (style == SHOW_LOGS) {
				printf("%d", z2ipolymod_log(fes[j], g, r));
			}
			else if (style == SHOW_LOG_QUOTS) {
				l2 = z2ipolymod_log(fes[j], g, r);
				if (i == 0) {
					quot = l2;
				}
				else if (l1 == 0) {
					quot = 0;
				}
				else {
					//quot = mod_div(l2, l1, pn1);
					quot = mod_div(l1, l2, pn1);
				}
				printf("%d", quot);
				l1 = l2;
			}
			else {
				z2ipoly_print(fes[j]);
			}
			if (i == max)
				printf("\n");
			else
				printf(" ");
			fes[j] = z2ipolymod_compose(f, fes[j], r);
		}
	}
	free(fes);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t f, r;
	int argb, argi, style = SHOW_ELTS;

	if ((argc >= 2) && (strcmp(argv[1], "-l") == 0)) {
		style = SHOW_LOGS;
		argb = 3;
	}
	else if ((argc >= 2) && (strcmp(argv[1], "-q") == 0)) {
		style = SHOW_LOG_QUOTS;
		argb = 3;
	}
	else {
		style = SHOW_ELTS;
		argb = 2;
	}
	if (argc <= argb)
		usage(argv[0]);
	if (!z2ipoly_scan(&f, argv[argb-1]))
		usage(argv[0]);

	for (argi = argb; argi < argc; argi++) {
		if (!z2ipoly_scan(&r, argv[argi]))
			usage(argv[0]);
		show_iterated_map(f, r, style);
		if (argi < (argc - 1))
			printf("\n");
	}

	return 0;
}
