#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z3polymod.h"
#include "z3npoly_factor.h"
#include "power.h"
#include "mod.h"
#include "malloc_check.h"

#define SHOW_ELTS 3
#define SHOW_LOGS 4
#define SHOW_LOG_QUOTS 5

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-c] {r ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void show_iterated_map(
	z3poly_t * pf,
	z3poly_t * pr,
	int        style)
{
	z3poly_t * fes;
	int n, nel, i, j, max;
	z3poly_t g = z3polymod_find_generator(pr);
	int l1 = 0, l2 = 0, quot;
	int pn1;

	fes = z3polymod_get_elements_frobord(pr, &nel);
	n = z3poly_get_degree(pr);
	pn1 = calc_power(3, z3poly_get_degree(pr)) - 1;
	max = n + 3;

	z3poly_print(pf);
	printf(" ");
	z3poly_print(pr);
	if (style != SHOW_ELTS) {
		printf(" ");
		z3poly_print(&g);
	}
	printf("\n");

	for (j = 0; j < nel; j++) {
		for (i = 0; i <= max; i++) {
			if (style == SHOW_LOGS) {
				printf("%d", z3polymod_log(&fes[j], &g, pr));
			}
			else if (style == SHOW_LOG_QUOTS) {
				l2 = z3polymod_log(&fes[j], &g, pr);
				if (i == 0) {
					quot = l2;
				}
				else if (l1 == 0) {
					quot = 0;
				}
				else {
					//quot = mod_div(l2, l1, pn1);
					quot = mod_div(l1, l2, pn1);
				}
				printf("%d", quot);
				l1 = l2;
			}
			else {
				z3poly_print(&fes[j]);
			}
			if (i == max)
				printf("\n");
			else
				printf(" ");
			fes[j] = z3polymod_compose(pf, &fes[j], pr);
		}
	}
	free(fes);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z3poly_t f, r;
	int argb, argi, style = SHOW_ELTS;

	if ((argc >= 2) && (strcmp(argv[1], "-l") == 0)) {
		style = SHOW_LOGS;
		argb = 3;
	}
	else if ((argc >= 2) && (strcmp(argv[1], "-q") == 0)) {
		style = SHOW_LOG_QUOTS;
		argb = 3;
	}
	else {
		style = SHOW_ELTS;
		argb = 2;
	}
	if (argc <= argb)
		usage(argv[0]);
	if (!z3poly_scan(&f, argv[argb-1]))
		usage(argv[0]);

	for (argi = argb; argi < argc; argi++) {
		if (!z3poly_scan(&r, argv[argi]))
			usage(argv[0]);
		show_iterated_map(&f, &r, style);
		if (argi < (argc - 1))
			printf("\n");
	}

	return 0;
}
