#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "zpnpoly_factor.h"
#include "power.h"
#include "mod.h"
#include "malloc_check.h"

#define SHOW_ELTS 3
#define SHOW_LOGS 4
#define SHOW_LOG_QUOTS 5

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-c] {r ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void show_iterated_map(
	zpoly_t * pf,
	zpoly_t * pr,
	int        p,
	int        style)
{
	zpoly_t * fes;
	int n, nel, i, j, max;
	zpoly_t g;
	int l1 = 0, l2 = 0, quot;
	int pn1;
	zppoly_aux_t aux;

	aux.p = p;
	aux.r = *pr;
	g = zppolymod_find_generator(&aux);

	fes = zppolymod_get_elements_frobord(pr, aux.p, &nel);
	n = pr->degree;
	pn1 = calc_power(aux.p, pr->degree) - 1;
	max = n + 3;

	zppoly_print(pf, aux.p);
	printf(" ");
	zppoly_print(pr, aux.p);
	if (style != SHOW_ELTS) {
		printf(" ");
		zppoly_print(&g, aux.p);
	}
	printf("\n");

	for (j = 0; j < nel; j++) {
		for (i = 0; i <= max; i++) {
			if (style == SHOW_LOGS) {
				printf("%d", zppolymod_log(&fes[j], &g, &aux));
			}
			else if (style == SHOW_LOG_QUOTS) {
				l2 = zppolymod_log(&fes[j], &g, &aux);
				if (i == 0) {
					quot = l2;
				}
				else if (l1 == 0) {
					quot = 0;
				}
				else {
					//quot = mod_div(l2, l1, pn1);
					quot = mod_div(l1, l2, pn1);
				}
				printf("%d", quot);
				l1 = l2;
			}
			else {
				zppoly_print(&fes[j], aux.p);
			}
			if (i == max)
				printf("\n");
			else
				printf(" ");
			fes[j] = zppolymod_compose(pf, &fes[j], pr, p);
		}
	}
	free(fes);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zpoly_t f, r;
	int p, argb, argi, style = SHOW_ELTS;

	if ((argc >= 3) && (strcmp(argv[1], "-l") == 0)) {
		style = SHOW_LOGS;
		argb = 4;
	}
	else if ((argc >= 3) && (strcmp(argv[1], "-q") == 0)) {
		style = SHOW_LOG_QUOTS;
		argb = 4;
	}
	else {
		style = SHOW_ELTS;
		argb = 3;
	}
	if (argc <= argb)
		usage(argv[0]);
	if (sscanf(argv[argb-2], "%d", &p) != 1)
		usage(argv[0]);
	if (!zppoly_scan(&f, argv[argb-1], p))
		usage(argv[0]);

	for (argi = argb; argi < argc; argi++) {
		if (!zppoly_scan(&r, argv[argi], p))
			usage(argv[0]);
		show_iterated_map(&f, &r, p, style);
		if (argi < (argc - 1))
			printf("\n");
	}

	return 0;
}
