// Computes Zech's (Jacobi's) log:  Given a generator g of the multiplicative
// group of a finite field:
//
//  1 + g^k = g^L(k)
// L(k) = log_g (g^k + 1)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "power.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r} {a}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zppoly_aux_t aux;
	zpoly_t g, a, one, a1;
	int pn, i, L;

	if ((argc != 3) && (argc < 4))
		usage(argv[0]);
	if (!zppolymod_auxscan(&aux, argv[1], argv[2]))
		usage(argv[0]);
	if (argc == 3) {
		if (!zppoly_scan(&g, argv[2], aux.p))
			usage(argv[0]);
	}
	else {
		g = zppolymod_find_generator(&aux);
	}
	pn = calc_power(aux.p, aux.r.degree);
	one = zppoly_from_base_rep(aux.p, 1);

	a = g;
	for (i = 0; i < pn-1; i++) {
		a1 = zppoly_add(&a, &one, aux.p);
		L = zppolymod_log(&a1, &g, &aux);
		zppolymod_print(&a, &aux);
		printf(" %d %d\n", i+1, L);
		a = zppolymod_mul(&a, &g, &aux);
	}

	return 0;
}
