#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2nipoly_factor.h"
#include "power.h"
#include "malloc_check.h"

typedef struct _z2ipair_t {
	z2ipoly_t z;
	z2ipoly_t r;
} z2ipair_t;

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {irrs ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t * irrs;
	z2ipoly_t * fes;
	int argi, i, j, nirrs, n, nel;

	if (argc < 2)
		usage(argv[0]);

	nirrs = argc - 1;
	irrs = malloc_check(nirrs * sizeof(z2ipoly_t));

	for (argi = 1, i = 0; argi < argc; argi++, i++) {
		if (!z2ipoly_scan(&irrs[i], argv[argi]))
			usage(argv[0]);
	}

	n = z2ipoly_get_degree(irrs[0]);
	for (i = 1; i < nirrs; i++) {
		int nother = z2ipoly_get_degree(irrs[i]);
		if (nother != n) {
			printf("Irrs must all be of the same degree.\n");
			printf("Got %llx (deg=%d) and %llx (deg=%d).\n",
				irrs[0], n, irrs[i], nother);
			exit(1);
		}
	}

	for (i = 0; i < nirrs; i++) {
		fes = z2ipolymod_get_elements_frobord(irrs[i], &nel);
		for (j = 0; j < nel; j++) {
			if (j > 0)
				printf(" ");
			z2ipolymod_print(fes[j], irrs[i]);
		}
		printf("\n");

		free(fes);
	}

	free(irrs);
	return 0;
}
