// Check for any p such that 2 is non-square, but is also not a generator.
#include <stdio.h>
#include <stdlib.h>
#include "isprime.h"
#include "legsym.h"
#include "mod.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {plo[-phi]} {glo[-ghi]}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static int isgen(
	int x,
	int p)
{
	int o = mod_mul_order(x, p);
	return (o == (p - 1));
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int plo, phi, glo, ghi;
	int p, g;

	if (argc != 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d-%d", &plo, &phi) == 2) {
	}
	else if (sscanf(argv[1], "%d", &plo) == 1) {
		phi = plo;
	}
	else {
		usage(argv[0]);
	}

	if (sscanf(argv[2], "%d-%d", &glo, &ghi) == 2) {
	}
	else if (sscanf(argv[2], "%d", &glo) == 1) {
		ghi = glo;
	}
	else {
		usage(argv[0]);
	}

	for (p = plo; p <= phi; p++) {
		if (!isprime(p))
			continue;
		for (g = glo; g <= ghi; g++) {
			if (calc_legsym_32(g, p) != -1)
				continue;
			// Now know g is non-square.
			if (isgen(g, p))
				continue;
			// Now know g is a non-generator.
			printf("%4d: %4d\n",
				p, g);
		}
		if ((phi != plo) && (ghi != glo))
			printf("\n");
	}

	return 0;
}
