#include <stdio.h>
#include <stdlib.h>
#include "isprime.h"
#include "mod.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {max}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static int isgen(
	int x,
	int p)
{
	int o = mod_mul_order(x, p);
	return (o == (p - 1));
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int max = 20;
	int p;
	int twois, threeis, n2 = 0, n3 = 0, nboth = 0, nneither = 0, np = 0;
	int verbose = 0;

	if ((argc != 2) && (argc != 3))
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &max) != 1)
		usage(argv[0]);
	if (argc == 3)
		verbose = 1;

	for (p = 2; p < max; p++) {
		if (!isprime(p))
			continue;
		twois   = isgen(2, p);
		threeis = isgen(3, p);
		np++;
		if (twois) {
			if (threeis)
				nboth++;
			else
				n2++;
		}
		else {
			if (threeis)
				n3++;
			else
				nneither++;
		}
		if (verbose) {
			printf("%3d: %c %c\n",
				p,
				twois ? '2' : '-',
				threeis ? '3' : '-');
		}
	}

	printf("\n");

	printf("np       = %4d\n", np);
	printf("n2       = %4d (%3.0lf%%)\n",
		n2, 100.0 * (double)n2 / (double)np);
	printf("n3       = %4d (%3.0lf%%)\n",
		n3, 100.0 * (double)n2 / (double)np);
	printf("nboth    = %4d (%3.0lf%%)\n",
		nboth, 100.0 * (double)nboth / (double)np);
	printf("nneither = %4d (%3.0lf%%)\n",
		nneither, 100.0 * (double)nneither / (double)np);

	return 0;
}
