#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_ascendant.h"

#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char Zi_buf[MAX_GROUP_BYTES];
static unsigned char Zi1_buf[MAX_GROUP_BYTES];

// ----------------------------------------------------------------
int ascendant_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t group;
	group_t Zi;  // current
	group_t Zi1; // next
	int i, klo, khi;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need ascendant depth.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "ascendant depth";

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (sscanf(opts.cmd_paramv[0], "%d-%d", &klo, &khi) == 2) {
	}
	else if (sscanf(opts.cmd_paramv[0], "%d", &klo) == 1) {
		khi = klo;
	}
	else {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			opts.cmd_paramv[1]);
		exit(1);
	}

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	if (opts.echo_inputs) {
		describe_group(&group);
		print_group_elements(&group);
		printf("\n");
	}

	Zi.pvelements = Zi_buf;
	Zi.pvaux      = group.pvaux;
	Zi.pspec      = group.pspec;

	Zi1.pvelements = Zi1_buf;
	Zi1.pvaux      = group.pvaux;
	Zi1.pspec      = group.pspec;

	// Z0(G) is {e}.
	Zi.num_elements = 1;
	Zi1.num_elements = 1;
	if (!find_identity(&group, Zi_buf)) {
		fprintf(stderr,
		"ascendant:  Couldn't find identity in parent group.\n");
		exit(1);
	}
	memcpy(Zi1_buf, Zi_buf, group.pspec->element_size);

	for (i = 0; i <= khi; i++) {

		if (i >= klo) {
			if (klo != khi) {
				printf("\n");
				printf("%d-ascendant subgroup:\n", i);
			}
			print_group_elements(&Zi1);
		}
		if (!make_ascendant_subgroup(&group, &Zi, &Zi1,
			Zi1_buf, MAX_GROUP_BYTES, opts.verbosity))
		{
			fprintf(stderr,
				"%s:  Couldn't make %d-ascendant subgroup.\n",
				argv[0], i);
			exit(1);
		}

		Zi.num_elements = Zi1.num_elements;
		memcpy(Zi.pvelements, Zi1.pvelements,
			Zi1.num_elements * group.pspec->element_size);
	}

	return 0;
}
