#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"
#include "pmt.h"

#include "group.h"
#include "group_check.h"
#include "group_file.h"
#include "group_aut_group.h"

#define MAX_AUX_BYTES   2048
#define MAX_GROUP_BYTES 0x100000
#define MAX_AUT_BYTES   0x100000
static unsigned char aux_buf [MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char aut_buf [MAX_AUT_BYTES];

// ----------------------------------------------------------------
int aut_group_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t group;
	group_t aut_group;
	int i, j;
	pmt_t * ppmt;
	void * pg;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 0;
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = 0;

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	if (opts.echo_inputs)
		print_group_elements(&group);

	make_aut_group(&group, &aut_group,
		aut_buf, MAX_AUT_BYTES, opts.verbosity);

	if (opts.echo_inputs) {
		printf("\n");
		printf("Automorphism group:\n");
	}

	if (opts.verbosity >= 1) {
		for (i = 0; i < aut_group.num_elements; i++) {
			ppmt = VGROUP_ELEMENT(&aut_group, i);
			for (j = 0; j < group.num_elements; j++) {
				pg = VGROUP_ELEMENT(&group, ppmt->images[j]);
				printf(" ");
				group.pspec->pvprint(pg, group.pvaux);
			}
			printf("\n");
		}
		printf("\n");
	}

	print_group_elements(&aut_group);

	return 0;
}
