#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"
#include "group.h"
#include "group_check.h"
#include "group_centralizer.h"
#include "malloc_check.h"

#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x100000

static unsigned char * aux_buf;
static unsigned char * elts_buf;
static unsigned char * sgr_buf;
static unsigned char * centralizer_buf;

// ----------------------------------------------------------------
int centralizer_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t group;
	group_t subgroup;
	group_t centralizer;

	aux_buf         = (unsigned char *)malloc_check(MAX_AUX_BYTES);
	elts_buf        = (unsigned char *)malloc_check(MAX_GROUP_BYTES);
	sgr_buf         = (unsigned char *)malloc_check(MAX_GROUP_BYTES);
	centralizer_buf = (unsigned char *)malloc_check(MAX_GROUP_BYTES);

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 0;
	ospec.num_file_names    = 2;
	ospec.cmd_param_desc    = 0;

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (opts.file_namec < 1) {
		fprintf(stderr, "Need 1 or 2 file names.\n");
		exit(1);
	}

	if (!make_group(&subgroup, opts.pgspec, aux_buf,
		opts.file_namev[0], sgr_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make subgroup in \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 1 ? 0 : opts.file_namev[1],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group in \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}


	if (opts.echo_inputs) {
		describe_group(&group);
		print_group_elements(&subgroup);
		printf("\n");
		print_group_elements(&group);
		printf("\n");
	}

	make_centralizer(&group, &subgroup, &centralizer,
		centralizer_buf, MAX_GROUP_BYTES);

	if (opts.echo_inputs) {
		printf("\n");
		printf("Centralizer:\n");
	}
	print_group_elements(&centralizer);

	free(aux_buf);
	free(elts_buf);
	free(sgr_buf);
	free(centralizer_buf);

	return 0;
}
