#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_commutator.h"

#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char sgr_buf[MAX_GROUP_BYTES];

// ----------------------------------------------------------------
int commsgr_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t group;
	group_t commutator_subgroup;
	int i, klo, khi;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need commutator depth.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "commutator depth";

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (sscanf(opts.cmd_paramv[0], "%d-%d", &klo, &khi) == 2) {
	}
	else if (sscanf(opts.cmd_paramv[0], "%d", &klo) == 1) {
		khi = klo;
	}
	else {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			opts.cmd_paramv[1]);
		exit(1);
	}

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	if (opts.echo_inputs) {
		describe_group(&group);
		print_group_elements(&group);
	}

	commutator_subgroup.pvelements = sgr_buf;
	commutator_subgroup.pvaux      = group.pvaux;
	commutator_subgroup.pspec      = group.pspec;

	commutator_subgroup.num_elements = group.num_elements;
	memcpy(commutator_subgroup.pvelements, group.pvelements,
		group.num_elements * group.pspec->element_size);

	for (i = 0; i <= khi; i++) {

		if (i >= klo) {
			if (klo != khi) {
				printf("\n");
				printf("%d-commutator subgroup:\n", i);
			}
			print_group_elements(&commutator_subgroup);
		}

		if (!make_commutator_subgroup(&group, &commutator_subgroup,
			sgr_buf, MAX_GROUP_BYTES, opts.verbosity))
		{
			fprintf(stderr,
				"%s:  Couldn't make %d-commutator subgroup.\n",
				argv[0], i);
			exit(1);
		}

		group.num_elements = commutator_subgroup.num_elements;
		memcpy(group.pvelements, commutator_subgroup.pvelements,
			commutator_subgroup.num_elements
			* group.pspec->element_size);

	}

	return 0;
}
