#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_conj_elt.h"

#define MAX_GEN_BYTES 2048
#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x1000000
static unsigned char elt_buf[MAX_GEN_BYTES];
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char sv_buf[MAX_GROUP_BYTES];

// ----------------------------------------------------------------
int conjelt_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t group;
	int i;
	void * pa, * pb;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need g as conjugator.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "element";

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (!opts.pgspec->pvscan(elt_buf, opts.cmd_paramv[0], aux_buf,
		sizeof(elt_buf)))
	{
		fprintf(stderr, "Couldn't scan \"%s\" as element.\n",
			opts.cmd_paramv[0]);
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	memcpy(sv_buf, elts_buf,
		group.num_elements * group.pspec->element_size);
	conjugate_by_element(&group, &elt_buf);

	for (i = 0; i < group.num_elements; i++) {
		pa = VARRAY_ELEMENT(sv_buf,   i, group.pspec->element_size);
		pb = VARRAY_ELEMENT(elts_buf, i, group.pspec->element_size);
		printf("  ");
		group.pspec->pvprint(pa, group.pvaux);
		printf("  ");
		group.pspec->pvprint(pb, group.pvaux);
		printf("\n");
	}
	printf("\n");

	return 0;
}
