#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_file.h"

#define MAX_AUX_BYTES  2048
#define MAX_ELT_BYTES 16384
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elt_buf1[MAX_ELT_BYTES];
static unsigned char elt_buf2[MAX_ELT_BYTES];
static unsigned char elt_buf3[MAX_ELT_BYTES];

// ----------------------------------------------------------------
int op_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 0;
	ospec.num_file_names    = 2;
	ospec.cmd_param_desc    = 0;

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (opts.file_namec != 2) {
		fprintf(stderr, "Need 2 elements on the command line.\n");
		exit(1);
	}

	if (!opts.pgspec->pvscan(elt_buf1, opts.file_namev[0], aux_buf,
		sizeof(elt_buf1)))
	{
		fprintf(stderr, "Couldn't scan \"%s\" as element.\n",
			opts.file_namev[0]);
		exit(1);
	}

	if (opts.file_namev[1][0] == '^') {
		int power;
		if (sscanf(&opts.file_namev[1][1], "%d", &power) != 1) {
			fprintf(stderr, "Couldn't scan \"%s\" as integer power.\n",
				opts.file_namev[1]);
			exit(1);
		}
		if (power <= 0) {
			fprintf(stderr, "Sorry, power must be postive.\n");
			exit(1);
		}

		if (opts.echo_inputs) {
			opts.pgspec->pvprint(elt_buf1, aux_buf);
			printf(" ^ %d = ", power);
		}

		memcpy(elt_buf3, elt_buf1, opts.pgspec->element_size);
		while (--power) {
			opts.pgspec->pvop(elt_buf3, elt_buf1, elt_buf3, aux_buf);
		}

	}
	else {

		if (!opts.pgspec->pvscan(elt_buf2, opts.file_namev[1], aux_buf,
			sizeof(elt_buf2)))
		{
			fprintf(stderr, "Couldn't scan \"%s\" as element.\n",
				opts.file_namev[1]);
			exit(1);
		}

		if (opts.echo_inputs) {
			opts.pgspec->pvprint(elt_buf1, aux_buf);
			printf(" * ");
			opts.pgspec->pvprint(elt_buf2, aux_buf);
			printf(" = ");
		}

		opts.pgspec->pvop(elt_buf3, elt_buf1, elt_buf2, aux_buf);

	}

	opts.pgspec->pvprint(elt_buf3, aux_buf);
	printf("\n");

	return 0;
}
