#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_commutator.h"

#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char sgr_buf[MAX_GROUP_BYTES];

// ----------------------------------------------------------------
int solvable_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t G;
	group_t Gprime;
	int i;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 0; // Need commutator depth.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = 0;

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (!make_group(&G, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	if (opts.echo_inputs) {
		describe_group(&G);
		print_group_elements(&G);
	}

	Gprime.pvelements = sgr_buf;
	Gprime.pvaux      = G.pvaux;
	Gprime.pspec      = G.pspec;

	Gprime.num_elements = G.num_elements;
	memcpy(Gprime.pvelements, G.pvelements,
		G.num_elements * G.pspec->element_size);

	for (i = 0; ; i++) {

		if (!make_commutator_subgroup(&G, &Gprime,
			sgr_buf, MAX_GROUP_BYTES, opts.verbosity))
		{
			fprintf(stderr,
				"%s:  Couldn't make %d-commutator subgroup.\n",
				argv[0], i);
			exit(1);
		}

		if (G.num_elements == Gprime.num_elements) // Stabilized
			break;

		G.num_elements = Gprime.num_elements;
		memcpy(G.pvelements, Gprime.pvelements,
			Gprime.num_elements * G.pspec->element_size);

	}

	if (Gprime.num_elements == 1)
		printf("solvable (depth %d)\n", i);
	else
		printf("NOT solvable\n");

	return 0;
}
