#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_stabilizer.h"

#define MAX_X_BYTES 2048
#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x100000
static unsigned char x_buf[MAX_X_BYTES];
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char stabilizer_buf[MAX_GROUP_BYTES];

// ----------------------------------------------------------------
int stabilizer_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t group;
	group_t stabilizer;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need g as generator.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "element";

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (!opts.pgspec->pvscan(x_buf, opts.cmd_paramv[0], aux_buf,
		sizeof(x_buf)))
	{
		fprintf(stderr, "Couldn't scan \"%s\" as group element.\n",
			opts.cmd_paramv[0]);
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}


	if (opts.echo_inputs) {
		describe_group(&group);
		print_group_elements(&group);
	}

	make_stabilizer(&group, x_buf, &stabilizer,
		stabilizer_buf, MAX_GROUP_BYTES);

	if (opts.echo_inputs) {
		printf("\n");
		printf("Centralizer of ");
		group.pspec->pvprint(x_buf, group.pvaux);
		printf(":\n");
	}
	print_group_elements(&stabilizer);

	return 0;
}
