#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_file.h"

#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char sgr_buf[MAX_GROUP_BYTES];

// ----------------------------------------------------------------
int subgroup_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t subgroup;
	group_t group;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 0;
	ospec.num_file_names    = 2;
	ospec.cmd_param_desc    = 0;

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (opts.file_namec < 1) {
		fprintf(stderr, "Need 1 or 2 file names.\n");
		exit(1);
	}

	if (!make_group(&subgroup, opts.pgspec, aux_buf,
		opts.file_namev[0], sgr_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make subgroup in \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 1 ? 0 : opts.file_namev[1],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group in \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

	if (opts.echo_inputs) {
		print_group_elements(&subgroup);
		printf("\n");
		print_group_elements(&group);
		printf("\n");
	}

	if (is_subgroup(&subgroup, &group, opts.verbosity))
		printf("Subgroup.\n");
	else
		printf("Not a subgroup.\n");

	return 0;
}
