#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "group_opts.h"

#include "group.h"
#include "group_check.h"
#include "group_torsion.h"

#define MAX_AUX_BYTES 2048
#define MAX_GROUP_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_GROUP_BYTES];
static unsigned char sgr_buf[MAX_GROUP_BYTES];

// ----------------------------------------------------------------
int torsion_cmd(int argc, char ** argv)
{
	group_opts_spec_t ospec;
	group_opts_t opts;
	group_t group;
	group_t torsion_subgroup;
	int k;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need k for k-torsion
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "k";

	if (!group_parse_opts(argc, argv, &ospec, &opts))
		group_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!sscanf(opts.cmd_paramv[0], "%d", &k)) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			opts.cmd_paramv[1]);
		exit(1);
	}

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse group aux.\n");
		exit(1);
	}

	if (!make_group(&group, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_GROUP_BYTES))
	{
		fprintf(stderr, "Couldn't make group \"%s\".\n",
			opts.group_spec_desc);
		exit(1);
	}

/*
	This is expensive to compute for larger groups.  Nonetheless the
	user (me) is warned that k-torsion subsets are only guaranteed to be
	subgroups if the parent group is abelian.  The isgroup command may be
	used, if desired, to verify that a k-torsion subset is actually a
	subgroup.
	if (!is_abelian(&group, 0)) {
		printf("Group is non-abelian; concept of k-torsion "
			"subgroup is meaningless.\n");
		exit(1);
	}
*/


	if (opts.echo_inputs) {
		describe_group(&group);
		print_group_elements(&group);
	}

	make_torsion_subgroup(&group, k, &torsion_subgroup,
		sgr_buf, MAX_GROUP_BYTES);

	if (opts.echo_inputs) {
		printf("\n");
		printf("%d-torsion subgroup:\n", k);
	}
	print_group_elements(&torsion_subgroup);

	return 0;
}
