#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "opts.h"
#include "common_math.h"
#include "cmps.h"

// ----------------------------------------------------------------
int common_gcd(
	void * g,
	void * a,
	void * b,
	opts_t * popts)
{
	char r[MAX_ATOM_SIZE];
	char t[MAX_ATOM_SIZE];
	char zero[MAX_ATOM_SIZE];

	if (popts->pops->pgcd_op) {
		popts->pops->pgcd_op(g, a, b, popts->aux_buf);
		return 1;
	}

	if (!popts->pops->pfrom_int_op)
		return 0;
	if (!popts->pops->pcmp_op)
		return 0;
	if (!popts->pops->pmod_op)
		return 0;

	popts->pops->pfrom_int_op(zero, 0);
	memcpy(g, zero, MAX_ATOM_SIZE);
	if (popts->pops->pcmp_op(a, CMP_EQ, zero, popts->aux_buf))
		return 1;
	if (popts->pops->pcmp_op(b, CMP_EQ, zero, popts->aux_buf))
		return 1;

	if (popts->pops->pcmp_op(a, CMP_LT, b, popts->aux_buf)) {
		memcpy(t, a, MAX_ATOM_SIZE);
		memcpy(a, b, MAX_ATOM_SIZE);
		memcpy(b, t, MAX_ATOM_SIZE);
	}

	//while (b > 0) {
	//	r = a % b;
	//	a = b;
	//	b = r;
	//}
	//return a;

	while(popts->pops->pcmp_op(b, CMP_GT, zero, popts->aux_buf)) {
		popts->pops->pmod_op(r, a, b, popts->aux_buf);
		memcpy(a, b, MAX_ATOM_SIZE);
		memcpy(b, r, MAX_ATOM_SIZE);
	}
	memcpy(g, a, MAX_ATOM_SIZE);

	return 1;
}

// ----------------------------------------------------------------
// (6 0) = 1
// (6 1) = 6       / 1
// (6 2) = 6 5     / 1 2
// (6 3) = 6 5 4   / 1 2 3
// (6 4) = 6 5 4 3 / 1 2 3 4

// (20 19) = (20 1)

int common_choose(
	void * c,
	void * a,
	void * b,
	opts_t * popts)
{
	char zero[MAX_ATOM_SIZE];
	char one[MAX_ATOM_SIZE];
	char two[MAX_ATOM_SIZE];
	char half_a[MAX_ATOM_SIZE];
	char hi [MAX_ATOM_SIZE];
	char lo [MAX_ATOM_SIZE];
	char ii [MAX_ATOM_SIZE];

	if (popts->pops->pchoose_op) {
		popts->pops->pchoose_op(c, a, b, popts->aux_buf);
		return 1;
	}
	if (!popts->pops->pfrom_int_op)
		return 0;
	if (!popts->pops->pcmp_op)
		return 0;
	if (!popts->pops->pmod_op)
		return 0;
	if (!popts->pops->pdiv_op)
		return 0;

	popts->pops->pfrom_int_op(zero, 0);
	popts->pops->pfrom_int_op(one, 1);
	popts->pops->pfrom_int_op(two, 2);
	popts->pops->pdiv_op(half_a, a, two, popts->aux_buf);
	if (popts->pops->pcmp_op(b, CMP_GT, half_a, popts->aux_buf))
		popts->pops->psub_op(b, a, b, popts->aux_buf);
	memcpy(hi, a,   MAX_ATOM_SIZE);
	memcpy(lo, one, MAX_ATOM_SIZE);
	popts->pops->pfrom_int_op(ii, 1);
	popts->pops->pfrom_int_op(c, 1);
	while (popts->pops->pcmp_op(ii, CMP_LE, b, popts->aux_buf)) {
		if (popts->pops->pcmp_op(c, CMP_EQ, zero, popts->aux_buf))
			break;
		popts->pops->pmul_op(c, c, hi, popts->aux_buf);
		popts->pops->pdiv_op(c, c, lo, popts->aux_buf);
		popts->pops->psub_op(hi, hi, one, popts->aux_buf);
		popts->pops->padd_op(lo, lo, one, popts->aux_buf);
		popts->pops->padd_op(ii, ii, one, popts->aux_buf);
	}

	return 1;
}

// ----------------------------------------------------------------
int common_factorial(
	void * c,
	void * a,
	opts_t * popts)
{
	char zero[MAX_ATOM_SIZE];
	char one[MAX_ATOM_SIZE];
	char loop[MAX_ATOM_SIZE];

	if (popts->pops->pfct_op) {
		popts->pops->pfct_op(c, a, popts->aux_buf);
		return 1;
	}

	if (!popts->pops->pfrom_int_op)
		return 0;
	if (!popts->pops->pcmp_op)
		return 0;

	//c = 1;
	//loop = a;
	//while (loop > 0) {
	//	c = c * loop;
	//	loop--;
	//}
	//return c;

	popts->pops->pfrom_int_op(zero, 0);
	popts->pops->pfrom_int_op(one, 1);
	memcpy(c,    one, MAX_ATOM_SIZE);
	memcpy(loop, a,   MAX_ATOM_SIZE);

	while (popts->pops->pcmp_op(loop, CMP_GT, zero, popts->aux_buf)) {
		popts->pops->pmul_op(c, c, loop, popts->aux_buf);
		popts->pops->psub_op(loop, loop, one, popts->aux_buf);
	}

	return 1;
}

// ----------------------------------------------------------------
char * cmp_name(
	int cmp)
{
	switch (cmp) {
	case CMP_LT: return "<" ; break;
	case CMP_LE: return "<="; break;
	case CMP_EQ: return "=="; break;
	case CMP_NE: return "!+"; break;
	case CMP_GT: return ">" ; break;
	case CMP_GE: return ">="; break;
	default:     return "??"; break;
	}
}
