#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "ops.h"
#include "lexan.h"

typedef struct _dcomplex_t {
	double re;
	double im;
} dcomplex_t;

// ----------------------------------------------------------------
void dcomplex_syms(symtab_t * pst)
{
	int stidx;
	dcomplex_t pi, e;

	e.re  = exp(1.0);
	e.im  = 0.0;

	pi.re = 4.0 * atan(1.0);
	pi.im = 0.0;

	st_insert(pst, "pi", L_ID, &stidx);
	st_set_val(pst, stidx, (char *)&pi);

	st_insert(pst, "e", L_ID, &stidx);
	st_set_val(pst, stidx, (char *)&e);
}

// ----------------------------------------------------------------
static int dcomplex_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (c == 'i')
		return 1;
	else if (isdigit(c))
		return 1;
	else if (c == '.')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int dcomplex_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	dcomplex_t dcomplex;
	int ch;

	ch = getc(fp);
	if (ch == 'i') {
		dcomplex.re = 0.0;
		dcomplex.im = 1.0;
	}
	else {
		ungetc(ch, fp);

		dcomplex.im = 0.0;
		if (fscanf(fp, "%lf", &dcomplex.re) != 1)
			return 0;

		ch = getc(fp);
		if (ch == 'i') {
			dcomplex.im = dcomplex.re;
			dcomplex.re = 0.0;
		}
		else {
			ungetc(ch, fp);
		}

		//ch = getc(fp);
		//if (ch != ',') {
			//ungetc(ch, fp);
		//}
		//else {
			//if (fscanf(fp, "%f", &dcomplex.im) != 1)
				//return 0;
		//}
	}
	memcpy(patom, &dcomplex, sizeof(dcomplex));
	return 1;
}


// ----------------------------------------------------------------
static void dcomplex_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	dcomplex_t x;
	memcpy(&x, patom, sizeof(x));

	if (x.re == 0.0) {
		if (x.im == 0.0) {
			fprintf(fp, "%lf", x.re);
		}
		else {
			if (x.im == 1.0)
				fprintf(fp, "i");
			else if (x.im == -1.0)
				fprintf(fp, "-i");
			else
				fprintf(fp, "%lfi", x.im);
		}
	}
	else {
		fprintf(fp, "%lf", x.re);
		if (x.im == 0.0) {
			;
		}
		else {
			if (x.im == 1.0)
				fprintf(fp, "+i");
			else if (x.im == -1.0)
				fprintf(fp, "-i");
			else if (x.im < 0.0)
				fprintf(fp, "-%lfi", -x.im);
			else
				fprintf(fp, "+%lfi", x.im);
		}
	}
}

// ----------------------------------------------------------------
// xxx needs cmd-line opt for conditional.
static void dcomplex_rectifier(void * buf, void * pvaux)
{
	dcomplex_t c;
	memcpy(&c, buf, sizeof(c));
	if (fabs(c.re) < 1e-7)
		c.re = 0.0;
	if (fabs(c.im) < 1e-7)
		c.im = 0.0;
	memcpy(buf, &c, sizeof(c));
}

// ----------------------------------------------------------------
static void dcomplex_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dcomplex_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c.re = a.re + b.re;
	c.im = a.im + b.im;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dcomplex_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dcomplex_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c.re = a.re - b.re;
	c.im = a.im - b.im;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dcomplex_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dcomplex_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	c.re = a.re * b.re - a.im * b.im;
	c.im = a.re * b.im + a.im * b.re;

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dcomplex_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dcomplex_t a, b, c;
	double bot;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	c.re = a.re * b.re + a.im * b.im;
	c.im = a.im * b.re - a.re * b.im;

	bot = b.re * b.re + b.im * b.im;

	c.re /= bot;
	c.im /= bot;

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dcomplex_uadd(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	memcpy(pvout, pvin, sizeof(dcomplex_t));
}

// ----------------------------------------------------------------
static void dcomplex_usub(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t c;
	memcpy(&c, pvin, sizeof(c));
	c.re = -c.re;
	c.im = -c.im;
	memcpy(pvout, &c, sizeof(c));
}

// ----------------------------------------------------------------
static void dcomplex_sqrt(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t a, c;
	double mag, phz;
	memcpy(&a, pvin, sizeof(a));
	mag = hypot(a.re, a.im);
	phz = atan2(a.im, a.re);
	mag = sqrt(mag);
	phz = phz * 0.5;
	if (a.im == 0.0) {
		phz = -phz;
	}
	c.re = mag * cos(phz);
	c.im = mag * sin(phz);
	memcpy(pvout, &c, sizeof(c));
}

// ----------------------------------------------------------------
// e ** (x + iy) = e**x * e**iy
static void dcomplex_exp(dcomplex_t * pc, dcomplex_t * pa)
{
	double expx;
	expx = exp(pa->re);
	pc->re = expx * cos(pa->im);
	pc->im = expx * sin(pa->im);
}

// ----------------------------------------------------------------
static void dcomplex_vexp(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t a, c;
	memcpy(&a, pvin, sizeof(a));
	dcomplex_exp(&c, &a);
	memcpy(pvout, &c, sizeof(c));
}

// ----------------------------------------------------------------
// i**i = exp(i * log i)
// i**i = exp(i * (ln mag i + i phz i))
// i**i = exp(i * (ln 1 + i pi/2))
// i**i = exp(i * (0 + i pi/2))
// i**i = exp(i * (i pi/2))
// i**i = exp(i * i * pi/2)
// i**i = exp(-pi/2)

// Log(z) = ln mag z + i phz z
static void dcomplex_log(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t a, c;
	double mag, phz;
	memcpy(&a, pvin, sizeof(a));
	mag = hypot(a.re, a.im);
	phz = atan2(a.im, a.re);
	c.re = log(mag);
	c.im = phz;
	memcpy(pvout, &c, sizeof(c));
}

// ----------------------------------------------------------------
// (exp(iz) - exp(-iz)) / 2i
static void dcomplex_sin(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t z, c, iz, negiz, expiz, expnegiz, i, negi, twoi, temp;
	memcpy(&z, pvin, sizeof(z));
	i.re    =  0.0; i.im    =  1.0;
	negi.re =  0.0; negi.im = -1.0;
	twoi.re =  0.0; twoi.im =  2.0;
	dcomplex_mul(&iz,    &i,    &z, pvaux);
	dcomplex_mul(&negiz, &negi, &z, pvaux);
	dcomplex_exp(&expiz, &iz);
	dcomplex_exp(&expnegiz, &negiz);
	dcomplex_sub(&temp, &expiz, &expnegiz, pvaux);
	dcomplex_div(&c, &temp, &twoi, pvaux);
	memcpy(pvout, &c, sizeof(c));
}

// ----------------------------------------------------------------
// (exp(iz) + exp(-iz)) / 2
static void dcomplex_cos(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t z, c, iz, negiz, expiz, expnegiz, i, negi, two, temp;
	memcpy(&z, pvin, sizeof(z));
	i.re    =  0.0; i.im    =  1.0;
	negi.re =  0.0; negi.im = -1.0;
	two.re  =  2.0; two.im  =  0.0;
	dcomplex_mul(&iz,    &i,    &z, pvaux);
	dcomplex_mul(&negiz, &negi, &z, pvaux);
	dcomplex_exp(&expiz, &iz);
	dcomplex_exp(&expnegiz, &negiz);
	dcomplex_add(&temp, &expiz, &expnegiz, pvaux);
	dcomplex_div(&c, &temp, &two, pvaux);
	memcpy(pvout, &c, sizeof(c));
}

// ----------------------------------------------------------------
static void dcomplex_tan(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t a, c, cos_part, sin_part;
	memcpy(&a, pvin, sizeof(a));
	dcomplex_sin(&sin_part, &a, pvaux);
	dcomplex_cos(&cos_part, &a, pvaux);
	dcomplex_div(&c, &sin_part, &cos_part, pvaux);
	memcpy(pvout, &c, sizeof(c));
}

// ----------------------------------------------------------------
// c = a^b = exp(b log a)
static void dcomplex_pow(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dcomplex_t a, b, c, temp1, temp2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	dcomplex_log(&temp1, &a, pvaux);
	dcomplex_mul(&temp2, &temp1, &b, pvaux);
	dcomplex_exp(&c, &temp2);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dcomplex_conj(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t c, a;
	memcpy(&a,  pvin, sizeof(a));
	c.re =  a.re;
	c.im = -a.im;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dcomplex_mag(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t c, a;
	memcpy(&a,  pvin, sizeof(a));
	c.re = hypot(a.re, a.im);
	c.im = 0.0;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dcomplex_phz(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dcomplex_t c, a;
	memcpy(&a,  pvin, sizeof(a));
	c.re = atan2(a.im, a.re);
	c.im = 0.0;
	memcpy(pvout, &c, sizeof(a));
}

// ================================================================
ops_t dcomplex_ops = {
	atom_desc          : "dcomplex",
	psymtab_populator  : dcomplex_syms,
	plex_first_func    : dcomplex_lex_first,
	pscan_func         : dcomplex_scan,
	pprint_func        : dcomplex_print,
	prectifier         : dcomplex_rectifier,
	padd_op            : dcomplex_add,
	psub_op            : dcomplex_sub,
	pmul_op            : dcomplex_mul,
	pdiv_op            : dcomplex_div,
	ppow_op            : dcomplex_pow,
	puadd_op           : dcomplex_uadd,
	pusub_op           : dcomplex_usub,
	psqrt_op           : dcomplex_sqrt,
	pexp_op            : dcomplex_vexp,
	plog_op            : dcomplex_log,
	psin_op            : dcomplex_sin,
	pcos_op            : dcomplex_cos,
	ptan_op            : dcomplex_tan,
	pconj_op           : dcomplex_conj,
	pmag_op            : dcomplex_mag,
	pphz_op            : dcomplex_phz,
	atom_size_in_bytes : sizeof(dcomplex_t),
	aux_size_in_bytes  : 0
};
