#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "ops.h"
#include "symtab.h"
#include "lexan.h"

// ----------------------------------------------------------------
void double_syms(symtab_t * pst)
{
	int stidx;
	double pi = 4.0 * atan(1.0);
	double e  = exp(1.0);

	st_insert(pst, "pi", L_ID, &stidx);
	st_set_val(pst, stidx, (char *)&pi);

	st_insert(pst, "e", L_ID, &stidx);
	st_set_val(pst, stidx, (char *)&e);
}

// ----------------------------------------------------------------
static int double_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == '.')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int double_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	return fscanf(fp, "%lf", (double *)patom);
}

// ----------------------------------------------------------------
static void double_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	double x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, "%lf", x);
}

// ----------------------------------------------------------------
static void double_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	double a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a + b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	double a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a - b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	double a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a * b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	double a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a / b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	double a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = pow(a, b);
	memcpy(pvout, &c, sizeof(a));
}


// ----------------------------------------------------------------
static void double_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = -a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_sqrt (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = sqrt(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_exp (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = exp(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_log (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = log(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_sin (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = sin(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_cos (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = cos(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_tan (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = tan(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void double_atan (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	double a, c;
	memcpy(&a, pvin, sizeof(a));
	c = atan(a);
	memcpy(pvout, &c, sizeof(a));
}

// ================================================================
ops_t double_ops = {
	atom_desc          : "double",
	psymtab_populator  : double_syms,
	plex_first_func    : double_lex_first,
	pscan_func         : double_scan,
	pprint_func        : double_print,
	por_op             : 0,
	pxor_op            : 0,
	pand_op            : 0,
	plsh_op            : 0,
	prsh_op            : 0,
	padd_op            : double_add,
	psub_op            : double_sub,
	pmul_op            : double_mul,
	pdiv_op            : double_div,
	ppow_op            : double_power,
	pmod_op            : 0,
	puadd_op           : double_uadd,
	pusub_op           : double_usub,
	punot_op           : 0,
	psqrt_op           : double_sqrt,
	pexp_op            : double_exp,
	plog_op            : double_log,
	psin_op            : double_sin,
	pcos_op            : double_cos,
	ptan_op            : double_tan,
	patan_op           : double_atan,
	atom_size_in_bytes : sizeof(double),
	aux_size_in_bytes  : 0
};
