#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "ops.h"

typedef struct _dquat_t {
	double re;
	double I;
	double J;
	double K;
} dquat_t;

// ----------------------------------------------------------------
static int dquat_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (c == 'i')
		return 1;
	else if (c == 'j')
		return 1;
	else if (c == 'k')
		return 1;

	else if (isdigit(c))
		return 1;
	else if (c == '.')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int dquat_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	dquat_t dquat;
	int ch;

	dquat.re = 0.0;
	dquat.I  = 0.0;
	dquat.J  = 0.0;
	dquat.K  = 0.0;

	ch = getc(fp);
	if (ch == 'i') {
		dquat.I = 1.0;
	}
	else if (ch == 'j') {
		dquat.J = 1.0;
	}
	else if (ch == 'k') {
		dquat.K = 1.0;
	}
	else {
		ungetc(ch, fp);

		if (fscanf(fp, "%lf", &dquat.re) != 1)
			return 0;

		ch = getc(fp);
		if (ch == 'i') {
			dquat.I = dquat.re;
			dquat.re = 0.0;
		}
		else if (ch == 'j') {
			dquat.J = dquat.re;
			dquat.re = 0.0;
		}
		else if (ch == 'k') {
			dquat.K = dquat.re;
			dquat.re = 0.0;
		}
		else {
			ungetc(ch, fp);
		}
	}
	memcpy(patom, &dquat, sizeof(dquat));
	return 1;
}
//{
//	dquat_t dquat;
//	int ch;
//	dquat.I = 0.0;
//	dquat.J = 0.0;
//	dquat.K = 0.0;
//	if (fscanf(fp, "%lf", &dquat.re) != 1)
//		return 0;
//
//	ch = getc(fp);
//	if (ch != ',') {
//		ungetc(ch, fp);
//		memcpy(patom, &dquat, sizeof(dquat));
//		return 1;
//	}
//	if (fscanf(fp, "%lf", &dquat.I) != 1)
//		return 0;
//
//	ch = getc(fp);
//	if (ch != ',') {
//		ungetc(ch, fp);
//		memcpy(patom, &dquat, sizeof(dquat));
//		return 1;
//	}
//	if (fscanf(fp, "%lf", &dquat.J) != 1)
//		return 0;
//
//	ch = getc(fp);
//	if (ch != ',') {
//		ungetc(ch, fp);
//		memcpy(patom, &dquat, sizeof(dquat));
//		return 1;
//	}
//	if (fscanf(fp, "%lf", &dquat.K) != 1)
//		return 0;
//
//	memcpy(patom, &dquat, sizeof(dquat));
//	return 1;
//}

// ----------------------------------------------------------------
static void dquat_print1(FILE * fp, double e, void * pvtopts, int is_re)
{
	if (is_re) {
		if (e == 1.0)
			fprintf(fp, "1");
		else if (e == -1.0)
			fprintf(fp, "-1");
		else
			fprintf(fp, "%lf", e);
	}
	else if (e == 1.0) {
	}
	else if (e == -1.0) {
		fprintf(fp, "-");
	}
	else {
		fprintf(fp, "%lf*", e);
	}
}

// ----------------------------------------------------------------
static void dquat_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	dquat_t x;
	memcpy(&x, patom, sizeof(x));

	if (x.re == 0.0) {
		if (x.I == 0.0) {
			if (x.J == 0.0) {
				if (x.K == 0.0) {
					// 0 0 0 0 => 0
					dquat_print1(fp, x.re, pvtopts, 1);
				}
				else {
					// 0 0 0 1 => k
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
			else {
				if (x.K == 0.0) {
					// 0 0 1 0 => j
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j");
				}
				else {
					// 0 0 1 1 => j+k
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j+");
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
		}
		else {
			if (x.J == 0.0) {
				if (x.K == 0.0) {
					// 0 1 0 0 => i
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i");
				}
				else {
					// 0 1 0 1 => i+k
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i+");
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
			else {
				if (x.K == 0.0) {
					// 0 1 1 0 => i+j
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i+");
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j");
				}
				else {
					// 0 1 1 1 => i+j+k
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i+");
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j+");
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
		}
	}
	else {
		if (x.I == 0.0) {
			if (x.J == 0.0) {
				if (x.K == 0.0) {
					// 1 0 0 0 => 1
					dquat_print1(fp, x.re, pvtopts, 1);
				}
				else {
					// 1 0 0 1 => 1+k
					dquat_print1(fp, x.re, pvtopts, 1);
					fprintf(fp, "+");
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
			else {
				if (x.K == 0.0) {
					// 1 0 1 0 => 1+j
					dquat_print1(fp, x.re, pvtopts, 1);
					fprintf(fp, "+");
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j");
				}
				else {
					// 1 0 1 1 => 1+j+k
					dquat_print1(fp, x.re, pvtopts, 1);
					fprintf(fp, "+");
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j+");
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
		}
		else {
			if (x.J == 0.0) {
				if (x.K == 0.0) {
					// 1 1 0 0 => 1+i
					dquat_print1(fp, x.re, pvtopts, 1);
					fprintf(fp, "+");
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i");
				}
				else {
					// 1 1 0 1 => 1+i+k
					dquat_print1(fp, x.re, pvtopts, 1);
					fprintf(fp, "+");
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i+");
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
			else {
				if (x.K == 0.0) {
					// 1 1 1 0 => 1+i+j
					dquat_print1(fp, x.re, pvtopts, 1);
					fprintf(fp, "+");
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i+");
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j");
				}
				else {
					// 1 1 1 1 => 1+i+j+k
					dquat_print1(fp, x.re, pvtopts, 1);
					fprintf(fp, "+");
					dquat_print1(fp, x.I, pvtopts, 0);
					fprintf(fp, "i+");
					dquat_print1(fp, x.J, pvtopts, 0);
					fprintf(fp, "j+");
					dquat_print1(fp, x.K, pvtopts, 0);
					fprintf(fp, "k");
				}
			}
		}
	}
}

// ----------------------------------------------------------------
static void dquat_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dquat_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c.re = a.re + b.re;
	c.I  = a.I  + b.I;
	c.J  = a.J  + b.J;
	c.K  = a.K  + b.K;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dquat_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dquat_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c.re = a.re - b.re;
	c.I  = a.I  - b.I;
	c.J  = a.J  - b.J;
	c.K  = a.K  - b.K;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
// i*i = -1   j*j = -1   k*k = -1
// i*j = k    j*k =  i   k*i =  j
// i*k = -j   j*i = -k   k*j = -i

// (a + bi + cj + dk) * (w + xi + yj + zk)

//   (a + bi + cj + dk)
// * (w + xi + yj + zk)
//  --------------------
//  a * w
//          b * wi
//                  c * wj
//                          d * wk
//          a * xi
// -b * x
//                         -c * xk
//                  d * xj
//                  a * yj
//                          b * yk
// -c * y
//         -d * yi
//                          a * zk
//                 -b * zj
//          c * zi
// -d * z

// (aw -bx -cy -dz)
// (bw  ax -dy  cz) i
// (cw  dx  ay -bz) j
// (dw -cx  by  az) k

// (a.re* b.re  -a.I * b.I  -a.J * b.J  -a.K * b.K)
// (a.I * b.re   a.re* b.I  -a.K * b.J   a.J * b.K) i
// (a.J * b.re   a.K * b.I   a.re* b.J  -a.I * b.K) j
// (a.K * b.re  -a.J * b.I   a.I * b.J   a.re* b.K) k

static void dquat_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dquat_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	c.re = a.re* b.re - a.I  * b.I - a.J  * b.J - a.K  * b.K;
	c.I  = a.I * b.re + a.re * b.I - a.K  * b.J + a.J  * b.K;
	c.J  = a.J * b.re + a.K  * b.I + a.re * b.J - a.I  * b.K;
	c.K  = a.K * b.re - a.J  * b.I + a.I  * b.J + a.re * b.K;

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
// (a + bi + cj + dk) * (a - bi - cj - dk) = a^2 + b^2 + c^2 + d^2

//  a + bi + cj + dk
// ------------------
//  w + xi + yj + zk

// (a + bi + cj + dk) * (w - xi - yj - zk)
// -----------------------------------------
// (w + xi + yj + zk) * (w - xi - yj - zk)

// (a + bi + cj + dk) * (w - xi - yj - zk)
// -----------------------------------------
// w^2 + x^2 + y^2 + z^2

// (a.re* b.re   a.I * b.I   a.J * b.J   a.K * b.K)
// (a.I * b.re  -a.re* b.I   a.K * b.J  -a.J * b.K) i
// (a.J * b.re  -a.K * b.I  -a.re* b.J   a.I * b.K) j
// (a.K * b.re   a.J * b.I  -a.I * b.J  -a.re* b.K) k
// -----------------------------------------
// w^2 + x^2 + y^2 + z^2

static void dquat_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	dquat_t a, b, c;
	double bot;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	c.re = a.re* b.re + a.I  * b.I + a.J  * b.J + a.K  * b.K;
	c.I  = a.I * b.re - a.re * b.I + a.K  * b.J - a.J  * b.K;
	c.J  = a.J * b.re - a.K  * b.I - a.re * b.J + a.I  * b.K;
	c.K  = a.K * b.re + a.J  * b.I - a.I  * b.J - a.re * b.K;

	bot = b.re * b.re + b.I * b.I + b.J * b.J + b.K * b.K;

	c.re /= bot;
	c.I  /= bot;
	c.J  /= bot;
	c.K  /= bot;

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void dquat_uadd(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	memcpy(pvout, pvin, sizeof(dquat_t));
}

// ----------------------------------------------------------------
static void dquat_usub(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	dquat_t c;
	memcpy(&c, pvin, sizeof(c));
	c.re = -c.re;
	c.I  = -c.I;
	c.J  = -c.J;
	c.K  = -c.K;
	memcpy(pvout, &c, sizeof(c));
}

// ================================================================
ops_t dquat_ops = {
	atom_desc          : "dquat",
	plex_first_func    : dquat_lex_first,
	pscan_func         : dquat_scan,
	pprint_func        : dquat_print,
	padd_op            : dquat_add,
	psub_op            : dquat_sub,
	pmul_op            : dquat_mul,
	pdiv_op            : dquat_div,
	puadd_op           : dquat_uadd,
	pusub_op           : dquat_usub,
	atom_size_in_bytes : sizeof(dquat_t),
	aux_size_in_bytes  : 0
};
