#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "ops.h"

typedef struct _fcomplex_t {
	float re;
	float im;
} fcomplex_t;

// ----------------------------------------------------------------
static int fcomplex_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (c == 'i')
		return 1;
	else if (isdigit(c))
		return 1;
	else if (c == '.')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int fcomplex_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	fcomplex_t fcomplex;
	int ch;

	ch = getc(fp);
	if (ch == 'i') {
		fcomplex.re = 0.0;
		fcomplex.im = 1.0;
	}
	else {
		ungetc(ch, fp);

		fcomplex.im = 0.0;
		if (fscanf(fp, "%f", &fcomplex.re) != 1)
			return 0;

		ch = getc(fp);
		if (ch == 'i') {
			fcomplex.im = fcomplex.re;
			fcomplex.re = 0.0;
		}
		else {
			ungetc(ch, fp);
		}


		//ch = getc(fp);
		//if (ch != ',') {
			//ungetc(ch, fp);
		//}
		//else {
			//if (fscanf(fp, "%f", &fcomplex.im) != 1)
				//return 0;
		//}
	}
	memcpy(patom, &fcomplex, sizeof(fcomplex));
	return 1;
}

// ----------------------------------------------------------------
static void fcomplex_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	fcomplex_t x;
	memcpy(&x, patom, sizeof(x));

	if (x.re == 0.0) {
		if (x.im == 0.0) {
			fprintf(fp, "%f", x.re);
		}
		else {
			if (x.im == 1.0)
				fprintf(fp, "i");
			else if (x.im == -1.0)
				fprintf(fp, "-i");
			else
				fprintf(fp, "%f*i", x.im);
		}
	}
	else {
		fprintf(fp, "%f", x.re);
		if (x.im == 0.0) {
			;
		}
		else {
			if (x.im == 1.0)
				fprintf(fp, "+i");
			else if (x.im == -1.0)
				fprintf(fp, "-i");
			else if (x.im < 0.0)
				fprintf(fp, "-%f*i", -x.im);
			else
				fprintf(fp, "+%f*i", x.im);
		}
	}
}

// ----------------------------------------------------------------
static void fcomplex_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	fcomplex_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c.re = a.re + b.re;
	c.im = a.im + b.im;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void fcomplex_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	fcomplex_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c.re = a.re - b.re;
	c.im = a.im - b.im;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void fcomplex_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	fcomplex_t a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	c.re = a.re * b.re - a.im * b.im;
	c.im = a.re * b.im + a.im * b.re;

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void fcomplex_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	fcomplex_t a, b, c;
	float bot;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	c.re = a.re * b.re + a.im * b.im;
	c.im = a.im * b.re - a.re * b.im;

	bot = b.re * b.re + b.im * b.im;

	c.re /= bot;
	c.im /= bot;

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void fcomplex_uadd(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	memcpy(pvout, pvin, sizeof(fcomplex_t));
}

// ----------------------------------------------------------------
static void fcomplex_usub(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	fcomplex_t c;
	memcpy(&c, pvin, sizeof(c));
	c.re = -c.re;
	c.im = -c.im;
	memcpy(pvout, &c, sizeof(c));
}

// ================================================================
ops_t fcomplex_ops = {
	atom_desc          : "fcomplex",
	plex_first_func    : fcomplex_lex_first,
	pscan_func         : fcomplex_scan,
	pprint_func        : fcomplex_print,
	padd_op            : fcomplex_add,
	psub_op            : fcomplex_sub,
	pmul_op            : fcomplex_mul,
	pdiv_op            : fcomplex_div,
	puadd_op           : fcomplex_uadd,
	pusub_op           : fcomplex_usub,
	atom_size_in_bytes : sizeof(fcomplex_t),
	aux_size_in_bytes  : 0
};
