#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "ops.h"
#include "lexan.h"

// ----------------------------------------------------------------
void float_syms(symtab_t * pst)
{
	int stidx;
	float pi = 4.0 * atan(1.0);
	float e  = exp(1.0);

	st_insert(pst, "pi", L_ID, &stidx);
	st_set_val(pst, stidx, (char *)&pi);

	st_insert(pst, "e", L_ID, &stidx);
	st_set_val(pst, stidx, (char *)&e);
}


// ----------------------------------------------------------------
static int float_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == '.')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int float_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	return fscanf(fp, "%f", (float *)patom);
}

// ----------------------------------------------------------------
static void float_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	float x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, "%f", x);
}

// ----------------------------------------------------------------
static void float_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	float a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a + b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	float a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a - b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	float a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a * b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	float a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a / b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	float a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = pow(a, b);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = -a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_sqrt (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = sqrt(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_exp (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = exp(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_log (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = log(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_sin (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = sin(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_cos (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = cos(a);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void float_tan (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	float a, c;
	memcpy(&a, pvin, sizeof(a));
	c = tan(a);
	memcpy(pvout, &c, sizeof(a));
}

// ================================================================
ops_t float_ops = {
	atom_desc          : "float",
	psymtab_populator  : float_syms,
	plex_first_func    : float_lex_first,
	pscan_func         : float_scan,
	pprint_func        : float_print,
	por_op             : 0,
	pand_op            : 0,
	plsh_op            : 0,
	prsh_op            : 0,
	padd_op            : float_add,
	psub_op            : float_sub,
	pmul_op            : float_mul,
	pdiv_op            : float_div,
	pmod_op            : 0,
	ppow_op            : float_power,
	puadd_op           : float_uadd,
	pusub_op           : float_usub,
	punot_op           : 0,
	psqrt_op           : float_sqrt,
	pexp_op            : float_exp,
	plog_op            : float_log,
	psin_op            : float_sin,
	pcos_op            : float_cos,
	ptan_op            : float_tan,
	atom_size_in_bytes : sizeof(float),
	aux_size_in_bytes  : 0
};
