#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ops.h"

typedef struct _s16_opts_t {
	char scan_fmt [16];
	int  hex_in;
	char print_fmt[16];
} s16_opts_t;

// ----------------------------------------------------------------
void * s16_opt_alloc(void)
{
	s16_opts_t * p = (s16_opts_t *)malloc(sizeof(s16_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(s16_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void s16_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	s16_opts_t * ps16_opts = pvtopts;

	strcpy(ps16_opts->scan_fmt,  "%hd");
	ps16_opts->hex_in = 0;
	strcpy(ps16_opts->print_fmt, "%hd");
}

// ----------------------------------------------------------------
int  s16_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	s16_opts_t * ps16_opts = pvtopts;

	if (strcmp(argv[argi], "-ix") == 0) {
		strcpy(ps16_opts->scan_fmt,  "%hx");
		ps16_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		strcpy(ps16_opts->print_fmt, "%04hx");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		strcpy(ps16_opts->scan_fmt,  "%hx");
		ps16_opts->hex_in = 1;
		strcpy(ps16_opts->print_fmt, "%04hx");
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		strcpy(ps16_opts->scan_fmt,  "%hd");
		ps16_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		strcpy(ps16_opts->print_fmt, "%hd");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		strcpy(ps16_opts->scan_fmt,  "%hd");
		strcpy(ps16_opts->print_fmt, "%hd");
		ps16_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}

	return 0;
}


// ----------------------------------------------------------------
static int s16_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	s16_opts_t * ps16_opts = pvtopts;
	if (ps16_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int s16_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	s16_opts_t * ps16_opts = pvtopts;
	return fscanf(fp, ps16_opts->scan_fmt, (short *)patom);
}

// ----------------------------------------------------------------
static void s16_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	s16_opts_t * ps16_opts = pvtopts;
	short x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, ps16_opts->print_fmt, x);
}

// ----------------------------------------------------------------
static void s16_or(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a | b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_xor(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a ^ b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_and(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a & b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_lsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a << b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_rsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a >> b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a + b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a - b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a * b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a / b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_mod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a % b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	short a, b, c;
	short i, p2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	// This is the naive implementation.
	//c = 1;
	//for (i = 0; i < b; i++)
	//	c *= a;

	// This is the repeated-squaring implementation.
	// xxx needs range check on b.

	c  = 1;
	p2 = a;

	for (i = 0; i < 16; i++) {
		if (b & 1)
			c *= p2;
		b >>= 1;
		p2 *= p2;
	}

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	short a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	short a, c;
	memcpy(&a, pvin, sizeof(a));
	c = -a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s16_unot (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	short a, c;
	memcpy(&a, pvin, sizeof(a));
	c = ~a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static int s16_cmp(
	void * pvin1,
	int    cmp,
	void * pvin2,
	void * pvaux)
{
	short a, b;
	int rv = 0;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	switch (cmp) {
	case CMP_LT: rv = (a <  b); break;
	case CMP_LE: rv = (a <= b); break;
	case CMP_EQ: rv = (a == b); break;
	case CMP_NE: rv = (a != b); break;
	case CMP_GT: rv = (a >  b); break;
	case CMP_GE: rv = (a >= b); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); break;
	}
	return rv;
}

// ----------------------------------------------------------------
static void s16_from_int(void * pvout, int i)
{
	short u = i;
	memcpy(pvout, &u, sizeof(u));
}


// ================================================================
ops_t s16_ops = {
	atom_desc          : "s16",
	plex_first_func    : s16_lex_first,
	pscan_func         : s16_scan,
	pprint_func        : s16_print,
	por_op             : s16_or,
	pxor_op            : s16_xor,
	pand_op            : s16_and,
	plsh_op            : s16_lsh,
	prsh_op            : s16_rsh,
	padd_op            : s16_add,
	psub_op            : s16_sub,
	pmul_op            : s16_mul,
	pdiv_op            : s16_div,
	pmod_op            : s16_mod,
	ppow_op            : s16_power,
	puadd_op           : s16_uadd,
	pusub_op           : s16_usub,
	punot_op           : s16_unot,
	pfrom_int_op       : s16_from_int,
	pcmp_op            : s16_cmp,
	ptopt_alloc        : s16_opt_alloc,
	parg_defaulter     : s16_opt_arg_defaulter,
	parg_handler       : s16_opt_arg_handler,
	atom_size_in_bytes : sizeof(short),
	aux_size_in_bytes  : 0
};
