#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ops.h"

typedef struct _s32_opts_t {
	char scan_fmt [16];
	int  hex_in;
	char print_fmt[16];
} s32_opts_t;

// ----------------------------------------------------------------
void * s32_opt_alloc(void)
{
	s32_opts_t * p = (s32_opts_t *)malloc(sizeof(s32_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(s32_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void s32_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	s32_opts_t * ps32_opts = pvtopts;

	strcpy(ps32_opts->scan_fmt,  "%d");
	ps32_opts->hex_in = 0;
	strcpy(ps32_opts->print_fmt, "%d");
}

// ----------------------------------------------------------------
int  s32_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	s32_opts_t * ps32_opts = pvtopts;

	if (strcmp(argv[argi], "-ix") == 0) {
		strcpy(ps32_opts->scan_fmt,  "%x");
		ps32_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		strcpy(ps32_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		strcpy(ps32_opts->scan_fmt,  "%x");
		ps32_opts->hex_in = 1;
		strcpy(ps32_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		strcpy(ps32_opts->scan_fmt,  "%d");
		ps32_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		strcpy(ps32_opts->print_fmt, "%d");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		strcpy(ps32_opts->scan_fmt,  "%d");
		strcpy(ps32_opts->print_fmt, "%d");
		ps32_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}

	return 0;
}


// ----------------------------------------------------------------
static int s32_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	s32_opts_t * ps32_opts = pvtopts;
	if (ps32_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}


// ----------------------------------------------------------------
static int s32_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	s32_opts_t * ps32_opts = pvtopts;
	return fscanf(fp, ps32_opts->scan_fmt, (int *)patom);
}

// ----------------------------------------------------------------
static void s32_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	s32_opts_t * ps32_opts = pvtopts;
	int x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, ps32_opts->print_fmt, x);
}

// ----------------------------------------------------------------
static void s32_or(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a | b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_xor(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a ^ b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_and(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a & b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_lsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a << b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_rsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a >> b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a + b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a - b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a * b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a / b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_mod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a % b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a, b, c;
	int i, p2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	// This is the naive implementation.
	//c = 1;
	//for (i = 0; i < b; i++)
	//	c *= a;

	// This is the repeated-squaring implementation.
	// xxx needs range check on b.

	c  = 1;
	p2 = a;

	for (i = 0; i < 32; i++) {
		if (b & 1)
			c *= p2;
		b >>= 1;
		p2 *= p2;
	}

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	int a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	int a, c;
	memcpy(&a, pvin, sizeof(a));
	c = -a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s32_unot (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	int a, c;
	memcpy(&a, pvin, sizeof(a));
	c = ~a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static int s32_cmp(
	void * pvin1,
	int    cmp,
	void * pvin2,
	void * pvaux)
{
	int a, b;
	int rv = 0;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	switch (cmp) {
	case CMP_LT: rv = (a <  b); break;
	case CMP_LE: rv = (a <= b); break;
	case CMP_EQ: rv = (a == b); break;
	case CMP_NE: rv = (a != b); break;
	case CMP_GT: rv = (a >  b); break;
	case CMP_GE: rv = (a >= b); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); break;
	}
	return rv;
}

// ----------------------------------------------------------------
static void s32_from_int(void * pvout, int i)
{
	int u = i;
	memcpy(pvout, &u, sizeof(u));
}


// ================================================================
ops_t s32_ops = {
	atom_desc          : "s32",
	plex_first_func    : s32_lex_first,
	pscan_func         : s32_scan,
	pprint_func        : s32_print,
	por_op             : s32_or,
	pxor_op            : s32_xor,
	pand_op            : s32_and,
	plsh_op            : s32_lsh,
	prsh_op            : s32_rsh,
	padd_op            : s32_add,
	psub_op            : s32_sub,
	pmul_op            : s32_mul,
	pdiv_op            : s32_div,
	pmod_op            : s32_mod,
	ppow_op            : s32_power,
	puadd_op           : s32_uadd,
	pusub_op           : s32_usub,
	punot_op           : s32_unot,
	pfrom_int_op       : s32_from_int,
	pcmp_op            : s32_cmp,
	ptopt_alloc        : s32_opt_alloc,
	parg_defaulter     : s32_opt_arg_defaulter,
	parg_handler       : s32_opt_arg_handler,
	atom_size_in_bytes : sizeof(int),
	aux_size_in_bytes  : 0
};

