#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ops.h"

typedef struct _s64_opts_t {
	char scan_fmt [16];
	int  hex_in;
	char print_fmt[16];
} s64_opts_t;

// ----------------------------------------------------------------
void * s64_opt_alloc(void)
{
	s64_opts_t * p = (s64_opts_t *)malloc(sizeof(s64_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(s64_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void s64_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	s64_opts_t * ps64_opts = pvtopts;

	strcpy(ps64_opts->scan_fmt,  "%lld");
	ps64_opts->hex_in = 0;
	strcpy(ps64_opts->print_fmt, "%lld");
}

// ----------------------------------------------------------------
int  s64_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	s64_opts_t * ps64_opts = pvtopts;

	if (strcmp(argv[argi], "-ix") == 0) {
		strcpy(ps64_opts->scan_fmt,  "%llx");
		ps64_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		strcpy(ps64_opts->print_fmt, "%016llx");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		strcpy(ps64_opts->scan_fmt,  "%llx");
		ps64_opts->hex_in = 1;
		strcpy(ps64_opts->print_fmt, "%016llx");
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		strcpy(ps64_opts->scan_fmt,  "%lld");
		ps64_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		strcpy(ps64_opts->print_fmt, "%lld");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		strcpy(ps64_opts->scan_fmt,  "%lld");
		strcpy(ps64_opts->print_fmt, "%lld");
		ps64_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int s64_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	s64_opts_t * ps64_opts = pvtopts;
	if (ps64_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int s64_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	s64_opts_t * ps64_opts = pvtopts;
	return fscanf(fp, ps64_opts->scan_fmt, (long long *)patom);
}

// ----------------------------------------------------------------
static void s64_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	s64_opts_t * ps64_opts = pvtopts;
	long long x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, ps64_opts->print_fmt, x);
}

// ----------------------------------------------------------------
static void s64_or(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a | b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_xor(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a ^ b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_and(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a & b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_lsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a << b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_rsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a >> b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a + b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a - b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a * b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a / b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_mod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a % b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	long long a, b, c;
	long long i, p2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	// This is the naive implementation.
	//c = 1;
	//for (i = 0; i < b; i++)
	//	c *= a;

	// This is the repeated-squaring implementation.
	// xxx needs range check on b.

	c  = 1;
	p2 = a;

	for (i = 0; i < 64; i++) {
		if (b & 1)
			c *= p2;
		b >>= 1;
		p2 *= p2;
	}

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	long long a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	long long a, c;
	memcpy(&a, pvin, sizeof(a));
	c = -a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s64_unot (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	long long a, c;
	memcpy(&a, pvin, sizeof(a));
	c = ~a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static int s64_cmp(
	void * pvin1,
	int    cmp,
	void * pvin2,
	void * pvaux)
{
	long long a, b;
	int rv = 0;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	switch (cmp) {
	case CMP_LT: rv = (a <  b); break;
	case CMP_LE: rv = (a <= b); break;
	case CMP_EQ: rv = (a == b); break;
	case CMP_NE: rv = (a != b); break;
	case CMP_GT: rv = (a >  b); break;
	case CMP_GE: rv = (a >= b); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); break;
	}
	return rv;
}

// ----------------------------------------------------------------
static void s64_from_int(void * pvout, int i)
{
	long long u = i;
	memcpy(pvout, &u, sizeof(u));
}


// ================================================================
ops_t s64_ops = {
	atom_desc          : "s64",
	plex_first_func    : s64_lex_first,
	pscan_func         : s64_scan,
	pprint_func        : s64_print,
	por_op             : s64_or,
	pxor_op            : s64_xor,
	pand_op            : s64_and,
	plsh_op            : s64_lsh,
	prsh_op            : s64_rsh,
	padd_op            : s64_add,
	psub_op            : s64_sub,
	pmul_op            : s64_mul,
	pdiv_op            : s64_div,
	pmod_op            : s64_mod,
	ppow_op            : s64_power,
	puadd_op           : s64_uadd,
	pusub_op           : s64_usub,
	punot_op           : s64_unot,
	pfrom_int_op       : s64_from_int,
	pcmp_op            : s64_cmp,
	ptopt_alloc        : s64_opt_alloc,
	parg_defaulter     : s64_opt_arg_defaulter,
	parg_handler       : s64_opt_arg_handler,
	atom_size_in_bytes : sizeof(long long),
	aux_size_in_bytes  : 0
};
