#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ops.h"

typedef struct _s8_opts_t {
	char scan_fmt [16];
	int  hex_in;
	char print_fmt[16];
} s8_opts_t;

// ----------------------------------------------------------------
void * s8_opt_alloc(void)
{
	s8_opts_t * p = (s8_opts_t *)malloc(sizeof(s8_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(s8_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void s8_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	s8_opts_t * ps8_opts = pvtopts;

	strcpy(ps8_opts->scan_fmt,  "%d");
	ps8_opts->hex_in = 0;
	strcpy(ps8_opts->print_fmt, "%d");
}

// ----------------------------------------------------------------
int  s8_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	s8_opts_t * ps8_opts = pvtopts;

	if (strcmp(argv[argi], "-ix") == 0) {
		strcpy(ps8_opts->scan_fmt,  "%x");
		ps8_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		strcpy(ps8_opts->print_fmt, "%02x");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		strcpy(ps8_opts->scan_fmt,  "%x");
		ps8_opts->hex_in = 1;
		strcpy(ps8_opts->print_fmt, "%02x");
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		strcpy(ps8_opts->scan_fmt,  "%d");
		ps8_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		strcpy(ps8_opts->print_fmt, "%d");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		strcpy(ps8_opts->scan_fmt,  "%d");
		strcpy(ps8_opts->print_fmt, "%d");
		ps8_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}

	return 0;
}


// ----------------------------------------------------------------
static int s8_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	s8_opts_t * ps8_opts = pvtopts;
	if (ps8_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}


// ----------------------------------------------------------------
static int s8_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	s8_opts_t * ps8_opts = pvtopts;
	int  i;
	char c;
	if (!fscanf(fp, ps8_opts->scan_fmt, &i))
		return 0;
	c = i;
	memcpy(patom, &c, sizeof(c));
	return 1;
}

// ----------------------------------------------------------------
static void s8_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	s8_opts_t * ps8_opts = pvtopts;
	char x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, ps8_opts->print_fmt, (int)x);
}

// ----------------------------------------------------------------
static void s8_or(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a | b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_xor(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a ^ b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_and(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a & b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_lsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a << b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_rsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a >> b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a + b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a - b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a * b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a / b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_mod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a % b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	char a, b, c;
	char i, p2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	// This is the naive implementation.
	//c = 1;
	//for (i = 0; i < b; i++)
	//	c *= a;

	// This is the repeated-squaring implementation.
	// xxx needs range check on b.

	c  = 1;
	p2 = a;

	for (i = 0; i < 8; i++) {
		if (b & 1)
			c *= p2;
		b >>= 1;
		p2 *= p2;
	}

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	char a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	char a, c;
	memcpy(&a, pvin, sizeof(a));
	c = -a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void s8_unot (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	char a, c;
	memcpy(&a, pvin, sizeof(a));
	c = ~a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static int s8_cmp(
	void * pvin1,
	int    cmp,
	void * pvin2,
	void * pvaux)
{
	char a, b;
	int rv = 0;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	switch (cmp) {
	case CMP_LT: rv = (a <  b); break;
	case CMP_LE: rv = (a <= b); break;
	case CMP_EQ: rv = (a == b); break;
	case CMP_NE: rv = (a != b); break;
	case CMP_GT: rv = (a >  b); break;
	case CMP_GE: rv = (a >= b); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); break;
	}
	return rv;
}

// ----------------------------------------------------------------
static void s8_from_int(void * pvout, int i)
{
	char u = i;
	memcpy(pvout, &u, sizeof(u));
}


// ================================================================
ops_t s8_ops = {
	atom_desc          : "s8",
	plex_first_func    : s8_lex_first,
	pscan_func         : s8_scan,
	pprint_func        : s8_print,
	por_op             : s8_or,
	pxor_op            : s8_xor,
	pand_op            : s8_and,
	plsh_op            : s8_lsh,
	prsh_op            : s8_rsh,
	padd_op            : s8_add,
	psub_op            : s8_sub,
	pmul_op            : s8_mul,
	pdiv_op            : s8_div,
	pmod_op            : s8_mod,
	ppow_op            : s8_power,
	puadd_op           : s8_uadd,
	pusub_op           : s8_usub,
	punot_op           : s8_unot,
	pfrom_int_op       : s8_from_int,
	pcmp_op            : s8_cmp,
	ptopt_alloc        : s8_opt_alloc,
	parg_defaulter     : s8_opt_arg_defaulter,
	parg_handler       : s8_opt_arg_handler,
	atom_size_in_bytes : sizeof(char),
	aux_size_in_bytes  : 0
};

