#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ops.h"
#include "cmps.h"
#include "common_math.h"

typedef struct _u32_opts_t {
	char scan_fmt [16];
	int  hex_in;
	char print_fmt[16];
} u32_opts_t;

// ----------------------------------------------------------------
void * u32_opt_alloc(void)
{
	u32_opts_t * p = (u32_opts_t *)malloc(sizeof(u32_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(u32_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void u32_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	u32_opts_t * pu32_opts = pvtopts;

	strcpy(pu32_opts->scan_fmt,  "%u");
	pu32_opts->hex_in = 0;
	strcpy(pu32_opts->print_fmt, "%u");
}

// ----------------------------------------------------------------
int  u32_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	u32_opts_t * pu32_opts = pvtopts;

	if (strcmp(argv[argi], "-ix") == 0) {
		strcpy(pu32_opts->scan_fmt,  "%x");
		pu32_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		strcpy(pu32_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		strcpy(pu32_opts->scan_fmt,  "%x");
		pu32_opts->hex_in = 1;
		strcpy(pu32_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		strcpy(pu32_opts->scan_fmt,  "%u");
		pu32_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		strcpy(pu32_opts->print_fmt, "%u");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		strcpy(pu32_opts->scan_fmt,  "%u");
		strcpy(pu32_opts->print_fmt, "%u");
		pu32_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int u32_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	u32_opts_t * pu32_opts = pvtopts;
	if (pu32_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int u32_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	u32_opts_t * pu32_opts = pvtopts;
	return fscanf(fp, pu32_opts->scan_fmt, (unsigned *)patom);
}

// ----------------------------------------------------------------
static void u32_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	u32_opts_t * pu32_opts = pvtopts;
	unsigned x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, pu32_opts->print_fmt, x);
}

// ----------------------------------------------------------------
static void u32_or(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a | b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_xor(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a ^ b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_and(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a & b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_lsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a << b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_rsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a >> b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a + b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a - b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a * b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a / b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_mod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a % b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned i, p2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	// This is the naive implementation.
	//c = 1;
	//for (i = 0; i < b; i++)
	//	c *= a;

	// This is the repeated-squaring implementation.
	// xxx needs range check on b.

	c  = 1;
	p2 = a;

	for (i = 0; i < 32; i++) {
		if (b & 1)
			c *= p2;
		b >>= 1;
		p2 *= p2;
	}

	memcpy(pvout, &c, sizeof(a));
}


// ----------------------------------------------------------------
static void u32_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, c;
	memcpy(&a, pvin, sizeof(a));
	c = -a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void u32_unot (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, c;
	memcpy(&a, pvin, sizeof(a));
	c = ~a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static int u32_cmp(
	void * pvin1,
	int    cmp,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b;
	int rv = 0;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	switch (cmp) {
	case CMP_LT: rv = (a <  b); break;
	case CMP_LE: rv = (a <= b); break;
	case CMP_EQ: rv = (a == b); break;
	case CMP_NE: rv = (a != b); break;
	case CMP_GT: rv = (a >  b); break;
	case CMP_GE: rv = (a >= b); break;
	default: fprintf(stderr, "Ack!\n"); exit(1); break;
	}
	return rv;
}

// ----------------------------------------------------------------
static void u32_from_int(void * pvout, int i)
{
	unsigned u = i;
	memcpy(pvout, &u, sizeof(u));
}

// ================================================================
ops_t u32_ops = {
	atom_desc          : "u32",
	plex_first_func    : u32_lex_first,
	pscan_func         : u32_scan,
	pprint_func        : u32_print,
	por_op             : u32_or,
	pxor_op            : u32_xor,
	pand_op            : u32_and,
	plsh_op            : u32_lsh,
	prsh_op            : u32_rsh,
	padd_op            : u32_add,
	psub_op            : u32_sub,
	pmul_op            : u32_mul,
	pdiv_op            : u32_div,
	pmod_op            : u32_mod,
	ppow_op            : u32_power,
	puadd_op           : u32_uadd,
	pusub_op           : u32_usub,
	punot_op           : u32_unot,
	pfrom_int_op       : u32_from_int,
	pcmp_op            : u32_cmp,
	ptopt_alloc        : u32_opt_alloc,
	parg_defaulter     : u32_opt_arg_defaulter,
	parg_handler       : u32_opt_arg_handler,
	atom_size_in_bytes : sizeof(unsigned),
	aux_size_in_bytes  : 0
};
