#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ops.h"
#include "ubig.h"

// ----------------------------------------------------------------
void * ubig_opt_alloc(void)
{
	ubig_opts_t * p = (ubig_opts_t *)malloc(sizeof(ubig_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(ubig_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void ubig_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	ubig_opts_t * pubig_opts = pvtopts;

	pubig_opts->hex_in    = 0;
	pubig_opts->hex_out   = 0;
	pubig_opts->zero_fill = 0;
	pubig_opts->underscore_out = 0;
}

// ----------------------------------------------------------------
int  ubig_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	ubig_opts_t * pubig_opts = pvtopts;

	if (strcmp(argv[argi], "-ix") == 0) {
		pubig_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		pubig_opts->hex_out = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		pubig_opts->hex_in = 1;
		pubig_opts->hex_out = 1;
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		pubig_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		pubig_opts->hex_out = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		pubig_opts->hex_in = 0;
		pubig_opts->hex_out = 0;
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-zf") == 0) {
		pubig_opts->zero_fill = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-nzf") == 0) {
		pubig_opts->zero_fill = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ou") == 0) {
		pubig_opts->underscore_out = 1;
		*pnum_args_used = 1;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int ubig_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	ubig_opts_t * pubig_opts = pvtopts;
	if (pubig_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int ubig_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	ubig_opts_t * pubig_opts = pvtopts;
	return ubig_fscan(fp, patom, pubig_opts);
}

// ----------------------------------------------------------------
static void ubig_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	ubig_opts_t * pubig_opts = pvtopts;

	ubig_print(fp, patom, pubig_opts);
}

// ----------------------------------------------------------------
static void ubig_vor(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_or(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vxor(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_xor(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vand(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_and(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vlsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_lsh(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vrsh(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_rsh(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_add(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_sub(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_mul(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_div(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_mod(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(ubig *)pvout = ubig_power(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(ubig *)pvout = ubig_uadd(pvin);
}

// ----------------------------------------------------------------
static void ubig_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(ubig *)pvout = ubig_usub(pvin);
}

// ----------------------------------------------------------------
static void ubig_vunot (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(ubig *)pvout = ubig_unot(pvin);
}

// ----------------------------------------------------------------
static int  ubig_vcmp (
	void * pvin1,
	int    cmp,
	void * pvin2,
	void * pvaux)
{
	return ubig_cmp(pvin1, cmp, pvin2);
}

// ----------------------------------------------------------------
static void ubig_vfrom_int (void * pvout, int ival)
{
	*(ubig *)pvout = ubig_from_unsigned((unsigned)ival);
}

// ================================================================
ops_t ubig_ops = {
	atom_desc          : "ubig",
	plex_first_func    : ubig_lex_first,
	pscan_func         : ubig_vscan,
	pprint_func        : ubig_vprint,
	por_op             : ubig_vor,
	pxor_op            : ubig_vxor,
	pand_op            : ubig_vand,
	plsh_op            : ubig_vlsh,
	prsh_op            : ubig_vrsh,
	padd_op            : ubig_vadd,
	psub_op            : ubig_vsub,
	pmul_op            : ubig_vmul,
	pdiv_op            : ubig_vdiv,
	pmod_op            : ubig_vmod,
	ppow_op            : ubig_vpower,
	puadd_op           : ubig_vuadd,
	pusub_op           : ubig_vusub,
	punot_op           : ubig_vunot,
	pfrom_int_op       : ubig_vfrom_int,
	pcmp_op            : ubig_vcmp,
	ptopt_alloc        : ubig_opt_alloc,
	parg_defaulter     : ubig_opt_arg_defaulter,
	parg_handler       : ubig_opt_arg_handler,
	atom_size_in_bytes : sizeof(ubig),
	aux_size_in_bytes  : 0
};
