#include <stdio.h>
#include <ctype.h>
#include "lexan.h"
#include "symtab.h"
#include "error.h"

#ifdef LEXAN_DEBUG
#define lexan_debug(x) lexan_debug_func(x)
#else
#define lexan_debug(x)
#endif

// ----------------------------------------------------------------
void lexctx_init(lexctx_t * pl)
{
	pl->lineno   = 1;
	pl->tokenval = L_NONE;
	pl->tokenbuf[0] = 0;
	pl->in_comment = 0;
}

// ----------------------------------------------------------------
#ifdef LEXAN_DEBUG
static void lexan_debug_func(lexctx_t * pl)
{
	if (pl->token == '\n')
		return;

	printf("-- lex tok ");
	if (pl->token < 128 && isprint(pl->token))
		printf("[%c]", pl->token);
	else
		printf("%3d", pl->token);
	printf(" val %3d\n", pl->tokenval);
}
#endif

// ----------------------------------------------------------------
void lexan(lexctx_t * pl, symtab_t * pst, FILE * infp, opts_t * popts,
	int in_pow_rhs)
{
	int  ch;
	int  ch2;
	char lexbuf[BSIZE];

	while (1) {
		ch = getc(infp);

		if (ch == '#')
			pl->in_comment = 1;
		else if (ch == '\n')
			pl->in_comment = 0;
		if (pl->in_comment)
			continue;


		if (ch == ' ' || ch == '\t') {
			;
		}
		else if (ch == '\n') {
			pl->lineno++;

			if (!popts->req_semi) {
				pl->tokenval = L_NONE;
				pl->token = ch;
				lexan_debug(pl);
				return;
			}
		}

		else if (popts->pops->plex_first_func(
			ch, popts->pvtopts, in_pow_rhs))
		{
			ungetc(ch, infp);
			// xxx if dynamic, alloc func; store ptr in
			// pl->tokenbuf.

			// xxx do not ignore the rv here!!!
			popts->pops->pscan_func(infp, &pl->tokenbuf,
				popts->pvtopts, popts->aux_buf, in_pow_rhs);
			pl->token = L_NUM;
			pl->tokenval = -1;
			lexan_debug(pl);
			return;
		}

		else if (isalpha(ch)) {
			int b = 0;
			int stidx;
			while (isalnum(ch)) {
				lexbuf[b] = ch;
				ch = getc(infp);

				b++;
				if (b >= BSIZE)
					lerror(pl->lineno, "compiler error");
			}
			lexbuf[b] = 0;

			if (ch != EOF)
				ungetc(ch, infp);

			if (!st_lookup(pst, lexbuf, &pl->token, &stidx)) {
				pl->token = L_ID;
				st_insert(pst, lexbuf, pl->token, &stidx);
			}
			pl->tokenval = stidx;
			lexan_debug(pl);
			return;
		}

		else if (ch == '<' || ch == '>') {
			ch2 = getc(infp);
			if (ch2 == ch) {
				if (ch == '<')
					pl->token = L_LSH;
				else
					pl->token = L_RSH;
				lexan_debug(pl);
				return;
			}
			else {
				if (ch2 != EOF)
					ungetc(ch2, infp);
				pl->token = ch;
				lexan_debug(pl);
				return;
			}
		}

		else if (ch == '*') {
			ch2 = getc(infp);
			if (ch2 == '*') {
				pl->token = L_POW;
				lexan_debug(pl);
				return;
			}
			else {
				if (ch2 != EOF)
					ungetc(ch2, infp);
				pl->token = ch;
				lexan_debug(pl);
				return;
			}
		}

		else if (ch == EOF) {
			pl->token = L_DONE;
			lexan_debug(pl);
			return;
		}
		else {
			pl->tokenval = L_NONE;
			pl->token = ch;
			lexan_debug(pl);
			return;
		}
	}
}

// ----------------------------------------------------------------
void lexan_desc_token(int tok, FILE * fp)
{
	switch (tok) {
	case L_NONE:   fprintf(fp, "NONE");   return;
	case L_DONE:   fprintf(fp, "DONE");   return;
	case L_NUM:    fprintf(fp, "NUM");    return;
	case L_ID:     fprintf(fp, "ID");     return;
	case L_LSH:    fprintf(fp, "LSH");    return;
	case L_RSH:    fprintf(fp, "RSH");    return;
	case L_POW:    fprintf(fp, "POW");    return;
	case L_GCD:    fprintf(fp, "GCD");    return;
	case L_CHOOSE: fprintf(fp, "CHOOSE"); return;
	case L_UNOT:   fprintf(fp, "UNOT");   return;
	case L_UPLUS:  fprintf(fp, "UPLUS");  return;
	case L_UMINUS: fprintf(fp, "UMINUS"); return;
	case L_SQRT:   fprintf(fp, "SQRT");   return;
	case L_EXP:    fprintf(fp, "EXP");    return;
	case L_LOG:    fprintf(fp, "LOG");    return;
	case L_SIN:    fprintf(fp, "SIN");    return;
	case L_COS:    fprintf(fp, "COS");    return;
	case L_TAN:    fprintf(fp, "TAN");    return;
	case L_ATAN:   fprintf(fp, "ATAN");   return;
	case L_CONJ:   fprintf(fp, "CONJ");   return;
	case L_MAG:    fprintf(fp, "MAG");    return;
	case L_PHZ:    fprintf(fp, "PHZ");    return;
	case L_ORD:    fprintf(fp, "ORD");    return;
	case L_DEFER:  fprintf(fp, "DEFER");  return;
	case '\n':     fprintf(fp, "carriage return"); return;
	}

	if (isprint(tok))
		fprintf(fp, "'%c'", tok);
	else
		fprintf(fp, "(token %d)", tok);
}
