#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "pmt.h"

// ----------------------------------------------------------------
void * pmt_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
void pmt_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	*(int *)pvaux = 4;
}

// ----------------------------------------------------------------
int  pmt_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	int n;

	if (strcmp(argv[argi], "-n") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr,
			"pmt_t:  -r requires an argument.\n");
			return 0;
		}
		if (sscanf(argv[argi+1], "%d", &n) != 1) {
			fprintf(stderr,
			"pmt:  Couldn't scan \"%s\" as integer.\n",
			argv[argi]);
			return 0;
		}
		*(int *)pvaux = n;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int pmt_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isxdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int kalk_pmt_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	int   n = *(int *)pvaux;
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	while (isdigit(ch) || (ch == ',')) {
		if (p >= end) {
			fprintf(stderr, "kalk_pmt_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return pmt_vscan(patom, mybuf, &n, sizeof(pmt_t));
}

// ----------------------------------------------------------------
static int kalk_pmt_vscan_cd(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	int    n = *(int *)pvaux;
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	while (isdigit(ch) || (ch == ',') || (ch == ':')) {
		if (p >= end) {
			fprintf(stderr, "kalk_pmt_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return pmt_vscan_cd(patom, mybuf, &n, sizeof(pmt_t));
}

// ----------------------------------------------------------------
static void pmt_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack

	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	// xxx need pvaux here!
	pmt_t a = *(pmt_t *)patom;
	int  n = *(int  *)pvaux;
	pmt_print(&a, n);
}

// ----------------------------------------------------------------
static void pmt_vprint_cd(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux)
{
	// xxx temp hack

	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	// xxx need pvaux here!
	pmt_t a = *(pmt_t *)patom;
	int  n = *(int  *)pvaux;
	pmt_print_cd(&a, n);
}

// ----------------------------------------------------------------
static void kalk_pmt_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	pmt_t * pa = (pmt_t *)pvin1;
	pmt_t * pb = (pmt_t *)pvin2;
	int  n = *(int  *)pvaux;
	pmt_t c = pmt_mul(pa, pb, n);
	*(pmt_t *)pvout = c;
}

// ----------------------------------------------------------------
static void pmt_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	pmt_t * pa = (pmt_t *)pvin1;
	pmt_t * pb = (pmt_t *)pvin2;
	int  n = *(int  *)pvaux;
	pmt_t   bi;
	pmt_t   c;

	pmt_invert(pb, &bi, n);
	c = pmt_mul(pa, &bi, n);
	*(pmt_t *)pvout = c;
}

// ----------------------------------------------------------------
static void pmt_vinvert(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	pmt_t * pa = (pmt_t *)pvin;
	int  n = *(int  *)pvaux;
	pmt_t c;
	pmt_invert(pa, &c, n);
	*(pmt_t *)pvout = c;
}

// ================================================================
ops_t pmt_ops = {
	atom_desc          : "pmt",
	plex_first_func    : pmt_lex_first,
	pscan_func         : kalk_pmt_vscan,
	pprint_func        : pmt_vprint,
	punot_op           : pmt_vinvert,
	pmul_op            : kalk_pmt_vmul,
	pdiv_op            : pmt_vdiv,
	ptopt_alloc        : pmt_opt_alloc,
	parg_defaulter     : pmt_opt_arg_defaulter,
	parg_handler       : pmt_opt_arg_handler,
	atom_size_in_bytes : sizeof(pmt_t),
	aux_size_in_bytes  : sizeof(int)
};

// ================================================================
ops_t pmtc_ops = {
	atom_desc          : "pmtc",
	plex_first_func    : pmt_lex_first,
	pscan_func         : kalk_pmt_vscan_cd,
	pprint_func        : pmt_vprint_cd,
	punot_op           : pmt_vinvert,
	pmul_op            : kalk_pmt_vmul,
	pdiv_op            : pmt_vdiv,
	ptopt_alloc        : pmt_opt_alloc,
	parg_defaulter     : pmt_opt_arg_defaulter,
	parg_handler       : pmt_opt_arg_handler,
	atom_size_in_bytes : sizeof(pmt_t),
	aux_size_in_bytes  : sizeof(int)
};
