#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "q8.h"

// ----------------------------------------------------------------
static int q8_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	switch (c) {
	case '1':
	case 'i':
	case 'j':
	case 'k':
		return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int kalk_q8_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	int    ch = getc(fp);
	int  * pa = (int *)patom;

	if (ch == '1') {
		*pa = Q8_EP;
		return 1;
	}
	if (ch == 'i') {
		*pa = Q8_IP;
		return 1;
	}
	if (ch == 'j') {
		*pa = Q8_JP;
		return 1;
	}
	if (ch == 'k') {
		*pa = Q8_KP;
		return 1;
	}

	fprintf(stderr, "q8_vscan: internal error.\n");
	exit(1);
	return 0;
}

// ----------------------------------------------------------------
static void q8_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	int a = *(int *)patom;
	q8_print(a);
}

// ----------------------------------------------------------------
static void kalk_q8_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int c = q8_mul(a, b);
	*(int *)pvout = c;
}

// ----------------------------------------------------------------
static void q8_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int a = *(int *)pvin1;
	int b = *(int *)pvin2;
	int bi;

	bi = q8_invert(b);
	*(int *)pvout = q8_mul(a, bi);
}

// ----------------------------------------------------------------
static void q8_vinvert(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	int a = *(int *)pvin;
	*(int *)pvout = q8_invert(a);
}

// ----------------------------------------------------------------
static void q8_vuadd(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	int a = *(int *)pvin;
	*(int *)pvout = a;
}

// ----------------------------------------------------------------
static void q8_vusub(
	void * pvout,
	void * pvin,
	void * pvaux)
{
	int a = *(int *)pvin;
	int c = Q8_EP;

	switch (a & 7) {
	case Q8_EP: c = Q8_EN; break;
	case Q8_EN: c = Q8_EP; break;
	case Q8_IP: c = Q8_IN; break;
	case Q8_IN: c = Q8_IP; break;
	case Q8_JP: c = Q8_JN; break;
	case Q8_JN: c = Q8_JP; break;
	case Q8_KP: c = Q8_KN; break;
	case Q8_KN: c = Q8_KP; break;

	}
	*(int *)pvout = c;
}


// ================================================================
ops_t q8_ops = {
	atom_desc          : "q8",
	plex_first_func    : q8_lex_first,
	pscan_func         : kalk_q8_vscan,
	pprint_func        : q8_vprint,
	puadd_op           : q8_vuadd,
	pusub_op           : q8_vusub,
	punot_op           : q8_vinvert,
	pmul_op            : kalk_q8_vmul,
	pdiv_op            : q8_vdiv,
	atom_size_in_bytes : sizeof(int),
	aux_size_in_bytes  : 0
};
