#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "ubig.h"
#include "isqrt.h"

// xxx hack; should be pvin bigmod_ops_t.
static ubig modulus;

static int isprime_uncached(ubig m);
static int isprime(ubig m);

typedef struct _mod_opts_t {
	char scan_fmt [16];
	int  hex_in;
	char print_fmt[16];
} bigmod_opts_t;

// ----------------------------------------------------------------
void * bigmod_opt_alloc(void)
{
	bigmod_opts_t * p = (bigmod_opts_t *)malloc(sizeof(bigmod_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(bigmod_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void bigmod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	bigmod_opts_t * pmod_opts = pvtopts;

	strcpy(pmod_opts->scan_fmt,  "%u");
	pmod_opts->hex_in = 0;
	strcpy(pmod_opts->print_fmt, "%u");
	modulus = 11;
}

// ----------------------------------------------------------------
int  bigmod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	bigmod_opts_t * pmod_opts = pvtopts;

	if (strcmp(argv[argi], "-ix") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%x");
		pmod_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		strcpy(pmod_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%x");
		pmod_opts->hex_in = 1;
		strcpy(pmod_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%u");
		pmod_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		strcpy(pmod_opts->print_fmt, "%u");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%u");
		strcpy(pmod_opts->print_fmt, "%u");
		pmod_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-m") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr, "bigmod:  -m requires an argument.\n");
			return 0;
		}
		if (sscanf(argv[argi+1], "%u", &modulus) != 1) {
			fprintf(stderr,
				"bigmod:  Couldn't parse \"%s\" as integer.\n",
				argv[argi+1]);
			return 0;
		}
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}


// ----------------------------------------------------------------
static int bigmod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	bigmod_opts_t * pmod_opts = pvtopts;
	if (pmod_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int bigmod_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	bigmod_opts_t * pmod_opts = pvtopts;
	return fscanf(fp, pmod_opts->scan_fmt, (unsigned *)patom);
}

// ----------------------------------------------------------------
static void bigmod_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	bigmod_opts_t * pmod_opts = pvtopts;
	unsigned x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, pmod_opts->print_fmt, x);
}

// ----------------------------------------------------------------
// Dilemma:
// (Assume mod 11 for concreteness.)
// 1.	User types pvin "13".  We should output 2, not 13 -- needs rectification.
// 2.	User types pvin "2 ** 13".  We should output 8 -- shouldn't rectify.
// Resolution:  Rectify expressions after evaluation; don't rectify numbers
// before evaluation.

static void bigmod_rectifier(void * buf)
{
	ubig temp;
	ubig c;
	memcpy(&temp, buf, sizeof(c));
	ubig_mod(&c, &temp, &modulus);
	memcpy(buf, &c, sizeof(c));
}

// ----------------------------------------------------------------
static void bigmod_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	ubig_mod(&a, &a, &modulus);
	ubig_mod(&b, &b, &modulus);
	ubig_add(&c, &a, &b);
	ubig_mod(&c, &c, &modulus);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void bigmod_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	ubig a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	ubig_mod(&a, &a, &modulus);
	ubig_mod(&b, &b, &modulus);
	if (ubig_lt(&a, &b))
		ubig_add(&a, &a, &modulus);
	ubig_sub(&c, &a, &b);
	ubig_mod(&c, &c, &modulus);

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void bigmod_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	ubig_mod(&a, &a, &modulus);
	ubig_mod(&b, &b, &modulus);
	ubig_mul(&c, &a, &b);
	ubig_mod(&c, &c, &modulus);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
// For prime p (this is Fermat's little theorem),
// b ^ p = b mod p.  Or, b ^ (p-1) = 1; and, 1/b = b^(-1) = b^(p-2) mod p.
// So, a / b = a * b^(p-2).

static void bigmod_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned q, i;
	unsigned p2, r;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	a %= modulus;
	b %= modulus;

	if (b == 0) {
		fprintf(stderr, "bigmod:  divide by zero.\n");
		exit(1);
	}

	if (!isprime(modulus)) {
		fprintf(stderr,
			"bigmod:  Can't divide when modulus (%u) is composite.\n",
			modulus);
		exit(1);
	}

	q  = 1;
	p2 = b;
	r  = modulus - 2;

	for (i = 0; i < 32; i++) {
		if (r & 1) {
			q = (q * p2) % modulus;
		}
		r >>= 1;
		p2 = (p2 * p2) % modulus;
	}

	c = (a * q) % modulus;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void bigmod_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned i, p2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	a %= modulus;
	// Don't rectify the exponent.  Madness lies that way. :)
	// Really, e.g. mod 11, we want 2 ** 13 to be 8.
	// *Really*, this program needs to have heterogeneous data types,
	// or at least have the possibility of the exponent being int
	// and the base being of modular type.

	c  = 1;
	p2 = a;

	for (i = 0; i < 32; i++) {
		if (b & 1) {
			c = (c * p2) % modulus;
		}
		b >>= 1;
		p2 = (p2 * p2) % modulus;
	}

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void bigmod_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, c;
	memcpy(&a, pvin, sizeof(a));
	ubig_mod(&c, &a, &modulus);
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void bigmod_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, c;
	memcpy(&a, pvin, sizeof(a));
	ubig_mod(&a, &a, &modulus);
	if (a == 0)
		c = 0;
	else
		c = modulus - a;
	memcpy(pvout, &c, sizeof(a));
}

// ================================================================
static int isprime_uncached(unsigned m)
{
	unsigned i;
	unsigned top;

	if ((m & 1) == 0)
		return 0;
	top = calc_isqrt_32(m, ICEIL);
	for (i = 3; i <= top; i += 2)
		if ((m % i) == 0)
			return 0;
	return 1;
}

// ----------------------------------------------------------------
static int isprime(unsigned m)
{
	static unsigned last_m      = 0;
	static unsigned have_last_m = 0;
	static int      last_answer = 0;

	if (have_last_m)
		if (m == have_last_m)
			return last_answer;

	last_answer = isprime_uncached(m);
	last_m = m;
	have_last_m = 1;

	return last_answer;
}

// ================================================================
ops_t bigmod_ops = {
	atom_desc          : "bigmod",
	plex_first_func    : bigmod_lex_first,
	pscan_func         : bigmod_scan,
	pprint_func        : bigmod_print,
	prectifier         : bigmod_rectifier,
	padd_op            : bigmod_add,
	psub_op            : bigmod_sub,
	pmul_op            : bigmod_mul,
	pdiv_op            : bigmod_div,
	ppow_op            : bigmod_power,
	puadd_op           : bigmod_uadd,
	pusub_op           : bigmod_usub,
	ptopt_alloc        : bigmod_opt_alloc,
	parg_defaulter     : bigmod_opt_arg_defaulter,
	parg_handler       : bigmod_opt_arg_handler,
	atom_size_in_bytes : sizeof(unsigned),
	aux_size_in_bytes  : 0
};
