#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "gcd.h"
#include "phi.h"
#include <math.h>

typedef struct _mod_opts_t {
	char scan_fmt [16];
	int  hex_in;
	char print_fmt[16];
} mod_opts_t;

// ----------------------------------------------------------------
void * mod_opt_alloc(void)
{
	mod_opts_t * p = (mod_opts_t *)malloc(sizeof(mod_opts_t));
	if (p == 0) {
		fprintf(stderr, "malloc(%d) failed.\n", sizeof(mod_opts_t));
		exit(1);
	}
	return p;
}

// ----------------------------------------------------------------
void mod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	mod_opts_t * pmod_opts = pvtopts;
	unsigned modulus = 11;

	strcpy(pmod_opts->scan_fmt,  "%u");
	pmod_opts->hex_in = 0;
	strcpy(pmod_opts->print_fmt, "%u");
	*(unsigned *)pvaux = modulus;
}

// ----------------------------------------------------------------
int  mod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	mod_opts_t * pmod_opts = pvtopts;
	unsigned modulus;

	if (strcmp(argv[argi], "-ix") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%x");
		pmod_opts->hex_in = 1;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-ox") == 0) {
		strcpy(pmod_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-x") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%x");
		pmod_opts->hex_in = 1;
		strcpy(pmod_opts->print_fmt, "%08x");
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-id") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%u");
		pmod_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-od") == 0) {
		strcpy(pmod_opts->print_fmt, "%u");
		*pnum_args_used = 1;
		return 1;
	}
	else if (strcmp(argv[argi], "-d") == 0) {
		strcpy(pmod_opts->scan_fmt,  "%u");
		strcpy(pmod_opts->print_fmt, "%u");
		pmod_opts->hex_in = 0;
		*pnum_args_used = 1;
		return 1;
	}

	else if (strcmp(argv[argi], "-m") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr, "mod:  -m requires an argument.\n");
			return 0;
		}
		if (sscanf(argv[argi+1], "%u", &modulus) != 1) {
			fprintf(stderr,
				"mod:  Couldn't parse \"%s\" as integer.\n",
				argv[argi+1]);
			return 0;
		}
		*(unsigned *)pvaux = modulus;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}


// ----------------------------------------------------------------
static int mod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	mod_opts_t * pmod_opts = pvtopts;
	if (pmod_opts->hex_in) {
		if (isxdigit(c))
			return 1;
	}
	else {
		if (isdigit(c))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int mod_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	mod_opts_t * pmod_opts = pvtopts;
	return fscanf(fp, pmod_opts->scan_fmt, (unsigned *)patom);
}

// ----------------------------------------------------------------
static void mod_print(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	mod_opts_t * pmod_opts = pvtopts;
	unsigned x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, pmod_opts->print_fmt, x);
}

// ----------------------------------------------------------------
// Dilemma:
// (Assume mod 11 for concreteness.)
// 1.	User types pvin "13".  We should output 2, not 13 -- needs rectification.
// 2.	User types pvin "2 ** 13".  We should output 8 -- shouldn't rectify.
// Resolution:  Rectify expressions after evaluation; don't rectify numbers
// before evaluation.

static void mod_rectifier(void * buf, void * pvaux)
{
	unsigned c;
	unsigned modulus = *(unsigned *)pvaux;
	memcpy(&c, buf, sizeof(c));
	c = c % modulus;
	memcpy(buf, &c, sizeof(c));
}

// ----------------------------------------------------------------
static void mod_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned modulus = *(unsigned *)pvaux;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	a %= modulus;
	b %= modulus;
	c = (a + b) % modulus;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void mod_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned modulus = *(unsigned *)pvaux;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	a %= modulus;
	b %= modulus;
	if (a < b)
		a += modulus;
	c = (a - b) % modulus;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void mod_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned modulus = *(unsigned *)pvaux;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	a %= modulus;
	b %= modulus;
	c = (a * b) % modulus;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
// For prime p (Fermat's little theorem),
// b ^ p = b mod p.  Or, b ^ (p-1) = 1; and, 1/b = b^(-1) = b^(p-2) mod p.
// So, a / b = a * b^(p-2).
//
// For modulus not necessarily prime, b ^ phi(m) = 1 mod m, with the phi
// function being Euler's.

static void mod_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned modulus = *(unsigned *)pvaux;
	unsigned q;
	unsigned p2, r;
	unsigned phi_m;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(b));

	a %= modulus;
	b %= modulus;

	if (b == 0) {
		fprintf(stderr, "mod:  divide by zero.\n");
		exit(1);
	}

	phi_m = calc_phi_unsigned(modulus);

	if (calc_gcd_unsigned(b, modulus) != 1) {
		fprintf(stderr,
		"mod:  divide by zero divisor (gcd(%u, %u) != 1).\n",
				b, modulus);
		exit(1);
	}

	q  = 1;
	p2 = b;
	r  = phi_m - 1;

	while (r) {
		if (r & 1) {
			q = (q * p2) % modulus;
		}
		r >>= 1;
		p2 = (p2 * p2) % modulus;
	}

	c = (a * q) % modulus;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void mod_power(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned a, b, c;
	unsigned modulus = *(unsigned *)pvaux;
	unsigned i, p2;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	a %= modulus;
	// Don't rectify the exponent.  Madness lies that way. :)
	// Really, e.g. mod 11, we want 2 ** 13 to be 8.
	// *Really*, this program needs to have heterogeneous data types,
	// or at least have the possibility of the exponent being int
	// and the base being of modular type.

	c  = 1;
	p2 = a;

	for (i = 0; i < 32; i++) {
		if (b & 1) {
			c = (c * p2) % modulus;
		}
		b >>= 1;
		p2 = (p2 * p2) % modulus;
	}

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void mod_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, c;
	unsigned modulus = *(unsigned *)pvaux;
	memcpy(&a, pvin, sizeof(a));
	c = a % modulus;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void mod_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, c;
	unsigned modulus = *(unsigned *)pvaux;
	memcpy(&a, pvin, sizeof(a));
	a %= modulus;
	if (a == 0)
		c = 0;
	else
		c = modulus - a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void mod_ord (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned a, apow, ord;
	unsigned modulus = *(unsigned *)pvaux;
	memcpy(&a, pvin, sizeof(a));
	a %= modulus;

	if (a == 0) {
		ord = calc_phi_unsigned(modulus);
	}
	else if (calc_gcd_unsigned(a, modulus) != 1) {
		fprintf(stderr,
			"Order of %u undefined since not relatively "
			"prime to modulus %u.\n",
			a, modulus);
		exit(1);
	}
	else {
		apow = 1;
		ord  = 0;
		do {
			apow = (a * apow) % modulus;
			ord++;
		} while (apow != 1);
	}
	memcpy(pvout, &ord, sizeof(a));
}

// ================================================================
ops_t mod_ops = {
	atom_desc          : "mod",
	plex_first_func    : mod_lex_first,
	pscan_func         : mod_scan,
	pprint_func        : mod_print,
	prectifier         : mod_rectifier,
	padd_op            : mod_add,
	psub_op            : mod_sub,
	pmul_op            : mod_mul,
	pdiv_op            : mod_div,
	ppow_op            : mod_power,
	puadd_op           : mod_uadd,
	pusub_op           : mod_usub,
	pord_op            : mod_ord,
	ptopt_alloc        : mod_opt_alloc,
	parg_defaulter     : mod_opt_arg_defaulter,
	parg_handler       : mod_opt_arg_handler,
	atom_size_in_bytes : sizeof(unsigned),
	aux_size_in_bytes  : sizeof(unsigned)
};
