#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "qnfe.h"

// ----------------------------------------------------------------
void * qnfe_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
void qnfe_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	qnfe_aux_t m;
	m.D = rat_from_int(2);  // Q(sqrt(2)) by default

	*(qnfe_aux_t *)pvaux = m;
}

// ----------------------------------------------------------------
int  qnfe_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	qnfe_aux_t m;

	if (strcmp(argv[argi], "-D") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr, "qnfe:  -D requires an argument.\n");
			return 0;
		}
		m.D.denom = 1;
		if (sscanf(argv[argi+1], "%d", &m.D.numer) != 1) {
			fprintf(stderr,
				"qnfe:  Couldn't parse \"%s\" as integer.\n",
				argv[argi+1]);
			return 0;
		}
		*(qnfe_aux_t *)pvaux = m;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int qnfe_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == 'd')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int qnfe_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	qnfe_t * pg = patom;
	int ch;

	ch = getc(fp);
	if (ch == 'd') {
		// xxx ctor
		pg->r = rat_from_int(0);
		pg->i = rat_from_int(1);
		return 1;
	}
	ungetc(ch, fp);

	// data_types directory stuff does string scan.
	// kalk does file scan.
	pg->r = rat_from_int(1);
	pg->i = rat_from_int(0);
	// xxx allow ","?
	return fscanf(fp, "%d", &pg->r.numer);
}

// ----------------------------------------------------------------
static void qnfe_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack

	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	qnfe_t * pg = patom;
	qnfe_print(*pg);
}

// ----------------------------------------------------------------
static void qnfe_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnfe_t a = *(qnfe_t *)pvin1;
	qnfe_t b = *(qnfe_t *)pvin2;
	qnfe_aux_t m = *(qnfe_aux_t *)pvaux;
	qnfe_t c = qnfe_add(a, b, m);
	*(qnfe_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnfe_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnfe_t a = *(qnfe_t *)pvin1;
	qnfe_t b = *(qnfe_t *)pvin2;
	qnfe_aux_t m = *(qnfe_aux_t *)pvaux;
	qnfe_t c = qnfe_sub(a, b, m);
	*(qnfe_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnfe_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnfe_t a = *(qnfe_t *)pvin1;
	qnfe_t b = *(qnfe_t *)pvin2;
	qnfe_aux_t m = *(qnfe_aux_t *)pvaux;
	qnfe_t c = qnfe_mul(a, b, m);
	*(qnfe_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnfe_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnfe_t a = *(qnfe_t *)pvin1;
	qnfe_t b = *(qnfe_t *)pvin2;
	qnfe_aux_t m = *(qnfe_aux_t *)pvaux;
	qnfe_t c = qnfe_quot(a, b, m);
	*(qnfe_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnfe_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnfe_t a = *(qnfe_t *)pvin1;
	qnfe_t b = *(qnfe_t *)pvin2;
	qnfe_aux_t m = *(qnfe_aux_t *)pvaux;
	qnfe_t c = qnfe_rem(a, b, m);
	*(qnfe_t *)pvout = c;
}

//// ----------------------------------------------------------------
//static void qnfe_vpower(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	unsigned a, b, c;
//	unsigned qnfeulus = *(unsigned *)pvaux;
//	unsigned i, p2;
//	memcpy(&a, pvin1, sizeof(a));
//	memcpy(&b, pvin2, sizeof(a));
//
//	a %= qnfeulus;
//	// Don't rectify the exponent.  Madness lies that way. :)
//	// Really, e.g. qnfe 11, we want 2 ** 13 to be 8.
//	// *Really*, this program needs to have heterogeneous data types,
//	// or at least have the possibility of the exponent being int
//	// and the base being of qnfeular type.
//
//	c  = 1;
//	p2 = a;
//
//	for (i = 0; i < 32; i++) {
//		if (b & 1) {
//			c = (c * p2) % qnfeulus;
//		}
//		b >>= 1;
//		p2 = (p2 * p2) % qnfeulus;
//	}
//
//	memcpy(pvout, &c, sizeof(a));
//}

// ----------------------------------------------------------------
static void qnfe_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(qnfe_t *)pvout = *(qnfe_t *)pvin;
}

// ----------------------------------------------------------------
static void qnfe_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	qnfe_t a = *(qnfe_t *)pvin;
	qnfe_t c;

	// rat usub
	c.r.numer = -a.r.numer;
	c.i.numer = -a.i.numer;

	*(qnfe_t *)pvout = c;
}

// ================================================================
ops_t qnfe_ops = {
	atom_desc          : "qnfe",
	plex_first_func    : qnfe_lex_first,
	pscan_func         : qnfe_vscan,
	pprint_func        : qnfe_vprint,
//	prectifier         : qnfe_vrectifier,
	padd_op            : qnfe_vadd,
	psub_op            : qnfe_vsub,
	pmul_op            : qnfe_vmul,
	pdiv_op            : qnfe_vdiv,
	pmod_op            : qnfe_vmod,
//	ppow_op            : qnfe_vpower,
	puadd_op           : qnfe_vuadd,
	pusub_op           : qnfe_vusub,
//	pord_op            : qnfe_vord,
	ptopt_alloc        : qnfe_opt_alloc,
	parg_defaulter     : qnfe_opt_arg_defaulter,
	parg_handler       : qnfe_opt_arg_handler,
	atom_size_in_bytes : sizeof(qnfe_t),
	aux_size_in_bytes  : sizeof(qnfe_aux_t)
};
