#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "qnre.h"

// ----------------------------------------------------------------
void * qnre_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
void qnre_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	qnre_aux_t m;
	m.D = 2;  // Q(sqrt(2)) by default

	*(qnre_aux_t *)pvaux = m;
}

// ----------------------------------------------------------------
int  qnre_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	qnre_aux_t m;

	if (strcmp(argv[argi], "-D") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr, "qnre:  -D requires an argument.\n");
			return 0;
		}
		if (sscanf(argv[argi+1], "%d", &m.D) != 1) {
			fprintf(stderr,
				"qnre:  Couldn't parse \"%s\" as integer.\n",
				argv[argi+1]);
			return 0;
		}
		*(qnre_aux_t *)pvaux = m;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int qnre_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == 'd')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int qnre_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	qnre_t * pg = patom;
	int ch;

	ch = getc(fp);
	if (ch == 'd') {
		// xxx ctor
		pg->r = 0;
		pg->i = 1;
		return 1;
	}
	ungetc(ch, fp);

	// data_types directory stuff does string scan.
	// kalk does file scan.
	return fscanf(fp, "%d,%d", &pg->r, &pg->i);
}

// ----------------------------------------------------------------
static void qnre_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack

	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	qnre_t * pg = patom;
	fprintf(fp, "%d,%d", pg->r, pg->i);
}

// ----------------------------------------------------------------
static void qnre_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnre_t a = *(qnre_t *)pvin1;
	qnre_t b = *(qnre_t *)pvin2;
	qnre_aux_t m = *(qnre_aux_t *)pvaux;
	qnre_t c = qnre_add(a, b, m);
	*(qnre_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnre_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnre_t a = *(qnre_t *)pvin1;
	qnre_t b = *(qnre_t *)pvin2;
	qnre_aux_t m = *(qnre_aux_t *)pvaux;
	qnre_t c = qnre_sub(a, b, m);
	*(qnre_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnre_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnre_t a = *(qnre_t *)pvin1;
	qnre_t b = *(qnre_t *)pvin2;
	qnre_aux_t m = *(qnre_aux_t *)pvaux;
	qnre_t c = qnre_mul(a, b, m);
	*(qnre_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnre_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnre_t a = *(qnre_t *)pvin1;
	qnre_t b = *(qnre_t *)pvin2;
	qnre_aux_t m = *(qnre_aux_t *)pvaux;
	qnre_t c = qnre_quot(a, b, m);
	*(qnre_t *)pvout = c;
}

// ----------------------------------------------------------------
static void qnre_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qnre_t a = *(qnre_t *)pvin1;
	qnre_t b = *(qnre_t *)pvin2;
	qnre_aux_t m = *(qnre_aux_t *)pvaux;
	qnre_t c = qnre_rem(a, b, m);
	*(qnre_t *)pvout = c;
}

//// ----------------------------------------------------------------
//static void qnre_vpower(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	unsigned a, b, c;
//	unsigned qnreulus = *(unsigned *)pvaux;
//	unsigned i, p2;
//	memcpy(&a, pvin1, sizeof(a));
//	memcpy(&b, pvin2, sizeof(a));
//
//	a %= qnreulus;
//	// Don't rectify the exponent.  Madness lies that way. :)
//	// Really, e.g. qnre 11, we want 2 ** 13 to be 8.
//	// *Really*, this program needs to have heterogeneous data types,
//	// or at least have the possibility of the exponent being int
//	// and the base being of qnreular type.
//
//	c  = 1;
//	p2 = a;
//
//	for (i = 0; i < 32; i++) {
//		if (b & 1) {
//			c = (c * p2) % qnreulus;
//		}
//		b >>= 1;
//		p2 = (p2 * p2) % qnreulus;
//	}
//
//	memcpy(pvout, &c, sizeof(a));
//}

// ----------------------------------------------------------------
static void qnre_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(qnre_t *)pvout = *(qnre_t *)pvin;
}

// ----------------------------------------------------------------
static void qnre_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	qnre_t a = *(qnre_t *)pvin;
	qnre_t c;

	c.r = -a.r;
	c.i = -a.i;

	*(qnre_t *)pvout = c;
}

// ================================================================
ops_t qnre_ops = {
	atom_desc          : "qnre",
	plex_first_func    : qnre_lex_first,
	pscan_func         : qnre_vscan,
	pprint_func        : qnre_vprint,
//	prectifier         : qnre_vrectifier,
	padd_op            : qnre_vadd,
	psub_op            : qnre_vsub,
	pmul_op            : qnre_vmul,
	pdiv_op            : qnre_vdiv,
	pmod_op            : qnre_vmod,
//	ppow_op            : qnre_vpower,
	puadd_op           : qnre_vuadd,
	pusub_op           : qnre_vusub,
//	pord_op            : qnre_vord,
	ptopt_alloc        : qnre_opt_alloc,
	parg_defaulter     : qnre_opt_arg_defaulter,
	parg_handler       : qnre_opt_arg_handler,
	atom_size_in_bytes : sizeof(qnre_t),
	aux_size_in_bytes  : sizeof(qnre_aux_t)
};
