#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ops.h"
#include "symtab.h"

// ----------------------------------------------------------------
// These are implemented in each *ops.c file, e.g. u32ops.c.


//miquat.h

//zmx.h

//z3poly.h
//z3polymod.h

//z2npoly.h
//z2npolymod.h

//z3npoly.h
//z3npolymod.h


extern ops_t u64_ops;
extern ops_t u32_ops;
extern ops_t u16_ops;
extern ops_t u8_ops;
extern ops_t ubig_ops;

extern ops_t s64_ops;
extern ops_t s32_ops;
extern ops_t s16_ops;
extern ops_t s8_ops;

extern ops_t float_ops;
extern ops_t double_ops;
extern ops_t dcomplex_ops;
extern ops_t dquat_ops;

extern ops_t mod_ops;
extern ops_t mod64_ops;
extern ops_t rat_ops;
extern ops_t rat64_ops;

extern ops_t aesff_ops;

extern ops_t zpoly_ops;
extern ops_t zpolymod_ops;
extern ops_t zppoly_ops;
extern ops_t zppolymod_ops;
extern ops_t z2ipoly_ops;
extern ops_t z2ipolymod_ops;
extern ops_t z2poly_ops;
extern ops_t z2polymod_ops;

extern ops_t zpnpoly_ops;
extern ops_t zpnpolymod_ops;
extern ops_t z2nipoly_ops;
extern ops_t z2nipolymod_ops;

extern ops_t qpoly_ops;
extern ops_t qpolymod_ops;

extern ops_t gint_ops;
extern ops_t qnre_ops;
extern ops_t qnfe_ops;
extern ops_t pmt_ops;
extern ops_t pmtc_ops;
extern ops_t q8_ops;

// Default data type:
//ops_t * pdefops = &ubig_ops;
ops_t * pdefops = &u32_ops;

// ----------------------------------------------------------------
static ops_t * op_list[] = {

	&u64_ops,
	&u32_ops,
	&u16_ops,
	&u8_ops,

	&ubig_ops,

	&s64_ops,
	&s32_ops,
	&s16_ops,
	&s8_ops,

	&float_ops,
	&double_ops,
	&dcomplex_ops,
	&dquat_ops,

	&mod_ops,
	&mod64_ops,
	&rat_ops,
	&rat64_ops,

	&aesff_ops,
	&zpoly_ops,    &zpolymod_ops,
	&zppoly_ops,   &zppolymod_ops,
	&z2ipoly_ops,  &z2ipolymod_ops,
	&z2poly_ops,   &z2polymod_ops,

	&zpnpoly_ops,  &zpnpolymod_ops,
	&z2nipoly_ops, &z2nipolymod_ops,

	&qpoly_ops,    &qpolymod_ops,

	&gint_ops,
	&qnre_ops,
	&qnfe_ops,

	&pmt_ops,
	&pmtc_ops,
	&q8_ops,

	0 // MUST be null-terminated

};

static void check_sizes(ops_t * pops);

// ----------------------------------------------------------------
void ops_default(ops_t ** ppops, void ** ppvtopts, void * pvaux)
{
	ops_t * pops;

	*ppops = pdefops;
	pops = *ppops;

	if (pops->ptopt_alloc) {
		*ppvtopts = pops->ptopt_alloc();
		if (pops->parg_defaulter) {
			pops->parg_defaulter(*ppvtopts, pvaux);
		}
	}
	else {
		*ppvtopts = 0;
	}
}

// ----------------------------------------------------------------
int  ops_set(ops_t ** ppops, void ** ppvtopts, char * name,
	symtab_t * pst, void * pvaux)
{
	ops_t * pops;
	int i;

	for (i = 0; op_list[i]; i++) {
		if (strcmp(op_list[i]->atom_desc, name) == 0) {
			check_sizes(op_list[i]);
			*ppops = op_list[i];
			pops = *ppops;

			if (*ppvtopts != 0) {
				free(*ppvtopts);
				*ppvtopts = 0;
			}

			if (pops->ptopt_alloc) {
				*ppvtopts = pops->ptopt_alloc();
				if (pops->parg_defaulter) {
					pops->parg_defaulter(*ppvtopts, pvaux);
				}
			}
			else {
				*ppvtopts = 0;
			}

			if (pops->psymtab_populator)
				pops->psymtab_populator(pst);

			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
void ops_release(ops_t ** ppops, void ** ppvtopts)
{
	if (*ppvtopts != 0) {
		free(*ppvtopts);
		*ppvtopts = 0;
	}
}

// ----------------------------------------------------------------
static void check_sizes(ops_t * pops)
{
	if (pops->atom_size_in_bytes > MAX_ATOM_SIZE) {
		fprintf(stderr,
			"Data type %s requires atom size %d; max is %d.\n",
			pops->atom_desc, pops->atom_size_in_bytes,
			MAX_ATOM_SIZE);
		fprintf(stderr,
			"Please increase MAX_ATOM_SIZE and recompile.\n");
		exit(1);
	}

	if (pops->aux_size_in_bytes > MAX_AUX_SIZE) {
		fprintf(stderr,
			"Data type %s requires aux size %d; max is %d.\n",
			pops->atom_desc, pops->aux_size_in_bytes,
			MAX_AUX_SIZE);
		fprintf(stderr,
			"Please increase MAX_AUX_SIZE and recompile.\n");
		exit(1);
	}
}
