#ifndef OPS_H
#define OPS_H

#include <stdio.h>
#include "ops.h"
#include "atom.h"
#include "cmps.h"
#include "symtab.h"

typedef int  lex_first_func_t(int c, void * pvtopts, int in_pow_rhs);

typedef int  scan_func_t(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs);

typedef void print_func_t(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux);

typedef void rectifier_t(
	void * c,
	void * pvaux);

typedef void binary_op_func_t(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux);

typedef void unary_op_func_t(
	void * pvout,
	void * pvin,
	void * pvaux);

typedef int  cmp_op_func_t(
	void * pvin1,
	int    cmp,
	void * pvin2,
	void * pvaux);

typedef void int_cvt_func_t  (void * out, int i);

typedef void symtab_populator_t(symtab_t * pst);
typedef void * topt_alloc_t  (void);

typedef void arg_defaulter_t(
	void * pvtopts,
	void * pvaux);

typedef int  arg_handler_t(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used);

typedef struct _ops_t {
	char             * atom_desc;

	lex_first_func_t * plex_first_func;
	scan_func_t      * pscan_func;
	print_func_t     * pprint_func;

	rectifier_t      * prectifier;

	binary_op_func_t * por_op;
	binary_op_func_t * pxor_op;
	binary_op_func_t * pand_op;
	binary_op_func_t * plsh_op;
	binary_op_func_t * prsh_op;
	binary_op_func_t * padd_op;
	binary_op_func_t * psub_op;
	binary_op_func_t * pmul_op;
	binary_op_func_t * pdiv_op;
	binary_op_func_t * pmod_op;
	binary_op_func_t * ppow_op;
	binary_op_func_t * pcomma_op;

	unary_op_func_t  * puadd_op;
	unary_op_func_t  * pusub_op; // E.g. unary minus; -3
	unary_op_func_t  * punot_op; // E.g. bitwise negation; ~3

	binary_op_func_t * pgcd_op;
	binary_op_func_t * pchoose_op;

	unary_op_func_t  * psqrt_op;
	unary_op_func_t  * pexp_op;
	unary_op_func_t  * plog_op;
	unary_op_func_t  * psin_op;
	unary_op_func_t  * pcos_op;
	unary_op_func_t  * ptan_op;
	unary_op_func_t  * patan_op;
	unary_op_func_t  * pconj_op;
	unary_op_func_t  * pmag_op;
	unary_op_func_t  * pphz_op;
	unary_op_func_t  * pord_op;
	unary_op_func_t  * pfct_op;

	cmp_op_func_t    * pcmp_op;
	int_cvt_func_t   * pfrom_int_op;

	symtab_populator_t * psymtab_populator;
	topt_alloc_t     * ptopt_alloc;
	arg_defaulter_t  * parg_defaulter;
	arg_handler_t    * parg_handler;

	int atom_size_in_bytes;
	int aux_size_in_bytes;

} ops_t;

void ops_default(ops_t ** ppops, void ** ppvtopts, void * pvaux);
int  ops_set    (ops_t ** ppops, void ** ppvtopts, char * name,
	symtab_t * pst, void * pvaux);
void ops_release(ops_t ** ppops, void ** ppvtopts);

#endif // OPS_H
