#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "ops.h"
#include "aesfftbls.h"

// ----------------------------------------------------------------
static int aesff_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isxdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int aesff_scan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	unsigned      i;
	unsigned char c;
	if (!fscanf(fp, "%x", &i))
		return 0;
	c = i;
	memcpy(patom, &c, sizeof(c));
	return 1;
}

// ----------------------------------------------------------------
static void aesff_print(FILE * fp, void * patom, void * pvtopts, void * pvaux)
{
	unsigned char x;
	memcpy(&x, patom, sizeof(x));
	fprintf(fp, "%02x", (unsigned)x);
}

// ----------------------------------------------------------------
static void aesff_add(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a ^ b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void aesff_sub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	c = a ^ b;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void aesff_mul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned char a, b, c;
	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));
	if (a && b)
		c = aesff_exp_table[aesff_log_table[a] + aesff_log_table[b]];
	else
		c = 0;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void aesff_div(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	unsigned char a, b, c;
	int idx;

	memcpy(&a, pvin1, sizeof(a));
	memcpy(&b, pvin2, sizeof(a));

	if (a && b) {
		idx = (aesff_log_table[a] - aesff_log_table[b] + 510) % 255;
		c = aesff_exp_table[idx];
	}
	else
		c = 0;

	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void aesff_uadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned char a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ----------------------------------------------------------------
static void aesff_usub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	unsigned char a, c;
	memcpy(&a, pvin, sizeof(a));
	c = a;
	memcpy(pvout, &c, sizeof(a));
}

// ================================================================
ops_t aesff_ops = {
	atom_desc          : "aesff",
	plex_first_func    : aesff_lex_first,
	pscan_func         : aesff_scan,
	pprint_func        : aesff_print,
	padd_op            : aesff_add,
	psub_op            : aesff_sub,
	pmul_op            : aesff_mul,
	pdiv_op            : aesff_div,
	puadd_op           : aesff_uadd,
	pusub_op           : aesff_usub,
	atom_size_in_bytes : sizeof(unsigned char),
	aux_size_in_bytes  : 0
};
