#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "qpolymod.h"
#include "cyclopoly.h"

// ----------------------------------------------------------------
static void * qpolymod_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void qpolymod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	int n = 8;
	qpoly_t r = qpoly_find_nth_cyclo(n);
	*(qpoly_t *)pvaux = r;
}

// ----------------------------------------------------------------
static int  qpolymod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	qpoly_t r;
	int n;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr,
				"qpolymod:  -r requires one arguments ("
				"n or red).\n");
			return 0;
		}

		if (sscanf(argv[argi+1], "%d", &n) != 1) {
			fprintf(stderr,
			"qpolymod: Couldn't scan \"%s\" as integer.\n",
			argv[argi]);
			return 0;
		}

		r = qpoly_find_nth_cyclo(n);
		*(qpoly_t *)pvaux = r;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int qpolymod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == 'x')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int qpolymod_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	qpoly_t * pz = patom;

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	if (ch == 'x') {
		*pz = qpoly_from_coeffs1(rat_from_int(1), rat_from_int(0));
		return 1;
	}

	while (isdigit(ch)) {
		if (p >= end) {
			fprintf(stderr, "qpolymod_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return qpoly_scan(pz, mybuf);
}

// ----------------------------------------------------------------
static void qpolymod_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	qpoly_t * pg = patom;
	qpoly_x_print(pg);
}

// ----------------------------------------------------------------
static void qpolymod_vrectifier(void * buf, void * pvaux)
{
	qpoly_t * pa = buf;
	qpoly_t * paux = pvaux;
	qpoly_t   quot;

	qpoly_quot_and_rem(pa, paux, &quot, pa);
}

// ----------------------------------------------------------------
static void qpolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpoly_add(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void qpolymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpoly_sub(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void qpolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpolymod_mul(pvin1, pvin2, pvaux);
}

// ----------------------------------------------------------------
static void qpolymod_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpolymod_div(pvin1, pvin2, pvaux);
}

// ----------------------------------------------------------------
static void qpolymod_vcomma(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	// c = ax + b
	qpoly_t x, ax;

	x = qpoly_from_coeffs1(rat_from_int(1), rat_from_int(0));
	ax = qpolymod_mul(pvin1, &x, pvaux);
	*(qpoly_t *)pvout = qpoly_add(&ax, pvin2);
}


// ----------------------------------------------------------------
static void qpolymod_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qpoly_t * p2 = pvin2;

	if (p2->degree > 0) {
		fprintf(stderr,
		"qpolymod_vpower:  Can only raise to scalar power.\n");
		exit(1);
	}

	if (p2->coeffs[0].denom != 1) {
		fprintf(stderr, "qpolymod_vpower:  Can only raise to a "
			"non-negative integer power.\n");
		exit(1);
	}

	if (p2->coeffs[0].numer < 0) {
		fprintf(stderr, "qpolymod_vpower:  Can only raise to a "
			"non-negative integer power.\n");
		exit(1);
	}

	*(qpoly_t *)pvout = qpolymod_power(pvin1,
		(unsigned)p2->coeffs[0].numer, pvaux);
}

// ----------------------------------------------------------------
static void qpolymod_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(qpoly_t *)pvout = *(qpoly_t *)pvin;
}

// ----------------------------------------------------------------
static void qpolymod_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpoly_usub(pvin);
}

// ================================================================
ops_t qpolymod_ops = {
	atom_desc          : "qpolymod",
	plex_first_func    : qpolymod_lex_first,
	pscan_func         : qpolymod_vscan,
	pprint_func        : qpolymod_vprint,
	prectifier         : qpolymod_vrectifier,
	padd_op            : qpolymod_vadd,
	psub_op            : qpolymod_vsub,
	pmul_op            : qpolymod_vmul,
	pdiv_op            : qpolymod_vdiv,
	ppow_op            : qpolymod_vpower,
	pcomma_op          : qpolymod_vcomma,
	puadd_op           : qpolymod_vuadd,
	pusub_op           : qpolymod_vusub,
//	pord_op            : qpolymod_vord,
	ptopt_alloc        : qpolymod_opt_alloc,
	parg_defaulter     : qpolymod_opt_arg_defaulter,
	parg_handler       : qpolymod_opt_arg_handler,
	atom_size_in_bytes : sizeof(qpoly_t),
	aux_size_in_bytes  : sizeof(qpoly_t)
};
