#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "qpoly.h"

// ----------------------------------------------------------------
static int qpoly_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == 'x')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int qpoly_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	qpoly_t * pz = patom;

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	if (ch == 'x') {
		*pz = qpoly_from_coeffs1(rat_from_int(1), rat_from_int(0));
		return 1;
	}

	while (isdigit(ch)) {
		if (p >= end) {
			fprintf(stderr, "qpoly_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return qpoly_scan(pz, mybuf);
}

// ----------------------------------------------------------------
static void qpoly_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	qpoly_t * pg = patom;
	qpoly_x_print(pg);
}

// ----------------------------------------------------------------
static void qpoly_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpoly_add(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void qpoly_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpoly_sub(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void qpoly_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpoly_mul(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void qpoly_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qpoly_t rem;
	qpoly_quot_and_rem(pvin1, pvin2, pvout, &rem);
}

// ----------------------------------------------------------------
static void qpoly_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qpoly_t quot;
	qpoly_quot_and_rem(pvin1, pvin2, &quot, pvout);
}

// ----------------------------------------------------------------
static void qpoly_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	qpoly_t * p2 = pvin2;

	if (p2->degree > 0) {
		fprintf(stderr,
		"qpoly_vpower:  Can only raise to scalar power.\n");
		exit(1);
	}

	if (p2->coeffs[0].denom != 1) {
		fprintf(stderr, "qpoly_vpower:  Can only raise to a "
			"non-negative integer power.\n");
		exit(1);
	}

	if (p2->coeffs[0].numer < 0) {
		fprintf(stderr, "qpoly_vpower:  Can only raise to a "
			"non-negative integer power.\n");
		exit(1);
	}

	*(qpoly_t *)pvout = qpoly_power(pvin1, (unsigned)p2->coeffs[0].numer);
}

// ----------------------------------------------------------------
static void qpoly_vcomma(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	// c = ax + b
	qpoly_t x, ax;

	x = qpoly_from_coeffs1(rat_from_int(1), rat_from_int(0));
	ax = qpoly_mul(pvin1, &x);
	*(qpoly_t *)pvout = qpoly_add(&ax, pvin2);
}

// ----------------------------------------------------------------
static void qpoly_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(qpoly_t *)pvout = *(qpoly_t *)pvin;
}

// ----------------------------------------------------------------
static void qpoly_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(qpoly_t *)pvout = qpoly_usub(pvin);
}

// ================================================================
ops_t qpoly_ops = {
	atom_desc          : "qpoly",
	plex_first_func    : qpoly_lex_first,
	pscan_func         : qpoly_vscan,
	pprint_func        : qpoly_vprint,
	padd_op            : qpoly_vadd,
	psub_op            : qpoly_vsub,
	pmul_op            : qpoly_vmul,
	pdiv_op            : qpoly_vdiv,
	pmod_op            : qpoly_vmod,
	ppow_op            : qpoly_vpower,
	pcomma_op          : qpoly_vcomma,
	puadd_op           : qpoly_vuadd,
	pusub_op           : qpoly_vusub,
	atom_size_in_bytes : sizeof(qpoly_t),
	aux_size_in_bytes  : 0
};
