#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "z2ipolymod.h"

// ----------------------------------------------------------------
static void * z2ipolymod_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void z2ipolymod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	z2ipoly_t r = 0x11d;
	*(z2ipoly_t *)pvaux = r;
}

// ----------------------------------------------------------------
static int  z2ipolymod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	z2ipoly_t r;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr,
			"z2ipoly_t:  -r requires an argument.\n");
			return 0;
		}
		if (!z2ipolymod_auxscan(&r, argv[argi+1])) {
			fprintf(stderr,
			"z2ipoly_t Couldn't scan \"%s\" as reductor.\n",
			argv[argi]);
			return 0;
		}
		*(z2ipoly_t *)pvaux = r;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int z2ipolymod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isxdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int z2ipolymod_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	// xxx temp hack

	// data_types directory stuff does string scan.
	// kalk does file scan.
	z2ipoly_t * pg = patom;
	if (in_pow_rhs)
		return fscanf(fp, "%lld", pg);
	else
		return fscanf(fp, "%llx", pg);
}

// ----------------------------------------------------------------
static void z2ipolymod_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack

	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	// xxx need pvaux here!
	z2ipoly_t a = *(z2ipoly_t *)patom;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	z2ipolymod_print(a, r);
}

// ----------------------------------------------------------------
static void z2ipolymod_vrectifier(void * buf, void * pvaux)
{
	z2ipoly_t a = *(z2ipoly_t *)buf;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	a = z2ipoly_mod(a, r);
	*(z2ipoly_t *)buf = a;
}

// ----------------------------------------------------------------
static void z2ipolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2ipoly_t a = *(z2ipoly_t *)pvin1;
	z2ipoly_t b = *(z2ipoly_t *)pvin2;
	z2ipoly_t c = z2ipoly_add(a, b);
	*(z2ipoly_t *)pvout = c;
}

// ----------------------------------------------------------------
static void z2ipolymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2ipoly_t a = *(z2ipoly_t *)pvin1;
	z2ipoly_t b = *(z2ipoly_t *)pvin2;
	z2ipoly_t c = z2ipoly_sub(a, b);
	*(z2ipoly_t *)pvout = c;
}

// ----------------------------------------------------------------
static void z2ipolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2ipoly_t a = *(z2ipoly_t *)pvin1;
	z2ipoly_t b = *(z2ipoly_t *)pvin2;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	z2ipoly_t c = z2ipolymod_mul(a, b, r);
	*(z2ipoly_t *)pvout = c;
}

// ----------------------------------------------------------------
static void z2ipolymod_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2ipoly_t a = *(z2ipoly_t *)pvin1;
	z2ipoly_t b = *(z2ipoly_t *)pvin2;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	z2ipoly_t c = z2ipolymod_div(a, b, r);
	*(z2ipoly_t *)pvout = c;
}

// ----------------------------------------------------------------
static void z2ipolymod_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2ipoly_t a = *(z2ipoly_t *)pvin1;
	z2ipoly_t b = *(z2ipoly_t *)pvin2;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	z2ipoly_t c = z2ipolymod_power(a, (unsigned)b, r);
	*(z2ipoly_t *)pvout = c;
}

// ----------------------------------------------------------------
static void z2ipolymod_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2ipoly_t *)pvout = *(z2ipoly_t *)pvin;
}

// ----------------------------------------------------------------
static void z2ipolymod_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2ipoly_t *)pvout = *(z2ipoly_t *)pvin;
}

// ----------------------------------------------------------------
// This is gross because kalk will print the result as a z2ipoly_t (in hex).
static void z2ipolymod_vord (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	z2ipoly_t a = *(z2ipoly_t *)pvin;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;
	z2ipoly_t c = (z2ipoly_t) (unsigned)z2ipolymod_element_order(a, r);
	*(z2ipoly_t *)pvout = c;
}

// ================================================================
ops_t z2ipolymod_ops = {
	atom_desc          : "z2ipolymod",
	plex_first_func    : z2ipolymod_lex_first,
	pscan_func         : z2ipolymod_vscan,
	pprint_func        : z2ipolymod_vprint,
	prectifier         : z2ipolymod_vrectifier,
	padd_op            : z2ipolymod_vadd,
	psub_op            : z2ipolymod_vsub,
	pmul_op            : z2ipolymod_vmul,
	pdiv_op            : z2ipolymod_vdiv,
//	pmod_op            : z2ipolymod_vmod,
	ppow_op            : z2ipolymod_vpower,
	puadd_op           : z2ipolymod_vuadd,
	pusub_op           : z2ipolymod_vusub,
	pord_op            : z2ipolymod_vord,
	ptopt_alloc        : z2ipolymod_opt_alloc,
	parg_defaulter     : z2ipolymod_opt_arg_defaulter,
	parg_handler       : z2ipolymod_opt_arg_handler,
	atom_size_in_bytes : sizeof(z2ipoly_t),
	aux_size_in_bytes  : sizeof(z2ipoly_t)
};
