#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "z2nipolymod.h"

typedef struct _z2nipolymod_aux_t {
	z2ipoly_t  innr;
	z2nipoly_t outr;
} z2nipolymod_aux_t;

// ----------------------------------------------------------------
static void * z2nipolymod_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void z2nipolymod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	z2nipolymod_aux_t aux;
	aux.innr = 0xbLL;
	aux.outr = z2nipoly_from_coeffs2(1LL, 2LL, 7LL);
	*(z2nipolymod_aux_t *)pvaux = aux;
}

// ----------------------------------------------------------------
static int  z2nipolymod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	z2nipolymod_aux_t aux;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 3) {
			fprintf(stderr, "z2nipolymod:  -r requires two "
				"arguments (inner and outer reds).\n");
			return 0;
		}
		if (!z2ipolymod_auxscan(&aux.innr, argv[argi+1])) {
			fprintf(stderr,
			"z2nipolymod: Couldn't scan \"%s\" as inner r.\n",
			argv[argi+1]);
			return 0;
		}
		if (!z2nipoly_scan(&aux.outr, argv[argi+2], aux.innr)) {
			fprintf(stderr,
			"z2nipolymod: Couldn't scan \"%s\" as outer r.\n",
			argv[argi+1]);
			return 0;
		}
		*(z2nipolymod_aux_t *)pvaux = aux;
		*pnum_args_used = 3;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int z2nipolymod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (in_pow_rhs) {
		if (isdigit(c))
			return 1;
	}
	else {
		if (isxdigit(c))
			return 1;
	}
	return 0;
}

// ----------------------------------------------------------------
static int z2nipolymod_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	z2nipoly_t * pz = patom;
	z2nipolymod_aux_t * paux = pvaux;

	if (in_pow_rhs) {
		*pz = z2nipoly_from_coeffs0(0);
		return fscanf(fp, "%llu", &pz->coeffs[0]);
	}

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

//	if (ch == 'x') {
//		*pz = z2nipolymod_from_coeffs1(1, 0);
//		return 1;
//	}

//	if (ch == 'u') {
//		xxx
//		*pz = z2nipolymod_from_coeffs0(2);
//		return 1;
//	}

	while (isxdigit(ch) || (ch == ':')) {
		if (p >= end) {
			fprintf(stderr,
				"z2nipolymod_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return z2nipoly_scan(pz, mybuf, paux->innr);
}

// ----------------------------------------------------------------
static void z2nipolymod_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	z2nipolymod_aux_t * paux = pvaux;

	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	z2nipoly_t * pa = (z2nipoly_t *)patom;
	z2nipoly_print(pa, paux->innr);
}

// ----------------------------------------------------------------
static void z2nipolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t * pa = pvin1;
	z2nipoly_t * pb = pvin2;
	z2nipolymod_aux_t * paux = pvaux;
	*(z2nipoly_t *)pvout = z2nipoly_add(pa, pb, paux->innr);
}

// ----------------------------------------------------------------
static void z2nipolymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t * pa = pvin1;
	z2nipoly_t * pb = pvin2;
	z2nipolymod_aux_t * paux = pvaux;
	*(z2nipoly_t *)pvout = z2nipoly_sub(pa, pb, paux->innr);
}

// ----------------------------------------------------------------
static void z2nipolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t * pa = pvin1;
	z2nipoly_t * pb = pvin2;
	z2nipolymod_aux_t * paux = pvaux;
	*(z2nipoly_t *)pvout = z2nipolymod_mul(pa, pb, &paux->outr, paux->innr);
}

//// ----------------------------------------------------------------
//static void z2nipolymod_vdiv(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	z2nipoly_t a = *(z2nipoly_t *)pvin1;
//	z2nipoly_t b = *(z2nipoly_t *)pvin2;
//	z2nipolymod_aux_t * paux = pvaux;
//	*(z2nipoly_t *)pvout = z2nipolymod_div(&a, &b, &paux->outr, paux->innr);
//}

//// ----------------------------------------------------------------
//static void z2nipolymod_vmod(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	z2nipoly_t a = *(z2nipoly_t *)pvin1;
//	z2nipoly_t b = *(z2nipoly_t *)pvin2;
//	z2nipolymod_aux_t * paux = pvaux;
//	*(z2nipoly_t *)pvout = z2nipolymod_mod(&a, &b, &paux->outr, paux->innr);
//}

// ----------------------------------------------------------------
static void z2nipolymod_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t * pin1 = (z2nipoly_t *)pvin1;
	z2nipoly_t * pin2 = (z2nipoly_t *)pvin2;
	z2nipolymod_aux_t * paux = pvaux;
	*(z2nipoly_t *)pvout = z2nipolymod_power(pin1, pin2->coeffs[0],
		&paux->outr, paux->innr);
}

// ----------------------------------------------------------------
static void z2nipolymod_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2nipoly_t *)pvout = *(z2nipoly_t *)pvin;
}

// ----------------------------------------------------------------
static void z2nipolymod_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2nipoly_t *)pvout = *(z2nipoly_t *)pvin;
}

//// ----------------------------------------------------------------
//// This is gross because kalk will print the result as a z2nipoly_t (in hex).
//static void z2nipolymod_vord (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//	z2nipoly_t a = *(z2nipoly_t *)pvin;
//	z2nipoly_t c = (z2nipoly_t) (unsigned)z2nipolymodmod_element_order(a, r);
//	*(z2nipoly_t *)pvout = c;
//}

// ================================================================
ops_t z2nipolymod_ops = {
	atom_desc          : "z2nipolymod",
	plex_first_func    : z2nipolymod_lex_first,
	pscan_func         : z2nipolymod_vscan,
	pprint_func        : z2nipolymod_vprint,
//	prectifier         : z2nipolymod_vrectifier,
	padd_op            : z2nipolymod_vadd,
	psub_op            : z2nipolymod_vsub,
	pmul_op            : z2nipolymod_vmul,
//	pdiv_op            : z2nipolymod_vdiv,
//	pmod_op            : z2nipolymod_vmod,
	ppow_op            : z2nipolymod_vpower,
	puadd_op           : z2nipolymod_vuadd,
	pusub_op           : z2nipolymod_vusub,
//	pord_op            : z2nipolymod_vord,
	ptopt_alloc        : z2nipolymod_opt_alloc,
	parg_defaulter     : z2nipolymod_opt_arg_defaulter,
	parg_handler       : z2nipolymod_opt_arg_handler,
	atom_size_in_bytes : sizeof(z2nipoly_t),
	aux_size_in_bytes  : sizeof(z2nipolymod_aux_t)
};
