#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "z2nipoly.h"

// ----------------------------------------------------------------
static void * z2nipoly_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void z2nipoly_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	*(z2ipoly_t *)pvaux = 0x13LL;
}

// ----------------------------------------------------------------
static int  z2nipoly_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	z2ipoly_t r;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr,
				"z2nipoly:  -r requires one argument (red).\n");
			return 0;
		}
		if (!z2ipoly_scan(&r, argv[argi+1])) {
			fprintf(stderr,
			"z2nipoly: Couldn't scan \"%s\" as reductor.\n",
			argv[argi+1]);
			return 0;
		}
		*(z2ipoly_t *)pvaux = r;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int z2nipoly_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isxdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int z2nipoly_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	z2nipoly_t * pz = patom;
	z2ipoly_t r = *(z2ipoly_t *)pvaux;

	if (in_pow_rhs) {
		*pz = z2nipoly_from_coeffs0(0);
		// xxx temp hack return fscanf(fp, "%u", &pz->parts[0]);
		return 0;
		//xxx
	}

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	if (ch == 'x') {
		*pz = z2nipoly_from_coeffs0(2);
		return 1;
	}

	while (isxdigit(ch) || (ch == ':')) {
		if (p >= end) {
			fprintf(stderr, "z2nipoly_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return z2nipoly_scan(pz, mybuf, r);
}

// ----------------------------------------------------------------
static void z2nipoly_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	z2ipoly_t r = *(z2ipoly_t *)pvaux;

	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	z2nipoly_t * pa = (z2nipoly_t *)patom;
	z2nipoly_print(pa, r);
}

// ----------------------------------------------------------------
static void z2nipoly_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t a = *(z2nipoly_t *)pvin1;
	z2nipoly_t b = *(z2nipoly_t *)pvin2;
	z2ipoly_t  r = *(z2ipoly_t *)pvaux;
	*(z2nipoly_t *)pvout = z2nipoly_add(&a, &b, r);
}

// ----------------------------------------------------------------
static void z2nipoly_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t a = *(z2nipoly_t *)pvin1;
	z2nipoly_t b = *(z2nipoly_t *)pvin2;
	z2ipoly_t  r = *(z2ipoly_t *)pvaux;
	*(z2nipoly_t *)pvout = z2nipoly_sub(&a, &b, r);
}

// ----------------------------------------------------------------
static void z2nipoly_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t a = *(z2nipoly_t *)pvin1;
	z2nipoly_t b = *(z2nipoly_t *)pvin2;
	z2ipoly_t  r = *(z2ipoly_t *)pvaux;
	*(z2nipoly_t *)pvout = z2nipoly_mul(&a, &b, r);
}

// ----------------------------------------------------------------
static void z2nipoly_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t a = *(z2nipoly_t *)pvin1;
	z2nipoly_t b = *(z2nipoly_t *)pvin2;
	z2ipoly_t  r = *(z2ipoly_t *)pvaux;
	*(z2nipoly_t *)pvout = z2nipoly_div(&a, &b, r);
}

// ----------------------------------------------------------------
static void z2nipoly_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2nipoly_t a = *(z2nipoly_t *)pvin1;
	z2nipoly_t b = *(z2nipoly_t *)pvin2;
	z2ipoly_t  r = *(z2ipoly_t *)pvaux;
	*(z2nipoly_t *)pvout = z2nipoly_mod(&a, &b, r);
}

//// ----------------------------------------------------------------
//static void z2nipoly_vpower(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	z2nipoly_t * pin1 = (z2nipoly_t *)pvin1;
//	//z2nipoly_t * pin2 = (z2nipoly_t *)pvin2;
//	//z2ipoly_t  r = *(z2ipoly_t *)pvaux;
//	//*(z2nipoly_t *)pvout = z2nipoly_power(pvin1, pin2->parts[0], r);
//	// xxx temp
//	*(z2nipoly_t *)pvout = *pin1;
//}

// ----------------------------------------------------------------
static void z2nipoly_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2nipoly_t *)pvout = *(z2nipoly_t *)pvin;
}

// ----------------------------------------------------------------
static void z2nipoly_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2nipoly_t *)pvout = *(z2nipoly_t *)pvin;
}

//// ----------------------------------------------------------------
//// This is gross because kalk will print the result as a z2nipoly_t (in hex).
//static void z2nipoly_vord (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//	z2nipoly_t a = *(z2nipoly_t *)pvin;
//	z2nipoly_t c = (z2nipoly_t) (unsigned)z2nipolymod_element_order(a, r);
//	*(z2nipoly_t *)pvout = c;
//}

// ================================================================
ops_t z2nipoly_ops = {
	atom_desc          : "z2nipoly",
	plex_first_func    : z2nipoly_lex_first,
	pscan_func         : z2nipoly_vscan,
	pprint_func        : z2nipoly_vprint,
//	prectifier         : z2nipoly_vrectifier,
	padd_op            : z2nipoly_vadd,
	psub_op            : z2nipoly_vsub,
	pmul_op            : z2nipoly_vmul,
	pdiv_op            : z2nipoly_vdiv,
	pmod_op            : z2nipoly_vmod,
//	ppow_op            : z2nipoly_vpower,
	puadd_op           : z2nipoly_vuadd,
	pusub_op           : z2nipoly_vusub,
//	pord_op            : z2nipoly_vord,
	ptopt_alloc        : z2nipoly_opt_alloc,
	parg_defaulter     : z2nipoly_opt_arg_defaulter,
	parg_handler       : z2nipoly_opt_arg_handler,
	atom_size_in_bytes : sizeof(z2nipoly_t),
	aux_size_in_bytes  : sizeof(z2ipoly_t)
};
