#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "z2polymod.h"

// ----------------------------------------------------------------
static void * z2polymod_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void z2polymod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	z2poly_t r = z2poly_from_base_rep(0x11d);
	*(z2poly_t *)pvaux = r;
}

// ----------------------------------------------------------------
static int  z2polymod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	z2poly_t r;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr,
			"z2poly_t:  -r requires an argument.\n");
			return 0;
		}
		if (!z2polymod_auxscan(&r, argv[argi+1])) {
			fprintf(stderr,
			"z2poly_t Couldn't scan \"%s\" as reductor.\n",
			argv[argi]);
			return 0;
		}
		*(z2poly_t *)pvaux = r;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int z2polymod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isxdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int z2polymod_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	z2poly_t * pz = patom;

	if (in_pow_rhs) {
		*pz = z2poly_from_base_rep(0);
		return fscanf(fp, "%u", &pz->parts[0]);
	}

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	if (ch == 'x') {
		*pz = z2poly_from_base_rep(2);
		return 1;
	}

	while (isxdigit(ch) || (ch == '_')) {
		if (p >= end) {
			fprintf(stderr,
				"z2polymod_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return z2poly_scan(pz, mybuf);
}

// ----------------------------------------------------------------
static void z2polymod_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack

	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	z2poly_t *pa = (z2poly_t *)patom;
	z2poly_t *pr = (z2poly_t *)pvaux;
	z2polymod_print(pa, pr);
}

// ----------------------------------------------------------------
static void z2polymod_vrectifier(void * buf, void * pvaux)
{
	z2poly_t * pa = (z2poly_t *)buf;
	z2poly_t * pr = (z2poly_t *)pvaux;
	*pa = z2poly_mod(pa, pr);
}

// ----------------------------------------------------------------
static void z2polymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(z2poly_t *)pvout = z2poly_add(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void z2polymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(z2poly_t *)pvout = z2poly_sub(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void z2polymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(z2poly_t *)pvout = z2polymod_mul(pvin1, pvin2, pvaux);
}

// ----------------------------------------------------------------
static void z2polymod_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(z2poly_t *)pvout = z2polymod_div(pvin1, pvin2, pvaux);
}

// ----------------------------------------------------------------
static void z2polymod_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2poly_t * pb = (z2poly_t *)pvin2;
	*(z2poly_t *)pvout = z2polymod_power(pvin1, pb->parts[0], pvaux);
}

// ----------------------------------------------------------------
static void z2polymod_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2poly_t *)pvout = *(z2poly_t *)pvin;
}

// ----------------------------------------------------------------
static void z2polymod_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2poly_t *)pvout = *(z2poly_t *)pvin;
}

// ================================================================
ops_t z2polymod_ops = {
	atom_desc          : "z2polymod",
	plex_first_func    : z2polymod_lex_first,
	pscan_func         : z2polymod_vscan,
	pprint_func        : z2polymod_vprint,
	prectifier         : z2polymod_vrectifier,
	padd_op            : z2polymod_vadd,
	psub_op            : z2polymod_vsub,
	pmul_op            : z2polymod_vmul,
	pdiv_op            : z2polymod_vdiv,
//	pmod_op            : z2polymod_vmod,
	ppow_op            : z2polymod_vpower,
	puadd_op           : z2polymod_vuadd,
	pusub_op           : z2polymod_vusub,
//	pord_op            : z2polymod_vord,
	ptopt_alloc        : z2polymod_opt_alloc,
	parg_defaulter     : z2polymod_opt_arg_defaulter,
	parg_handler       : z2polymod_opt_arg_handler,
	atom_size_in_bytes : sizeof(z2poly_t),
	aux_size_in_bytes  : sizeof(z2poly_t)
};
