#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "z2poly.h"

// ----------------------------------------------------------------
static int z2poly_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isxdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int z2poly_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	z2poly_t * pz = patom;

	if (in_pow_rhs) {
		*pz = z2poly_from_base_rep(0);
		return fscanf(fp, "%u", &pz->parts[0]);
	}

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	if (ch == 'x') {
		*pz = z2poly_from_base_rep(2);
		return 1;
	}

	while (isxdigit(ch) || (ch == '_')) {
		if (p >= end) {
			fprintf(stderr, "z2poly_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return z2poly_scan(pz, mybuf);
}

// ----------------------------------------------------------------
static void z2poly_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack

	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	z2poly_t * pa = (z2poly_t *)patom;
	z2poly_print(pa);
}

// ----------------------------------------------------------------
static void z2poly_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2poly_t a = *(z2poly_t *)pvin1;
	z2poly_t b = *(z2poly_t *)pvin2;
	*(z2poly_t *)pvout = z2poly_add(&a, &b);
}

// ----------------------------------------------------------------
static void z2poly_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2poly_t a = *(z2poly_t *)pvin1;
	z2poly_t b = *(z2poly_t *)pvin2;
	*(z2poly_t *)pvout = z2poly_sub(&a, &b);
}

// ----------------------------------------------------------------
static void z2poly_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2poly_t a = *(z2poly_t *)pvin1;
	z2poly_t b = *(z2poly_t *)pvin2;
	*(z2poly_t *)pvout = z2poly_mul(&a, &b);
}

// ----------------------------------------------------------------
static void z2poly_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2poly_t a = *(z2poly_t *)pvin1;
	z2poly_t b = *(z2poly_t *)pvin2;
	*(z2poly_t *)pvout = z2poly_div(&a, &b);
}

// ----------------------------------------------------------------
static void z2poly_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2poly_t a = *(z2poly_t *)pvin1;
	z2poly_t b = *(z2poly_t *)pvin2;
	*(z2poly_t *)pvout = z2poly_mod(&a, &b);
}

//// ----------------------------------------------------------------
static void z2poly_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	z2poly_t * pin2 = (z2poly_t *)pvin2;
	*(z2poly_t *)pvout = z2poly_power(pvin1, pin2->parts[0]);
}

// ----------------------------------------------------------------
static void z2poly_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2poly_t *)pvout = *(z2poly_t *)pvin;
}

// ----------------------------------------------------------------
static void z2poly_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(z2poly_t *)pvout = *(z2poly_t *)pvin;
}

//// ----------------------------------------------------------------
//// This is gross because kalk will print the result as a z2poly_t (in hex).
//static void z2poly_vord (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//	z2poly_t a = *(z2poly_t *)pvin;
//	z2poly_t c = (z2poly_t) (unsigned)z2polymod_element_order(a, r);
//	*(z2poly_t *)pvout = c;
//}

// ================================================================
ops_t z2poly_ops = {
	atom_desc          : "z2poly",
	plex_first_func    : z2poly_lex_first,
	pscan_func         : z2poly_vscan,
	pprint_func        : z2poly_vprint,
//	prectifier         : z2poly_vrectifier,
	padd_op            : z2poly_vadd,
	psub_op            : z2poly_vsub,
	pmul_op            : z2poly_vmul,
	pdiv_op            : z2poly_vdiv,
	pmod_op            : z2poly_vmod,
	ppow_op            : z2poly_vpower,
	puadd_op           : z2poly_vuadd,
	pusub_op           : z2poly_vusub,
//	pord_op            : z2poly_vord,
//	ptopt_alloc        : z2poly_opt_alloc,
//	parg_defaulter     : z2poly_opt_arg_defaulter,
//	parg_handler       : z2poly_opt_arg_handler,
	atom_size_in_bytes : sizeof(z2poly_t),
	aux_size_in_bytes  : 0
};
