#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "zpnpoly.h"
#include "zpnpolymod.h"

typedef struct _zpnpolymod_aux_t {
	zppoly_aux_t innaux;
	zpnpoly_t    outr;
} zpnpolymod_aux_t;

// ----------------------------------------------------------------
static void * zpnpolymod_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void zpnpolymod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	zpnpolymod_aux_t aux;
	zpoly_t c2 = zppoly_from_base_rep(2, 1);
	zpoly_t c1 = zppoly_from_base_rep(2, 2);
	zpoly_t c0 = zppoly_from_base_rep(2, 7);

	aux.innaux.p = 2;
	aux.innaux.r = zppoly_from_base_rep(2, 0xb);
	aux.outr     = zpnpoly_from_coeffs2(&c2, &c1, &c0);
	*(zpnpolymod_aux_t *)pvaux = aux;
}

// ----------------------------------------------------------------
static int  zpnpolymod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	zpnpolymod_aux_t aux;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 4) {
			fprintf(stderr,
				"zpnpoly:  -r requires 3 arguments "
				"(p, inn red, out red).\n");
			return 0;
		}
		if (!zppolymod_auxscan(&aux.innaux,argv[argi+1],argv[argi+2])){
			fprintf(stderr,
				"zpnpoly: Couldn't scan \"%s\" and \"%s\" "
				"as p and inner reductor.\n",
				argv[argi+1],
				argv[argi+2]);
			return 0;
		}
		if (!zpnpoly_scan(&aux.outr, argv[argi+3], &aux.innaux)) {
			fprintf(stderr,
				"zpnpoly: Couldn't scan \"%s\" "
				"as outer reductor.\n",
				argv[argi+3]);
			return 0;
		}

		*(zpnpolymod_aux_t *)pvaux = aux;
		*pnum_args_used = 4;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int zpnpolymod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int kzpnpolymod_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	zpnpoly_t * pz = patom;
	zpnpolymod_aux_t * paux = pvaux;

	if (in_pow_rhs) {
		*pz = zpnpoly_from_int(paux->innaux.p, 0);
		return fscanf(fp, "%d", &pz->coeffs[0].coeffs[0]);
	}

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	//if (ch == 'x') {
		//*pz = zpnpoly_from_coeffs0(2);
		//return 1;
	//}

	while (isdigit(ch) || (ch == ',') || (ch == ':')) {
		if (p >= end) {
			fprintf(stderr, "kzpnpoly_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return zpnpoly_scan(pz, mybuf, &paux->innaux);
}

// ----------------------------------------------------------------
static void kzpnpolymod_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	zpnpoly_t * pa = (zpnpoly_t *)patom;
	zpnpolymod_aux_t * paux = pvaux;

	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	zpnpoly_print(pa, &paux->innaux);
}

// ----------------------------------------------------------------
static void zpnpolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zpnpolymod_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpoly_add(pa, pb, &paux->innaux);
}

// ----------------------------------------------------------------
static void zpnpolymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zpnpolymod_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpoly_sub(pa, pb, &paux->innaux);
}

// ----------------------------------------------------------------
static void zpnpolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zpnpolymod_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpolymod_mul(pa, pb,
		&paux->outr, &paux->innaux);
}

//// ----------------------------------------------------------------
//static void zpnpolymod_vdiv(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	zpnpoly_t * pa = pvin1;
//	zpnpoly_t * pb = pvin2;
//	zpnpolymod_aux_t * paux = pvaux;
//	// xxx temp *(zpnpoly_t *)pvout = zpnpolymod_div(pa, pb, paux);
//}

//// ----------------------------------------------------------------
//static void zpnpolymod_vmod(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	zpnpoly_t * pa = pvin1;
//	zpnpoly_t * pb = pvin2;
//	zpnpolymod_aux_t * paux = pvaux;
//	*(zpnpoly_t *)pvout = zpnpolymod_mod(pa, pb, paux);
//}

// ----------------------------------------------------------------
static void zpnpolymod_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pin1 = pvin1;
	zpnpoly_t * pin2 = pvin2;
	zpnpolymod_aux_t  * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpolymod_power(pin1,
		pin2->coeffs[0].coeffs[0],
		&paux->outr, &paux->innaux);
}

// ----------------------------------------------------------------
static void zpnpolymod_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpnpoly_t *)pvout = *(zpnpoly_t *)pvin;
}

//// ----------------------------------------------------------------
//static void zpnpolymod_vusub (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//	*(zpnpoly_t *)pvout = *(zpnpoly_t *)pvin;
//}

//// ----------------------------------------------------------------
//// This is gross because kalk will print the result as a zpnpoly_t (in hex).
//static void zpnpolymod_vord (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//	zpnpoly_t a = *(zpnpoly_t *)pvin;
//	zpnpoly_t c = (zpnpoly_t) (unsigned)zpnpolymod_element_order(a, r);
//	*(zpnpoly_t *)pvout = c;
//}

// ================================================================
ops_t zpnpolymod_ops = {
	atom_desc          : "zpnpolymod",
	plex_first_func    : zpnpolymod_lex_first,
	pscan_func         : kzpnpolymod_vscan,
	pprint_func        : kzpnpolymod_vprint,
//	prectifier         : zpnpolymod_vrectifier,
	padd_op            : zpnpolymod_vadd,
	psub_op            : zpnpolymod_vsub,
	pmul_op            : zpnpolymod_vmul,
//	pdiv_op            : zpnpolymod_vdiv,
//	pmod_op            : zpnpolymod_vmod,
	ppow_op            : zpnpolymod_vpower,
	puadd_op           : zpnpolymod_vuadd,
//	pusub_op           : zpnpolymod_vusub,
//	pord_op            : zpnpolymod_vord,
	ptopt_alloc        : zpnpolymod_opt_alloc,
	parg_defaulter     : zpnpolymod_opt_arg_defaulter,
	parg_handler       : zpnpolymod_opt_arg_handler,
	atom_size_in_bytes : sizeof(zpnpoly_t),
	aux_size_in_bytes  : sizeof(zpnpolymod_aux_t)
};
