#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "zpnpoly.h"

// ----------------------------------------------------------------
static void * zpnpoly_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void zpnpoly_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	zppoly_aux_t aux;
	aux.p = 2;
	aux.r = zppoly_from_base_rep(2, 0x13);
	*(zppoly_aux_t *)pvaux = aux;
}

// ----------------------------------------------------------------
static int  zpnpoly_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	zppoly_aux_t aux;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 3) {
			fprintf(stderr,
			"zpnpoly:  -r requires two arguments (p and red).\n");
			return 0;
		}
		if (!zppolymod_auxscan(&aux, argv[argi+1], argv[argi+2])) {
			fprintf(stderr,
				"zpnpoly: Couldn't scan \"%s\" and \"%s\" "
				"as p and reductor.\n",
				argv[argi+1],
				argv[argi+2]);
			return 0;
		}
		*(zppoly_aux_t *)pvaux = aux;
		*pnum_args_used = 3;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int zpnpoly_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int kzpnpoly_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	zpnpoly_t * pz = patom;
	zppoly_aux_t * paux = pvaux;

	if (in_pow_rhs) {
		*pz = zpnpoly_from_int(paux->p, 0);
		return 0;
		//xxx
	}

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	//if (ch == 'x') {
		//*pz = zpnpoly_from_coeffs0(2);
		//return 1;
	//}

	while (isdigit(ch) || (ch == ',') || (ch == ':')) {
		if (p >= end) {
			fprintf(stderr, "kzpnpoly_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return zpnpoly_scan(pz, mybuf, paux);
}

// ----------------------------------------------------------------
static void kzpnpoly_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	zpnpoly_t * pa = (zpnpoly_t *)patom;
	zppoly_aux_t * paux = pvaux;

	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	zpnpoly_print(pa, paux);
}

// ----------------------------------------------------------------
static void zpnpoly_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zppoly_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpoly_add(pa, pb, paux);
}

// ----------------------------------------------------------------
static void zpnpoly_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zppoly_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpoly_sub(pa, pb, paux);
}

// ----------------------------------------------------------------
static void zpnpoly_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zppoly_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpoly_mul(pa, pb, paux);
}

// ----------------------------------------------------------------
static void zpnpoly_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zppoly_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpoly_div(pa, pb, paux);
}

// ----------------------------------------------------------------
static void zpnpoly_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpnpoly_t * pa = pvin1;
	zpnpoly_t * pb = pvin2;
	zppoly_aux_t * paux = pvaux;
	*(zpnpoly_t *)pvout = zpnpoly_mod(pa, pb, paux);
}

//// ----------------------------------------------------------------
//static void zpnpoly_vpower(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	zpnpoly_t * pin1 = pvin1;
//	//zpnpoly_t * pin2 = pvin2;
//	//zppoly_aux_t  * paux = pvaux;
//	//*(zpnpoly_t *)pvout = zpnpoly_power(pvin1, pin2->parts[0], r);
//	// xxx temp
//	*(zpnpoly_t *)pvout = *pin1;
//}

// ----------------------------------------------------------------
static void zpnpoly_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpnpoly_t *)pvout = *(zpnpoly_t *)pvin;
}

//// ----------------------------------------------------------------
//static void zpnpoly_vusub (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//	*(zpnpoly_t *)pvout = *(zpnpoly_t *)pvin;
//}

//// ----------------------------------------------------------------
//// This is gross because kalk will print the result as a zpnpoly_t (in hex).
//static void zpnpoly_vord (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//	zpnpoly_t a = *(zpnpoly_t *)pvin;
//	zpnpoly_t c = (zpnpoly_t) (unsigned)zpnpolymod_element_order(a, r);
//	*(zpnpoly_t *)pvout = c;
//}

// ================================================================
ops_t zpnpoly_ops = {
	atom_desc          : "zpnpoly",
	plex_first_func    : zpnpoly_lex_first,
	pscan_func         : kzpnpoly_vscan,
	pprint_func        : kzpnpoly_vprint,
//	prectifier         : zpnpoly_vrectifier,
	padd_op            : zpnpoly_vadd,
	psub_op            : zpnpoly_vsub,
	pmul_op            : zpnpoly_vmul,
	pdiv_op            : zpnpoly_vdiv,
	pmod_op            : zpnpoly_vmod,
//	ppow_op            : zpnpoly_vpower,
	puadd_op           : zpnpoly_vuadd,
//	pusub_op           : zpnpoly_vusub,
//	pord_op            : zpnpoly_vord,
	ptopt_alloc        : zpnpoly_opt_alloc,
	parg_defaulter     : zpnpoly_opt_arg_defaulter,
	parg_handler       : zpnpoly_opt_arg_handler,
	atom_size_in_bytes : sizeof(zpnpoly_t),
	aux_size_in_bytes  : sizeof(zppoly_aux_t)
};
