#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "zpolymod.h"
#include "cyclopoly.h"

// ----------------------------------------------------------------
static void * zpolymod_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void zpolymod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	int n = 8;
	zpoly_t r = zpoly_find_nth_cyclo(n);
	*(zpoly_t *)pvaux = r;
}

// ----------------------------------------------------------------
static int  zpolymod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	zpoly_t r;
	int n;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr,
				"zpolymod:  -r requires one arguments ("
				"n or red).\n");
			return 0;
		}

		if (sscanf(argv[argi+1], "%d", &n) != 1) {
			fprintf(stderr,
			"zpolymod: Couldn't scan \"%s\" as integer.\n",
			argv[argi]);
			return 0;
		}

		// xxx arb poly; -cyc or sth if they want cyclo.
		r = zpoly_find_nth_cyclo(n);
		*(zpoly_t *)pvaux = r;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int zpolymod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == 'x')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int zpolymod_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	zpoly_t * pz = patom;

	if (in_pow_rhs) {
		*pz = zpoly_from_coeffs0(0);
		return fscanf(fp, "%d", &pz->coeffs[0]);
	}


	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	if (ch == 'x') {
		*pz = zpoly_from_coeffs1(1, 0);
		return 1;
	}

	while (isdigit(ch)) {
		if (p >= end) {
			fprintf(stderr, "zpolymod_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return zpoly_scan(pz, mybuf);
}

// ----------------------------------------------------------------
static void zpolymod_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.
	zpoly_t * pg = patom;
	zpoly_print(pg);
}

// ----------------------------------------------------------------
static void zpolymod_vrectifier(void * buf, void * pvaux)
{
	zpoly_t * pa = buf;
	zpoly_t * paux = pvaux;
	zpoly_t   quot;

	zpoly_quot_and_rem(pa, paux, &quot, pa);
}

// ----------------------------------------------------------------
static void zpolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpoly_add(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void zpolymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpoly_sub(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void zpolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpolymod_mul(pvin1, pvin2, pvaux);
}

// ----------------------------------------------------------------
static void zpolymod_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	// xxx not right ...
	zpoly_t rem;
	zpoly_quot_and_rem(pvin1, pvin2, pvout, &rem);
}

// ----------------------------------------------------------------
static void zpolymod_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	// xxx not right ...
	zpoly_t quot;
	zpoly_quot_and_rem(pvin1, pvin2, &quot, pvout);
}

// ----------------------------------------------------------------
static void zpolymod_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t * p2 = pvin2;

	if (p2->degree > 0) {
		fprintf(stderr,
		"zpolymod_vpower:  Can only raise to scalar power.\n");
		exit(1);
	}

	if (p2->coeffs[0] < 0) {
		fprintf(stderr,
		"zpolymod_vpower:  Can only raise to a non-negative power.\n");
		exit(1);
	}

	*(zpoly_t *)pvout = zpolymod_power(pvin1, (unsigned)p2->coeffs[0], pvaux);
}

// ----------------------------------------------------------------
static void zpolymod_vcomma(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	// c = ax + b
	zpoly_t x, ax;

	x = zpoly_from_coeffs1(1, 0);
	ax = zpolymod_mul(pvin1, &x, pvaux);
	*(zpoly_t *)pvout = zpoly_add(&ax, pvin2);
}


// ----------------------------------------------------------------
static void zpolymod_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpoly_t *)pvout = *(zpoly_t *)pvin;
}

// ----------------------------------------------------------------
static void zpolymod_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpoly_usub(pvin);
}

// ================================================================
ops_t zpolymod_ops = {
	atom_desc          : "zpolymod",
	plex_first_func    : zpolymod_lex_first,
	pscan_func         : zpolymod_vscan,
	pprint_func        : zpolymod_vprint,
	prectifier         : zpolymod_vrectifier,
	padd_op            : zpolymod_vadd,
	psub_op            : zpolymod_vsub,
	pmul_op            : zpolymod_vmul,
	pdiv_op            : zpolymod_vdiv,
	pmod_op            : zpolymod_vmod,
	ppow_op            : zpolymod_vpower,
	pcomma_op          : zpolymod_vcomma,
	puadd_op           : zpolymod_vuadd,
	pusub_op           : zpolymod_vusub,
//	pord_op            : zpolymod_vord,
	ptopt_alloc        : zpolymod_opt_alloc,
	parg_defaulter     : zpolymod_opt_arg_defaulter,
	parg_handler       : zpolymod_opt_arg_handler,
	atom_size_in_bytes : sizeof(zpoly_t),
	aux_size_in_bytes  : sizeof(zpoly_t)
};
