#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "zpoly.h"

// ----------------------------------------------------------------
static int zpoly_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	if (c == 'x')
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int zpoly_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	zpoly_t * pz = patom;

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);

	if (ch == 'x') {
		*pz = zpoly_from_coeffs1(1, 0);
		return 1;
	}

	while (isdigit(ch)) {
		if (p >= end) {
			fprintf(stderr, "zpoly_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return zpoly_scan(pz, mybuf);
}

// ----------------------------------------------------------------
static void zpoly_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	zpoly_t * pg = patom;
	zpoly_print(pg);
}

// ----------------------------------------------------------------
static void zpoly_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpoly_add(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void zpoly_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpoly_sub(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void zpoly_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpoly_mul(pvin1, pvin2);
}

// ----------------------------------------------------------------
static void zpoly_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t rem;
	zpoly_quot_and_rem(pvin1, pvin2, pvout, &rem);
}

// ----------------------------------------------------------------
static void zpoly_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t quot;
	zpoly_quot_and_rem(pvin1, pvin2, &quot, pvout);
}

// ----------------------------------------------------------------
static void zpoly_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t * p2 = pvin2;

	if (p2->degree > 0) {
		fprintf(stderr,
		"zpoly_vpower:  Can only raise to scalar power.\n");
		exit(1);
	}

	if (p2->coeffs[0] < 0) {
		fprintf(stderr,
		"zpoly_vpower:  Can only raise to a non-negative power.\n");
		exit(1);
	}

	*(zpoly_t *)pvout = zpoly_power(pvin1, (unsigned)p2->coeffs[0]);
}

// ----------------------------------------------------------------
static void zpoly_vcomma(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	// c = ax + b
	zpoly_t x, ax;

	x = zpoly_from_coeffs1(1, 0);
	ax = zpoly_mul(pvin1, &x);
	*(zpoly_t *)pvout = zpoly_add(&ax, pvin2);
}


// ----------------------------------------------------------------
static void zpoly_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpoly_t *)pvout = *(zpoly_t *)pvin;
}

// ----------------------------------------------------------------
static void zpoly_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpoly_t *)pvout = zpoly_usub(pvin);
}

// ================================================================
ops_t zpoly_ops = {
	atom_desc          : "zpoly",
	plex_first_func    : zpoly_lex_first,
	pscan_func         : zpoly_vscan,
	pprint_func        : zpoly_vprint,
	padd_op            : zpoly_vadd,
	psub_op            : zpoly_vsub,
	pmul_op            : zpoly_vmul,
	pdiv_op            : zpoly_vdiv,
	pmod_op            : zpoly_vmod,
	ppow_op            : zpoly_vpower,
	pcomma_op          : zpoly_vcomma,
	puadd_op           : zpoly_vuadd,
	pusub_op           : zpoly_vusub,
	atom_size_in_bytes : sizeof(zpoly_t),
	aux_size_in_bytes  : 0
};
