#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "zppolymod.h"

// ----------------------------------------------------------------
static void * zppolymod_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void zppolymod_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	int p = 2;
	int m = 3;
	zppoly_aux_t r = zppoly_find_irreducible(p, m);
	*(zppoly_aux_t *)pvaux = r;
}

// ----------------------------------------------------------------
static int  zppolymod_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	zppoly_aux_t r;

	if (strcmp(argv[argi], "-r") == 0) {
		if ((argc - argi) < 3) {
			fprintf(stderr,
				"zppolymod:  -r requires two arguments (p, "
				"and m or red).\n");
			return 0;
		}
		if (!zppolymod_auxscan(&r, argv[argi+1], argv[argi+2])) {
			fprintf(stderr,
			"zppolymod: Couldn't scan \"%s\" as reductor.\n",
			argv[argi]);
			return 0;
		}
		*(zppoly_aux_t *)pvaux = r;
		*pnum_args_used = 3;
		return 1;
	}

	return 0;
}


// ----------------------------------------------------------------
static int zppolymod_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int zppolymod_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * p = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	zpoly_t * pz = patom;
	zppoly_aux_t * paux = pvaux;

	if (in_pow_rhs) {
		*pz = zppoly_from_base_rep(paux->p, 0);
		return fscanf(fp, "%d", &pz->coeffs[0]);
	}

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);
	while (isdigit(ch)) {
		if (p >= end) {
			fprintf(stderr, "zppolymod_vscan:  Overflow on read.\n");
			exit(1);
		}
		*p = ch;
		p++;
		ch = getc(fp);
	}
	*p = 0;
	ungetc(ch, fp);

	return zppoly_scan(pz, mybuf, paux->p);
}

// ----------------------------------------------------------------
static void zppolymod_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	zpoly_t      * pg = patom;
	zppoly_aux_t * pr = pvaux;
	zppolymod_print(pg, pr);
}

// ----------------------------------------------------------------
static void zppolymod_vrectifier(void * buf, void * pvaux)
{
	zpoly_t      * pa = buf;
	zppoly_aux_t * paux = pvaux;

	*pa = zppoly_mod(pa, &paux->r, paux->p);
}

// ----------------------------------------------------------------
static void zppolymod_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zppoly_aux_t * paux = pvaux;
	zpoly_t * pout = pvout;
	*pout = zppoly_add(pvin1, pvin2, paux->p);
}

// ----------------------------------------------------------------
static void zppolymod_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zppoly_aux_t * paux = pvaux;
	zpoly_t * pout = pvout;
	*pout = zppoly_sub(pvin1, pvin2, paux->p);
}

// ----------------------------------------------------------------
static void zppolymod_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zppolymod_mul(pvin1, pvin2, pvaux);
}

// ----------------------------------------------------------------
static void zppolymod_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	*(zpoly_t *)pvout = zppolymod_div(pvin1, pvin2, pvaux);
}

// ----------------------------------------------------------------
static void zppolymod_vcomma(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	// c = ax + b
	zpoly_t x, ax;
	zppoly_aux_t * paux = pvaux;

	x = zpoly_from_coeffs1(1, 0);
	ax = zppolymod_mul(pvin1, &x, pvaux);
	*(zpoly_t *)pvout = zppoly_add(&ax, pvin2, paux->p);
}


// ----------------------------------------------------------------
static void zppolymod_vpower(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t * pa = pvin1;
	zpoly_t * pb = pvin2;
	zppoly_aux_t * paux = pvaux;
	*(zpoly_t *)pvout = zppolymod_power(pa, pb->coeffs[0], paux);
}

// ----------------------------------------------------------------
static void zppolymod_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpoly_t *)pvout = *(zpoly_t *)pvin;
}

// ----------------------------------------------------------------
static void zppolymod_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	zppoly_aux_t * paux = pvaux;
	zpoly_t * pout = pvout;
	*pout = zppoly_usub(pvin, paux->p);
}

// ----------------------------------------------------------------
// xxx homogeneity problem here too.
//static void zppolymod_vord (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//
//int zppolymod_element_order(
//	zpoly_t * pz,
//	zppoly_aux_t * paux);
//
//}

// ================================================================
ops_t zppolymod_ops = {
	atom_desc          : "zppolymod",
	plex_first_func    : zppolymod_lex_first,
	pscan_func         : zppolymod_vscan,
	pprint_func        : zppolymod_vprint,
	prectifier         : zppolymod_vrectifier,
	padd_op            : zppolymod_vadd,
	psub_op            : zppolymod_vsub,
	pmul_op            : zppolymod_vmul,
	pdiv_op            : zppolymod_vdiv,
	ppow_op            : zppolymod_vpower,
	pcomma_op          : zppolymod_vcomma,
	puadd_op           : zppolymod_vuadd,
	pusub_op           : zppolymod_vusub,
//	pord_op            : zppolymod_vord,
	ptopt_alloc        : zppolymod_opt_alloc,
	parg_defaulter     : zppolymod_opt_arg_defaulter,
	parg_handler       : zppolymod_opt_arg_handler,
	atom_size_in_bytes : sizeof(zpoly_t),
	aux_size_in_bytes  : sizeof(zppoly_aux_t)
};
