#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ops.h"
#include "zppoly.h"

// ----------------------------------------------------------------
static void * zppoly_opt_alloc(void)
{
	return 0;
}

// ----------------------------------------------------------------
static void zppoly_opt_arg_defaulter(
	void * pvtopts,
	void * pvaux)
{
	int p = 2;
	*(int *)pvaux = p;
}

// ----------------------------------------------------------------
static int  zppoly_opt_arg_handler(
	void  * pvtopts,
	void  * pvaux,
	int     argi,
	int     argc,
	char ** argv,
	int   * pnum_args_used)
{
	int p;

	if (strcmp(argv[argi], "-p") == 0) {
		if ((argc - argi) < 2) {
			fprintf(stderr,
				"zppoly:  -p requires one argument (p).\n");
			return 0;
		}
		if (sscanf(argv[argi+1], "%d", &p) != 1) {
			fprintf(stderr,
				"zppoly: Couldn't scan \"%s\" as p.\n",
				argv[argi+1]);
			return 0;
		}
		if (p < 2) {
			fprintf(stderr,
				"zppoly: p must be >= 2; got %d.\n", p);
			return 0;
		}
		*(int *)pvaux = p;
		*pnum_args_used = 2;
		return 1;
	}

	return 0;
}

// ----------------------------------------------------------------
static int zppoly_lex_first(int c, void * pvtopts, int in_pow_rhs)
{
	if (isdigit(c))
		return 1;
	return 0;
}

// ----------------------------------------------------------------
static int zppoly_vscan(
	FILE * fp,
	void * patom,
	void * pvtopts,
	void * pvaux,
	int    in_pow_rhs)
{
	char   mybuf[256];
	int    ch;
	char * pb = mybuf;
	char * end = &mybuf[sizeof(mybuf) - 1];
	zpoly_t * pz = patom;
	int p = *(int *)pvaux;

	// data_types directory stuff does string scan; kalk does file scan.
	ch = getc(fp);
	while (isdigit(ch)) {
		if (pb >= end) {
			fprintf(stderr, "zppoly_vscan:  Overflow on read.\n");
			exit(1);
		}
		*pb = ch;
		pb++;
		ch = getc(fp);
	}
	*pb = 0;
	ungetc(ch, fp);

	return zppoly_scan(pz, mybuf, p);
}

// ----------------------------------------------------------------
static void zppoly_vprint(FILE * fp, void * patom, void * pvtopts,
	void * pvaux)
{
	// xxx temp hack
	// data_types directory stuff does print to stdout.
	// kalk does print to specified fp.

	zpoly_t     * pg = patom;
	int p = *(int *)pvaux;
	zppoly_print(pg, p);
}

// ----------------------------------------------------------------
static void zppoly_vadd(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int p = *(int *)pvaux;
	zpoly_t * pout = pvout;
	*pout = zppoly_add(pvin1, pvin2, p);
}

// ----------------------------------------------------------------
static void zppoly_vsub(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int p = *(int *)pvaux;
	zpoly_t * pout = pvout;
	*pout = zppoly_sub(pvin1, pvin2, p);
}

// ----------------------------------------------------------------
static void zppoly_vmul(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int p = *(int *)pvaux;
	*(zpoly_t *)pvout = zppoly_mul(pvin1, pvin2, p);
}

// ----------------------------------------------------------------
static void zppoly_vdiv(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int p = *(int *)pvaux;
	*(zpoly_t *)pvout = zppoly_div(pvin1, pvin2, p);
}

// ----------------------------------------------------------------
static void zppoly_vmod(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	int p = *(int *)pvaux;
	*(zpoly_t *)pvout = zppoly_mod(pvin1, pvin2, p);
}

// ----------------------------------------------------------------
// c = ax + b
static void zppoly_vcomma(
	void * pvout,
	void * pvin1,
	void * pvin2,
	void * pvaux)
{
	zpoly_t x, ax;
	zpoly_t * pout = pvout;
	int p = *(int *)pvaux;

	x = zpoly_from_coeffs1(1, 0);
	ax = zppoly_mul(pvin1, &x, p);
	*pout = zppoly_add(&ax, pvin2, p);
}


//// ----------------------------------------------------------------
// xxx need a way to store power as int in this program.
// remember kalk has homogeneous data types ...
// xxx a hack would be an extra attr in the zpp struct.

//static void zppoly_vpower(
//	void * pvout,
//	void * pvin1,
//	void * pvin2,
//	void * pvaux)
//{
//	zpoly_t a = *(zpoly_t *)pvin1;
//	zpoly_t b = *(zpoly_t *)pvin2;
//	int p = *(int *)pvaux
//	zpoly_t c = zppoly_power(a, (unsigned)b, p);

// ----------------------------------------------------------------
static void zppoly_vuadd (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	*(zpoly_t *)pvout = *(zpoly_t *)pvin;
}

// ----------------------------------------------------------------
static void zppoly_vusub (
	void * pvout,
	void * pvin,
	void * pvaux)
{
	zpoly_t * pout = pvout;
	int p = *(int *)pvaux;
	*pout = zppoly_usub(pvin, p);
}

// ----------------------------------------------------------------
// xxx homogeneity problem here too.
//static void zppoly_vord (
//	void * pvout,
//	void * pvin,
//	void * pvaux)
//{
//}

// ================================================================
ops_t zppoly_ops = {
	atom_desc          : "zppoly",
	plex_first_func    : zppoly_lex_first,
	pscan_func         : zppoly_vscan,
	pprint_func        : zppoly_vprint,
//	prectifier         : zppoly_vrectifier,
	padd_op            : zppoly_vadd,
	psub_op            : zppoly_vsub,
	pmul_op            : zppoly_vmul,
	pdiv_op            : zppoly_vdiv,
	pmod_op            : zppoly_vmod,
//	ppow_op            : zppoly_vpower,
	pcomma_op          : zppoly_vcomma,
	puadd_op           : zppoly_vuadd,
	pusub_op           : zppoly_vusub,
//	pord_op            : zppoly_vord,
	ptopt_alloc        : zppoly_opt_alloc,
	parg_defaulter     : zppoly_opt_arg_defaulter,
	parg_handler       : zppoly_opt_arg_handler,
	atom_size_in_bytes : sizeof(zpoly_t),
	aux_size_in_bytes  : sizeof(int)
};
