#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "symtab.h"
#include "error.h"

static void st_bounds_check(symtab_t * pst, int stidx);

// ----------------------------------------------------------------
void st_init(symtab_t * pst)
{
	pst->nchars = 0;
	pst->nents  = 0;
}

// ----------------------------------------------------------------
int st_lookup(symtab_t * pst, char * string, int * ptok, int * pidx)
{
	int enti;

	for (enti = pst->nents - 1; enti >= 0; enti--) {
		if (strcmp(pst->entries[enti].lexptr, string) == 0) {
			if (pidx)
				*pidx = enti;
			if (ptok)
				*ptok = pst->entries[enti].token;
			return 1;
		}
	}
	return 0;
}

// ----------------------------------------------------------------
int st_insert(symtab_t * pst, char * string, int tok, int * pidx)
{
	int len_with_null = strlen(string) + 1;

	if (pst->nents >= SYMMAX)
		error("symbol table full");
	if (pst->nchars + len_with_null >= STRMAX)
		error("lexemes array full");

	pst->entries[pst->nents].token = tok;
	pst->entries[pst->nents].lexptr = &pst->lexemes[pst->nchars];
	strcpy(pst->entries[pst->nents].lexptr, string);
	pst->entries[pst->nents].has_val = 0;

	if (pidx)
		*pidx = pst->nents;
	pst->nents++;
	pst->nchars += len_with_null;

	return 1;
}

// ----------------------------------------------------------------
static void st_bounds_check(symtab_t * pst, int stidx)
{
	if ((stidx < 0) || (stidx >= pst->nents)) {
		fprintf(stderr,
			"symbol table index %d out of bounds 0-%d.\n",
			stidx, pst->nents-1);
		exit(1);
	}
}

// ----------------------------------------------------------------
char * st_name(symtab_t * pst, int stidx)
{
	st_bounds_check(pst, stidx);
	return pst->entries[stidx].lexptr;
}

// ----------------------------------------------------------------
int  st_get_val(symtab_t * pst, int stidx, char buf[MAX_ATOM_SIZE])
{
	st_bounds_check(pst, stidx);
	if (pst->entries[stidx].has_val) {
		memcpy(buf, pst->entries[stidx].atombuf, MAX_ATOM_SIZE);
		return 1;
	}
	else {
		return 0;
	}
}

// ----------------------------------------------------------------
void st_set_val(symtab_t * pst, int stidx, char buf[MAX_ATOM_SIZE])
{
	st_bounds_check(pst, stidx);
	pst->entries[stidx].has_val = 1;
	memcpy(pst->entries[stidx].atombuf, buf, MAX_ATOM_SIZE);
}

// ----------------------------------------------------------------
void st_dump(symtab_t * pst)
{
	int enti;

	printf("Symbol table:\n");
	printf("  # entries: %d\n", pst->nents);
	printf("  # chars:   %d\n", pst->nchars);
	printf("  &lexemes : %p\n", pst->lexemes);

	for (enti = 0; enti < pst->nents; enti++) {
		printf("  [%2d] tok %3d string <<%s>> at %p\n",
			enti,
			pst->entries[enti].token,
			pst->entries[enti].lexptr,
			pst->entries[enti].lexptr);
	}
	printf("\n");
}
