#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z3polymod.h"
#include "malloc_check.h"
#include "power.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int n;
	z3poly_t r, cpoly, rc;
	int pn, irlo, i, c;
	char * marks;

	if (argc != 2)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &n) != 1)
		usage(argv[0]);

	pn = calc_power(3, n);
	marks = malloc_check(pn * sizeof(char));
	irlo = pn;

	for (i = 0; i < pn; i++) {
		marks[i] = 0;
	}

	for (i = 0; i < pn; i++) {
		if (marks[i])
			continue;
		r = z3poly_from_base_rep(irlo + i);
		if (!z3poly_is_irreducible(&r))
			continue;

		for (c = 3; c < 6; c++) {
			cpoly = z3poly_from_base_rep(c);
			rc = z3poly_compose(&r, &cpoly);
			if (c == 3) {
				z3poly_print(&rc);
			}
			else if (!z3poly_equal(&rc, &r)) {
				printf(" ");
				z3poly_print(&rc);
			}
			marks[z3poly_to_base_rep(&rc) - irlo] = 1;
		}
		printf("\n");
	}

	free(marks);
	return 0;
}
