#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2nipolymod.h"
#include "power.h"
#include "mod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// for each reductor r
//   for each possible generator g
//     if  g^5  mod r == 0:10
//     and g^10 mod r == 0:11
//       wahoo!

// Inputs:
// * Reductor for intermediate field, w/ degree implicit
// * Degree of extension field over the intermediate field
// * Array of exponents & fit points

void seek_min_poly(
	z2ipoly_t    int_r,   // reductor for intermediate field
	int          ext_int_deg, // degree of extension field over intermed
	int          fit_log,
	z2nipoly_t * pfit_val,
	int          verbose)
{
	int int_deg = z2ipoly_get_degree(int_r);
	int tot_deg = int_deg * ext_int_deg;
	int rep_max = 1LL << tot_deg;
	int i, k;
	z2nipoly_t g = z2nipoly_from_coeffs1(1LL, 0LL);
	z2nipoly_t rr;
	z2nipoly_t gpow;

	for (i = 0; i < rep_max; i++) {
		rr = z2nipoly_from_base_rep(i + rep_max, int_deg);

		gpow = z2nipolymod_power(&g, fit_log, &rr, int_r);
		if (!z2nipoly_equal(&gpow, pfit_val, int_r))
			continue;

		printf("rr=");
		z2nipoly_print(&rr, int_r);
		printf(" g=");
		z2nipoly_print(&g, int_r);
		printf("\n");

		if (verbose) {
			for (k = 0; k < rep_max; k++) {
				z2nipoly_t gpow = z2nipolymod_power(
					&g, k, &rr, int_r);
				z2nipoly_print(&g, int_r);
				printf("^%d = ", (int)k);
				z2nipoly_print(&gpow, int_r);
				printf("\n");
			}
			printf("\n");
		}
	}
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s {r} {ext'n/intermed deg} {log} {val} [verbose]\n",
		argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t  int_r = 0xb;
	int        ext_int_deg = 2;
	int        fit_log = 9;
	z2nipoly_t fit_val = z2nipoly_from_coeffs1(0LL, 2LL);
	int        verbose = 0;

	if ((argc != 5) && (argc != 6))
		usage(argv[0]);

	if (!z2ipoly_scan(&int_r, argv[1]))
		usage(argv[0]);
	if (sscanf(argv[2], "%d", &ext_int_deg) != 1)
		usage(argv[0]);
	if (sscanf(argv[3], "%d", &fit_log) != 1)
		usage(argv[0]);
	if (!z2nipoly_scan(&fit_val, argv[4], int_r))
		usage(argv[0]);
	if (argc == 6)
		verbose = 1;

	seek_min_poly(int_r, ext_int_deg, fit_log, &fit_val, verbose);

	return 0;
}
