#include <stdio.h>
#include <stdlib.h>
#include "rat.h"
#include "mod.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {numbers ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int argi;
	int i, prec = 40;
	int p, d;
	int num_fract;
	rat_t q, drat, diff, prat;

	if (argc < 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);

	for (argi = 2; argi < argc; argi++) {

		if (!rat_sscan(argv[argi], &q)) {
			fprintf(stderr,
				"Couldn't parse \"%s\"; skipping.\n",
				argv[argi]);
			continue;
		}

		rat_print(q);
		printf(": ");

		num_fract = 0;
		while ((q.denom % p) == 0) {
			num_fract++;
			q.denom /= p;
		}

		// q = a/b
		// Pick d0 in 0 .. p-1 such that |a/b - d0|_p < 1
		// I.e. p | a/b - d0
		// i.e. d0 equiv a/b (mod p)
		// i.e. d0 = a * b^{-1} mod p
		// q := (q - d0) / p
		// Repeat to get d1, d2, ... until q := 0.

		for (i = 0; i < prec; i++) {
			d = mod_mul(
				q.numer,
				mod_recip(q.denom, p),
				p);

			if (i == num_fract)
				printf(".");
			else if (p > 10)
				printf(",");
			printf("%d", d);

			// q := (q - d) / p
			drat = rat_from_int(d);
			diff = rat_sub(q, drat);
			prat = rat_from_int(p);
			q    = rat_div(diff, prat);

			if (i >= num_fract)
				if (rat_is_zero(q))
					break;
		}

		if (!rat_is_zero(q))
			printf(" ...");

		printf("\n");
	}

	return 0;
}
