#include <stdio.h>
#include <string.h>
#include <iostream.h>
#include <unistd.h>
#include "qpoly.h"
#include "power.h"

// ----------------------------------------------------------------
void read_coeffs(
	int * coeffs,
	int * pdegree)
{
	char line[256];
	int scanrc;

	if (fgets(line, sizeof(line), stdin) == 0) {
		fprintf(stderr, "read_coeffs(): fgets null.\n");
		exit(1);
	}

	// xxx need to tokenize, then scan "%d/%d", else scan "%d".
	scanrc = sscanf(line, "%d %d %d %d %d %d %d %d %d %d %d %d %d",
		&coeffs[0], &coeffs[1], &coeffs[2],  &coeffs[3],
		&coeffs[4], &coeffs[5], &coeffs[6],  &coeffs[7],
		&coeffs[8], &coeffs[9], &coeffs[10], &coeffs[11]);

	if (scanrc < 1) {
		fprintf(stderr, "read_coeffs(): need >= 2 ints.\n");
		exit(1);
	}
	*pdegree = scanrc - 1;
}

// ----------------------------------------------------------------
void get_poly(
	char * name,
	qpoly & poly)
{
	int icoeffs[20];
	rat rcoeffs[20];
	int degree;
	int i;

	if (isatty(fileno(stdin))) {
		printf("Coeffs for %s: ", name);
		fflush(stdout);
	}
	read_coeffs(icoeffs, &degree);
	for (i = 0; i <= degree; i++)
		rcoeffs[i] = icoeffs[i];
	poly = qpoly(degree, rcoeffs);
}

// ----------------------------------------------------------------
void test_ops(void)
{
	qpoly a, b, r;

	get_poly("a", a);
	get_poly("b", b);
	get_poly("r", r);

	qpoly sum = a + b;
	cout << a << " + " << b << " = " << sum << endl;

	qpoly neg = -a;
	cout << "- " << a << " = " << neg << endl;

	qpoly diff = a - b;
	cout << a << " - " << b << " = " << diff << endl;

	qpoly prod = a.mul(b, r);
	cout << a << " * " << b << " = " << prod << " mod " << r << endl;

	qpoly quot = a.div(b, r);
	cout << a << " / " << b << " = " << quot << " mod " << r << endl;

	// a/b = c  =>  b*c = a
	qpoly check = b.mul(quot, r);
	cout << "check: " << check << endl;
	if (check == a)
		cout << "OK" << endl;
	else {
		cout << "Error!!!!" << endl;
		exit(1);
	}
}

//// ----------------------------------------------------------------
//void show_Z3_3_mul_tables(void)
//{
//	int p = 2;
//	int m = 3;
//	int pm = calc_power(p, m);
//	int i, j;
//	qpoly r1(p, 1, 1, 0, 1);
//	qpoly r2(p, 1, 0, 1, 1);
//
//	for (i = 1; i < pm; i++) {
//		for (j = 1; j < pm; j++) {
//			qpoly a  = qpoly::from_int(p, i);
//			qpoly b  = qpoly::from_int(p, j);
//			qpoly c1 = a.mul(b, r1);
//			cout << " " << c1;
//		}
//		cout << endl;
//	}
//	cout << endl;
//
//	for (i = 1; i < pm; i++) {
//		for (j = 1; j < pm; j++) {
//			qpoly a  = qpoly::from_int(p, i);
//			qpoly b  = qpoly::from_int(p, j);
//			qpoly c2 = a.mul(b, r2);
//			cout << " " << c2;
//		}
//		cout << endl;
//	}
//	cout << endl;
//}
//
//// ----------------------------------------------------------------
//void show_order(void)
//{
//	qpoly a, r;
//
//	get_poly("a", a);
//	get_poly("r", r);
//
//	int p = a.get_modulus();
//
//	qpoly one = qpoly(p, 1);
//	qpoly zero = qpoly(p, 0);
//	qpoly apower = a;
//	int order = 1;
//
//	while (apower != one) {
//		if (apower == zero) {
//			cout << "ZERO!" << endl;
//			exit(1);
//		}
//		cout << a << " ^ " << order << " = " << apower << endl;
//		apower = apower.mul(a, r);
//		order++;
//	}
//	cout << a << " ^ " << order << " = " << apower << endl;
//
//	cout << "order = " << order << endl;
//}
//
//// ----------------------------------------------------------------
//// Note:  The k-torsion subgroup of Z_p^m always seems to have
//// k elements.  E.g.:
//// Z_5^3:
////
//// Order   #elts w/ord    #elts in k-torsion subgroup
//// -----  ------------    ---------------------
////     1       1           1
////     2       1           1+1 = 2
////     4       2           1+1+2 = 4
////    31      30           30+1  = 31
////    62      30           30+30+1+1 = 62
////   124      60           124
//
//void show_orders(void)
//{
//	qpoly r;
//
//	get_poly("r", r);
//
//	int m = r.get_degree();
//	int p = r.get_modulus();
//	int pm = calc_power(p, m);
//	qpoly one = qpoly(p, 1);
//	qpoly zero = qpoly(p, 0);
//	int i;
//
//	for (i = 1; i < pm; i++) {
//		qpoly a = qpoly::from_int(p, i);
//
//		qpoly apower = a;
//		int order = 1;
//
//		while (apower != one) {
//			if (apower == zero) {
//				cout << "ZERO!" << endl;
//				exit(1);
//			}
//			apower = apower.mul(a, r);
//			order++;
//		}
//
//		cout << "|" << a << "| = " << order << endl;
//	}
//}
//
//// ----------------------------------------------------------------
//void test_phy_reciprocation(int argc, char ** argv)
//{
//	qpoly r;
//	int i;
//	int nerr = 0;
//	int ngood = 0;
//
//	get_poly("r", r);
//
//	int m = r.get_degree();
//	int p = r.get_modulus();
//	int pm = calc_power(p, m);
//	int ok;
//
//	for (i = 1; i < pm; i++) {
//		qpoly a = qpoly::from_int(p, i);
//		qpoly ainv1 = a.recip(r);
//		qpoly ainv2 = a.phi_recip(r);
//
//		ok = (ainv1 == ainv2);
//
//		if (argc > 1) {
//			cout
//				<< "a = " << a << "\t"
//				<< "gcd recip = " << ainv1 << "\t"
//				<< "phi recip = " << ainv2 << "\t";
//			if (ok)
//				cout << "OK";
//			else
//				cout << "Error!!!";
//			cout << endl;
//		}
//
//		if ((i & 127) == 0)
//			cout << "." << flush;
//
//		if (ok)
//			ngood++;
//		else
//			nerr++;
//
//	}
//	cout << endl;
//	cout << "ngood: " << ngood << endl;
//	cout << "nerr:  " << nerr  << endl;
//}
//
//// ----------------------------------------------------------------
//void time_phy_reciprocation(int argc, char ** argv)
//{
//	qpoly r;
//	int i;
//	int argi;
//	int verbose = 0;
//	int do_ext = 1;
//
//	for (argi = 1; argi < argc; argi++) {
//		if (strcmp(argv[argi], "-v") == 0)
//			verbose = 1;
//		else if (strcmp(argv[argi], "-e") == 0)
//			do_ext = 1;
//		else if (strcmp(argv[argi], "-p") == 0)
//			do_ext = 0;
//	}
//
//	get_poly("r", r);
//
//	int m = r.get_degree();
//	int p = r.get_modulus();
//	int pm = calc_power(p, m);
//
//	if (do_ext) {
//		cout << "Ext GCD reciprocation" << endl;
//		for (i = 1; i < pm; i++) {
//			qpoly a = qpoly::from_int(p, i);
//			qpoly ainv1 = a.recip(r);
//			if (verbose)
//				cout << a << "^-1 = " << ainv1
//					<< " : " << a.mul(ainv1, r)
//					<< endl;
//		}
//	}
//	else {
//		cout << "Phi reciprocation" << endl;
//		for (i = 1; i < pm; i++) {
//			qpoly a = qpoly::from_int(p, i);
//			qpoly ainv2 = a.phi_recip(r);
//			if (verbose)
//				cout << a << "^-1 = " << ainv2 << endl;
//		}
//	}
//}
//
//// ----------------------------------------------------------------
//void list_irreds(int argc, char ** argv)
//{
//	int p = 2;
//	int m = 3;
//	int lo, hi;
//	int i;
//	qpoly r;
//
//	if (argc == 3) {
//		(void)sscanf(argv[1], "%d", &p);
//		(void)sscanf(argv[2], "%d", &m);
//	}
//	lo = calc_power(p, m);
//	hi = lo + lo - 1;
//	for (i = lo; i <= hi; i++) {
//		r = qpoly::from_int(p, i);
//		// xxx if r.is_irreducible() ...
//		cout << r << endl;
//	}
//}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	test_ops();
	//test_int_conv();
	//show_Z3_3_mul_tables();
	//show_order();
	//show_orders();
	//test_phy_reciprocation(argc, argv);
	//time_phy_reciprocation(argc, argv);
	//list_irreds(argc, argv);

	return 0;
}
