#include <stdio.h>
#include <stdlib.h>
#include "qpoly.h"
#include "power.h"
#include "malloc_check.h"

#define NEG_ONE_POW(n) ((n) & 1 ? -1 : 1)

#define DEBUG 0

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {poly1} {poly2}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
rat_t qpoly_resultant(
	qpoly_t * pf,
	qpoly_t * pg)
{
	qpoly_t f, g, quot, rem, * pswap;
	rat_t res = rat_from_int(1);
	int degfr;
	rat_t bm, bmpow;

	if (pf->degree < pg->degree) {
		pswap = pf;
		pf    = pg;
		pg    = pswap;
	}
	f = *pf;
	g = *pg;

	for (;;) {

		if (g.degree == 0) {
			// R(f, c) = c^n
			bm = g.coeffs[0];
			bmpow = rat_power(bm, f.degree);
			res = rat_mul(res, bmpow);

#if DEBUG
		printf("\n");
		printf("@f="); qpoly_print(&f); printf("\n");
		printf("g="); qpoly_print(&g); printf("\n");
		printf("bmp="); rat_print(bmpow); printf("\n");
		printf("res="); rat_print(res); printf("\n");
#endif
			return res;
		}

		// R(f,g) = (-1)^nm R(g,f)
		// R(g,f) = b_m^{deg f - deg r} R(g, r), where f = qg + r
		// so,
		// R(f,g) = (-1)^nm b_m^{deg f - deg r} R(g, r)

		quot = qpoly_div(&f, &g);
		rem  = qpoly_mod(&f, &g);

		if ((f.degree * g.degree) & 1)
			res = rat_mul(res, rat_from_int(-1));
		degfr = f.degree - rem.degree;
		if (degfr < 0) {
			fprintf(stderr, "ZPRES ABEND!\n");
			exit(1);
		}
		bm = g.coeffs[g.degree];
		bmpow = rat_power(bm, degfr);
		res = rat_mul(res, bmpow);

#if DEBUG
		printf("\n");
		printf("f="); qpoly_print(&f); printf("\n");
		printf("g="); qpoly_print(&g); printf("\n");
		printf("q="); qpoly_print(&quot); printf("\n");
		printf("r="); qpoly_print(&rem); printf("\n");
		printf("bmp="); rat_print(bmpow); printf("\n");
		printf("res="); rat_print(res); printf("\n");
#endif

		f = g;
		g = rem;
	}
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	qpoly_t f, g;
	rat_t res;

	if (argc != 3)
		usage(argv[0]);
	if (!qpoly_scan(&f, argv[1]))
		usage(argv[0]);
	if (!qpoly_scan(&g, argv[2]))
		usage(argv[0]);

	res = qpoly_resultant(&f, &g);
	rat_print(res);
	printf("\n");

	return 0;
}
