// Parameters:
// p, m, n; nreps
//
//int bins[MAX_PM];
//init bins to 0
//for (i = 0; i < nreps; i++) {
//	make rand nxn matrix over p^m
//	det = calc_det
//	base_rep = zppoly_to_base_rep(det);
//	bins[base_rep]++;
//}

#include <stdio.h>
#include <stdlib.h>
#include "zppoly.h"
#include "vmatrix.h"
#include "power.h"
#include "posmod.h"

#define MAX_PM 100000
#define MAX_M    30
#define MAX_N    10

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// ----------------------------------------------------------------
#define BUFSZ 10000
static int get_rand_int(void)
{
	static int fd = -1;
	static int buf[BUFSZ];
	static int bufpos = 0;
	int rv;
	if (fd == -1) {
		fd = open("/dev/urandom", O_RDONLY);
		if (fd < 0) {
			fprintf(stderr, "Couldn't open /dev/urandom.\n");
			perror("Reason");
			exit(1);
		}
	}

	if (bufpos == 0) {
		if (read(fd, buf, BUFSZ * sizeof(int)) < 0) {
			fprintf(stderr, "Couldn't read /dev/urandom.\n");
			perror("Reason");
			exit(1);
		}
	}
	rv = buf[bufpos];
	bufpos++;
	if (bufpos >= BUFSZ)
		bufpos = 0;
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int argi;
	int p = 11;//3;
	int m = 1;//3;
	int n = 4;
	int rbr = -1;//34;
	int max_reps = 1000;
	int coeffs[MAX_M+1];
	int bins[MAX_PM];
	zppoly_aux_t aux;
	zpoly_t mat[MAX_N][MAX_N];
	zpoly_t det;
	int verbose = 0;
	int pm;
	int rep;
	int br;
	int list_irrs = 0;
	int i, j, k;
	int do_print_status = isatty(fileno(stderr));

	for (argi = 1; argi < argc; argi++) {
		if (sscanf(argv[argi], "p=%d", &p) == 1)
			;
		else if (sscanf(argv[argi], "m=%d", &m) == 1)
			;
		else if (sscanf(argv[argi], "n=%d", &n) == 1)
			;
		else if (sscanf(argv[argi], "reps=%d", &max_reps) == 1)
			;
		else if (sscanf(argv[argi], "rbr=%i", &rbr) == 1)
			;
		else if (sscanf(argv[argi], "v=%d", &verbose) == 1)
			;
		else if (sscanf(argv[argi], "li=%d", &list_irrs) == 1)
			;
		else {
			fprintf(stderr, "Huh?\n");
			exit(1);
		}
	}

	pm = calc_power(p, m);
	if (pm > MAX_PM) {
		fprintf(stderr, "p^m too big.\n");
		exit(1);
	}
	if (n > MAX_N) {
		fprintf(stderr, "n too big.\n");
		exit(1);
	}
	if (m > MAX_M) {
		fprintf(stderr, "m too big.\n");
		exit(1);
	}

	// aux = find irred(p, m)
	if (rbr == -1) {
		aux = zppoly_find_irreducible(p, m);
		printf("Using reductor ");
		zppoly_vfops.pvprint(&aux, &aux);
		printf("\n");

		if (list_irrs)
			zppoly_list_irreducibles(p, m);
	}
	else {
		aux = zppoly_aux_from_base_rep(p, m, rbr);
		if (!zppoly_is_irreducible(&aux.r, p, m)) {
			fprintf(stderr, "aux is reducible.\n");
			exit(1);
		}
	}

	if (max_reps == 0)
		exit(0);

	for (i = 0; i < pm; i++)
		bins[i] = 0;

	for (rep = 0; rep < max_reps; rep++) {
		// Make a random nxn matrix over p^m
		for (i = 0; i < n; i++) {
			for (j = 0; j < n; j++) {
#if 0
				for (k = 0; k < m; k++) {
					static int foo = 0xe9c61201;
					coeffs[k] =
						//xxx rand % p;
						posmod(foo >> 21, p);
					foo *= 0xdeadbeef;
				}
#else
				for (k = 0; k < m; k++) {
					coeffs[k] = get_rand_int();
					coeffs[k] = posmod(coeffs[k], p);
				}
#endif
				mat[i][j] = zppoly_from_coeffs(p, m-1, coeffs);
			}
		}
		if (verbose)
			vmatrix_print(mat, &aux, &zppoly_vfops, n, n, MAX_N);

		vmatrix_det(mat, &det, &aux, &zppoly_vfops, n, MAX_N);

		if (verbose) {
			printf("det = ");
			zppoly_vfops.pvprint(&det, &aux);
			printf("\n");
		}

		br = zppoly_to_base_rep(&det, p);

		if (verbose)
			printf("%d\n", br);

		bins[br]++;

		if (do_print_status) {
			if ((rep & 0xff) == 0xff) {
				fprintf(stderr, "\r%5.2lf%%",
					(double)(100.0 * rep) / max_reps);
				fflush(stderr);
			}
		}
	}

	if (do_print_status) {
		fprintf(stderr, "\r        \r\n");
		fflush(stderr);
	}
	//for (i = 0; i < pm; i++)
		//printf("%4d: %6d\n", i, bins[i]);
	for (i = 0; i < pm; i++) {
		printf("%3d:%3d", i, bins[i]);
		if ((i & 7) == 7)
			printf("\n");
		else
			printf("  ");
	}
	printf("\n");

	return 0;
}
