#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "power.h"
#include "isprime.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {n} [rk]\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void print_compositions(
	zpoly_t * pr,
	zpoly_t * prk,
	int       p,
	int       n,
	int       do_mod,
	int       show_ip,
	int       izlo,
	int       izhi)
{
	int iz;
	zpoly_t z, rz;

	if (show_ip) {
		if (zppoly_is_irreducible(pr, p)) {
			if (zppoly_is_primitive(pr, p, n)) {
				printf("p");
			}
			else {
				printf("i");
			}
		}
		else {
			printf("-");
		}
	}

	printf(" ");
	zppoly_print(pr, p);
	for (iz = izlo; iz <= izhi; iz++) {
		z = zppoly_from_base_rep(p, iz);
		if (do_mod)
			rz = zppolymod_compose(pr, &z, prk, p);
		else
			rz = zppoly_compose(pr, &z, p);
		printf(" ");
		zppoly_print(&rz, p);
	}
	printf("\n");
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zpoly_t ri, z, xc;
	zpoly_t * rjs;
	zpoly_t rk = zppoly_from_base_rep(1, 0);
	unsigned char * flags;
	int iri, irlo, irhi, iz, izlo, izhi;
	int do_mod = 0;
	int show_ip = 0;
	int p, n, c, idx;
	int npolys;
	int num_self_compl = 0;

	if (argc < 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	if (sscanf(argv[2], "%d", &n) != 1)
		usage(argv[0]);
	if (argc > 3) {
		if (!zppoly_scan(&rk, argv[3], p))
			usage(argv[0]);
		do_mod = 1;
	}
	if (isprime(p))
		show_ip = 1;
	else
		show_ip = 0;

	irlo = calc_power(p, n);
	//irhi = 2 * irlo - 1;
	irhi = p * irlo - 1;
	npolys = irhi - irlo + 1;
	izlo = 0;
	//izhi = irlo - 1;
	izhi = 4 * p;

	if (show_ip)
		printf(". . ");
	else
		printf(". ");
	for (iz = izlo; iz <= izhi; iz++) {
		printf(" ");
		z = zppoly_from_base_rep(p, iz);
		zppoly_print(&z, p);
	}
	printf("\n");
	printf("\n");

	flags = malloc_check(npolys);
	rjs = malloc_check(p * sizeof(zpoly_t));
	for (iri = irlo; iri <= irhi; iri++) {
		flags[iri-irlo] = 0;
	}

	for (iri = irlo; iri <= irhi; iri++) {
		if (flags[iri-irlo])
			continue;

		ri = zppoly_from_base_rep(p, iri);
		for (c = 0; c < p; c++) {
			xc = zppoly_from_coeffs1(1, c, p);
			rjs[c] = zppoly_compose(&ri, &xc, p);
		}

		// This is an insufficient check when p is composite.
		if (zppoly_equal(&rjs[0], &rjs[1], p)) {
			num_self_compl++;
		}

		for (c = 0; c < p; c++) {
			idx = zppoly_to_base_rep(&rjs[c], p) - irlo;
			if (flags[idx])
				continue;
			print_compositions(&rjs[c], &rk, p, n, do_mod, show_ip,
				izlo, izhi);
			flags[idx] = 1;
		}
		printf("\n");

	}
	printf("\n");
	printf("# self-complementary: %d\n", num_self_compl);

	free(flags);
	free(rjs);

	return 0;
}
