#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "isprime.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n} [rk]\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void print_compositions(
	z2ipoly_t r,
	z2ipoly_t rk,
	int       do_mod,
	z2ipoly_t clo,
	z2ipoly_t chi)
{
	z2ipoly_t c, rc;

	if (z2ipoly_is_irreducible(r)) {
		if (z2ipoly_is_primitive(r)) {
			printf("p");
		}
		else {
			printf("i");
		}
	}
	else {
		printf("-");
	}

	printf(" ");
	z2ipoly_print(r);
	for (c = clo; c <= chi; c++) {
		if (do_mod)
			rc = z2ipolymod_compose(r, c, rk);
		else
			rc = z2ipoly_compose(r, c);
		printf(" ");
		z2ipoly_print(rc);
	}
	printf("\n");
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t ri, rj, c, rk;
	z2ipoly_t rlo, rhi, clo, chi;
	int do_mod = 0;
	int n;

	if (argc < 2)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &n) != 1)
		usage(argv[0]);
	if (argc > 2) {
		if (!z2ipoly_scan(&rk, argv[2]))
			usage(argv[0]);
		do_mod = 1;
	}

	rlo = 1LL << n;
	rhi = (rlo << 1) - 1;
	clo = 0LL;
	chi = rlo - 1;

	printf(". . ");
	for (c = clo; c <= chi; c++) {
		printf(" ");
		z2ipolymod_print(c, rlo);
	}
	printf("\n");
	printf("\n");

	for (ri = rlo; ri <= rhi; ri++) {
		rj = z2ipoly_compose(ri, 3LL);
		if (rj < ri)
			continue;

		print_compositions(ri, rk, do_mod, clo, chi);
		if (ri != rj)
			print_compositions(rj, rk, do_mod, clo, chi);
		printf("\n");

	}

	return 0;
}
