#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z3polymod.h"
#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n} [rk]\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
static void print_compositions(
	z3poly_t * pr,
	z3poly_t * prk,
	int       n,
	int       do_mod,
	int       izlo,
	int       izhi)
{
	int iz;
	z3poly_t z, rz;

	if (z3poly_is_irreducible(pr)) {
		if (z3poly_is_primitive(pr, n)) {
			printf("p");
		}
		else {
			printf("i");
		}
	}
	else {
		printf("-");
	}

	printf(" ");
	z3poly_print(pr);
	for (iz = izlo; iz <= izhi; iz++) {
		z = z3poly_from_base_rep(iz);
		if (do_mod)
			rz = z3polymod_compose(pr, &z, prk);
		else
			rz = z3poly_compose(pr, &z);
		printf(" ");
		z3poly_print(&rz);
	}
	printf("\n");
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z3poly_t ri, rj, z, xc;
	z3poly_t rk = z3poly_from_base_rep(0);
	unsigned char * flags;
	int iri, irlo, irhi, iz, izlo, izhi;
	int do_mod = 0;
	int n, c;
	int npolys;
	int num_self_compl = 0;

	if (argc < 2)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &n) != 1)
		usage(argv[0]);
	if (argc > 2) {
		if (!z3poly_scan(&rk, argv[2]))
			usage(argv[0]);
		do_mod = 1;
	}

	irlo = calc_power(3, n);
	irhi = 2 * irlo - 1;
	npolys = irhi - irlo + 1;
	izlo = 0;
	//izhi = irlo - 1;
	izhi = 20;

	printf(". . ");
	for (iz = izlo; iz <= izhi; iz++) {
		printf(" ");
		z = z3poly_from_base_rep(iz);
		z3poly_print(&z);
	}
	printf("\n");
	printf("\n");

	flags = malloc_check(npolys);
	for (iri = irlo; iri <= irhi; iri++) {
		flags[iri-irlo] = 0;
	}

	for (iri = irlo; iri <= irhi; iri++) {
		if (flags[iri-irlo])
			continue;

		ri = z3poly_from_base_rep(iri);
		for (c = 0; c < 3; c++) {
			xc = z3poly_from_coeffs1(1, c);
			rj = z3poly_compose(&ri, &xc);
			if (z3poly_equal(&ri, &rj)) {
				if (c > 0) {
					if (c == 1)
						num_self_compl++;
					continue;
				}
			}
			flags[z3poly_to_base_rep(&rj)-irlo] = 1;
			print_compositions(&rj, &rk, n, do_mod, izlo, izhi);
		}
		printf("\n");

	}
	printf("\n");
	printf("# self-complementary: %d\n", num_self_compl);

	free(flags);

	return 0;
}
