// alloc src marks
// lfe
// sort by order
// for each element in src array
//   if marked, continue
//   for frobit
//     push to dst array

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2ipoly_factor.h"
#include "z2nipoly_factor.h"
#include "power.h"
#include "malloc_check.h"

typedef struct _z2ipair_t {
	z2ipoly_t z;
	z2ipoly_t r;
} z2ipair_t;

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {irrs ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t * irrs;
	z2ipoly_t ** fes;
	z2ipoly_t rjai;
	int argi, i, j, k, nirrs, n, nel = 0;
	int fi, ci;
	z2ipoly_factor_info_t finfo;

	if (argc < 2)
		usage(argv[0]);

	nirrs = argc - 1;
	irrs = malloc_check(nirrs * sizeof(z2ipoly_t));
	fes  = malloc_check(nirrs * sizeof(z2ipoly_t *));

	for (argi = 1, j = 0; argi < argc; argi++, j++) {
		if (!z2ipoly_scan(&irrs[j], argv[argi]))
			usage(argv[0]);
	}

	n = z2ipoly_get_degree(irrs[0]);
	nel = 1 << n;
	for (j = 1; j < nirrs; j++) {
		int nother = z2ipoly_get_degree(irrs[j]);
		if (nother != n) {
			printf("Irrs must all be of the same degree.\n");
			printf("Got %llx (deg=%d) and %llx (deg=%d).\n",
				irrs[0], n, irrs[j], nother);
			exit(1);
		}
	}

	for (j = 0; j < nirrs; j++)
		fes[j] = z2ipolymod_get_elements_frobord(irrs[j], &nel);

	for (i = 0; i < nel; i++) {
		for (j = 0; j < nirrs; j++) {
			// print ai
			z2ipolymod_print(fes[j][i], irrs[j]);
			printf(": ");

			// print rj[ai]
			rjai = z2ipoly_compose(irrs[j], fes[j][i]);
			z2ipoly_print(rjai);
			printf(" ");

			// print factors of ri[aj]
			printf(" = ");
			z2ipoly_factor(rjai, &finfo);
			for (fi = 0; fi < finfo.num_distinct; fi++) {
			for (ci = 0; ci < finfo.factor_counts[fi].count; ci++) {
				if ((fi > 0) || (ci > 0))
					printf(",");
				z2ipoly_print(finfo.factor_counts[fi].f);
			}
			}

			// print rj[ai] mod each rk
			for (k = 0; k < nirrs; k++) {
				z2ipoly_t m = z2ipoly_mod(rjai, irrs[k]);
				printf(" ");
				z2ipolymod_print(m, irrs[k]);
			}
			if (j < (nirrs - 1))
				printf(" | ");
		}
		printf("\n");
	}

	for (j = 0; j < nirrs; j++)
		free(fes[j]);
	free(fes);
	free(irrs);
	return 0;
}

//0: 1    = 1       1  |  0: 1    = 1         1  |  0: 1    = 1         1
//1: 1    = 1       1  |  1: 1    = 1         1  |  1: 1    = 1         1
//
//7: 117  = 19 1f   1  |  b: 12f7 = 13 11d    b  |  c: 129d = 13 11b    d
//6: 117  = 19 1f   1  |  a: 12b9 = 12b9      a  |  d: 12c1 = 12c1      c
//
//4: 105  = 13^2    0  |  d: 129d = 13 11b    5  |  b: 12b9 = 12b9      8
//3: 13   = 13      0  |  7: 17b  = 17b       8  |  7: 169  = 169       2
//5: 105  = 13^2    0  |  c: 12c1 = 12c1      f  |  a: 12f7 = 13 11d    4
//2: 13   = 13      0  |  6: 169  = 169       3  |  6: 17b  = 17b       f
//
//d: 110d = 19 1f5  7  |  4: 141  = 19^2      0  |  3: 19   = 19        6
//e: 111f = 111f    6  |  9: 1249 = 13 19 1f  0  |  5: 141  = 19^2      b
//b: 101b = 1f 18d  7  |  e: 1249 = 13 19 1f  0  |  e: 1213 = 1213      7
//9: 1009 = 1009    6  |  2: 19   = 19        0  |  9: 1201 = 1201      a
//
//c: 110d = 19 1f5  7  |  5: 155  = 1f^2      d  |  8: 1249  = 13 19 1f 0
//f: 111f = 111f    6  |  8: 1201 = 1201      7  |  2: 1f    = 1f       0
//a: 101b = 1f 18d  7  |  f: 1213 = 1213      c  |  4: 155   = 1f^2     0
//8: 1009 = 1009    6  |  3: 1f   = 1f        6  |  f: 1249  = 13 19 1f 0
