#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"
#include "ring.h"
#include "ring_check.h"
#include "ring_centralizer.h"
#include "malloc_check.h"

#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000

static unsigned char * aux_buf;
static unsigned char * elts_buf;
static unsigned char * sgr_buf;
static unsigned char * centralizer_buf;

// ----------------------------------------------------------------
int centralizer_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t ring;
	ring_t subring;
	ring_t centralizer;

	aux_buf         = (unsigned char *)malloc_check(MAX_AUX_BYTES);
	elts_buf        = (unsigned char *)malloc_check(MAX_RING_BYTES);
	sgr_buf         = (unsigned char *)malloc_check(MAX_RING_BYTES);
	centralizer_buf = (unsigned char *)malloc_check(MAX_RING_BYTES);

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 0;
	ospec.num_file_names    = 2;
	ospec.cmd_param_desc    = 0;

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (opts.file_namec < 1) {
		fprintf(stderr, "Need 1 or 2 file names.\n");
		exit(1);
	}

	if (!make_ring(&subring, opts.pgspec, aux_buf,
		opts.file_namev[0], sgr_buf, MAX_RING_BYTES))
	{
		fprintf(stderr, "Couldn't make subring in \"%s\".\n",
			opts.ring_spec_desc);
		exit(1);
	}

	if (!make_ring(&ring, opts.pgspec, aux_buf,
		opts.file_namec == 1 ? 0 : opts.file_namev[1],
		elts_buf, MAX_RING_BYTES))
	{
		fprintf(stderr, "Couldn't make ring in \"%s\".\n",
			opts.ring_spec_desc);
		exit(1);
	}


	if (opts.echo_inputs) {
		describe_ring(&ring);
		print_ring_elements(&subring);
		printf("\n");
		print_ring_elements(&ring);
		printf("\n");
	}

	make_centralizer(&ring, &subring, &centralizer,
		centralizer_buf, MAX_RING_BYTES);

	if (opts.echo_inputs) {
		printf("\n");
		printf("Centralizer:\n");
	}
	print_ring_elements(&centralizer);

	free(aux_buf);
	free(elts_buf);
	free(sgr_buf);
	free(centralizer_buf);

	return 0;
}
