#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"

#include "ring.h"
#include "ring_check.h"
#include "ring_commutator.h"

#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_RING_BYTES];
static unsigned char sgr_buf[MAX_RING_BYTES];

// ----------------------------------------------------------------
int commsgr_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t ring;
	ring_t commutator_subring;
	int i, klo, khi;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need commutator depth.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "commutator depth";

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (sscanf(opts.cmd_paramv[0], "%d-%d", &klo, &khi) == 2) {
	}
	else if (sscanf(opts.cmd_paramv[0], "%d", &klo) == 1) {
		khi = klo;
	}
	else {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			opts.cmd_paramv[1]);
		exit(1);
	}

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (!make_ring(&ring, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_RING_BYTES))
	{
		fprintf(stderr, "Couldn't make ring \"%s\".\n",
			opts.ring_spec_desc);
		exit(1);
	}

	if (opts.echo_inputs) {
		describe_ring(&ring);
		print_ring_elements(&ring);
	}

	commutator_subring.pvelements = sgr_buf;
	commutator_subring.pvaux      = ring.pvaux;
	commutator_subring.pspec      = ring.pspec;

	commutator_subring.num_elements = ring.num_elements;
	memcpy(commutator_subring.pvelements, ring.pvelements,
		ring.num_elements * ring.pspec->element_size);

	for (i = 0; i <= khi; i++) {

		if (i >= klo) {
			if (klo != khi) {
				printf("\n");
				printf("%d-commutator subring:\n", i);
			}
			print_ring_elements(&commutator_subring);
		}

		if (!make_commutator_subring(&ring, &commutator_subring,
			sgr_buf, MAX_RING_BYTES, opts.verbosity))
		{
			fprintf(stderr,
				"%s:  Couldn't make %d-commutator subring.\n",
				argv[0], i);
			exit(1);
		}

		ring.num_elements = commutator_subring.num_elements;
		memcpy(ring.pvelements, commutator_subring.pvelements,
			commutator_subring.num_elements
			* ring.pspec->element_size);

	}

	return 0;
}
