#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"

#include "ring.h"
#include "ring_check.h"
#include "ring_conj_elt.h"

#define MAX_GEN_BYTES 2048
#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000
static unsigned char elt_buf[MAX_GEN_BYTES];
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_RING_BYTES];
static unsigned char sv_buf[MAX_RING_BYTES];

// ----------------------------------------------------------------
int conjelt_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t ring;
	int i;
	void * pa, * pb;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need g as conjugator.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "element";

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (!opts.pgspec->pvscan(elt_buf, opts.cmd_paramv[0], aux_buf,
		sizeof(elt_buf)))
	{
		fprintf(stderr, "Couldn't scan \"%s\" as element.\n",
			opts.cmd_paramv[0]);
		exit(1);
	}

	if (!make_ring(&ring, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_RING_BYTES))
	{
		fprintf(stderr, "Couldn't make ring \"%s\".\n",
			opts.ring_spec_desc);
		exit(1);
	}

	memcpy(sv_buf, elts_buf,
		ring.num_elements * ring.pspec->element_size);
	conjugate_by_element(&ring, &elt_buf);

	for (i = 0; i < ring.num_elements; i++) {
		pa = VARRAY_ELEMENT(sv_buf,   i, ring.pspec->element_size);
		pb = VARRAY_ELEMENT(elts_buf, i, ring.pspec->element_size);
		printf("  ");
		ring.pspec->pvprint(pa, ring.pvaux);
		printf("  ");
		ring.pspec->pvprint(pb, ring.pvaux);
		printf("\n");
	}
	printf("\n");

	return 0;
}
