#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"

#include "ring.h"
#include "ring_check.h"
#include "ring_file.h"
#include "ring_core.h"

#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000
static unsigned char aux_buf [MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_RING_BYTES];
static unsigned char sgr_buf [MAX_RING_BYTES];
static unsigned char core_buf[MAX_RING_BYTES];

// ----------------------------------------------------------------
int core_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t ring;
	ring_t subring;
	ring_t core;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 0;
	ospec.num_file_names    = 2;
	ospec.cmd_param_desc    = 0;

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (opts.file_namec < 1) {
		fprintf(stderr, "Need 1 or 2 file names.\n");
		exit(1);
	}

	if (opts.file_namec == 1) {
		if (!make_ring(&ring, opts.pgspec, aux_buf,
			0, sgr_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make ring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
		if (!make_ring(&subring, opts.pgspec, aux_buf,
			opts.file_namev[0], elts_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make subring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
	}
	else {
		if (!make_ring(&ring, opts.pgspec, aux_buf,
			opts.file_namev[0], sgr_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make ring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
		if (!make_ring(&subring, opts.pgspec, aux_buf,
			opts.file_namev[1], elts_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make subring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
	}

	if (opts.echo_inputs) {
		print_ring_elements(&subring);
		printf("\n");
		print_ring_elements(&ring);
		printf("\n");
	}

	make_core(&ring, &subring, &core, core_buf, MAX_RING_BYTES,
		opts.verbosity);
	print_ring_elements(&core);

	return 0;
}
