#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"

#include "ring.h"
#include "ring_check.h"
#include "ring_file.h"
#include "ring_cosets.h"

#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000
static unsigned char aux_buf [MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_RING_BYTES];
static unsigned char sgr_buf [MAX_RING_BYTES];
static unsigned char quot_buf[MAX_RING_BYTES];

// ----------------------------------------------------------------
int cosets_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t subring;
	ring_t ring;
	ring_cosets_t cosets;
	int direction = LEFT_COSETS;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // "left" or "right"
	ospec.num_file_names    = 2;
	ospec.cmd_param_desc    = "left or right";

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (opts.cmd_paramv[0][0] == 'l') {
		direction = LEFT_COSETS;
	}
	else if (opts.cmd_paramv[0][0] == 'r') {
		direction = RIGHT_COSETS;
	}
	else {
		fprintf(stderr, "Argument \"%s\" is neither \"l\" or \"r\".\n",
			opts.cmd_paramv[0]);
		exit(1);
	}

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (opts.file_namec < 1) {
		fprintf(stderr, "Need 1 or 2 file names.\n");
		exit(1);
	}

	if (opts.file_namec == 1) {
		if (!make_ring(&ring, opts.pgspec, aux_buf,
			0, sgr_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make ring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
		if (!make_ring(&subring, opts.pgspec, aux_buf,
			opts.file_namev[0], elts_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make subring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
	}
	else {
		if (!make_ring(&ring, opts.pgspec, aux_buf,
			opts.file_namev[0], sgr_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make ring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
		if (!make_ring(&subring, opts.pgspec, aux_buf,
			opts.file_namev[1], elts_buf, MAX_RING_BYTES))
		{
			fprintf(stderr, "Couldn't make subring in \"%s\".\n",
				opts.ring_spec_desc);
			exit(1);
		}
	}

	if (opts.echo_inputs) {
		print_ring_elements(&subring);
		printf("\n");
		print_ring_elements(&ring);
		printf("\n");
	}

	if (!make_cosets(&ring, &subring, direction, &cosets,
		quot_buf, MAX_RING_BYTES, opts.verbosity))
	{
		fprintf(stderr, "Couldn't make cosets.\n");
		exit(1);
	}
	ring_print_cosets(&cosets);

	return 0;
}
