#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"

#include "ring.h"
#include "ring_check.h"
#include "ring_powers.h"

#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_RING_BYTES];

// ----------------------------------------------------------------
int powers_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t ring;
	int max_power;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need max power to raise to.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "max power";

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!sscanf(opts.cmd_paramv[0], "%d", &max_power)) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			opts.cmd_paramv[1]);
		exit(1);
	}
	if (max_power < 0) {
		fprintf(stderr, "Max power must be >= 0; got %d.\n",
			max_power);
		exit(1);
	}

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (!make_ring(&ring, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_RING_BYTES))
	{
		fprintf(stderr, "Couldn't make ring \"%s\".\n",
			opts.ring_spec_desc);
		exit(1);
	}

	if (opts.echo_inputs) {
		describe_ring(&ring);
		print_ring_elements(&ring);
		printf("\n");
	}

	show_powers(&ring, max_power);

	return 0;
}
