#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"

#include "ring.h"
#include "ring_check.h"
#include "ring_stabilizer.h"

#define MAX_X_BYTES 2048
#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000
static unsigned char x_buf[MAX_X_BYTES];
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_RING_BYTES];
static unsigned char stabilizer_buf[MAX_RING_BYTES];

// ----------------------------------------------------------------
int stabilizer_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t ring;
	ring_t stabilizer;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need g as generator.
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "element";

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (!opts.pgspec->pvscan(x_buf, opts.cmd_paramv[0], aux_buf,
		sizeof(x_buf)))
	{
		fprintf(stderr, "Couldn't scan \"%s\" as ring element.\n",
			opts.cmd_paramv[0]);
		exit(1);
	}

	if (!make_ring(&ring, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_RING_BYTES))
	{
		fprintf(stderr, "Couldn't make ring \"%s\".\n",
			opts.ring_spec_desc);
		exit(1);
	}


	if (opts.echo_inputs) {
		describe_ring(&ring);
		print_ring_elements(&ring);
	}

	make_stabilizer(&ring, x_buf, &stabilizer,
		stabilizer_buf, MAX_RING_BYTES);

	if (opts.echo_inputs) {
		printf("\n");
		printf("Centralizer of ");
		ring.pspec->pvprint(x_buf, ring.pvaux);
		printf(":\n");
	}
	print_ring_elements(&stabilizer);

	return 0;
}
