#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ring_opts.h"

#include "ring.h"
#include "ring_check.h"
#include "ring_torsion.h"

#define MAX_AUX_BYTES 2048
#define MAX_RING_BYTES 0x100000
static unsigned char aux_buf[MAX_AUX_BYTES];
static unsigned char elts_buf[MAX_RING_BYTES];
static unsigned char sgr_buf[MAX_RING_BYTES];

// ----------------------------------------------------------------
int torsion_cmd(int argc, char ** argv)
{
	ring_opts_spec_t ospec;
	ring_opts_t opts;
	ring_t ring;
	ring_t torsion_subring;
	int k;

	ospec.default_verbosity = 0;
	ospec.num_cmd_params    = 1; // Need k for k-torsion
	ospec.num_file_names    = 1;
	ospec.cmd_param_desc    = "k";

	if (!ring_parse_opts(argc, argv, &ospec, &opts))
		ring_opts_usage(argv[0], ospec.cmd_param_desc);

	if (!sscanf(opts.cmd_paramv[0], "%d", &k)) {
		fprintf(stderr, "Couldn't scan \"%s\" as int.\n",
			opts.cmd_paramv[1]);
		exit(1);
	}

	if (!opts.pgspec->pvauxscan(aux_buf, opts.auxc, opts.auxv,
		MAX_AUX_BYTES))
	{
		fprintf(stderr, "Couldn't parse ring aux.\n");
		exit(1);
	}

	if (!make_ring(&ring, opts.pgspec, aux_buf,
		opts.file_namec == 0 ? 0 : opts.file_namev[0],
		elts_buf, MAX_RING_BYTES))
	{
		fprintf(stderr, "Couldn't make ring \"%s\".\n",
			opts.ring_spec_desc);
		exit(1);
	}

/*
	This is expensive to compute for larger rings.  Nonetheless the
	user (me) is warned that k-torsion subsets are only guaranteed to be
	subrings if the parent ring is abelian.  The isring command may be
	used, if desired, to verify that a k-torsion subset is actually a
	subring.
	if (!is_abelian(&ring, 0)) {
		printf("Ring is non-abelian; concept of k-torsion "
			"subring is meaningless.\n");
		exit(1);
	}
*/


	if (opts.echo_inputs) {
		describe_ring(&ring);
		print_ring_elements(&ring);
	}

	make_torsion_subring(&ring, k, &torsion_subring,
		sgr_buf, MAX_RING_BYTES);

	if (opts.echo_inputs) {
		printf("\n");
		printf("%d-torsion subring:\n", k);
	}
	print_ring_elements(&torsion_subring);

	return 0;
}
