#include <stdio.h>
#include <stdlib.h>
#include "z2npoly_factor.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {polys ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
z2poly_t z2polymod_discriminant(
	z2poly_t * pz,
	z2poly_t * pr)
{
	z2poly_t * roots = z2poly_get_roots(pz, pr);
	int deg = pz->degree;
	int i, j;
	z2poly_t diff;
	z2poly_t rv = z2poly_from_base_rep(1);

	for (i = 0; i < deg; i++) {
		for (j = i+1; j < deg; j++) {
			diff = z2poly_sub(&roots[i], &roots[j]);
			rv = z2polymod_mul(&rv, &diff, pr);
		}
	}

	free(roots);
	return rv;
}

// ----------------------------------------------------------------
z2poly_t z2poly_discriminant(
	z2poly_t * pz,
	z2poly_t * pr)
{
	z2poly_t * roots = z2poly_get_roots(pz, pr);
	int deg = pz->degree;
	int i, j;
	z2poly_t diff;
	z2poly_t rv = z2poly_from_base_rep(1);

	for (i = 0; i < deg; i++) {
		for (j = i+1; j < deg; j++) {
			diff = z2poly_sub(&roots[i], &roots[j]);
			rv = z2poly_mul(&rv, &diff);
		}
	}

	free(roots);
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2poly_t * r;
	z2poly_t resm, res;
	int numr, argi, i, j;

	if (argc < 2)
		usage(argv[0]);
	numr = argc - 1;
	r = malloc_check(numr * sizeof(z2poly_t));
	for (i = 0, argi = 1; argi < argc; i++, argi++) {
		if (!z2poly_scan(&r[i], argv[argi]))
			usage(argv[0]);
	}

	for (i = 0; i < numr; i++) {
		for (j = 0; j < numr; j++) {
			resm = z2polymod_discriminant(&r[i], &r[j]);
			res  = z2poly_discriminant(&r[i], &r[j]);

			z2poly_print(&r[i]); printf(" ");
			z2poly_print(&r[j]); printf(" ");
			z2poly_print(&resm); printf(" ");
			z2poly_print(&res);  printf("\n");
		}
		printf("\n");
	}

	free(r);
	return 0;
}
