#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2nipolymod.h"
#include "z2nilinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r} {polys}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int nrows, argi, polyi;
	z2ipoly_t r;
	z2nipoly_t * polys;
	z2ipoly_t ** M;
	int i;
	int rank;
	int curr_deg, max_deg = 0, ncols;
	z2nipoly_t max_poly = z2nipoly_from_coeffs0(0LL);

	if (argc < 3)
		usage(argv[0]);
	if (!z2ipolymod_auxscan(&r, argv[1]))
		usage(argv[0]);
	nrows = argc - 2;
	polys = malloc_check(nrows * sizeof(z2nipoly_t));

	for (argi = 2, polyi = 0; argi < argc; argi++, polyi++) {
		if (!z2nipoly_scan(&polys[polyi], argv[argi], r)) {
			fprintf(stderr, "%s:  couldn't scan \"%s\"\n",
				argv[0], argv[argi]);
			exit(1);
		}

		curr_deg = polys[polyi].degree;
		if (curr_deg > max_deg) {
			max_deg = curr_deg;
			max_poly = polys[polyi];
		}
	}
	ncols = max_deg + 1;
	M = z2nila_alloc_matrix(nrows, ncols);
	z2nila_matrix_from_z2nipolys(M, nrows, ncols, polys);

	z2nila_row_echelon_form(M, r, nrows, ncols);
	rank = z2nila_find_rank(M, nrows, ncols);
	z2nila_z2nipolys_from_matrix(polys, M, rank, ncols);

	for (i = 0; i < rank; i++) {
		z2nipoly_print(&polys[i], r);
		printf("\n");
	}

	free(polys);
	z2nila_free_matrix(M, nrows, ncols);
	return 0;
}
