#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2ipoly_factor.h"

// ----------------------------------------------------------------
void foo_print_factors(
	z2ipoly_t r)
{
	z2ipoly_factor_info_t finfo;
	int i, j, k;

	z2ipoly_factor(r, &finfo);

	k = 0;
	for (i = finfo.num_distinct - 1; i >= 0; i--) {
		for (j = 0; j < finfo.factor_counts[i].count; j++) {
			if (k > 0)
				printf(" ");
			z2ipoly_print(finfo.factor_counts[i].f);
			k++;
		}
	}
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t r, hi, a;
	int verbose = 0;
	int nr = 0;
	int argi, n, elt_order, max_order, group_order;

	if (argc < 2)
		usage(argv[0]);

	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "-v") == 0) {
			verbose = 1;
			continue;
		}

		if (++nr > 1)
			printf("\n");
		if (!z2ipoly_scan(&r, argv[argi]))
			usage(argv[0]);
		n = z2ipoly_get_degree(r);
		hi = 1LL << n;

		z2ipoly_print(r);
		printf(" = [");
		z2ipoly_print_factors(r);
		printf("]\n");

		group_order = 0;
		max_order = 0;
		for (a = 0; a < hi; a++) {
			if (z2ipoly_gcd(a,r) == 1) {
				elt_order = z2ipolymod_element_order(a, r);
				if (elt_order > max_order)
					max_order = elt_order;
				if (verbose) {
					printf(" i ");
					z2ipolymod_print(a, r);
					printf(" %d\n", elt_order);
				}
				group_order++;
			}
			else {
				if (verbose) {
					printf(" r ");
					z2ipolymod_print(a, r);
					printf(" = [");
					//z2ipoly_print_factors(a);
					foo_print_factors(a);
					printf("]\n");
				}
			}
		}
		printf("Group order %d; max elt order %d\n",
			group_order, max_order);
	}

	return 0;
}
