#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "linescan.h"
#include "z2ipolymod.h"
#include "z2ilinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {polys}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int nrows, argi, polyi;
	z2ipoly_t * M;
	int i;
	int rank;
	int curr_deg, max_deg = 0;
	z2ipoly_t max_poly =  0LL;

	if (argc < 2)
		usage(argv[0]);
	nrows = argc - 1;
	M = malloc_check(nrows * sizeof(z2ipoly_t));

	for (argi = 1, polyi = 0; argi < argc; argi++, polyi++) {
		if (!z2ipoly_scan(&M[polyi], argv[argi])) {
			fprintf(stderr, "%s:  couldn't scan \"%s\"\n",
				argv[0], argv[argi]);
			exit(1);
		}
	}

	z2ila_row_echelon_form(M, nrows);
	rank = z2ila_find_rank(M, nrows);

	for (i = 0; i < rank; i++) {
		curr_deg = z2ipoly_get_degree(M[i]);
		if (curr_deg > max_deg) {
			max_deg = curr_deg;
			max_poly = M[i];
		}
	}
	max_poly <<= 1;

	for (i = 0; i < rank; i++) {
		z2ipolymod_print(M[i], max_poly);
		printf("\n");
	}

	free(M);
	return 0;
}
