#include <stdio.h>
#include <stdlib.h>
#include "z2ipolymod.h"
#include "z2nipoly.h"
#include "z2ni_convert.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {r1} {r2} {polys ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
// Degrees d1, d2; reductors r1, r2; generators g1, g2; input element u1,
// output u2.

int main(int argc, char ** argv)
{
	z2ipoly_t r1, r2, g1 = 0LL, g2 = 0LL;
	int d1, d2, pd1, pd2;
	z2nipoly_t f1, f2;
	int argi;

	if (argc < 4)
		usage(argv[0]);
	if (!z2ipolymod_auxscan(&r1, argv[1]))
		usage(argv[0]);
	if (!z2ipolymod_auxscan(&r2, argv[2]))
		usage(argv[0]);

	d1 = z2ipoly_get_degree(r1);
	d2 = z2ipoly_get_degree(r2);
	pd1 = (1 << d1) - 1;
	pd2 = (1 << d2) - 1;

	if ((d2 % d1) == 0) {
		// field 1 is a subfield of field 2.
		if (!z2ni_convert_find_generators(r1, r2, &g1, &g2)) {
			fprintf(stderr, "%s:  Couldn't find generators.\n",
				argv[0]);
			exit(1);
		}
	}
	else if ((d1 % d2) == 0) {
		// field 1 is an extension field of field 2.
		if (!z2ni_convert_find_generators(r2, r1, &g2, &g1)) {
			fprintf(stderr, "%s:  Couldn't find generators.\n",
				argv[0]);
			exit(1);
		}
	}
	else {
		fprintf(stderr, "%s: Immisicible degrees %d, %d\n",
			argv[0], d1, d2);
		exit(1);
	}

	for (argi = 3; argi < argc; argi++) {

		if (!z2nipoly_scan(&f1, argv[argi], r1))
			usage(argv[0]);

		if (!z2ni_convert_poly(&f1, r1, g1, &f2, r2, g2)) {
			printf("%s: Incontractible element ", argv[0]);
			z2nipoly_print(&f1, r1);
			printf("\n");
			continue;
		}

		if (argi > 3)
			printf(" ");
		z2nipoly_print(&f2, r2);
	}
	printf("\n");

	return 0;
}
