#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2nipolymod.h"
#include "z2ni_convert.h"
#include "power.h"
#include "mod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s {int r} {ext r} [verbose]\n",
		argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t  int_r, ext_r;
	z2ipoly_t  int_g, ext_sub_g, ext_g;
	int        verbose = 0;
	int        d, n, i, power;
	z2nipoly_t ext_min;
	z2nipoly_t int_min;
	z2ipoly_t  gpower = 0LL;
	z2nipoly_t factor;

	if ((argc != 3) && (argc != 4))
		usage(argv[0]);

	if (!z2ipoly_scan(&int_r, argv[1]))
		usage(argv[0]);
	if (!z2ipoly_scan(&ext_r, argv[2]))
		usage(argv[0]);
	if (argc == 4)
		verbose = 1;

	d = z2ipoly_get_degree(int_r);
	n = z2ipoly_get_degree(ext_r);
	if ((n % d) != 0) {
		fprintf(stderr, "Indivisible extension: deg %d over %d\n",
			n, d);
		exit(1);
	}

	if (!z2ni_convert_find_generators(int_r, ext_r, &int_g, &ext_sub_g)) {
		fprintf(stderr,
			"%s:  Couldn't find subfield generators.\n", argv[0]);
		exit(1);
	}
	ext_g = z2ipolymod_find_generator(ext_r);

	if (verbose) {
		printf("int r = ");
		z2ipolymod_print(int_r,int_r<<1);
		printf("  ");

		printf("ext r = ");
		z2ipolymod_print(ext_r,ext_r<<1);
		printf("\n");

		printf("int g = ");
		z2ipolymod_print(int_g,int_r<<1);
		printf("  ");

		printf("ext gs= ");
		z2ipolymod_print(ext_sub_g,ext_r<<1);
		printf("  ");

		printf("ext g = ");
		z2ipolymod_print(ext_g,ext_r<<1);
		printf("\n");
	}

	// (x - beta) (x - beta^4) (x - beta^16)

	gpower = ext_g;
	ext_min = z2nipoly_extend_z2ipoly(1LL);
	power = 1;
	for (i = 0; i < n; i += d) {
		factor.degree = 1;
		factor.coeffs[1] = 1LL;
		//factor.coeffs[0] = z2ipoly_usub(&gpower);
		//Z2 self-negates
		factor.coeffs[0] = gpower;
		if (verbose) {
			if (i > 0)
				printf(" * ");
			z2nipoly_print(&factor, ext_r);
			printf("[x-");
			z2ipolymod_print(ext_g, ext_r);
			printf("^%d]", power);
		}
		ext_min = z2nipoly_mul(&ext_min, &factor, ext_r);
		gpower = z2ipolymod_power(gpower, 1 << d, ext_r);
		power *= 1 << d;
	}
	if (verbose) {
		printf("\n");
		z2nipoly_print(&ext_min, ext_r);
		printf("\n");
	}
	if (!z2ni_convert_poly(&ext_min, ext_r, ext_sub_g, &int_min,
		int_r, int_g))
	{
		fprintf(stderr, "%s:  Couldn't unembed.\n", argv[0]);
		exit(1);
	}

	z2nipoly_print(&int_min, int_r);
	printf("\n");

	return 0;
}
