#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2npolymod.h"
#include "power.h"
#include "mod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// for each reductor r
//   for each possible generator g
//     if  g^5  mod r == 0:10
//     and g^10 mod r == 0:11
//       wahoo!

// Inputs:
// * Reductor for intermediate field, w/ degree implicit
// * Degree of extension field over the base field
// * Array of exponents & fit points

void seek_min_poly(
	z2poly_t  * pint_r,  // reductor for intermediate field
	z2poly_t  * pext_r,  // reductor for extension field over base
	int         fit_log,
	z2npoly_t * pfit_val,
	int         verbose)
{
	int int_deg = z2poly_get_degree(pint_r);
	int ext_deg = z2poly_get_degree(pext_r);
	int rep_max = 1LL << ext_deg;
	int i, k;
	z2npoly_t g = z2npoly_from_two_ints(1, 0);
	z2npoly_t rr;
	z2npoly_t gpow;
	z2poly_t  rem;

	for (i = 0; i < rep_max; i++) {
		rr = z2npoly_from_base_rep(i + rep_max, int_deg);
		gpow = z2npolymod_power(&g, fit_log, &rr, pint_r);
		if (!z2npoly_equal(&gpow, pfit_val, pint_r))
			continue;

		rem = z2npoly_flatten(&rr, pint_r, pext_r, fit_log);
		if (!z2poly_is_zero(&rem))
			continue;

		printf("  rr=");
		z2npoly_print(&rr, pint_r);
		printf(" g=");
		z2npoly_print(&g, pint_r);
		printf("\n");

		if (verbose) {
			for (k = 0; k < rep_max; k++) {
				z2npoly_t gpow = z2npolymod_power(
					&g, k, &rr, pint_r);
				printf("    ");
				z2npoly_print(&g, pint_r);
				printf("^%d = ", (int)k);
				z2npoly_print(&gpow, pint_r);
				printf("\n");
			}
			printf("\n");
		}
	}
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s {intermediate r} {extension r} [verbose]\n",
		argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2poly_t  int_r, ext_r;
	int       d, n, pd1, pn1;
	int       fit_log, fit_log_init;
	z2npoly_t fit_val;
	int       verbose = 0;

	if ((argc != 3) && (argc != 4))
		usage(argv[0]);

	if (!z2poly_scan(&int_r, argv[1]))
		usage(argv[0]);
	if (!z2poly_scan(&ext_r, argv[2]))
		usage(argv[0]);
	if (argc == 4)
		verbose = 1;

	fit_val = z2npoly_from_two_ints(0, 2);
	d = z2poly_get_degree(&int_r);
	n = z2poly_get_degree(&ext_r);
	if ((n % d) != 0) {
		fprintf(stderr, "Indivisible extension: deg %d over %d\n",
			n, d);
		exit(1);
	}
	pd1 = (1 << d) - 1;
	pn1 = (1 << n) - 1;
	fit_log_init = pn1 / pd1;

	//for (fit_log = fit_log_init; fit_log < pn1; fit_log += fit_log_init) {
	fit_log = fit_log_init;

	printf("n=%d d=%d pd1=%d pn1=%d fitlog=%d\n",
		n, d, pd1, pn1, fit_log);
	seek_min_poly(&int_r, &ext_r, fit_log, &fit_val, verbose);
	//}

	return 0;
}
