#include <stdio.h>
#include <stdlib.h>
#include "zpnpoly_factor.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {polys ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
zpoly_t zppolymod_discriminant(
	zpoly_t * pz,
	zppoly_aux_t * paux)
{
	zpoly_t * roots = zppoly_get_roots(pz, paux);
	int deg = pz->degree;
	int i, j;
	zpoly_t diff;
	zpoly_t rv = zppoly_from_base_rep(paux->p, 1);

	for (i = 0; i < deg; i++) {
		zppoly_print(&roots[i], paux->p);
		printf(" ");
	}
	printf("|");
	for (i = 0; i < deg; i++) {
		for (j = i+1; j < deg; j++) {
			diff = zppoly_sub(&roots[i], &roots[j], paux->p);
			printf(" ");
			zppoly_print(&diff, paux->p);
			rv = zppolymod_mul(&rv, &diff, paux);
		}
	}
	printf(" |");

	free(roots);
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zpoly_t * r;
	zpoly_t resm;//, res;
	int p, numr, argi, i, j;
	zppoly_aux_t aux;

	if (argc < 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	numr = argc - 2;
	r = malloc_check(numr * sizeof(zpoly_t));
	for (i = 0, argi = 2; argi < argc; i++, argi++) {
		if (!zppoly_scan(&r[i], argv[argi], p))
			usage(argv[0]);
	}

	for (i = 0; i < numr; i++) {
		zppoly_print(&r[i], p);
		printf(": ");
		for (j = 0; j < numr; j++) {
			aux.p = p;
			aux.r = r[j];
			resm = zppolymod_discriminant(&r[i], &aux);
		}
		printf("\n");
	}

	free(r);
	return 0;
}
