#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zpnpolymod.h"
#include "zpnlinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r} {polys ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int argi, polyi;
	zppoly_aux_t aux;
	zpnpoly_t * polys;
	zpoly_t ** M;
	int i;
	int nrows = 0, ncols;
	int rank;
	int curr_deg, max_deg = 0;
	zpnpoly_t max_poly = zpnpoly_from_int(1, 0);

	if (argc < 3)
		usage(argv[0]);
	if (!zppolymod_auxscan(&aux, argv[1], argv[2]))
		usage(argv[0]);

	nrows = argc - 3;
	polys = malloc_check(nrows * sizeof(zpnpoly_t));

	for (argi = 3, polyi = 0; argi < argc; argi++, polyi++) {
		if (!zpnpoly_scan(&polys[polyi], argv[argi], &aux)) {
			fprintf(stderr, "%s:  couldn't scan \"%s\"\n",
				argv[0], argv[argi]);
			exit(1);
		}
		curr_deg = polys[polyi].degree;
		if (curr_deg > max_deg) {
			max_deg = curr_deg;
			max_poly = polys[polyi];
		}
	}
	ncols = max_deg + 1;

	M = zpnla_alloc_matrix(nrows, ncols);
	zpnla_matrix_from_zpnpolys(M, nrows, ncols, polys);

	zpnla_row_echelon_form(M, &aux, nrows, ncols);
	rank = zpnla_find_rank(M, nrows, ncols);

	zpnla_zpnpolys_from_matrix(polys, M, rank, ncols);

	for (i = 0; i < rank; i++) {
		zpnpoly_print(&polys[i], &aux);
		printf("\n");
	}

	zpnla_free_matrix(M, nrows, ncols);

	return 0;
}
