#include <stdio.h>
#include <stdlib.h>
#include "zppolymod.h"
#include "zpnpoly.h"
#include "zpn_convert.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r1} {r2} {polys ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
// Degrees d1, d2; reductors r1, r2; generators g1, g2; input element u1,
// output u2.

int main(int argc, char ** argv)
{
	zppoly_aux_t r1, r2;
	zpoly_t g1, g2;
	int d1, d2, pd1, pd2;
	zpnpoly_t f1, f2;
	int argi;

	if (argc < 5)
		usage(argv[0]);
	if (!zppolymod_auxscan(&r1, argv[1], argv[2]))
		usage(argv[0]);
	if (!zppolymod_auxscan(&r2, argv[1], argv[3]))
		usage(argv[0]);

	d1 = r1.r.degree;
	d2 = r2.r.degree;
	pd1 = (1 << d1) - 1;
	pd2 = (1 << d2) - 1;

	if ((d2 % d1) == 0) {
		// field 1 is a subfield of field 2.
		if (!zpn_convert_find_generators(&r1, &r2, &g1, &g2)) {
			fprintf(stderr, "%s:  Couldn't find generators.\n",
				argv[0]);
			exit(1);
		}
	}
	else if ((d1 % d2) == 0) {
		// field 1 is an extension field of field 2.
		if (!zpn_convert_find_generators(&r2, &r1, &g2, &g1)) {
			fprintf(stderr, "%s:  Couldn't find generators.\n",
				argv[0]);
			exit(1);
		}
	}
	else {
		fprintf(stderr, "%s: Immisicible degrees %d, %d\n",
			argv[0], d1, d2);
		exit(1);
	}

	for (argi = 4; argi < argc; argi++) {

		if (!zpnpoly_scan(&f1, argv[argi], &r1))
			usage(argv[0]);

		if (!zpn_convert_poly(&f1, &r1, &g1, &f2, &r2, &g2)) {
			printf("%s: Incontractible element ", argv[0]);
			zpnpoly_print(&f1, &r1);
			printf("\n");
			continue;
		}

		zpnpoly_print(&f2, &r2);
		printf("\n");
	}

	return 0;
}
