#include <stdio.h>
#include <stdlib.h>
#include "zpnpoly.h"

typedef zpnpoly_t my_vop_t(
	zpnpoly_t * pin1,
	zpnpoly_t * pin2,
	zppoly_aux_t * paux);

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr,
		"Usage: %s {p} {m or red} {op} {poly1} {poly2}\n",
		argv0);
	fprintf(stderr,
	"Op is one of +, -, *, / or %%; . may be used in place of *.\n");
	exit(1);
}

// ----------------------------------------------------------------
void show_op(
	char * op,
	char * in1,
	char * in2,
	zppoly_aux_t * paux)
{
	zpnpoly_t a, b, c, extra;

	if (!zpnpoly_vscan(&a, in1, paux, sizeof(a))) {
		fprintf(stderr, "Scan failure on \"%s\".\n",
			in1);
		exit(1);
	}
	if (!zpnpoly_vscan(&b, in2, paux, sizeof(b))) {
		fprintf(stderr, "Scan failure on \"%s\".\n",
			in2);
		exit(1);
	}
	switch (op[0]) {
	case '+':
		c = zpnpoly_add(&a, &b, paux);
		break;
	case '-':
		c = zpnpoly_sub(&a, &b, paux);
		break;
	case '.':
	case '*':
		c = zpnpoly_mul(&a, &b, paux);
		break;
	case '/':
		zpnpoly_quot_and_rem(&a, &b, &c, &extra, paux);
		break;
	case '%':
		zpnpoly_quot_and_rem(&a, &b, &extra, &c, paux);
		break;
	default:
		fprintf(stderr, "Unknown operator \"%s\".\n",
			op);
		exit(1);
		break;
	}

	zpnpoly_vprintsmall(&c, paux);
	printf("\n");
}

// ----------------------------------------------------------------
void show_ops(
	char  * op,
	char ** args,
	int     num_args,
	zppoly_aux_t * paux)
{
	my_vop_t * pvop;
	int i;
	zpnpoly_t curr, accum;
	int idrep = 0;

	switch (op[0]) {
	case '+':
		pvop = zpnpoly_add;
		idrep = 0;
		break;
	case '.':
	case '*':
		pvop = zpnpoly_mul;
		idrep = 1;
		break;
	case '-':
	case '/':
	case '%':
		if (num_args != 2) {
			fprintf(stderr, "Need two arguments for /.\n");
			exit(1);
		}
		show_op(op, args[0], args[1], paux);
		return;
		break;
	default:
		fprintf(stderr, "Unknown operator \"%s\".\n", op);
		exit(1);
		break;
	}

	accum.degree = 0;
	accum.coeffs[0] = zppoly_from_base_rep(paux->p, idrep);
	for (i = 0; i < num_args; i++) {
		if (!zpnpoly_vscan(&curr, args[i], paux, sizeof(curr))) {
			fprintf(stderr, "Scan failure on \"%s\".\n",
				args[i]);
			exit(1);
		}
		accum = pvop(&accum, &curr, paux);
	}
	zpnpoly_vprintsmall(&accum, paux);
	printf("\n");
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zppoly_aux_t aux;

	if (argc < 4)
		usage(argv[0]);

	// argv[1] and argv[2]
	if (!zppolymod_auxscan(&aux, argv[1], argv[2]))
		usage(argv[0]);

	show_ops(argv[3], &argv[4], argc - 4, &aux);

	return 0;
}
