#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "zpnpoly.h"

typedef struct _factor_list_elt_t {
	zpnpoly_t factor;
	struct _factor_list_elt_t * pnext;
} factor_list_elt_t;

typedef struct _poly_factor_info_t {
	int count;
	zppoly_aux_t aux;
	factor_list_elt_t * phead;
} poly_factor_info_t;

static poly_factor_info_t poly_factor_infos[ZPNPOLY_MAX_DEGREE];

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {poly}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p, m, pm, i;
	int count;
	int verbose = 0;
	zppoly_aux_t aux, aux1;
	zpnpoly_t z, factor, remnant, quot, rem;
	int num_divisors;
	int max_m;
	factor_list_elt_t * pelt;
	factor_list_elt_t * pwalk;
	zpnpoly_t check, mcheck;

	if (argc != 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);
	if (p < 2) {
		fprintf(stderr, "p must be >= 2; got %d.\n", p);
	}

	aux.p = p;
	aux.r = zppoly_from_base_rep(p, p+1);
	aux1 = aux;

	if (!zpnpoly_vscan(&z, argv[2], &aux, sizeof(z))) {
		fprintf(stderr, "Scan failure on \"%s\".\n",
			argv[2]);
		exit(1);
	}

	printf("Factoring ");
	zpnpoly_vprint(&z, &aux);
	printf("\n");

	remnant = z;
	num_divisors = 0;
	for (m = 0; m <= ZPNPOLY_MAX_DEGREE; m++) {
		poly_factor_infos[m].count = 0;
		poly_factor_infos[m].phead = 0;
	}

	for (m = 1; ; m++) {
		max_m = m;
		if (m == 1)
			printf("Factors over GF(%d):\n", p);
		else
			printf("Factors over GF(%d^%d):\n", p, m);
		aux = zppoly_find_irreducible(p, m);
		pm = calc_power(p, m);

		// Try each possible linear factor (X - a) for a in GF(p^m).
		factor.degree = 1;
		factor.coeffs[1] = zppoly_from_base_rep(aux.p, 1);
		for (i = 0; i < pm; i++) {
			factor.coeffs[0] = zppoly_from_base_rep(aux.p, i);

			// Keep dividing out this factor as many times as
			// possible (i.e. up to the multiplicity of the
			// factor).
			for (count = 0; ; count++) {

				zpnpoly_quot_and_rem(&remnant, &factor,
					&quot, &rem, &aux);

				if (verbose) {
					printf("  Remnant:    ");
					zpnpoly_vprint(&remnant, &aux);
					printf("\n");
					printf("  Factor:     ");
					zpnpoly_vprint(&factor, &aux);
					printf("\n");
					printf("  Quot:       ");
					zpnpoly_vprint(&quot, &aux);
					printf("\n");
					printf("  Rem:        ");
					zpnpoly_vprint(&rem, &aux);
					printf("\n");
					printf("\n");
				}

				if (!zpnpoly_is_zero(&rem))
					break;

				printf("  ");
				zpnpoly_vprint(&factor, &aux);
				printf("\n");

				if (poly_factor_infos[m].count == 0)
					poly_factor_infos[m].aux = aux;
				poly_factor_infos[m].count++;
				pelt = (factor_list_elt_t *)
					malloc(sizeof(factor_list_elt_t));
				if (pelt == 0) {
					fprintf(stderr, "malloc(%d) failed.\n",
						sizeof(factor_list_elt_t));
					exit(1);
				}
				pelt->factor = factor;
				pelt->pnext = 0;

				if (poly_factor_infos[m].phead == 0) {
					poly_factor_infos[m].phead = pelt;
				}
				else {
					pwalk = poly_factor_infos[m].phead;
					while (pwalk->pnext != 0)
						pwalk = pwalk->pnext;
					pwalk->pnext = pelt;
				}

				remnant = quot;
			}
		}

		if (zpnpoly_is_one(&remnant)) {
			break;
		}
	}

	// Check -- how?  This is mixing math across fields of different
	// degree.  Need to multiply together all factors of a given degree,
	// then de-extend to Fp[X].  & make sure this is OK.

	// Example:  Factor X^4 + X^3 + X^2 + 1 in the algebraic closure
	// of GF(2).  Obtain the single factor X^1 in GF(2) and the three
	// factors X + 0,1,0, X + 1,0,0 and X + 1,1,0 in GF(2^3).  The four
	// together are immisicible.  However, the three GF(2^3) factors
	// multiply to X^3 + X + 1.  This is miscible with the linear factor
	// X + 1 -- with coefficients in GF(2) -- with product equal to the
	// original polynomial.

	printf("\n");
	check.degree = 0;
	check.coeffs[0] = zppoly_from_base_rep(aux.p, 1);

	for (m = 1; m <= max_m; m++) {
		if (poly_factor_infos[m].count == 0)
			continue;

		mcheck.degree = 0;
		mcheck.coeffs[0] = zppoly_from_base_rep(aux.p, 1);

		pwalk = poly_factor_infos[m].phead;
		while (pwalk) {
			mcheck = zpnpoly_mul(&mcheck, &pwalk->factor,
				&poly_factor_infos[m].aux);
			pwalk = pwalk->pnext;
		}
		if (m == 1)
			printf("Product of factors over GF(%d):\n", p);
		else
			printf("Product of factors over GF(%d^%d):\n", p, m);
		printf("  ");
		zpnpoly_vprintsmall(&mcheck, &poly_factor_infos[m].aux);
		printf("\n");

		// Make sure all coefficients are really in GF(p),
		// for miscibility of elements of different fields.
		for (i = 0; i <= mcheck.degree; i++) {
			if (mcheck.coeffs[i].degree != 0) {
				printf(
				"WARNING: Product not confined to GF(%d)[X].\n",
				aux.p);
				break;
			}
		}
		printf("\n");

		check = zpnpoly_mul(&check, &mcheck, &aux1);
	}

	printf("Product of all factors over GF(%d):\n", p);
	zpnpoly_vprintsmall(&check, &poly_factor_infos[m].aux);
	printf("\n");

	// xxx call zpn veq to make sure check == z
	if (!zpnpoly_equal(&z, &check, &aux1)) {
		printf("Error:  unequal to original polynomial.\n");
	}
	else {
		printf("OK\n");
	}

	return 0;
}
