#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "power.h"

typedef zpoly_t my_zpop_t(
	zpoly_t * pin1,
	zpoly_t * pin2,
	zppoly_aux_t * paux);

// ----------------------------------------------------------------
void show_table(
	char * op,
	zppoly_aux_t * paux,
	int long_fmt)
{
	zpoly_t a, b, c;
	int i, j;
	int lo;
	int pm;
	my_zpop_t * pzpop;

	pm = calc_power(paux->p, paux->r.degree);

	switch (op[0]) {
	case '+': pzpop = zppolymod_add; lo   = 0; break;
	case '-': pzpop = zppolymod_sub; lo   = 0; break;
	case '.':
	case '*': pzpop = zppolymod_mul; lo   = 0; break;
	case '/': pzpop = zppolymod_div; lo   = 1; break;
	default:
		fprintf(stderr, "Unknown operator \"%s\".\n", op);
		exit(1);
		break;
	}

	if (long_fmt) {
		for (i = lo; i < pm; i++) {
			a = zppoly_from_base_rep(paux->p, i);
			for (j = lo; j < pm; j++) {
				b = zppoly_from_base_rep(paux->p, j);

				c = pzpop(&a, &b, paux);

				zppolymod_print(&a, paux);
				printf(" %s ", op);
				zppolymod_print(&b, paux);
				printf(" = ");
				zppolymod_print(&c, paux);
				printf("\n");
			}
			printf("\n");
		}
	}

	else {
		for (i = lo; i < pm; i++) {
			a = zppoly_from_base_rep(paux->p, i);
			for (j = lo; j < pm; j++) {
				b = zppoly_from_base_rep(paux->p, j);
				c = pzpop(&a, &b, paux);
				if (j > lo)
					printf(" ");
				zppolymod_print(&c, paux);
			}
			printf("\n");
		}
	}
}

// ----------------------------------------------------------------
void show_order(
	zpoly_t      * pz,
	zppoly_aux_t * paux)
{
	int order;

	if (zppoly_is_zero(pz))
		order = 0;
	else
		order = zppolymod_element_order(pz, paux);
	printf("%3d ", order);
	zppolymod_print(pz, paux);
	printf("\n");
}

// ----------------------------------------------------------------
void show_orders(
	char ** args,
	int     num_args,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!zppoly_scan(&z, args[i], paux->p)) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			show_order(&z, paux);
		}
	}
	else {
		pm = calc_power(paux->p, paux->r.degree);
		for (i = 1; i < pm; i++) {
			z = zppoly_from_base_rep(paux->p, i);
			show_order(&z, paux);
		}
	}
}

// ----------------------------------------------------------------
void show_orbit(
	zpoly_t      * pz,
	zppoly_aux_t * paux)
{
	int count;
	zpoly_t zpower;
	zpoly_t zero = zppoly_from_base_rep(paux->p, 0);
	zpoly_t one  = zppoly_from_base_rep(paux->p, 1);

	zpower = *pz;
	count = 0;
	while (1) {
		if (count > 0)
			printf(" ");
		count++;
		zppolymod_print(&zpower, paux);
		if (zppoly_equal(&zpower, &zero, paux->p))
			break;
		if (zppoly_equal(&zpower, &one, paux->p))
			break;
		zpower = zppolymod_mul(&zpower, pz, paux);
	}
	printf("\n");
}

// ----------------------------------------------------------------
void show_orbits(
	char ** args,
	int     num_args,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!zppoly_scan(&z, args[i], paux->p)) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			show_orbit(&z, paux);
		}
	}
	else {
		pm = calc_power(paux->p, paux->r.degree);
		for (i = 1; i < pm; i++) {
			z = zppoly_from_base_rep(paux->p, i);
			show_orbit(&z, paux);
		}
	}
}

// ----------------------------------------------------------------
void show_frobit(
	zpoly_t      * pz,
	zppoly_aux_t * paux)
{
	int count;
	zpoly_t zpower;
	zpoly_t zero = zppoly_from_coeffs0(0, paux->p);
	zpoly_t one  = zppoly_from_coeffs0(1, paux->p);

	zpower = *pz;
	count = 0;
	while (1) {
		if (count > 0)
			printf(" ");
		count++;
		zppolymod_print(&zpower, paux);
		if (zppoly_equal(&zpower, &zero, paux->p))
			break;
		if (zppoly_equal(&zpower, &one, paux->p))
			break;
		if (zppoly_equal(&zpower, pz, paux->p) && (count > 1))
			break;
		zpower = zppolymod_power(&zpower, paux->p, paux);
	}
	printf("\n");
}

// ----------------------------------------------------------------
void show_frobits(
	char ** args,
	int     num_args,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!zppoly_scan(&z, args[i], paux->p)) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			show_frobit(&z, paux);
		}
	}
	else {
		pm = calc_power(paux->p, paux->r.degree);
		for (i = 1; i < pm; i++) {
			z = zppoly_from_base_rep(paux->p, i);
			show_frobit(&z, paux);
		}
	}
}

// ----------------------------------------------------------------
void show_rep_power(
	zpoly_t      * pz,
	int            power,
	zppoly_aux_t * paux)
{
	int i, n = paux->r.degree;
	zpoly_t zpower;

	zpower = *pz;
	for (i = 0; i < n; i++) {
		zppolymod_print(&zpower, paux);
		printf("\n");
		zpower = zppolymod_power(&zpower, power, paux);
	}
	printf("\n");
}

// ----------------------------------------------------------------
void show_rep_powers(
	char ** args,
	int     num_args,
	int            power,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!zppoly_scan(&z, args[i], paux->p)) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			show_rep_power(&z, power, paux);
		}
	}
	else {
		pm = calc_power(paux->p, paux->r.degree);
		for (i = 1; i < pm; i++) {
			z = zppoly_from_base_rep(paux->p, i);
			show_rep_power(&z, power, paux);
		}
	}
}

// ----------------------------------------------------------------
void show_power(
	zpoly_t      * pz,
	int            power,
	zppoly_aux_t * paux)
{
	zpoly_t zpower;

	zpower = zppolymod_power(pz, power, paux);
	zppolymod_print(&zpower, paux);
	printf("\n");
}

// ----------------------------------------------------------------
void show_powers(
	char ** args,
	int     num_args,
	int     power,
	zppoly_aux_t * paux)
{
	int i;
	zpoly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!zppoly_scan(&z, args[i], paux->p)) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			show_power(&z, power, paux);
		}
	}
	else {
		pm = calc_power(paux->p, paux->r.degree);
		for (i = 1; i < pm; i++) {
			z = zppoly_from_base_rep(paux->p, i);
			show_power(&z, power, paux);
		}
	}
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {m or red} tbl {op}\n", argv0);
	fprintf(stderr, "  Op is one of +, -, * or /; . may be used in "
		"place of *.\n");
	fprintf(stderr, "Or:    %s {m or red} order  [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} orbit  [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} frobit [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} power {n} [...]\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	zppoly_aux_t aux;
	int long_tbl_fmt = 0;

	if (argc < 4)
		usage(argv[0]);

	if (!zppolymod_auxscan(&aux, argv[1], argv[2]))
		usage(argv[0]);

	if (strcmp(argv[3], "tbl") == 0) {
		if (argc < 5)
			usage(argv[0]);
		if (argc == 6)
			long_tbl_fmt = 1;
		show_table(argv[4], &aux, long_tbl_fmt);
	}
	else if (strcmp(argv[3], "order") == 0) {
		show_orders(&argv[4], argc - 4, &aux);
	}
	else if (strcmp(argv[3], "orbit") == 0) {
		show_orbits(&argv[4], argc - 4, &aux);
	}
	else if (strcmp(argv[3], "frobit") == 0) {
		show_frobits(&argv[4], argc - 4, &aux);
	}
	else if (strcmp(argv[3], "repp") == 0) {
		int power;
		if (argc < 5)
			usage(argv[0]);
		if (sscanf(argv[4], "%d", &power) != 1)
			usage(argv[0]);
		show_rep_powers(&argv[5], argc - 5, power, &aux);
	}
	else if (strcmp(argv[3], "power") == 0) {
		int power;
		if (argc < 5)
			usage(argv[0]);
		if (sscanf(argv[4], "%d", &power) != 1)
			usage(argv[0]);
		show_powers(&argv[5], argc - 5, power, &aux);
	}
	else {
		usage(argv[0]);
	}

	return 0;
}
