#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "power.h"
#include "malloc_check.h"

typedef z2ipoly_t my_z2iop_t(
	z2ipoly_t a,
	z2ipoly_t b,
	z2ipoly_t r);

// ----------------------------------------------------------------
void show_table(
	char * op,
	z2ipoly_t r,
	int long_fmt)
{
	z2ipoly_t a, b, c;
	int i, j;
	int lo;
	int pm;
	my_z2iop_t * pz2iop;

	pm = 1 << z2ipoly_get_degree(r);

	switch (op[0]) {
	case '+':
		pz2iop = z2ipolymod_add;
		lo   = 0;
		break;
	case '-':
		pz2iop = z2ipolymod_sub;
		lo   = 0;
		break;
	case '.':
	case '*':
		pz2iop = z2ipolymod_mul;
		lo   = 0;
		break;
	case '/':
		pz2iop = z2ipolymod_div;
		lo   = 1;
		break;
	default:
		fprintf(stderr, "Unknown operator \"%s\".\n",
			op);
		exit(1);
		break;
	}

	if (long_fmt) {
		for (i = lo; i < pm; i++) {
			a = i;
			for (j = lo; j < pm; j++) {
				b = j;
				c = pz2iop(a, b, r);

				z2ipolymod_print(a, r);
				printf(" %s ", op);
				z2ipolymod_print(b, r);
				printf(" = ");
				z2ipolymod_print(c, r);
				printf("\n");
			}
			printf("\n");
		}
	}

	else {
		for (i = lo; i < pm; i++) {
			a = i;
			for (j = lo; j < pm; j++) {
				b = j;
				c = pz2iop(a, b, r);
				if (j > lo)
					printf(" ");
				z2ipolymod_print(c, r);
			}
			printf("\n");
		}
	}
}

// ----------------------------------------------------------------
void show_order(
	z2ipoly_t z,
	z2ipoly_t r)
{
	int order;

	if (z == 0)
		order = 0;
	else
		order = z2ipolymod_element_order(z, r);
	printf("%3d ", order);
	z2ipolymod_print(z, r);
	printf("\n");
}

// ----------------------------------------------------------------
void show_orders(
	char ** args,
	int     num_args,
	z2ipoly_t r)
{
	int i;
	z2ipoly_t z;
	z2ipoly_t pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z2ipoly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z2ipoly_mod(z, r);
			show_order(z, r);
		}
	}
	else {
		pm = 1LL << z2ipoly_get_degree(r);
		for (z = 1; z < pm; z++) {
			show_order(z, r);
		}
	}
}

// ----------------------------------------------------------------
void show_orbit(
	z2ipoly_t z,
	z2ipoly_t r)
{
	int count;
	z2ipoly_t zpower;
	z2ipoly_t zero = 0LL;
	z2ipoly_t one  = 1LL;

	zpower = z;
	count = 0;
	while (1) {
		if (count > 0)
			printf(" ");
		count++;
		z2ipolymod_print(zpower, r);
		if (zpower == zero)
			break;
		if (zpower == one)
			break;
		zpower = z2ipolymod_mul(zpower, z, r);
	}
	printf("\n");
}

// ----------------------------------------------------------------
void show_orbits(
	char ** args,
	int     num_args,
	z2ipoly_t r)
{
	int i;
	z2ipoly_t z;
	z2ipoly_t pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z2ipoly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z2ipoly_mod(z, r);
			show_orbit(z, r);
		}
	}
	else {
		pm = 1LL << z2ipoly_get_degree(r);
		for (z = 1; z < pm; z++) {
			show_orbit(z, r);
		}
	}
}

// ----------------------------------------------------------------
void show_frobit(
	z2ipoly_t z,
	z2ipoly_t r)
{
	int count;
	z2ipoly_t zpower;
	z2ipoly_t zero = 0LL;
	z2ipoly_t one  = 1LL;

	zpower = z;
	count = 0;
	while (1) {
		if (count > 0)
			printf(" ");
		count++;
		z2ipolymod_print(zpower, r);
		if (zpower == zero)
			break;
		if (zpower == one)
			break;
		if ((zpower == z) && (count > 1))
			break;
		zpower = z2ipolymod_mul(zpower, zpower, r);
	}
	printf("\n");
}

// ----------------------------------------------------------------
void show_frobits(
	char ** args,
	int     num_args,
	z2ipoly_t r)
{
	int i;
	z2ipoly_t z;
	z2ipoly_t pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z2ipoly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z2ipoly_mod(z, r);
			show_frobit(z, r);
		}
	}
	else {
		char * marks;
		int count;
		z2ipoly_t zpower;
		z2ipoly_t zero = 0LL;
		z2ipoly_t one  = 1LL;

		pm = 1LL << z2ipoly_get_degree(r);
		marks = malloc_check(pm);

		for (z = 0; z < pm; z++)
			marks[z] = 0;
		for (z = 1; z < pm; z++) {
			if (marks[z])
				continue;

			zpower = z;
			marks[zpower] = 1;
			count = 0;
			while (1) {
				if (count > 0)
					printf(" ");
				count++;
				if (zpower == zero)
					break;
				if (zpower == one)
					break;
				if ((zpower == z) && (count > 1))
					break;
				z2ipolymod_print(zpower, r);
				zpower = z2ipolymod_mul(zpower, zpower, r);
				marks[zpower] = 1;
			}
			if (count == 1)
				z2ipolymod_print(zpower, r);
			printf("\n");
		}
		free(marks);
	}
}

// ----------------------------------------------------------------
void show_log(
	z2ipoly_t z,
	z2ipoly_t r)
{
	int log, got_it = 0;
	z2ipoly_t xp;
	int deg = z2ipoly_get_degree(r);
	int pm = (1 << deg);

	// find p such that x^p == z mod r

	xp = 1LL;
	log = 0;
	while (1) {
		if (z == xp) {
			got_it = 1;
			break;
		}
		xp = z2ipolymod_mul(xp, 2LL, r);
		log++;
		if (log > pm)
			break;
	}
	z2ipolymod_print(z, r);
	printf(":  ");
	if (got_it) {
		printf("%3d\n", log);
	}
	else {
		printf("  not a power of x (");
		z2ipoly_print(r);
		printf(" is not primitive)\n");
	}
}

// ----------------------------------------------------------------
void show_logs(
	char ** args,
	int     num_args,
	z2ipoly_t r)
{
	int i;
	z2ipoly_t z;
	z2ipoly_t pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z2ipoly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z2ipoly_mod(z, r);
			show_log(z, r);
		}
	}
	else {
		pm = 1LL << z2ipoly_get_degree(r);
		for (z = 1; z < pm; z++) {
			show_log(z, r);
		}
	}
}

// ----------------------------------------------------------------
void show_power(
	z2ipoly_t z,
	int       power,
	z2ipoly_t r)
{
	z2ipoly_t zpower = z2ipolymod_power(z, power, r);
	z2ipolymod_print(zpower, r);
	printf("\n");
}

// ----------------------------------------------------------------
void show_powers(
	char   ** args,
	int       num_args,
	int       power,
	z2ipoly_t r)
{
	int i;
	z2ipoly_t z;
	z2ipoly_t pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z2ipoly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z2ipoly_mod(z, r);
			show_power(z, power, r);
		}
	}
	else {
		pm = 1LL << z2ipoly_get_degree(r);
		for (z = 1; z < pm; z++) {
			show_power(z, power, r);
		}
	}
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {m or red} tbl {op}\n", argv0);
	fprintf(stderr, "  Op is one of +, -, * or /; . may be used in "
		"place of *.\n");
	fprintf(stderr, "Or:    %s {m or red} order  [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} orbit  [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} frobit [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} power {n} [...]\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t r;
	int long_tbl_fmt = 0;

	if (argc < 3)
		usage(argv[0]);

	if (!z2ipolymod_auxscan(&r, argv[1]))
		usage(argv[0]);

	if (strcmp(argv[2], "tbl") == 0) {
		if (argc < 4)
			usage(argv[0]);
		if (argc == 5)
			long_tbl_fmt = 1;
		show_table(argv[3], r, long_tbl_fmt);
	}
	else if (strcmp(argv[2], "order") == 0) {
		show_orders(&argv[3], argc - 3, r);
	}
	else if (strcmp(argv[2], "orbit") == 0) {
		show_orbits(&argv[3], argc - 3, r);
	}
	else if (strcmp(argv[2], "frobit") == 0) {
		show_frobits(&argv[3], argc - 3, r);
	}
	else if (strcmp(argv[2], "log") == 0) {
		show_logs(&argv[3], argc - 3, r);
	}
	else if (strcmp(argv[2], "power") == 0) {
		int power;
		if (argc < 4)
			usage(argv[0]);
		if (sscanf(argv[3], "%d", &power) != 1)
			usage(argv[0]);
		show_powers(&argv[4], argc - 4, power, r);
	}
	else {
		usage(argv[0]);
	}

	return 0;
}
