#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z3polymod.h"
#include "power.h"

typedef z3poly_t my_z3op_t(
	z3poly_t * pa,
	z3poly_t * pb,
	z3poly_t * pr);

// ----------------------------------------------------------------
void show_table(
	char * op,
	z3poly_t * pr,
	int long_fmt)
{
	z3poly_t a, b, c;
	int i, j;
	int lo;
	int pm;
	my_z3op_t * pz3op;

	pm = calc_power(3, z3poly_get_degree(pr));

	switch (op[0]) {
	case '+':
		pz3op = z3polymod_add;
		lo   = 0;
		break;
	case '-':
		pz3op = z3polymod_sub;
		lo   = 0;
		break;
	case '.':
	case '*':
		pz3op = z3polymod_mul;
		lo   = 0;
		break;
	case '/':
		pz3op = z3polymod_div;
		lo   = 1;
		break;
	default:
		fprintf(stderr, "Unknown operator \"%s\".\n",
			op);
		exit(1);
		break;
	}

	if (long_fmt) {
		for (i = lo; i < pm; i++) {
			a = z3poly_from_base_rep(i);
			for (j = lo; j < pm; j++) {
				b = z3poly_from_base_rep(j);
				c = pz3op(&a, &b, pr);

				z3polymod_print(&a, pr);
				printf(" %s ", op);
				z3polymod_print(&b, pr);
				printf(" = ");
				z3polymod_print(&c, pr);
				printf("\n");
			}
			printf("\n");
		}
	}

	else {
		for (i = lo; i < pm; i++) {
			a = z3poly_from_base_rep(i);
			for (j = lo; j < pm; j++) {
				b = z3poly_from_base_rep(j);
				c = pz3op(&a, &b, pr);
				if (j > lo)
					printf(" ");
				z3polymod_print(&c, pr);
			}
			printf("\n");
		}
	}
}

// ----------------------------------------------------------------
void show_order(
	z3poly_t * pz,
	z3poly_t * pr)
{
	int order;

	if (z3poly_is_zero(pz))
		order = 0;
	else
		order = z3polymod_element_order(pz, pr);
	printf("%3d ", order);
	z3polymod_print(pz, pr);
	printf("\n");
}

// ----------------------------------------------------------------
void show_orders(
	char ** args,
	int     num_args,
	z3poly_t * pr)
{
	int i;
	z3poly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z3poly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z3poly_mod(&z, pr);
			show_order(&z, pr);
		}
	}
	else {
		pm = calc_power(3, z3poly_get_degree(pr));
		for (i = 1; i < pm; i++) {
			z = z3poly_from_base_rep(i);
			show_order(&z, pr);
		}
	}
}

// ----------------------------------------------------------------
void show_orbit(
	z3poly_t * pz,
	z3poly_t * pr)
{
	int count;
	z3poly_t zpower;

	zpower = *pz;
	count = 0;
	while (1) {
		if (count > 0)
			printf(" ");
		count++;
		z3polymod_print(&zpower, pr);
		if (z3poly_is_zero(&zpower))
			break;
		if (z3poly_is_one(&zpower))
			break;
		zpower = z3polymod_mul(&zpower, pz, pr);
	}
	printf("\n");
}

// ----------------------------------------------------------------
void show_orbits(
	char ** args,
	int     num_args,
	z3poly_t * pr)
{
	int i;
	z3poly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z3poly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z3poly_mod(&z, pr);
			show_orbit(&z, pr);
		}
	}
	else {
		pm = calc_power(3, z3poly_get_degree(pr));
		for (i = 1; i < pm; i++) {
			z = z3poly_from_base_rep(i);
			show_orbit(&z, pr);
		}
	}
}

// ----------------------------------------------------------------
void show_frobit(
	z3poly_t * pz,
	z3poly_t * pr)
{
	int count;
	z3poly_t zpower;

	zpower = *pz;
	count = 0;
	while (1) {
		if (count > 0)
			printf(" ");
		count++;
		z3polymod_print(&zpower, pr);
		if (z3poly_is_zero(&zpower))
			break;
		if (z3poly_is_one(&zpower))
			break;
		if (z3poly_equal(&zpower, pz) && (count > 1))
			break;
		//z3polymod_mul(&zpower, &zpower, &zpower, pr);
		zpower = z3polymod_power(&zpower, 3, pr);
	}
	printf("\n");
}

// ----------------------------------------------------------------
void show_frobits(
	char ** args,
	int     num_args,
	z3poly_t * pr)
{
	int i;
	z3poly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z3poly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z3poly_mod(&z, pr);
			show_frobit(&z, pr);
		}
	}
	else {
		pm = calc_power(3, z3poly_get_degree(pr));
		for (i = 1; i < pm; i++) {
			z = z3poly_from_base_rep(i);
			show_frobit(&z, pr);
		}
	}
}

// ----------------------------------------------------------------
void show_log(
	z3poly_t * pz,
	z3poly_t * pr)
{
	int log, got_it = 0;
	z3poly_t xp;
	z3poly_t x = z3poly_from_base_rep(3);
	int deg = z3poly_get_degree(pr);
	int pm = calc_power(3, deg);

	// find p such that x^p == z mod r

	xp = z3poly_from_base_rep(1);
	log = 0;
	while (1) {
		if (z3poly_equal(pz, &xp)) {
			got_it = 1;
			break;
		}
		xp = z3polymod_mul(&xp, &x, pr);
		log++;
		if (log > pm)
			break;
	}
	z3polymod_print(pz, pr);
	printf(":  ");
	if (got_it) {
		printf("%3d\n", log);
	}
	else {
		printf("  not a power of x (");
		z3poly_print(pr);
		printf(" is not primitive)\n");
	}
}

// ----------------------------------------------------------------
void show_logs(
	char ** args,
	int     num_args,
	z3poly_t * pr)
{
	int i;
	z3poly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z3poly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z3poly_mod(&z, pr);
			show_log(&z, pr);
		}
	}
	else {
		pm = calc_power(3, z3poly_get_degree(pr));
		for (i = 1; i < pm; i++) {
			z = z3poly_from_base_rep(i);
			show_log(&z, pr);
		}
	}
}

// ----------------------------------------------------------------
void show_power(
	z3poly_t * pz,
	int       power,
	z3poly_t * pr)
{
	z3poly_t zpower;
	zpower = z3polymod_power(pz, power, pr);
	z3polymod_print(&zpower, pr);
	printf("\n");
}

// ----------------------------------------------------------------
void show_powers(
	char   ** args,
	int       num_args,
	int       power,
	z3poly_t * pr)
{
	int i;
	z3poly_t z;
	int pm;

	if (num_args > 0) {
		for (i = 0; i < num_args; i++) {
			if (!z3poly_scan(&z, args[i])) {
				fprintf(stderr, "Scan failure on \"%s\".\n",
					args[i]);
				exit(1);
			}
			z = z3poly_mod(&z, pr);
			show_power(&z, power, pr);
		}
	}
	else {
		pm = calc_power(3,  z3poly_get_degree(pr));
		for (i = 1; i < pm; i++) {
			z = z3poly_from_base_rep(i);
			show_power(&z, power, pr);
		}
	}
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {m or red} tbl {op}\n", argv0);
	fprintf(stderr, "  Op is one of +, -, * or /; . may be used in "
		"place of *.\n");
	fprintf(stderr, "Or:    %s {m or red} order  [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} orbit  [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} frobit [...]\n", argv0);
	fprintf(stderr, "Or:    %s {m or red} power {n} [...]\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z3poly_t r;
	int long_tbl_fmt = 0;

	if (argc < 3)
		usage(argv[0]);

	if (!z3poly_scan(&r, argv[1]))
		usage(argv[0]);

	if (strcmp(argv[2], "tbl") == 0) {
		if (argc < 4)
			usage(argv[0]);
		if (argc == 5)
			long_tbl_fmt = 1;
		show_table(argv[3], &r, long_tbl_fmt);
	}
	else if (strcmp(argv[2], "order") == 0) {
		show_orders(&argv[3], argc - 3, &r);
	}
	else if (strcmp(argv[2], "orbit") == 0) {
		show_orbits(&argv[3], argc - 3, &r);
	}
	else if (strcmp(argv[2], "frobit") == 0) {
		show_frobits(&argv[3], argc - 3, &r);
	}
	else if (strcmp(argv[2], "log") == 0) {
		show_logs(&argv[3], argc - 3, &r);
	}
	else if (strcmp(argv[2], "power") == 0) {
		int power;
		if (argc < 4)
			usage(argv[0]);
		if (sscanf(argv[3], "%d", &power) != 1)
			usage(argv[0]);
		show_powers(&argv[4], argc - 4, power, &r);
	}
	else {
		usage(argv[0]);
	}

	return 0;
}
