#include <stdio.h>
#include <stdlib.h>
#include "mod.h"
#include "posmod.h"
#include "zppoly.h"
#include "malloc_check.h"

#define NEG_ONE_POW(n) ((n) & 1 ? -1 : 1)

#define DEBUG 0

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {poly1} {poly2} {p}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int zppoly_resultant(
	zpoly_t * pf,
	zpoly_t * pg,
	int       p)
{
	zpoly_t f, g, quot, rem, * pswap;
	int res = 1;
	int degfr, bm, bmpow;

	if (pf->degree < pg->degree) {
		pswap = pf;
		pf    = pg;
		pg    = pswap;
	}
	f = *pf;
	g = *pg;

	for (;;) {
		if (g.degree == 0) {
			// R(f, c) = c^n
			bm = g.coeffs[0];
			bmpow = mod_power(bm, f.degree, p);
			res = mod_mul(res, bmpow, p);

#if DEBUG
			printf("\n");
			printf("f="); zppoly_print(&f, p); printf("\n");
			printf("@g="); zppoly_print(&g, p); printf("\n");
			printf("bmp=%d\n", bmpow);
			printf("res=%d\n", res);
#endif

			return res;
		}

		// R(f,g) = (-1)^nm R(g,f)
		// R(g,f) = b_m^{deg f - deg r} R(g, r), where f = qg + r
		// so,
		// R(f,g) = (-1)^nm b_m^{deg f - deg r} R(g, r)

		quot = zppoly_div(&f, &g, p);
		rem  = zppoly_mod(&f, &g, p);

		if ((f.degree * g.degree) & 1)
			res = posmod(-res, p);
		degfr = f.degree - rem.degree;
		if (degfr < 0) {
			fprintf(stderr, "ZPRES ABEND!\n");
			exit(1);
		}
		bm = g.coeffs[g.degree];
		bmpow = mod_power(bm, degfr, p);
		res = mod_mul(res, bmpow, p);

#if DEBUG
		printf("\n");
		printf("f="); zppoly_print(&f, p); printf("\n");
		printf("g="); zppoly_print(&g, p); printf("\n");
		printf("q="); zppoly_print(&quot, p); printf("\n");
		printf("r="); zppoly_print(&rem, p); printf("\n");
		printf("bmp=%d\n", bmpow);
		printf("res=%d\n", res);
#endif

		f = g;
		g = rem;
	}
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p;
	zpoly_t f, g;

	if (argc != 4)
		usage(argv[0]);
	if (sscanf(argv[3], "%d", &p) != 1)
		usage(argv[0]);
	if (!zppoly_scan(&f, argv[1], p))
		usage(argv[0]);
	if (!zppoly_scan(&g, argv[2], p))
		usage(argv[0]);

	printf("%d\n", zppoly_resultant(&f, &g, p));

	return 0;
}
