#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "power.h"
#include "zppolymod.h"
#include "zppoly_factor.h"

// ----------------------------------------------------------------
void foo_print_factors(
	zpoly_t * pr,
	int       p)
{
	zppoly_factor_info_t finfo;
	int i, j, k;

	zppoly_factor(pr, p, &finfo);

	k = 0;
	for (i = finfo.num_distinct - 1; i >= 0; i--) {
		for (j = 0; j < finfo.factor_counts[i].count; j++) {
			if (k > 0)
				printf(" ");
			zppoly_print(&finfo.factor_counts[i].f, p);
			k++;
		}
	}
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {r ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int p, argi, verbose = 0, nr = 0, n;
	zpoly_t r, a, g;
	zppoly_aux_t aux;
	int ihi, ia, group_order, elt_order, max_order;

	if (argc < 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);

	for (argi = 2; argi < argc; argi++) {
		if (strcmp(argv[argi], "-v") == 0) {
			verbose = 1;
			continue;
		}

		if (++nr > 1)
			printf("\n");
		if (!zppoly_scan(&r, argv[argi], p))
			usage(argv[0]);
		aux.r = r;
		aux.p = p;
		n = r.degree;
		ihi = calc_power(p, n);

		zppoly_print(&r, p);
		printf(" = [");
		zppoly_print_factors(&r, p);
		printf("]\n");

		group_order = 0;
		max_order = 0;
		for (ia = 0; ia < ihi; ia++) {
			a = zppoly_from_base_rep(p, ia);
			g = zppoly_gcd(&a, &r, p);
			if ((g.degree == 0) && !zppoly_is_zero(&g)) {
				elt_order = zppolymod_element_order(&a, &aux);
				if (elt_order > max_order)
					max_order = elt_order;
				if (verbose) {
					printf(" i ");
					zppolymod_print(&a, &aux);
					printf(" %d\n", elt_order);
				}
				group_order++;
			}
			else {
				if (verbose) {
					printf(" r ");
					zppolymod_print(&a, &aux);
					printf(" = [");
					foo_print_factors(&a, p);
					printf("]\n");
				}
			}
		}
		printf("Group order %d; max elt order %d\n",
			group_order, max_order);
	}

	return 0;
}
