#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "zppolymod.h"
#include "zplinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {p} {polys ...}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int argi, polyi;
	int p;
	zpoly_t * polys;
	int ** M;
	int i;
	int nrows = 0, ncols;
	int rank;
	int curr_deg, max_deg = 0;
	zpoly_t max_poly = zppoly_from_coeffs0(0, 1);
	zppoly_aux_t aux;

	if (argc < 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &p) != 1)
		usage(argv[0]);

	nrows = argc - 2;
	polys = malloc_check(nrows * sizeof(zpoly_t));

	for (argi = 2, polyi = 0; argi < argc; argi++, polyi++) {
		if (!zppoly_scan(&polys[polyi], argv[argi], p)) {
			fprintf(stderr, "%s:  couldn't scan \"%s\"\n",
				argv[0], argv[argi]);
			exit(1);
		}
		curr_deg = polys[polyi].degree;
		if (curr_deg > max_deg) {
			max_deg = curr_deg;
			max_poly = polys[polyi];
		}
	}
	ncols = max_deg + 1;

	M = zpla_alloc_matrix(nrows, ncols);
	zpla_matrix_from_zpolys(M, nrows, ncols, polys);

	zpla_row_echelon_form(M, p, nrows, ncols);
	rank = zpla_find_rank(M, nrows, ncols);

	zpla_zpolys_from_matrix(polys, M, rank, ncols);

	aux.p = p;
	aux.r = max_poly;
	aux.r.degree++;

	for (i = 0; i < rank; i++) {
		zppolymod_print(&polys[i], &aux);
		printf("\n");
	}

	zpla_free_matrix(M, nrows, ncols);
	free(polys);

	return 0;
}
