#include <stdio.h>
#include <stdlib.h>
#include "z2ipolymod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
z2ipoly_t get_fST(
	z2ipoly_t S1,
	z2ipoly_t T)
{
	z2ipoly_t S;
	z2ipoly_t xnk, Smk, fST;
	int m, n, k;

	m = z2ipoly_get_degree(T);
	n = z2ipoly_get_degree(S1);
	if (m < 1) {
		fprintf(stderr, "m must be >= 1.\n");
		exit(1);
	}
	if (n < 1) {
		fprintf(stderr, "n must be >= 1.\n");
		exit(1);
	}
	if ((m * n) > 63) {
		fprintf(stderr, "mn must be <= 63.\n");
		exit(1);
	}
	S = S1 ^ (1LL << n);

	// sum k=0 to m of t^k x^nk fS(m-k)

	fST = 0;
	for (k = 0; k <= m; k++) {

		// Check t^k
		if (!(T & (1LL << k)))
			continue;

		// Compute x^nk
		xnk = 1LL << (n * k);

		// Compute fS^{m-k}
		Smk = z2ipoly_power(S, m-k);

		fST ^= z2ipoly_mul(xnk, Smk);
	}

	// xxx temp hack
	//printf("%04llx %04llx ", S, T);

	return fST;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{

	printf("%04llx\n", get_fST(0x11LL, 0x13LL));
	printf("%04llx\n", get_fST(0x13LL, 0x13LL));
	printf("%04llx\n", get_fST(0x15LL, 0x13LL));
	printf("%04llx\n", get_fST(0x17LL, 0x13LL));
	printf("%04llx\n", get_fST(0x19LL, 0x13LL));
	printf("%04llx\n", get_fST(0x1bLL, 0x13LL));
	printf("%04llx\n", get_fST(0x1dLL, 0x13LL));
	printf("%04llx\n", get_fST(0x1fLL, 0x13LL));
	printf("\n");

	printf("%04llx\n", get_fST(0x11LL, 0x19LL));
	printf("%04llx\n", get_fST(0x13LL, 0x19LL));
	printf("%04llx\n", get_fST(0x15LL, 0x19LL));
	printf("%04llx\n", get_fST(0x17LL, 0x19LL));
	printf("%04llx\n", get_fST(0x19LL, 0x19LL));
	printf("%04llx\n", get_fST(0x1bLL, 0x19LL));
	printf("%04llx\n", get_fST(0x1dLL, 0x19LL));
	printf("%04llx\n", get_fST(0x1fLL, 0x19LL));
	printf("\n");

	return 0;
}
