#include <stdio.h>
#include <stdlib.h>
#include "z2ipolymod.h"
#include "z2ilinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void print_mx(
	z2ipoly_t * M,
	int n)
{
	int i;
	for (i = 0; i < n; i++) {
		if (i > 0)
			printf(" ");
		z2ipoly_dprint(M[i], n);
	}
	printf("\n");
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t Tp, * T, * Tk;
	int n, k, i;
	int niter = 20;
	int do_galois_cfg = 0;
	int isid;

	Tp = 0x13LL;

	if (argc >= 2)
		(void)sscanf(argv[1], "%llx", &Tp);
	if (argc >= 3)
		(void)sscanf(argv[2], "%d", &do_galois_cfg);
	if (argc >= 4)
		(void)sscanf(argv[3], "%d", &niter);

	n = z2ipoly_get_degree(Tp);
	if (n < 1) {
		fprintf(stderr, "n must be >= 1.\n");
		exit(1);
	}
	T = z2ipoly_companion_matrix(Tp, do_galois_cfg);
	Tk = malloc_check(n * sizeof(z2ipoly_t));

	printf("n = %d Tp = %llx\n", n, Tp);
	print_mx(T, n);
	printf("\n");

	for (i = 0; i < n; i++)
		Tk[i] = 1LL << (n-i-1);

	for (k = 0; k < niter; k++) {
		printf("[%3d] ", k);

		isid = 1;
		for (i = 0; i < n; i++) {
			if (Tk[i] != (1LL << (n-i-1))) {
				isid = 0;
				break;
			}
		}
		printf("%c ", isid ? '*' : '.');

		print_mx(Tk, n);
		z2ila_sq_mx_times_sq_mx(Tk, Tk, T, n);
	}

	free(T);
	free(Tk);
	return 0;
}
