#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "z2ipolymod.h"
#include "z2ipoly_factor.h"
#include "z2ilinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
z2ipoly_t find_xe1(
	z2ipoly_t T)
{
	int e;
	z2ipoly_t xe1, rem;

	for (e = 1; e <= Z2IPOLY_MAX_DEGREE; e++) {
		xe1 = (1LL << e) | 1;
		rem = z2ipoly_mod(xe1, T);
		if (rem == 0LL)
			return xe1;
	}

	//fprintf(stderr, "find_xe1: internal error or overflow at %llx.\n", T);
	//exit(1);
	return 0LL;
}

// ----------------------------------------------------------------
void do_T(
	z2ipoly_t T,
	int       do_galois_cfg,
	int       show_short_orbits)
{
	z2ipoly_t * M;
	z2ipoly_t a, Ta, hi;
	int i, n, orbit_length;
	char * marks;

	M = z2ipoly_companion_matrix(T, do_galois_cfg);
	n = z2ipoly_get_degree(T);
	hi = 1LL << n;

	marks = malloc_check(hi);

	printf("-- ");
	z2ipoly_print(T);

	//printf("-- %llx: rows = [", T);
	//for (i = 0; i < n; i++) {
	//	if (i > 0)
	//		printf(" ");
	//	z2ipolymod_print(M[i], T);
	//}
	//printf("]");

	if (T & 1)
		printf(" o=%d ", z2ipolymod_element_order(2LL, T));

	printf(" factors=[");
	z2ipoly_print_factors(T);
	printf("]");

	printf("\n");

	for (a = 0; a < hi; a++) {
		marks[a] = 0;
	}

	for (a = 0; a < hi; a++) {
		if (marks[a])
			continue;
		Ta = a;
		marks[Ta] = 1;
		orbit_length = 1;
		for (i = 0; i < hi; i++) {
			if (!show_short_orbits) {
				if (i > 0)
					printf(" ");
				z2ipolymod_print(Ta, T);
			}
			Ta = z2ila_mx_times_vr(M, n, Ta);
			marks[Ta] = 1;
			if (Ta == a)
				break;
			orbit_length++;
		}
		if (show_short_orbits)
			printf(" %d", orbit_length);
		else
			printf("\n");
	}
	if (show_short_orbits)
		printf("\n");

	free(M);
	free(marks);
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s [-f|-g] {r}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int argi;
	int do_galois_cfg = 0;
	int show_short_orbits = 0;
	int nT = 0;
	z2ipoly_t T;

	if (argc < 2)
		usage(argv[0]);

	for (argi = 1; argi < argc; argi++) {
		if (strcmp(argv[argi], "-f") == 0) {
			do_galois_cfg = 0;
			continue;
		}
		else if (strcmp(argv[argi], "-g") == 0) {
			do_galois_cfg = 1;
			continue;
		}
		else if (strcmp(argv[argi], "-s") == 0) {
			show_short_orbits = 1;
			continue;
		}

		if (nT > 0)
			printf("\n");

		if (sscanf(argv[argi], "%llx", &T) != 1) {
			fprintf(stderr, "%s: Couldn't scan \"%s\".\n",
				argv[0], argv[argi]);
			exit(1);
		}
		nT++;
		do_T(T, do_galois_cfg, show_short_orbits);
	}

	return 0;
}
