#include <stdio.h>
#include <stdlib.h>
#include "z2ipolymod.h"
#include "z2ilinalg.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void print_regs(
	z2ipoly_t * R,
	int         m,
	int         n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (i > 0)
			printf(" ");
		z2ipoly_dprint(R[i], m-1);
	}
	printf("\n");
}

// ----------------------------------------------------------------
static void shift(
	z2ipoly_t * T,
	z2ipoly_t   S,
	z2ipoly_t * R,
	int         m,
	int         n)
{
	z2ipoly_t sum = 0LL, T_sum;
	int i;

	for (i = 0; i < n; i++)
		if (S & (1LL << i))
			sum ^= R[i];
	T_sum = z2ila_mx_times_vr(T, m, sum);

	for (i = 0; i < n-1; i++)
		R[i] = R[i+1];
	R[n-1] = T_sum;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t Tp, S, * T, * R;
	int i, m, n;
	int do_galois_cfg = 0;

	Tp = 0x13LL;
	S  = 0x1fLL;

	if (argc >= 3) {
		(void)sscanf(argv[1], "%llx", &Tp);
		(void)sscanf(argv[2], "%llx", &S);
	}
	if (argc == 4)
		do_galois_cfg = 1;

	m = z2ipoly_get_degree(Tp);
	n = z2ipoly_get_degree(S);
	if (m < 1) {
		fprintf(stderr, "m must be >= 1.\n");
		exit(1);
	}
	if (n < 1) {
		fprintf(stderr, "n must be >= 1.\n");
		exit(1);
	}
	T = z2ipoly_companion_matrix(Tp, do_galois_cfg);
	R = malloc_check(n * sizeof(z2ipoly_t));

	printf("m = %d Tp = %llx n = %d S = %llx\n", m, Tp, n, S);

	for (i = 0; i < n-1; i++)
		R[i] = 0LL;
	R[n-1] = 1;
	if (argc >= 4) {
		(void)sscanf(argv[3], "%llx", &R[n-1]);
	}

	for (i = 0; i <= 20; i++) {
		print_regs(R, m, n);
		shift(T, S, R, m, n);
	}

	free(T);
	free(R);
	return 0;
}
