#include <stdio.h>
#include <stdlib.h>
#include "z2ipolymod.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
static void print_regs(
	z2ipoly_t * R,
	int         m,
	int         n)
{
	int i;

	for (i = 0; i < n; i++) {
		if (i > 0)
			printf(" ");
		z2ipoly_dprint(R[i], m-1);
	}
	printf("\n");
}

// ----------------------------------------------------------------
static void shift(
	z2ipoly_t   T,
	z2ipoly_t   S,
	z2ipoly_t * R,
	int         m,
	int         n)
{
	z2ipoly_t sum = 0LL, T_sum;
	int i;

	for (i = 0; i < n; i++)
		if (S & (1LL << i))
			sum ^= R[i];
	T_sum = z2ipolymod_mul(sum, 2LL, T);

	for (i = 0; i < n-1; i++)
		R[i] = R[i+1];
	R[n-1] = T_sum;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	z2ipoly_t T, S, * R;
	int i, m, n, niter = 20;

	T = 0x13LL;
	S = 0x1fLL;

	if (argc >= 3) {
		(void)sscanf(argv[1], "%llx", &T);
		(void)sscanf(argv[2], "%llx", &S);
	}

	m = z2ipoly_get_degree(T);
	n = z2ipoly_get_degree(S);
	if (m < 1) {
		fprintf(stderr, "m must be >= 1.\n");
		exit(1);
	}
	if (n < 1) {
		fprintf(stderr, "n must be >= 1.\n");
		exit(1);
	}
	R = malloc_check(n * sizeof(z2ipoly_t));

	printf("m = %d T = %llx n = %d S = %llx\n", m, T, n, S);

	for (i = 0; i < n-1; i++)
		R[i] = 0LL;
	R[n-1] = 1;
	if (argc >= 4) {
		(void)sscanf(argv[3], "%llx", &R[n-1]);
	}
	if (argc >= 5) {
		(void)sscanf(argv[4], "%d", &niter);
	}

	for (i = 0; i <= niter; i++) {
		print_regs(R, m, n);
		shift(T, S, R, m, n);
	}

	free(R);
	return 0;
}
