\\ Given modulus m and relatively prime factors f1, f2, ..., fn,
\\ find the Chinese coefficients of f and test them out.
\\
\\ To solve the n equations
\\
\\   x = r_i mod f_i, i = 1 .. n,
\\
\\ use
\\
\\   x = sum_i r_i (c_i (m/f_i) mod m)
\\
\\ where c_i = f_i/m mod f_i.

chirem(f)=
{
	local(i, j, n, c, C, r, m, rex);

	\\ Get factors of m from user
	n=length(f);
	for (i = 1, n,
		for (j=i+1,n,
			if (gcd(f[i], f[j]) != 1,
				error("Factors ", f[i], ", ", f[j],
					" are not relatively prime.")
			)
		)
	);

	\\ Compute m (product of factors)
	m = 1;
	for (i = 1, n,
		m *= f[i];
	);
	print("m = ", m);

	\\ Echo back factors
	print1("f = (");
	for (i = 1, n,
		if (i > 1, print1(" "));
		print1(f[i]);
	);
	print(")");


	\\ Compute c_i (Chinese coefficients)
	c=vector(n);
	C=vector(n);
	for (i = 1, n,
		\\ E.g. 1/110 mod 9 = 5
		c[i] = (1/(m/f[i])) % f[i];
		\\ E.g. 1/110 mod 9, times 110 = 550
		C[i] = c[i] * m/f[i];
	);

\\	ngood = 0;
\\	nbad  = 0;

	r=vector(n);
	for (x = 1, m,
		if (gcd(x, m) != 1, next);

		\\ Split x into remainders.
		for (i = 1, n,
			r[i] = x % f[i];
		);

		\\ Reconstitute x in terms of its remainders.
		\\ Use x = sum_i r_i c_i (m/f_i) mod m.
		rex = 0;
		for (i = 1, n,
			rex += r[i] * C[i];
			rex %= m;
		);

		if (verbose > 0,
			print1("  x=", x, " ");
			print1(" r=(");
			for (i = 1, n,
				print1(r[i]);
				if (i < n,
					print1(",");
				);
			);
			print1(")");
			print1(" rex=", rex);

			if (x != rex,
				print1(" !");
			);
			print("");
		);

		\\ Check that x has been reconstituted.
		if (x != rex,
			print("Error!");
		);
\\			ngood++;
\\		else
\\			nbad++;
	);
\\	if (nbad != 0)
\\		printf("Fail: Good %d, bad %d.\n", ngood, nbad);

	print1("C = (");
	for (i = 1, n,
		if (i != 1,
			print1(" ");
		);
		print1(C[i]);
	);
	print(")");

	print1("c = (");
	for (i = 1, n,
		if (i != 1,
			print1(" ");
		);
		print1(c[i]);
	);
	print(")");
}
