#include <stdio.h>
#include <stdlib.h>
#include "gcd.h"
#include "phi.h"

// ----------------------------------------------------------------
static char usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {m} [-v]\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
// Assumes x is relatively prime to m.
int calc_order(
	int x,
	int m)
{
	int px = x;
	int order;

	for (order = 1; px != 1; order++) {
		px = (px * x) % m;
	}
	return order;
}

// ----------------------------------------------------------------
int calc_max_order(
	int m,
	int do_list)
{
	int x;
	int max_order = 1;
	int order;

	if (do_list)
		printf("\n");
	for (x = 1; x < m; x++) {
		if (calc_gcd(x, m) != 1)
			continue;
		order = calc_order(x, m);
		if (order > max_order)
			max_order = order;
		if (do_list)
			printf("%03d:%02d\n", x, order);
	}
	if (do_list)
		printf("\n");
	return max_order;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int x;
	int xp;
	int m;

	if (argc != 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%d", &x) != 1)
		usage(argv[0]);
	if (sscanf(argv[2], "%d", &m) != 1)
		usage(argv[0]);

	if (calc_gcd(x, m) != 1) {
		fprintf(stderr, "%d is not relatively prime to %d.\n", x, m);
		exit(1);
	}

	for (xp = x; xp != 1; xp = (xp * x) % m)
		printf("%d\n", xp);
	printf("%d\n", xp);

	return 0;
}
