// xxx make moduli just 32 bits
// xxx get data types straightened out.  What really needs to be 64 bits?

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "isqrt.h"
#include "lcm.h"
#include "gcd.h"
#include "malloc_check.h"

#define MAX_M 32
#define MAX_R 32

static unsigned char * make_marks(
	unsigned length);

static unsigned * make_first_diffs(
	unsigned char * marks,
	unsigned   num_marks,
	int * pfirst_offset,
	int * pnum_diffs);

void report_marks_and_diffs(
	unsigned char * marks,
	int num_marks,
	unsigned * diffs,
	int num_diffs,
	int diff_first_offset);

// ----------------------------------------------------------------
unsigned long long factor_trial_no_sieve(
	unsigned long long N)
{
	unsigned long long top = calc_isqrt_64(N, IFLOOR);
	unsigned long long d;
	unsigned long long rv = N; // In case N is prime.
	unsigned long long num_trials = 0;

	for (d = 2; d <= top; d++) {
		num_trials++;
		if ((N % d) == 0) {
			rv = d;
			break;
		}
	}
	return rv;
}

// ----------------------------------------------------------------
unsigned long long factor_trial_linear_sieve(
	unsigned long long   N,
	unsigned * moduli,
	int        num_moduli)
{
	unsigned long long top = calc_isqrt_64(N, IFLOOR);
	unsigned lcm;
	unsigned long long d, i, j;
	unsigned char * marks;
	unsigned * diffs;
	int diff_first_offset, num_diffs, di;
	unsigned long long rv = N; // In case N is prime.
	unsigned long long num_trials = 0;

	lcm = calc_lcm_unsigned_multi(moduli, num_moduli);

	// Check for divisors which are 0 mod m[i]
	for (i = 0; i < num_moduli; i++) {
		if ((N % (unsigned long long)moduli[i]) == 0) {
		//if (calc_gcd_unsigned_ll(i, moduli[i]) != 1)
			rv = (unsigned long long)moduli[i];
			return rv;
		}
	}

	marks = make_marks(lcm);

	// Cast out trial divisors which are 0 mod m[i]
	for (i = 0; i < num_moduli; i++) {
		for (j = 0; j < lcm; j += moduli[i]) {
			marks[j] = 0;
		}
	}

	// Compute first differences
	diffs = make_first_diffs(marks, lcm, &diff_first_offset, &num_diffs);

	//report_marks_and_diffs(marks, lcm, diffs, num_diffs,
		//diff_first_offset);

	// Sieve
	for (d = diff_first_offset, di = 0; d <= top; ) {

		num_trials++;
		if ( ((N % d) == 0) && (d > 1)) {
			rv = d;
			break;
		}

		d += diffs[di];
		di++;
		if (di >= num_diffs)
			di = 0;
	}

	free(marks);
	free(diffs);
	return rv;
}

// ----------------------------------------------------------------
unsigned long long factor_fermat_no_sieve(
	unsigned long long N)
{
	unsigned long long x = calc_isqrt_64(N, IFLOOR);
	unsigned long long y, x2_minus_N;
	unsigned long long num_trials = 0;

	if ((N & 1) == 0) {
		return N >> 1;
	}

	while (1) {
		num_trials++;
		x2_minus_N = x * x - N;
		y = calc_isqrt_64(x2_minus_N, IFLOOR);
		if (y*y == x2_minus_N) {
			//return x - y;
			return x + y; // Return the bigger factor
		}
		x++;
	}
}

// ----------------------------------------------------------------
unsigned long long factor_fermat_quadratic_sieve(
	unsigned long long N,
	unsigned * moduli,
	int        num_moduli)
{
	unsigned long long x = calc_isqrt_64(N, IFLOOR);
	unsigned long long y, x2_minus_N, rv;
	unsigned lcm;
	unsigned char * marks;
	unsigned * diffs;
	int diff_first_offset, num_diffs, di;
	int i, j, k, is_square, j2_minus_N;
	unsigned long long num_trials = 0;
	unsigned long long squares[MAX_R][MAX_M];
	unsigned long long lag;

	if ((N & 1) == 0) {
		return N >> 1;
	}

	// Check for divisors which are 0 mod m[i]
	for (i = 0; i < num_moduli; i++) {
		if ((N % moduli[i]) == 0) {
		//if (calc_gcd_unsigned_ll(i, moduli[i]) != 1)
			rv = moduli[i];
			return rv;
		}
	}

	// Find the list of quadratic residues mod each m[i]
	lcm = calc_lcm_unsigned_multi(moduli, num_moduli);
	marks = make_marks(lcm);

	for (i = 0; i < num_moduli; i++) {
		for (k = 0; k < moduli[i]; k++)
			squares[i][k] = (k * k) % moduli[i];
	}

	// Unmark x mod m[i] where x**2 - N is not quadratic mod each m[i].
	// Do this as follows:  For each modulus m[i], for each x, unmark
	// if x is not quadratic mod that m[i].  All values still marked
	// after all moduli have been processed are those that will be
	// tried for factorization.

	for (i = 0; i < num_moduli; i++) {
		for (j = 0; j < moduli[i]; j++) {
			// Want to know if k**2 - N is quadratic mod m[i].
			//j2_minus_N = ((j*j) - (int)N) % (int)moduli[i];
			int Nmi = N % (long long)moduli[i];
			j2_minus_N = ((j*j) - Nmi) % (int)moduli[i];

			if (j2_minus_N < 0)
				j2_minus_N += moduli[i];
			is_square = 0;
			for (k = 0; k < moduli[i]; k++) {
				if (j2_minus_N == squares[i][k]) {
					is_square = 1;
					break;
				}
			}

			if (!is_square) {
				for (k = j; k < lcm; k+=moduli[i]) {
					marks[k] = 0;
				}
			}
		}
	}

	// Compute first differences of the marks
	diffs = make_first_diffs(marks, lcm, &diff_first_offset, &num_diffs);

	//report_marks_and_diffs(marks, lcm, diffs, num_diffs,
		//diff_first_offset);

	//printf("Work factor %3.0lf%% (%u marks with LCM %u)\n",
		//100.0 * (double)num_diffs / lcm, num_diffs, lcm);

	lag = (unsigned long long)lcm - (x % (unsigned long long)lcm);
	// xxx compute lag diff lcm - x
	// for x = . < nearest multiple of lcm
	//	test x
	//	x++

	// Now use diffs
	// for loop is kind of as above
	//	test x
	//	x++

	for (i = 0; i < lag; i++) {
		if (!marks[x % (unsigned long long)lcm]) {
			x++;
			continue;
		}
		num_trials++;
		x2_minus_N = x * x - N;
		y = calc_isqrt_64(x2_minus_N, IFLOOR);
		if (y*y == x2_minus_N) {
			free(marks);
			free(diffs);
			return x + y; // Return the bigger factor
		}
		x++;
	}

	// Sieve
	x += diff_first_offset;
	di = 0;
	while (1) {
		num_trials++;

		x2_minus_N = x * x - N;
		y = calc_isqrt_64(x2_minus_N, IFLOOR);
		if (y*y == x2_minus_N) {
			free(marks);
			free(diffs);
			return x + y; // Return the bigger factor
		}

		x += diffs[di];
		di++;
		if (di >= num_diffs)
			di = 0;
		if (x > N) {
			fprintf(stderr, "BOOM!\n");
			exit(1);
		}
	}

}

// ----------------------------------------------------------------
static unsigned char * make_marks(
	unsigned length)
{
	unsigned i;
	unsigned char * marks = malloc_check(length);

	for (i = 0; i < length; i++)
		marks[i] = 1;
	return marks;
}

// ----------------------------------------------------------------
// E.g. moduli are 2, 5.  LCM = 10
// 0 1 2 3 4 5 6 7 8 9
//   *   *       *   * <-- marks
// \ /\  /\     /\   /\ / (wrap to first mark)
//  1   2    4      2  2  <-- diffs
static unsigned * make_first_diffs(
	unsigned char * marks,
	unsigned   num_marks,
	int * pfirst_offset,
	int * pnum_diffs)
{
	int si, di, previ;
	unsigned * diffs;
	int first_offset, num_diffs;

	first_offset = -1;
	num_diffs = 0;
	for (si = 0; si < num_marks; si++) {
		if (marks[si] == 1) {
			if (first_offset == -1)
				first_offset = si;
			num_diffs++;
		}
	}
	if (num_diffs == 0) {
		fprintf(stderr,
			"Feh!  Haven't thought this case through yet.\n");
		exit(1);
	}

	diffs = (int *)malloc_check(num_diffs * sizeof(int));

	previ = first_offset;
	for (si = first_offset + 1, di = 0; si < num_marks; si++) {
		if (marks[si]) {
			diffs[di] = si -  previ;
			di++;
			previ = si;
		}
	}
	diffs[di] = first_offset + num_marks - previ;

	*pfirst_offset = first_offset;
	*pnum_diffs    = num_diffs;
	return diffs;
}

// ----------------------------------------------------------------
void report_marks_and_diffs(
	unsigned char * marks,
	int num_marks,
	unsigned * diffs,
	int num_diffs,
	int diff_first_offset)
{
	int i;

	printf("Marks [0-%d]: ", num_marks-1);
	for (i = 0; i < num_marks; i++) {
		//printf("%d", marks[i]);
		if (marks[i])
			printf(" %d", i);
	}
	printf("\n");

	printf("1do %d\n", diff_first_offset);
	printf("diffs: ");
	for (i = 0; i < num_diffs; i++)
		printf(" %d", diffs[i]);
	printf("\n");
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	unsigned long long N = 697;
	//unsigned moduli[] = { 9, 10, 11 };
	unsigned moduli[] = { 2, 3, 5, 7, 11, 13 };
	//unsigned moduli[] = { 4, 9, 5, 7, 11, 13 };
	int num_moduli = sizeof(moduli) / sizeof(moduli[0]);
	unsigned long long tns, tls, fns, fqs;
	int hex_io = 0;

	if (argc >= 2) {
		if (strncmp(argv[1], "0x", 2) == 0) {
			(void)sscanf(argv[1], "%llx", &N);
			hex_io = 1;
		}
		else {
			(void)sscanf(argv[1], "%llu", &N);
		}
	}

//	tns = factor_trial_no_sieve(N);
//	printf("Factor w/ trial,  no        sieve: %9llu * %9llu = %9llu\n",
//		tns, N/tns, tns * (N/tns));

	tls = factor_trial_linear_sieve(N, moduli, num_moduli);
	if (hex_io)
	printf("Factor w/ trial,  linear    sieve: 0x%llx * 0x%llx = 0x%llx\n",
		tls, N/tls, tls * (N/tls));
	else
	printf("Factor w/ trial,  linear    sieve: %9llu * %9llu = %9llu\n",
		tls, N/tls, tls * (N/tls));

//	fns = factor_fermat_no_sieve(N);
//	printf("Factor w/ fermat, no        sieve: %9llu * %9llu = %9llu\n",
//		fns, N/fns, fns * (N/fns));

	fqs = factor_fermat_quadratic_sieve(N, moduli, num_moduli);
	if (hex_io)
	printf("Factor w/ fermat, quadratic sieve: 0x%llx * 0x%llx = 0x%llx\n",
		fqs, N/fqs, fqs * (N/fqs));
	else
	printf("Factor w/ fermat, quadratic sieve: %9llu * %9llu = %9llu\n",
		fqs, N/fqs, fqs * (N/fqs));

	return 0;
}
