#ifndef FACTORIZE_H
#define FACTORIZE_H

#define FACTORIZE_MAX 20
// To factor an arbitrary 64-bit integer, the worst case (in terms
// of numbers of distinct primes) is the product of the first primes up
// to about 53.  53 is the 16th prime.  20 is a nice round number.
// If the number in question has a larger prime, then the remaining
// factors must be smaller, or fewer in number.  If there are repeated
// factors, then there will be fewer distinct ones.

// ----------------------------------------------------------------
typedef struct _factor_count_t {
	unsigned p;
	unsigned count;
} factor_count_t;
typedef struct _factor_info_t {
	factor_count_t factors[FACTORIZE_MAX];
	int  num_distinct;
} factor_info_t;

void calc_factors(
	unsigned n,
	factor_info_t * pfinfo);

void print_factors(
	unsigned n);

unsigned num_divisors(
	factor_info_t * pfinfo);

unsigned kth_divisor(
	factor_info_t * pfinfo,
	unsigned k);

// The caller must free the return value.
unsigned * get_all_divisors(
	unsigned   n,
	unsigned * pnum_divisors);

// ----------------------------------------------------------------
typedef struct _factor_count_ull_t {
	unsigned long long p;
	unsigned count;
} factor_count_ull_t;
typedef struct _factor_info_ull_t {
	factor_count_ull_t factors[FACTORIZE_MAX];
	int  num_distinct;
} factor_info_ull_t;

void calc_factors_ull(
	unsigned long long n,
	factor_info_ull_t * pfinfo);

unsigned num_divisors_ull(
	factor_info_ull_t * pfinfo);

unsigned long long kth_divisor_ull(
	factor_info_ull_t * pfinfo,
	unsigned k);

// The caller must free the return value.
unsigned long long * get_all_divisors_ull(
	unsigned long long n,
	unsigned * pnum_divisors);

#endif  // FACTORIZE_H
