#include "gcd.h"

// ----------------------------------------------------------------
int calc_gcd_signed(int a, int b)
{
	int r;

	if (a == 0)
		return b;
	if (b == 0)
		return a;

	if (a < 0) // Breaks for most negative int (-2^31); I don't care.
		a = -a;
	if (b < 0)
		b = -b;

	while (1) {
		r = a % b;
		if (r == 0)
			break;
		a = b;
		b = r;
	}

	return b;
}

// ----------------------------------------------------------------
int calc_ext_gcd_signed(int a, int b, int * pm, int * pn)
{
	int mprime, nprime, c, q, r, t, qm, qn;
	int d;   // Return value.

	// Initialize
	mprime = 1;
	*pn    = 1;
	*pm    = 0;
	nprime = 0;
	c = a;
	d = b;

	while (1) {

		// Divide
		q = c / d;
		r = c % d;
		// Note:  now c = qd + r and 0 <= r < d

		// Remainder zero?
		if (r == 0)
			break;

		// Recycle
		c = d;
		d = r;

		t      = mprime;
		mprime = *pm;
		qm     = q * *pm;
		*pm    = t - qm;

		t      = nprime;
		nprime = *pn;
		qn     = q * *pn;
		*pn    = t - qn;
	}
	return d;
}

// ----------------------------------------------------------------
unsigned calc_gcd_unsigned(unsigned a, unsigned b)
{
	unsigned r;

	if (a == 0)
		return b;
	if (b == 0)
		return a;

	while (1) {
		r = a % b;
		if (r == 0)
			break;
		a = b;
		b = r;
	}

	return b;
}

// ----------------------------------------------------------------
long long calc_gcd_signed_ll(long long a, long long b)
{
	long long r;

	if (a == 0)
		return b;
	if (b == 0)
		return a;

	if (a < 0) // Breaks for most negative int (-2^63); I don't care.
		a = -a;
	if (b < 0)
		b = -b;

	while (1) {
		r = a % b;
		if (r == 0)
			break;
		a = b;
		b = r;
	}

	return b;
}

// ----------------------------------------------------------------
unsigned long long calc_gcd_unsigned_ll(
	unsigned long long a,
	unsigned long long b)
{
	unsigned long long r;

	if (a == 0)
		return b;
	if (b == 0)
		return a;

	while (1) {
		r = a % b;
		if (r == 0)
			break;
		a = b;
		b = r;
	}

	return b;
}
