#include <stdio.h>
#include "isqrt.h"

// ----------------------------------------------------------------
unsigned calc_isqrt_32(
	unsigned n,
	int want_ceil)
{
	return calc_isqrt_bisect_32(n, want_ceil);
}

// ----------------------------------------------------------------
unsigned calc_isqrt_bisect_32(
	unsigned n,
	int want_ceil)
{
	unsigned nsv, half_log;
	unsigned lower, upper, diff;
	unsigned half_diff, middle, middle_sq;

	if (n == 0)
		return 0;

	for (nsv = n, half_log = 0; nsv != 0; nsv >>= 2, half_log++)
		;
	half_log--;
	lower = 1 << half_log;
	upper = lower << 1;
	diff  = upper - lower;

	while (diff > 0) {
		half_diff = diff >> 1;
		middle = lower + half_diff;
		middle_sq = middle * middle;
		if (middle_sq == n)
			return middle;
		else if (middle_sq < n)
			lower += half_diff;
		else
			upper -= half_diff;
		diff = half_diff;
	}
	if (want_ceil == ICEIL)
		return upper;
	else
		return lower;
}

// ----------------------------------------------------------------
unsigned long long calc_isqrt_64(
	unsigned long long n,
	int want_ceil)
{
	return calc_isqrt_bisect_64(n, want_ceil);
}

// ----------------------------------------------------------------
unsigned long long calc_isqrt_bisect_64(
	unsigned long long n,
	int want_ceil)
{
	unsigned long long nsv, half_log;
	unsigned long long lower, upper, diff;
	unsigned long long half_diff, middle, middle_sq;

	if (n == 0)
		return 0;

	for (nsv = n, half_log = 0; nsv != 0; nsv >>= 2, half_log++)
		;
	half_log--;
	lower = 1ULL << half_log;
	upper = lower << 1;
	diff  = upper - lower;

	while (diff > 0) {
		half_diff = diff >> 1;
		middle = lower + half_diff;
		middle_sq = middle * middle;
		if (middle_sq == n)
			return middle;
		else if (middle_sq < n)
			lower += half_diff;
		else
			upper -= half_diff;
		diff = half_diff;
	}
	if (want_ceil == ICEIL)
		return upper;
	else
		return lower;
}
