#include <stdio.h>
#include <stdlib.h>
#include "jacsym.h"
#include "legsym.h"
#include "power.h"
#include "factorize.h"

// ----------------------------------------------------------------
// Given (x|m), with m factoring into product of pi**ai,
// returns product of (x|pi)**ai, with (x|pi) being the
// Legendre symbol of x and pi.

int calc_jacsym_32(
	unsigned x,
	unsigned m)
{
	factor_info_t finfo;
	int i, l, rv = 1;

	if (!(m & 1)) {
		fprintf(stderr,
			"calc_jacsym_32:  Undefined for even m (%u).\n", m);
		exit(1);
	}

	calc_factors(m, &finfo);
	for (i = 0; i < finfo.num_distinct; i++) {
		l = calc_legsym_32(x, finfo.factors[i].p);
		if ((l == -1) && (finfo.factors[i].count & 1))
			rv = -rv;
	}

	return rv;
}

// ----------------------------------------------------------------
int calc_jacsym_64(
	unsigned long long x,
	unsigned long long m)
{
	factor_info_ull_t finfo;
	int i, l, rv = 1;

	if (!(m & 1LL)) {
		fprintf(stderr,
			"calc_jacsym_64:  Undefined for even m (%llu).\n", m);
		exit(1);
	}

	calc_factors_ull(m, &finfo);
	for (i = 0; i < finfo.num_distinct; i++) {
		l = calc_legsym_64(x, finfo.factors[i].p);
		if ((l == -1) && (finfo.factors[i].count & 1))
			rv = -rv;
	}

	return rv;
}

// ----------------------------------------------------------------
int main(void)
{
	unsigned x, m;
	int j;

	for (m = 3; m < 30; m+=2) {
		for (x = 0; x < m; x++) {
			j = calc_jacsym_32(x, m);
			printf("%3u %3u %2d\n", x, m, j);
		}
		printf("\n");
	}
	return 0;
}
