#include "log2.h"

static int u4_table[] = {
	0, /* 0x0 */ 0, /* 0x1 */ 1, /* 0x2 */ 1, /* 0x3 */
	2, /* 0x4 */ 2, /* 0x5 */ 2, /* 0x6 */ 2, /* 0x7 */
	3, /* 0x8 */ 3, /* 0x9 */ 3, /* 0xa */ 3, /* 0xb */
	3, /* 0xc */ 3, /* 0xd */ 3, /* 0xe */ 3, /* 0xf */
};

static int l4_table[] = {
	0, /* 0x0 */ 0, /* 0x1 */ 1, /* 0x2 */ 0, /* 0x3 */
	2, /* 0x4 */ 0, /* 0x5 */ 1, /* 0x6 */ 0, /* 0x7 */
	3, /* 0x8 */ 0, /* 0x9 */ 1, /* 0xa */ 0, /* 0xb */
	2, /* 0xc */ 0, /* 0xd */ 1, /* 0xe */ 0, /* 0xf */
};

// ----------------------------------------------------------------
unsigned find_msb_16(
	unsigned short n)
{
	unsigned rv = 0;
	unsigned short upper;

	upper = n >> 8;
	if (upper) {
		rv += 8;
		n = upper;
	}
	else
		n &= 0xff;

	upper = n >> 4;
	if (upper) {
		rv += 4;
		n = upper;
	}
	else
		n &= 0x0f;

	rv += u4_table[n];
	return rv;
}

// ----------------------------------------------------------------
unsigned find_msb_32(
	unsigned n)
{
	unsigned rv = 0;
	unsigned upper;

	upper = n >> 16;
	if (upper) {
		rv += 16;
		n = upper;
	}
	else
		n &= 0xffff;

	upper = n >> 8;
	if (upper) {
		rv += 8;
		n = upper;
	}
	else
		n &= 0xff;

	upper = n >> 4;
	if (upper) {
		rv += 4;
		n = upper;
	}
	else
		n &= 0x0f;

	rv += u4_table[n];
	return rv;
}

// ----------------------------------------------------------------
unsigned find_msb_64(
	unsigned long long n)
{
	unsigned upper = n >> 32;
	if (upper)
		return find_msb_32(upper) + 32;
	else
		return find_msb_32((unsigned)n);
}

// ----------------------------------------------------------------
unsigned find_lsb_32(
	unsigned n)
{
	unsigned rv = 0;
	unsigned lower;

	lower = n & 0xffff;
	if (lower) {
		n = lower;
	}
	else {
		n >>= 16;
		rv += 16;
	}

	lower = n & 0xff;
	if (lower) {
		n = lower;
	}
	else {
		n >>= 8;
		rv += 8;
	}

	lower = n & 0xf;
	if (lower) {
		n = lower;
	}
	else {
		n >>= 4;
		rv += 4;
	}

	rv += l4_table[n];
	return rv;
}

// ----------------------------------------------------------------
unsigned find_lsb_64(
	unsigned long long n)
{
	unsigned lower = n;
	if (lower)
		return find_lsb_32(lower);
	else
		return find_lsb_32((unsigned)(n >> 32)) + 32;
}

// ----------------------------------------------------------------
unsigned calc_log2_unsigned(
	unsigned n,
	int want_ceil)
{
	unsigned l = find_msb_32(n);

	if (want_ceil == IFLOOR) {
		return l;
	}
	else {
		if (((unsigned)1 << l) == n)
			return l;
		else
			return l+1;
	}
}

// ----------------------------------------------------------------
unsigned long long calc_log2_unsigned_ll(
	unsigned long long n,
	int want_ceil)
{
	unsigned long long l = (unsigned long long)find_msb_64(n);

	if (want_ceil == IFLOOR) {
		return l;
	}
	else {
		if (((unsigned long long)1LL << l) == n)
			return l;
		else
			return l+1;
	}
}
