#include <stdio.h>
#include <stdlib.h>
#include "gcd.h"

#define MAX_NUM_MODULI 32
#define MAX_MODULUS    32

// ----------------------------------------------------------------
unsigned compute_lcm_multi(
	unsigned * m,
	int        count)
{
	unsigned product, gcd, lcm, j;

	gcd = m[0];
	for (j = 1; j < count; j++) {
		gcd = calc_gcd_unsigned(gcd, m[j]);
	}
	product = 1;
	for (j = 0; j < count; j++) {
		product *= m[j];
	}
	lcm = product / gcd;
	if (count == 1) {
		gcd = m[0];
		lcm = m[0];
		product = m[0];
	}
	printf("GCD  %u\n", gcd);
	printf("prod %u\n", product);
	printf("LCM  %u\n", lcm);

	return lcm;
}

// ----------------------------------------------------------------
void report_linear_sieve(
	unsigned * moduli,
	int        num_moduli)
{
	unsigned lcm;
	int ncompos, nmaybe_prime;
	int known_composite;
	int x, i;

	lcm = compute_lcm_multi(moduli, num_moduli);

	ncompos = 0;
	nmaybe_prime = 0;
	for (x = 0; x < lcm; x++) {
		known_composite = 0;
		for (i = 0; i < num_moduli; i++) {
			if (calc_gcd_unsigned(x, moduli[i]) != 1) {
				known_composite = 1;
				break;
			}
		}
		if (known_composite) {
			printf(" %u\n", x);
			ncompos++;
		}
		else {
			printf(" %u *\n", x);
			nmaybe_prime++;
		}
	}
	printf("\n");
	printf("# known composite   %d\n", ncompos);
	printf("# maybe prime       %d\n", nmaybe_prime);
	printf("Contraction         %5.3lf%%\n",
		100.0 * (double)nmaybe_prime / (ncompos + nmaybe_prime));
}

// ----------------------------------------------------------------
void report_quadratic_sieve(
	unsigned * moduli,
	int        num_moduli)
{
	unsigned N = 0;
	char * pnstring;
	unsigned lcm;
	int nwill_divide, nto_skip;
	int is_quadratic_mod_all;
	int is_quadratic_mod_this;
	unsigned x, i, j;
	unsigned x2_minus_N;
	unsigned squares[MAX_NUM_MODULI][MAX_MODULUS];

	pnstring = getenv("MYN");
	if (pnstring)
		(void)sscanf(pnstring, "%u", &N);
	printf("N=%u\n", N);

	lcm = compute_lcm_multi(moduli, num_moduli);

	// is x^2 - N quadratic mod N?

	for (i = 0; i < num_moduli; i++) {
		for (j = 0; j < moduli[i]; j++) {
			squares[i][j] = (j*j) % moduli[i];
		}
	}

	printf("Squares table:\n");
	for (i = 0; i < num_moduli; i++) {
		printf("mod %3u: ", moduli[i]);
		for (j = 0; j < moduli[i]; j++) {
			printf(" %2u", squares[i][j]);
		}
		printf("\n");
	}
	printf("\n");

	nwill_divide = 0;
	nto_skip = 0;
	for (x = 0; x < lcm; x++) {
		x2_minus_N = x*x - N;
		x2_minus_N %= lcm;
		if (x2_minus_N < 0)
			x2_minus_N += lcm;

		printf(" x=%3u  x^2-N=%3u: ", x, x2_minus_N);
		is_quadratic_mod_all = 1;
		for (i = 0; i < num_moduli; i++) {

			is_quadratic_mod_this = 0;
			for (j = 0; j < moduli[i]; j++) {
				if ((x2_minus_N % moduli[i]) == squares[i][j]) {
					printf(" sq mod %u (%u=%u**2)",
						moduli[i],
						squares[i][j],
						x);
					is_quadratic_mod_this = 1;
					break;
				}
			}
			if (!is_quadratic_mod_this) {
				is_quadratic_mod_all = 0;
				//break;

			}
		}
		printf("\n");
		if (is_quadratic_mod_all) {
			nwill_divide++;
		}
		else {
			printf(" %u *\n", x);
			nto_skip++;
		}
	}
	printf("\n");
	printf("# will divide %d\n", nwill_divide);
	printf("# will skip   %d\n", nto_skip);
	printf("Contraction   %5.3lf%%\n",
		100.0 * (double)nwill_divide / (nwill_divide + nto_skip));
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	unsigned moduli[MAX_NUM_MODULI];
	int num_moduli = argc - 1;
	int j;

	if (argc < 2) {
		printf("Bleah.\n");
		exit(1);
	}

	for (j = 1; j < argc; j++) {
		(void)sscanf(argv[j], "%u", &moduli[j-1]);
	}

	//report_linear_sieve(moduli, num_moduli);
	report_quadratic_sieve(moduli, num_moduli);

	return 0;
}
