#include <stdio.h>
#include <stdlib.h>
#include "isqrt.h"
#include "isqrt_ubig.h"

// ----------------------------------------------------------------
int test_32(int argc, char ** argv)
{
	unsigned lo, hi, step;
	unsigned n, snf, snc;

	if (argc == 2) {
		(void)sscanf(argv[1], "%u", &lo);
		hi = lo;
		step = 1;
	}
	else if (argc == 3) {
		(void)sscanf(argv[1], "%u", &lo);
		(void)sscanf(argv[2], "%u", &hi);
		step = 1;
	}
	else if (argc == 4) {
		(void)sscanf(argv[1], "%u", &lo);
		(void)sscanf(argv[2], "%u", &hi);
		(void)sscanf(argv[3], "%u", &step);
	}
	else {
		fprintf(stderr, "Bleah.\n");
		exit(1);
	}

	for (n = lo; n <= hi; n += step) {
		snf = calc_isqrt_bisect_32(n, IFLOOR);
		snc = calc_isqrt_bisect_32(n, ICEIL);
		printf("isqrt(%6u) = %6u %6u%s\n",
			n, snf, snc, (snf * snf) == n ? " *" : "");
	}
	return 0;
}

// ----------------------------------------------------------------
int test_64(int argc, char ** argv)
{
	unsigned long long lo, hi, step;
	unsigned long long n, snf, snc;

	if (argc == 2) {
		(void)sscanf(argv[1], "%llu", &lo);
		hi = lo;
		step = 1;
	}
	else if (argc == 3) {
		(void)sscanf(argv[1], "%llu", &lo);
		(void)sscanf(argv[2], "%llu", &hi);
		step = 1;
	}
	else if (argc == 4) {
		(void)sscanf(argv[1], "%llu", &lo);
		(void)sscanf(argv[2], "%llu", &hi);
		(void)sscanf(argv[3], "%llu", &step);
	}
	else {
		fprintf(stderr, "Bleah.\n");
		exit(1);
	}

	for (n = lo; n <= hi; n += step) {
		snf = calc_isqrt_bisect_64(n, IFLOOR);
		snc = calc_isqrt_bisect_64(n, ICEIL);
		printf("isqrt(%16llu) = %16llu %16llu%s\n",
			n, snf, snc, (snf * snf) == n ? " *" : "");
	}
	return 0;
}

// ----------------------------------------------------------------
int test_ubig(int argc, char ** argv)
{
	ubig lo, hi, step;
	ubig n, snf, snc, snf_sq;
	ubig_opts_t uo;

	uo.hex_in = 0;
	uo.hex_out = 0;
	uo.zero_fill = 0;
	uo.underscore_out = 0;

	if (argc == 2) {
		(void)ubig_sscan(argv[1], &lo, &uo);
		hi = lo;
		step = ubig_from_unsigned(1);
	}
	else if (argc == 3) {
		(void)ubig_sscan(argv[1], &lo, &uo);
		(void)ubig_sscan(argv[2], &hi, &uo);
		step = ubig_from_unsigned(1);
	}
	else if (argc == 4) {
		(void)ubig_sscan(argv[1], &lo, &uo);
		(void)ubig_sscan(argv[2], &hi, &uo);
		(void)ubig_sscan(argv[3], &step, &uo);
	}
	else {
		fprintf(stderr, "Bleah.\n");
		exit(1);
	}

	for (n = lo; ubig_le(&n, &hi); ubig_add(&n, &n, &step)) {
		snf = calc_isqrt_bisect_ubig(&n, IFLOOR);
		snc = calc_isqrt_bisect_ubig(&n, ICEIL);
		snf_sq = ubig_mul(&snf, &snf);

		printf("isqrt(");
		ubig_print(stdout, &n, &uo);
		printf(") = ");
		ubig_print(stdout, &snf, &uo);
		printf(" ");
		ubig_print(stdout, &snc, &uo);
		printf("%s\n",
			ubig_eq(&snf_sq, &n) ? " *" : "");

	}
	return 0;
}

// ----------------------------------------------------------------
int compare_bis_new(int argc, char ** argv)
{
	ubig lo, hi, step;
	ubig n;
	ubig bis_snf, bis_snc, bis_snf_sq;
	ubig new_snf, new_snc, new_snf_sq;
	ubig_opts_t uo;
	unsigned ngood = 0, nbad = 0;

	uo.hex_in = 0;
	uo.hex_out = 0;
	uo.zero_fill = 0;
	uo.underscore_out = 0;

	if (argc == 2) {
		(void)ubig_sscan(argv[1], &lo, &uo);
		hi = lo;
		step = ubig_from_unsigned(1);
	}
	else if (argc == 3) {
		(void)ubig_sscan(argv[1], &lo, &uo);
		(void)ubig_sscan(argv[2], &hi, &uo);
		step = ubig_from_unsigned(1);
	}
	else if (argc == 4) {
		(void)ubig_sscan(argv[1], &lo, &uo);
		(void)ubig_sscan(argv[2], &hi, &uo);
		(void)ubig_sscan(argv[3], &step, &uo);
	}
	else {
		fprintf(stderr, "Bleah.\n");
		exit(1);
	}

	for (n = lo; ubig_le(&n, &hi); n = ubig_add(&n, &step)) {

		bis_snf = calc_isqrt_bisect_ubig(&n, IFLOOR);
		bis_snc = calc_isqrt_bisect_ubig(&n, ICEIL);
		bis_snf = ubig_mul(&bis_snf, &bis_snf);

		new_snf = calc_isqrt_newton_ubig(&n, IFLOOR);
		new_snc = calc_isqrt_newton_ubig(&n, ICEIL);
		new_snf_sq = ubig_mul(&new_snf, &new_snf);

		if (ubig_eq(&bis_snf, &new_snf)) {
			ngood++;
		}
		else {
			nbad++;
			printf("@ ");
			ubig_print(stdout, &n, &uo);
			printf(" bis_snf=");
			ubig_print(stdout, &bis_snf, &uo);
			printf(" new_snf=");
			ubig_print(stdout, &new_snf, &uo);
			printf("\n");
		}

		if (ubig_eq(&bis_snc, &new_snc)) {
			ngood++;
		}
		else {
			nbad++;
			printf("@ ");
			ubig_print(stdout, &n, &uo);
			printf(" bis_snc=");
			ubig_print(stdout, &bis_snc, &uo);
			printf(" new_snc=");
			ubig_print(stdout, &new_snc, &uo);
			printf("\n");
		}

	}
	printf("%u good, %u bad.\n", ngood, nbad);

	return (nbad == 0) ? 0 : 1;
}

// ----------------------------------------------------------------
int prof_bis_new(int argc, char ** argv)
{
	ubig lo, hi, step;
	ubig n;
	ubig snf;
	ubig sum;
	ubig_opts_t uo;
	int do_bis;

	uo.hex_in = 0;
	uo.hex_out = 0;
	uo.zero_fill = 0;
	uo.underscore_out = 0;

	if ((argc >= 2) && (argv[1][0] == 'b'))
		do_bis = 1;
	else
		do_bis = 0;
	if (argc == 3) {
		(void)ubig_sscan(argv[2], &lo, &uo);
		hi = lo;
		step = ubig_from_unsigned(1);
	}
	else if (argc == 4) {
		(void)ubig_sscan(argv[2], &lo, &uo);
		(void)ubig_sscan(argv[3], &hi, &uo);
		step = ubig_from_unsigned(1);
	}
	else if (argc == 5) {
		(void)ubig_sscan(argv[2], &lo, &uo);
		(void)ubig_sscan(argv[3], &hi, &uo);
		(void)ubig_sscan(argv[4], &step, &uo);
	}
	else {
		fprintf(stderr, "Bleah.\n");
		exit(1);
	}

	sum = ubig_from_unsigned(0);

	if (do_bis) {
		printf("bisect: ");
		for (n = lo; ubig_le(&n, &hi); n = ubig_add(&n, &step)) {
			snf = calc_isqrt_bisect_ubig(&n, IFLOOR);
			sum = ubig_add(&snf, &snf);
		}
	}
	else {
		printf("newton: ");
		for (n = lo; ubig_le(&n, &hi); n = ubig_add(&n, &step)) {
			snf = calc_isqrt_newton_ubig(&n, IFLOOR);
			sum = ubig_add(&snf, &snf);
		}
	}

	printf("sum ");
	ubig_dprint(&sum);
	printf("\n");

	return 0;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	//return test_32(argc, argv);
	//return test_64(argc, argv);
	//return test_ubig(argc, argv);
	//return compare_bis_new(argc, argv);
	return prof_bis_new(argc, argv);
}
