#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ubig_factorize.h"
#include "malloc_check.h"

typedef struct _ubig_string_factors_t {
	char * strings[UBIG_FACTORIZE_MAX + 1];
} ubig_string_factors_t;

#include "ubig_f2n1.dat"

#define MAXN_2N1 (sizeof(string_factors_2n1) / sizeof(string_factors_2n1[0]))

ubig_factor_info_t ubig_factors_2n1[MAXN_2N1];
static int init_done = 0;

// ----------------------------------------------------------------
static void myscan(
	ubig_factor_info_t * pfinfo,
	ubig_string_factors_t * pstrings,
	ubig   * temps,
	int      ntemps)
{
	int si, di;
	int nfact;

	// Scan strings into temp array.
	for (si = 0, di = 0; pstrings->strings[si]; si++, di++) {
		if (di >= ntemps) { // bounds check
			fprintf(stderr,
				"ubig_fac2n1_init():  buffer overflow.\n");
			exit(1);
		}
		if (!ubig_sdscan(pstrings->strings[si], &temps[di]))
			exit(1);
	}
	nfact = di;

	// Sort the temp array.
	qsort(temps, nfact, sizeof(ubig), ubig_qcmp);

	// Uniq the temp array into the fifo struct.
	pfinfo->num_distinct = 0;
	for (si = 0; si < nfact; ) {
		pfinfo->factors[pfinfo->num_distinct].count = 0;
		pfinfo->factors[pfinfo->num_distinct].p = temps[si];
		do {
			pfinfo->factors[pfinfo->num_distinct].count++;
			si++;
		} while ((si < nfact) && ubig_eq(&temps[si-1], &temps[si]));
		pfinfo->num_distinct++;
	}
}

// ----------------------------------------------------------------
static ubig unfactor(
	ubig_factor_info_t * pfinfo)
{
	ubig rv = ubig_from_unsigned(1);
	int i;

	for (i = 0; i < pfinfo->num_distinct; i++) {
		if (pfinfo->factors[i].count == 1) {
			rv = ubig_mul(&rv, &pfinfo->factors[i].p);
		}
		else {
			ubig pow = ubig_ipower(&pfinfo->factors[i].p,
				pfinfo->factors[i].count);
			rv = ubig_mul(&rv, &pow);
		}
	}
	ubig_incr(&rv);

	return rv;
}

// ----------------------------------------------------------------
static void self_test(void)
{
	int n;
	ubig one = ubig_from_unsigned(1);
	ubig expect, actual;
	int ngood = 0, nbad = 0;

	for (n = 0; n < MAXN_2N1; n++) {
		expect = ubig_ilsh(&one, n);
		actual = unfactor(&ubig_factors_2n1[n]);
		if (ubig_eq(&expect, &actual)) {
			ngood++;
		}
		else {
			printf("bad at %d\n", n);
			printf("actual: "); ubig_dprint(&actual); printf("\n");
			printf("expect: "); ubig_dprint(&expect); printf("\n");
			nbad++;
		}
	}

	if ((ngood != MAXN_2N1) || (nbad != 0)) {
		fprintf(stderr, "ubig_fac2n1 self-test failed!\n");
		exit(1);
	}
}

// ----------------------------------------------------------------
static void ubig_fac2n1_init(void)
{
	int i;
	ubig * temps = malloc_check(UBIG_FACTORIZE_MAX * sizeof(ubig));

	for (i = 0; i < MAXN_2N1; i++) {
		myscan(&ubig_factors_2n1[i], &string_factors_2n1[i],
			temps, UBIG_FACTORIZE_MAX);
	}
	self_test();
	init_done = 1;
	free(temps);
}

// ----------------------------------------------------------------
ubig_factor_info_t * ubig_fac2n1_get(
	int n)
{
	if (!init_done)
		ubig_fac2n1_init();

	if ((n < 0) || (n >= MAXN_2N1)) {
		fprintf(stderr,
			"ubig_fac2n1_get():  n (%d) out of bounds 0-%d\n",
			n, MAXN_2N1-1);
		exit(1);
	}

	return &ubig_factors_2n1[n];
}

//// ----------------------------------------------------------------
//static void usage(char * argv0)
//{
//	fprintf(stderr, "Usage: %s {n}\n", argv0);
//	exit(1);
//}
//
//// ----------------------------------------------------------------
//int main(int argc, char ** argv)
//{
//	int argi;
//	int n, i;
//	ubig_factor_info_t * pfinfo;
//
//	if (argc < 2)
//		usage(argv[0]);
//
//	for (argi = 1; argi < argc; argi++) {
//		if (sscanf(argv[1], "%d", &n) != 1) {
//			fprintf(stderr,
//				"%s:  couldn't scan \"%s\".\n",
//				argv[0], argv[argi]);
//			exit(1);
//		}
//		pfinfo = ubig_fac2n1_get(n);
//
//		if (pfinfo->num_distinct == 0) {
//			printf("(none)\n");
//		}
//		else {
//			for (i = 0; i < pfinfo->num_distinct; i++) {
//				printf(" ");
//				ubig_dprint(&pfinfo->factors[i].p);
//				if (pfinfo->factors[i].count > 1)
//					printf("^%u", pfinfo->factors[i].count);
//			}
//		}
//		printf("\n");
//	}
//
//	return 0;
//}
