#include <stdio.h>
#include <stdlib.h>
#include "ubig_factorize.h"
//#include "primes16.h"
//#include "power.h"
#include "malloc_check.h"

// ----------------------------------------------------------------
// Given the prime factorization p1^m1 ... pk^mk of n, how to enumerate all
// factors of n?
//
// Example 72 = 2^3 * 3^2.  Exponent on 2 is one of 0, 1, 2, 3.  Exponent on
// 3 is one of 0, 1, 2.  Number of possibilities:  product over i of (mi + 1).
// Factors are:
//
//	2^0 3^0    1
//	2^0 3^1    3
//	2^0 3^2    9
//	2^1 3^0    2
//	2^1 3^1    6
//	2^1 3^2   18
//	2^2 3^0    4
//	2^2 3^1   12
//	2^2 3^2   36
//	2^3 3^0    8
//	2^3 3^1   24
//	2^3 3^2   72

unsigned ubig_num_divisors(
	ubig_factor_info_t * pfinfo)
{
	unsigned rv;
	int i;

	if ((pfinfo->num_distinct < 0)
		|| (pfinfo->num_distinct > UBIG_FACTORIZE_MAX))
	{
		fprintf(stderr,
			"num_divisors:  num distinct (%d) out of bounds 0-%d\n",
			pfinfo->num_distinct, UBIG_FACTORIZE_MAX);
		exit(1);
	}

	rv = 1;
	for (i = 0; i < pfinfo->num_distinct; i++)
		rv *= pfinfo->factors[i].count + 1;
	return rv;
}

// ----------------------------------------------------------------
// See comments to the above.  k is treated as a multibase representation
// over the bases mi+1.

ubig ubig_kth_divisor(
	ubig_factor_info_t * pfinfo,
	unsigned k)
{
	ubig pow, rv;
	int i;
	unsigned base;
	unsigned e;

	rv = ubig_from_unsigned(1);
	for (i = 0; i < pfinfo->num_distinct; i++) {
		base = pfinfo->factors[i].count + 1;
		e = k % base;
		k = k / base;
		pow = ubig_ipower(&pfinfo->factors[i].p, e);
		rv = ubig_mul(&rv, &pow);
	}
	return rv;
}
