#ifndef UBIG_FACTORIZE_H
#define UBIG_FACTORIZE_H

#include "ubig.h"

// To factor an arbitrary 400-bit integer, the worst case (in terms
// of numbers of distinct primes) is the product of the first primes up
// to about 307.  307 is the 63rd prime.  65 is a nice round number.
// If the number in question has a larger prime, then the remaining
// factors must be smaller, or fewer in number.  If there are repeated
// factors, then there will be fewer distinct ones.

// top = 63;
// ps = primes(top);
// myprod = 1;
// for (i=1,top,myprod *= ps[i]);
// log(myprod) / log(2)

#define UBIG_FACTORIZE_MAX 65

typedef struct _ubig_factor_count_t {
	ubig p;
	unsigned count;
} ubig_factor_count_t;

typedef struct _ubig_factor_info_t {
	ubig_factor_count_t factors[UBIG_FACTORIZE_MAX];
	int  num_distinct;
} ubig_factor_info_t;

// ----------------------------------------------------------------
// No general-purpose ubig factorization routine is provided as of this
// writing.  ubig_f2n1.c may be used to obtain precomputed factors of
// 2^n-1 for various n.

unsigned ubig_num_divisors(
	ubig_factor_info_t * pfinfo);

ubig ubig_kth_divisor(
	ubig_factor_info_t * pfinfo,
	unsigned k);

#endif  // UBIG_FACTORIZE_H
