// Want a (p-1) x p matrix of ints.  For specified number of iterations:
// * Fill with units.
// * Permute the rows
// * Compute column products (PPUPs = permuted partial unit products).
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "factorize.h"
#include "isprime.h"
#include "posmod.h"
#include "mod.h"
#include "zlinalg.h"

static int get_rand_int(void);
static void permute_array(int * A, int len);

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	int lo    = 2;
	int hi    = 20;
	int maxiter = 100000, iter;
	int p, p2, i, j, ppup, up;
	int nyes, nno, cur;
	int ** M;
	int verbose = 0;

	if (argc == 2) {
		(void)sscanf(argv[1], "%d", &hi);
		lo = hi;
	}
	if (argc == 3) {
		(void)sscanf(argv[1], "%d", &lo);
		(void)sscanf(argv[2], "%d", &hi);
	}
	if (argc == 4) {
		(void)sscanf(argv[1], "%d", &lo);
		(void)sscanf(argv[2], "%d", &hi);
		(void)sscanf(argv[3], "%d", &maxiter);
	}
	if (argc == 5) {
		(void)sscanf(argv[1], "%d", &lo);
		(void)sscanf(argv[2], "%d", &hi);
		(void)sscanf(argv[3], "%d", &maxiter);
		verbose = 1;
	}

	for (p = lo; p <= hi; p++) {
		if (!isprime(p))
			continue;
		p2 = p * p;

		M = zla_alloc_matrix(p-1, p);

		nyes = 0;
		nno = 0;
		for (iter = 0; iter < maxiter; iter++) {

			// Initialize matrix.
			for (i = 0; i < p-1; i++) {
				for (j = 0; j < p; j++) {
					M[i][j] = i + 1 + p*j;
				}
			}

			// Permute rows
			//for (i = 0; i < p-1; i++)
			for (i = p-2; i < p-1; i++) {
				permute_array(M[i], p);
				// rand pmt
			}

			if (verbose) {
				printf("\n");
				zla_print_matrix(M, p-1, p);
				printf("\n");
			}

			// Compute PPUPs.
			up = 1;
			cur = 1;
			for (j = 0; j < p; j++) {
				ppup = 1;
				for (i = 0; i < p-1; i++) {
					ppup *= M[i][j];
					ppup %= p2;
				}
				if (ppup != (p2-1))
					cur = 0;
				if (verbose)
					printf(" %d", ppup);
				up *= ppup;
				up %= p2;
			}
			if (verbose)
				printf(" | %d (%d)\n", up, cur);

			// xxx temp
			if (cur) {
				zla_print_matrix(M, p-1, p);
				printf("\n");
			}

#if 0
			// xxx temp
			// Compute what the last row *should* be like.
			cur = 1;
			for (j = 0; j < p; j++) {
				int want;
				ppup = 1;
				for (i = 0; i < p-2; i++) {
					ppup *= M[i][j];
					ppup %= p2;
				}
				// abcd = -1; want d = -1/abc
				want = mod_recip(ppup, p2);
				want *= p2-1;
				want %= p2;
				printf(" %d", want);

				// xxx temp
				M[p-2][j] = want;
			}
			printf(" <-- want\n");
			// xxx temp

			// Compute PPUPs.
			up = 1;
			cur = 1;
			for (j = 0; j < p; j++) {
				ppup = 1;
				for (i = 0; i < p-1; i++) {
					ppup *= M[i][j];
					ppup %= p2;
				}
				if (ppup != (p2-1))
					cur = 0;
				if (verbose)
					printf(" %d", ppup);
				up *= ppup;
				up %= p2;
			}
			if (verbose)
				printf(" | %d (%d)\n", up, cur);
#endif

			if (cur == 1)
				nyes++;
			else
				nno++;

		}
		printf("nyes = %d  nno = %d\n", nyes, nno);

		zla_free_matrix(M, p-1, p);
	}

	return 0;
}

// ----------------------------------------------------------------
static void permute_array(int * A, int len)
{
	int i, j, t;

	for (i = 0; i < len; i++) {
		j = get_rand_int() % len;
		if (j < 0)
			j += len;
		t    = A[i];
		A[i] = A[j];
		A[j] = t;
	}
}

// ----------------------------------------------------------------
#define BUFSZ 10000
static int get_rand_int(void)
{
	static int fd = -1;
	static int buf[BUFSZ];
	static int bufpos = 0;
	int rv;
	if (fd == -1) {
		fd = open("/dev/urandom", O_RDONLY);
		if (fd < 0) {
			fprintf(stderr, "Couldn't open /dev/urandom.\n");
			perror("Reason");
			exit(1);
		}
	}

	if (bufpos == 0) {
		if (read(fd, buf, BUFSZ * sizeof(int)) < 0) {
			fprintf(stderr, "Couldn't read /dev/urandom.\n");
			perror("Reason");
			exit(1);
		}
	}
	rv = buf[bufpos];
	bufpos++;
	if (bufpos >= BUFSZ)
		bufpos = 0;
	return rv;
}

