// Want a (p-1) x p matrix of ints.  For specified number of iterations:
// * Fill with units.
// * Permute the rows
// * Compute column products (PPUPs = permuted partial unit products).
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "factorize.h"
#include "isprime.h"
#include "posmod.h"
#include "mod.h"

typedef long long i64;

// ----------------------------------------------------------------
i64 calc_factorial_mod(i64 a, i64 m)
{
	i64 rv = 1;

	for ( ; a > 1; a--) {
		rv *= a;
		rv %= m;
	}
	return rv;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	i64 lo    = 2;
	i64 hi    = 20;
	i64 p, p2, pm2fact, quot, k, w;
	int verbose = 0;

	if (argc == 2) {
		(void)sscanf(argv[1], "%lld", &hi);
		lo = hi;
	}
	if (argc == 3) {
		(void)sscanf(argv[1], "%lld", &lo);
		(void)sscanf(argv[2], "%lld", &hi);
	}
	if (argc == 4) {
		(void)sscanf(argv[1], "%lld", &lo);
		(void)sscanf(argv[2], "%lld", &hi);
		verbose = 1;
	}

	for (p = lo; p <= hi; p++) {
		if (!isprime(p))
			continue;

		// Find k such that (p-2)! * (kp-1) = -1 mod p^2.
		//
		// (p-2)! * (kp-1) = -1   mod p^2.
		// kp-1 = -1 / (p-2)!     mod p^2.
		// kp   = -1 / (p-2)! + 1 mod p^2.
		// k    = -1 / (p-2)! + 1 mod p

		p2 = p * p;
		pm2fact = calc_factorial_mod(p-2, p2);
		quot = mod64_div(p2-1, pm2fact, p2);
		quot++;
		quot %= p2;
		if ((quot % p) != 0) {
			printf("ABEND!\n");
			exit(1);
		}
		k = quot / p;
		w = ((pm2fact * (p-1) + 1) % p2) / p;
		//printf("p=%8lld (p-2)!=%8lld quot=%8lld k=%8lld w=%8lld",
			//p, pm2fact, quot, k, w);
		printf("%8lld %8lld %8lld %8lld", p, k, w, (k+w)%p);
		if (k == 1)
			printf(" *");
		printf("\n");
	}

	return 0;
}
