//a.out 1 2 | tail -1
//a.out 2 1 | tail -1
//echo ""
//
//a.out 1 3 | tail -1
//a.out 3 1 | tail -1
//echo ""
//
//a.out 1 4 | tail -1
//a.out 4 1 | tail -1
//echo ""
//
//a.out 2 3 | tail -1
//a.out 3 2 | tail -1
//echo ""
//
//a.out 2 4 | tail -1
//a.out 4 2 | tail -1
//echo ""

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

typedef unsigned long long u64;

// ----------------------------------------------------------------
// I want to count all linear transformations from Z2^n to Z2^m.
// There are 2^n inputs, each n bits wide.  (These are just n-tuples over Z2.)
// For a given function (not necessarily linear), there are 2^n m-bit outputs,
// one for each of the 2^n inputs.
// There are (2^m)^(2^n) = 2^(m*2^n) functions from Z2^n to Z2^m.

// ----------------------------------------------------------------
// Write fno as an m*2^n-bit integer.
// Split into 2^n m-bit parts; take the xth.

// E.g. n = 3, m = 2:
// 2^3 = 8 elements of Z_2^3
// 2^3 = 8 images
// Each element image has 2 bits
// m * 2^n = 2 * 2^3 = 16
// There are 2^16 = 65536 functions.
// Look at function number 0xbeef

// x:    0      1      2      3      4      5      6      7
//    f(000) f(001) f(010) f(011) f(100) f(101) f(110) f(111)
//       =      =      =      =      =      =      =      =
// y:   10     11     11     10     11     10     11     11 (0xbeef)
//
// E.g. get fi(3) by right-shifting fno by 4*2, then mask with binary 11.

u64 nmi_func(u64 fno, u64 n, u64 m, u64 x)
{
	u64 two_to_the_n = (u64)1 << n;
	u64 two_to_the_m = (u64)1 << m;
	u64 nx    = two_to_the_n - x - 1;
	u64 shamt = nx * m;
	u64 mask  = two_to_the_m - 1;
	u64 y     = (fno >> shamt) & mask;
	return y;
}

// ----------------------------------------------------------------
char * u64_to_binary(u64 u, char * buf, u64 nbits)
{
	int bit_pos;
	int str_pos;
	for (
		bit_pos = nbits - 1, str_pos = 0;
		bit_pos >= 0;
		bit_pos--, str_pos++)
	{
		if (u & ((u64)1 << bit_pos))
			buf[str_pos] = '1';
		else
			buf[str_pos] = '0';
	}
	buf[nbits] = 0;
	return buf;
}

// ----------------------------------------------------------------
static void usage(char * argv0)
{
	fprintf(stderr, "Usage: %s {n} {m}\n", argv0);
	exit(1);
}

// ----------------------------------------------------------------
void check_overflow(u64 n, u64 m)
{
	u64 itwo_to_the_n;
	u64 itwo_to_the_m;
	u64 infunc;
	double dtwo_to_the_n;
	double dtwo_to_the_m;
	double dnfunc;

	itwo_to_the_n = (u64)1 << n;
	itwo_to_the_m = (u64)1 << m;
	infunc = (u64)1 << (m * itwo_to_the_n);

	dtwo_to_the_n = pow(2.0, (double)n);
	dtwo_to_the_m = pow(2.0, (double)m);
	dnfunc        = pow(2.0, (double)m * dtwo_to_the_n);

	if ((u64)dtwo_to_the_n != itwo_to_the_n) {
		printf("Overflow\n");
		exit(1);
	}

	if ((u64)dtwo_to_the_m != itwo_to_the_m) {
		printf("Overflow\n");
		exit(1);
	}

	if ((u64)dnfunc != infunc) {
		printf("Overflow\n");
		exit(1);
	}
}

// ----------------------------------------------------------------
void report_sizes(u64 n, u64 m)
{
	u64 two_to_the_n;
	u64 two_to_the_m;
	u64 nfunc;

	two_to_the_n = (u64)1 << n;
	two_to_the_m = (u64)1 << m;
	nfunc = (u64)1 << (m * two_to_the_n);

	printf("n=%2llu  ", n);
	printf("m=%2llu  ", m);
	printf("2^n=%2llu  ", two_to_the_n);
	printf("2^m=%2llu  ", two_to_the_m);
	printf("m2n=%4llu  ", m * two_to_the_n);
	printf("nfu=%10llu\n", nfunc);
}

// ----------------------------------------------------------------
void print_all_xy(u64 n, u64 m)
{
	u64 two_to_the_n;
	u64 two_to_the_m;
	u64 nfunc;
	u64 x, y, fno;
	char xbuf[70];
	char ybuf[70];
	int width = (m > n) ? m : n;

	two_to_the_n = (u64)1 << n;
	two_to_the_m = (u64)1 << m;
	nfunc = (u64)1 << (m * two_to_the_n);

	for (x = 0; x < two_to_the_n; x++)
		printf(" %*s", width, u64_to_binary(x, xbuf, n));
	printf("\n");
	printf("\n");
	for (fno = 0; fno < nfunc; fno++) {
		for (x = 0; x < two_to_the_n; x++) {
			y = nmi_func(fno, n, m, x);
			printf(" %*s", width, u64_to_binary(y, ybuf, m));
		}
		printf("  (# %llu/%llu)\n", fno, nfunc);
	}
}

// ----------------------------------------------------------------
void print_one_xy(u64 fno, u64 n, u64 m)
{
	u64 two_to_the_n;
	u64 two_to_the_m;
	u64 nfunc;
	u64 x, y;
	char xbuf[70];
	char ybuf[70];

	two_to_the_n = (u64)1 << n;
	two_to_the_m = (u64)1 << m;
	nfunc = (u64)1 << (m * two_to_the_n);

	for (x = 0; x < two_to_the_n; x++) {
		y = nmi_func(fno, n, m, x);
		printf(" %s:%s",
			u64_to_binary(x, xbuf, n),
			u64_to_binary(y, ybuf, m));
	}
	printf("\n");
}

// ----------------------------------------------------------------
int preserves_zero(u64 fno, u64 n, u64 m)
{
	if (nmi_func(fno, n, m, 0) == 0)
		return 1;
	else
		return 0;
}

// ----------------------------------------------------------------
// f(x) + f(y) =?= f(x+y)
int preserves_sums(u64 fno, u64 n, u64 m)
{
	u64 x, y, fx, fy, fxy;
	u64 two_to_the_n = (u64)1 << n;

	for (x = 0; x < two_to_the_n; x++) {
		for (y = 0; y < two_to_the_n; y++) {
			fx  = nmi_func(fno, n, m, x);
			fy  = nmi_func(fno, n, m, y);
			fxy = nmi_func(fno, n, m, x^y);

			if ((fx^fy) != fxy)
				return 0;
		}
	}
	return 1;
}

// ----------------------------------------------------------------
// f(cx) =?= cf(x)
int preserves_scalar_product(u64 fno, u64 n, u64 m)
{
	u64 x, c, cx, fx, fcx, cfx;
	u64 two_to_the_n = (u64)1 << n;

	for (x = 0; x < two_to_the_n; x++) {
		for (c = 0; c <= 1; c++) {
			cx  = c * x;
			fx  = nmi_func(fno, n, m, x);
			cfx = c * fx;
			fcx = nmi_func(fno, n, m, cx);

			if (cfx != fcx)
				return 0;
		}
	}
	return 1;
}

// ----------------------------------------------------------------
int main(int argc, char ** argv)
{
	u64 n, m;
	u64 two_to_the_n;
	u64 two_to_the_m;
	u64 fno;
	u64 nfunc;
	u64 ltcount;

	if (argc != 3)
		usage(argv[0]);
	if (sscanf(argv[1], "%llu", &n) != 1)
		usage(argv[0]);
	if (sscanf(argv[2], "%llu", &m) != 1)
		usage(argv[0]);
	two_to_the_n = (u64)1 << n;
	two_to_the_m = (u64)1 << m;
	nfunc = (u64)1 << (m * two_to_the_n);

	check_overflow(n, m);
	//report_sizes(n, m);
	//print_xy(n, m);
	//printf("\n");

	ltcount = 0;
	for (fno = 0; fno < nfunc; fno++) {
		//printf("%6llu: ", fno);

		if (!preserves_zero(fno, n, m)) {
			//printf("doesn't preserve zero.\n");
			continue;
		}
		if (!preserves_sums(fno, n, m)) {
			//printf("doesn't preserve sums.\n");
			continue;
		}

		//printf("linear.\n");
		ltcount++;
		printf("#0x%06llx: ", fno);
		print_one_xy(fno, n, m);
	}
	printf("\n");
	printf("n=%llu m=%llu: %6llu of %6llu %s linear.\n",
		n, m, ltcount, nfunc, ltcount == 1 ? "is" : "are");

	return 0;
}
